#include <enoki/autodiff.h>
#include <enoki/cuda.h>
#include <mitsuba/render/autodiff.h>
#include <mitsuba/optix/renderer.h>
#include <tinyformat.h>

#define MTS_HANDLE_PARAM_CASES(func, ptr, size)                                                    \
    switch (size) {                                                                                \
        case 1:                                                                                    \
            func((FloatD *) ptr);                                                                  \
            break;                                                                                 \
        case 2:                                                                                    \
            func((Vector2fD *) ptr);                                                               \
            break;                                                                                 \
        case 3:                                                                                    \
            func((Vector3fD *) ptr);                                                               \
            break;                                                                                 \
        case 4:                                                                                    \
            func((Vector4fD *) ptr);                                                               \
            break;                                                                                 \
        case 16:                                                                                   \
            func((Matrix4fD *) ptr);                                                               \
            break;                                                                                 \
        default:                                                                                   \
            throw std::runtime_error("DifferentiableParameters: unsupported type!");               \
    }

NAMESPACE_BEGIN(mitsuba)

void set_requires_gradients_for_param(const DifferentiableParameters &params, const std::string &key) {
    const auto &entry = params[key];
    void *ptr = std::get<1>(entry);
    size_t size = std::get<2>(entry);

    auto set_requires = [](auto *array) {
        set_requires_gradient(*array);
    };
    MTS_HANDLE_PARAM_CASES(set_requires, ptr, size);
}

void set_requires_gradients_for_param(const DifferentiableParameters &params, const std::vector<std::string> &keys) {
    for (const auto &key : keys)
        set_requires_gradients_for_param(params, key);
}

/**
 * Returns a pointer to the gradient buffer that was actually accumulated to,
 * which may or may not be the given `out_grad`.
 */
void *accumulate_param_gradient_to_buffer(const DifferentiableParameters &params,
                                          const std::string &key, const std::string &prefix,
                                          FloatC *out_grad) {

    const auto &entry = params[key];
    void *ptr         = std::get<1>(entry);
    size_t size       = std::get<2>(entry);
    void *actual_grad_buffer_ptr = out_grad->data();

    auto scatter_grad = [&prefix, &out_grad, &actual_grad_buffer_ptr](auto *array) {
        using T = expr_t<decltype(*array)>;

        const auto &grad_of_param = gradient(*array);
        set_label(grad_of_param, (prefix + "grad_of_param").c_str());

        if constexpr (array_depth_v<T> == 1) {
            // For example, T = FloatD.
            // If this parameter is a texture, it is most likely gathered from when
            // it is used. In the backward pass, this leads to a `scatter_add`. This
            // means that the buffer is `dirty`, and we can use it directly as our output.
            // Otherwise, we perform the `scatter_add` ourselves.
            if (cuda_var_is_dirty(grad_of_param.index_())) {
                actual_grad_buffer_ptr = (void *) grad_of_param.data();
            } else {
                auto index = arange<UInt64C>(1);
                scatter_add(*out_grad, grad_of_param, index);
            }
        } else if constexpr (array_depth_v<T> == 2) {
            // For example, T = Vector3fD. Linearized storage (e.g. RGBRGBRGBRGB...)
            // TODO: any way to avoid the multiple scatter_add?
            auto index = arange<UInt64C>(1);
            set_label(index, (prefix + "_index").c_str());
            for (size_t k = 0; k < T::Size; ++k) {
                auto active = neq(grad_of_param[k], 0.f);
                scatter_add(*out_grad, grad_of_param[k], index + k, active);
            }
        } else {
            throw std::runtime_error("Not implemented: accumulating param gradients for arrays of "
                                     "depth > 2 (e.g. matrices).");
        }
    };

    MTS_HANDLE_PARAM_CASES(scatter_grad, ptr, size);
    return actual_grad_buffer_ptr;
}

void add_gradient_for_param(const DifferentiableParameters &params, const std::string &key, const FloatC &grad, bool add_to_graph) {
    const auto &entry = params[key];
    void *ptr         = std::get<1>(entry);
    size_t size       = std::get<2>(entry);

    auto set_grad = [&grad, &add_to_graph](auto *array) {
        using T = expr_t<decltype(detach(*array))>;
#if 0
        if constexpr (array_depth_v<T> == 1) {
            // For example, T = FloatD
            set_gradient(*array, grad, /*backward*/ true, add_to_graph);
            // TODO: add back support for this
            // set_gradient(*array, (prev_grad.size() > 0 ? grad + prev_grad : grad),
            //              /*backward*/ true, add_to_graph);
        } else if constexpr (array_depth_v<T> == 2) {
            // For example, T = Vector3fD.
            // Gradient was stored linearized (e.g. RGBRGBRGBRGB...).
            T grad_reshaped;
            auto index = arange<UInt64C>(array->x().size());
            for (size_t k = 0; k < T::Size; ++k) {
                grad_reshaped[k] = gather<FloatC>(grad, index + k);
            }

            // TODO: add back support for this
            // auto prev_grad = gradient(*array);
            // set_gradient(*array, (prev_grad.size() > 0 ? grad_reshaped + prev_grad : grad_reshaped),
            //              /*backward*/ true, add_to_graph);
            set_gradient(*array, grad_reshaped, /*backward*/ true, add_to_graph);
        } else {
            throw std::runtime_error("Not implemented: setting param gradients for arrays of "
                                     "depth > 2 (e.g. matrices).");
        }
#else
        if constexpr (array_depth_v<T> == 1) {
            // For example, T = FloatD
            set_gradient(*array, grad, /*backward*/ true, add_to_graph);
        } else if constexpr (array_depth_v<T> == 2) {
            // For example, T = Vector3fD.
            // Gradient was stored linearized (e.g. RGBRGBRGBRGB...).
            T grad_reshaped;
            auto index = arange<UInt64C>(array->x().size());
            for (size_t k = 0; k < T::Size; ++k) {
                grad_reshaped[k] = gather<FloatC>(grad, index + k);
            }
            set_gradient(*array, grad_reshaped, /*backward*/ true, add_to_graph);
        } else {
            throw std::runtime_error("Not implemented: setting param gradients for arrays of "
                                     "depth > 2 (e.g. matrices).");
        }
#endif
    };

    MTS_HANDLE_PARAM_CASES(set_grad, ptr, size);
}
#undef MTS_HANDLE_PARAM_CASES
NAMESPACE_END(mitsuba)
