#pragma once

#include <unordered_map>
#include <vector>
#include <optix.h>

#include <mitsuba/render/fwd.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/optix/common.h>

// Forward declaration
struct EmitterProperties;

NAMESPACE_BEGIN(mitsuba)

static constexpr size_t kOptixGlobalsCount = 22;
static constexpr size_t kOptixStandardVariableCount = 30;
static constexpr size_t kOptixBuffersCount = 1;
static constexpr size_t kOptixRbBuffersCount = 1;

struct MTS_EXPORT_OPTIX BSDFPrograms {
    size_t hash;
    const BSDF *extracted_from;
    std::vector<std::string> ptx;
};

struct MTS_EXPORT_OPTIX MyOptixState {
    MyOptixState(Scene::OptixState *other)
        : state(other) {}

    ~MyOptixState() {}

    Scene::OptixState *state;
    RTvariable global_vars[kOptixGlobalsCount];
    RTbuffer buffer_vars[kOptixBuffersCount];
    RTbuffer rb_buffer_vars[kOptixRbBuffersCount];
    RTbuffer material_params_buffer;
    RTbuffer emitter_params_buffer;

    /// Additional extracted PTX common to all BSDFs, e.g. for envmaps.
    std::string extra_ptx;

    /// For each BSDF, the corresponding OptiX material handle
    std::unordered_map<const void *, RTmaterial> materials;

    /** For each differentiable parameter, a tuple with:
     * - a pointer to the object from which it was extracted (possibly a parent)
     * - the label name given to the corresponding gradient output buffer
     * - a buffer with the right size
     */
    std::unordered_map<std::string, std::tuple<const void *, std::string, FloatC>> gradient_buffers;
    /// Objects (e.g. BSDFs) which have enabled differentiable parameters
    std::set<const void *> enabled_differentiable_objects;

    /// For each BSDF, the hash of the JIT-compiled programs
    std::unordered_map<const void *, size_t> bsdf_program_hashes;
    /// Extracted BSDF programs (sample, eval, pdf, etc), keyed by program hash
    std::unordered_map<size_t, BSDFPrograms> bsdf_programs;

};

class MTS_EXPORT_OPTIX OptiXRenderer : public Object {
public:
    enum EIntegratorType : uint32_t {
        EDepth = 0,
        EDirect,
        EPath,
        EVolPath,
        ERadiativeBackprop,
        ERadiativeBackpropVol,
        ERadiativeBackpropVolSimple,
    };

    constexpr bool supports_volumes(EIntegratorType tp) const {
        switch (tp) {
            case EVolPath:
            case ERadiativeBackpropVol:
            case ERadiativeBackpropVolSimple:
                return true;
            default:
                return false;
        }
    }
    constexpr bool supports_gradients(EIntegratorType tp) const {
        switch (tp) {
            case ERadiativeBackprop:
            case ERadiativeBackpropVol:
            case ERadiativeBackpropVolSimple:
                return true;
            default:
                return false;
        }
    }

public:
    /**
     * \param scene       Scene loaded in RGB mode. Will be kept in a ref<Scene>.
     * \param integrator  Determines which integration algorithm to use.
     * \param params      Holds the differentiable parameters in the scene. Does
     *                    not take ownership. Mandatory in differentiable mode.
     */
    OptiXRenderer(Scene *scene, EIntegratorType integrator = EPath,
                  DifferentiableParameters *params = nullptr);
    virtual ~OptiXRenderer();

    void init();

    /// Creates and register a BSDF for each shape in the scene
    void create_bsdfs_and_emitters_props();

    EmitterProperties create_emitter_props(const Shape &shape, const Emitter &emitter);

    /// JIT-compile relevant BSDFs, Emitters, etc using Enoki. Stores programs in the context.
    void extract_programs();

    /// JIT-compile sample(), eval() and pdf() methods of the given BSDF.
    BSDFPrograms extract_bsdf_programs(BSDF &bsdf, Shape &shape) const;

    /// JIT-compile envmap methods.
    std::string extract_common_programs() const;

    /// Upload material-specific global OptiX variables for all BSDFs in the scene.
    void upload_bsdf_globals() const;

    /**
     * Render the given scene on the GPU with a basic OptiX path tracer.
     */
    ref<Bitmap> render(size_t seed = 1, size_t _spp = (size_t) -1, size_t pass_count = 1);

    /**
     * Make a differentiable render of the scene with radiative
     * backpropagation, using the given image partial derivatives
     * as input.
     *
     * If `return_image` is true, an image will be computed at the same time
     * as gradients are propagated and returned. Otherwise, returns nullptr.
     */
    ref<Bitmap> delta_render(size_t seed, const FloatC &image_derivatives,
                             size_t _spp = (size_t) -1, size_t pass_count = 1,
                             bool return_image = false);
    /// Override of \ref delta_render, provided for convenience
    ref<Bitmap> delta_render(size_t seed, const Bitmap *image_derivatives,
                             size_t _spp = (size_t) -1, size_t pass_count = 1,
                             bool return_image = false);

    float *trace(size_t nx, size_t ny, size_t spp, size_t pass_count, size_t seed,
                 bool return_image = true, const FloatC *image_derivatives = nullptr);

    /**
     * Set gradients accumulated from radiative backprop to the
     * `gradient` field of each parameter.
     * Also resets gradient accumulators back to zero.
     */
    void unpack_gradients(DifferentiableParameters *params) const;

    // ----- Other setters and getters
    bool use_approx_li_1() const { return m_use_approx_li_1; }
    void set_use_approx_li_1(bool v) { m_use_approx_li_1 = v; }

    MTS_DECLARE_CLASS()
private:
    ref<Scene> m_scene;
    MyOptixState *m_optix_state = nullptr;
    EIntegratorType m_integrator_type;
    DifferentiableParameters *m_params = nullptr;  // Not owned
    struct {
        const char *hit;   // PTX source for the ray hit function
        const char *rest;  // PTX source for the rest of the functions
    } m_program_source = { nullptr, nullptr };

    // Pre-allocated image output buffer (must be allocated with `cuda_host_malloc`)
    float *m_image_buffer = nullptr;
    size_t m_image_buffer_size = 0;

    // Optimization flags and other parameters
    bool m_use_approx_li_1 = false;
};


/// Dummy function which can be called to ensure that the liboptix shared library is loaded
void MTS_EXPORT_OPTIX liboptix_nop();

NAMESPACE_END(mitsuba)
