// -----------------------------------------------------------------------
//! @{ \name Enoki support for vectorized function calls
// -----------------------------------------------------------------------

// Enable usage of array pointers for our types
ENOKI_CALL_SUPPORT_BEGIN(mitsuba::Medium)
ENOKI_CALL_SUPPORT_METHOD(sample_distance)
ENOKI_CALL_SUPPORT_METHOD(eval_transmittance)
ENOKI_CALL_SUPPORT_METHOD(phase_function)
ENOKI_CALL_SUPPORT_METHOD(use_emitter_sampling)
ENOKI_CALL_SUPPORT_END(mitsuba::Medium)

// //! @}
// // -----------------------------------------------------------------------

/*
 * \brief This macro should be used in the definition of Medium
 * plugins to instantiate concrete versions of the \c sample_distance and \c
 * eval_transmittance.
 */
#define MTS_IMPLEMENT_MEDIUM_SCALAR()                                                              \
    using Medium::sample_distance;                                                                 \
    using Medium::sample_distance_with_si;                                                         \
    using Medium::eval_transmittance;                                                              \
    using Medium::eval_scattering;                                                                 \
    using Medium::eval_albedo;                                                                     \
    using Medium::pdf_distance;                                                                    \
    using Medium::density;                                                                         \
    std::tuple<SurfaceInteraction3f, MediumInteraction3f, Spectrumf> sample_distance(              \
        const Scene *scene, const Ray3f &ray, const Point2f &sample, Sampler *sampler)             \
        const override {                                                                           \
        ScopedPhase p(EProfilerPhase::EMediumSample);                                              \
        return sample_distance_impl(scene, ray, sample, sampler, 0u, true);                        \
    }                                                                                              \
    std::pair<MediumInteraction3f, Spectrumf> sample_distance_with_si(                             \
        const SurfaceInteraction3f &si, const Ray3f &ray, const Point2f &sample, Sampler *sampler) \
        const override {                                                                           \
        ScopedPhase p(EProfilerPhase::EMediumSample);                                              \
        return sample_distance_with_si_impl(si, ray, sample, sampler, 0u, true);                   \
    }                                                                                              \
                                                                                                   \
    Spectrumf eval_transmittance(const Ray3f &ray, Sampler *sampler) const override {              \
        ScopedPhase p(EProfilerPhase::ETransmittanceEval);                                         \
        return eval_transmittance_impl(ray, sampler, 0u, true);                                    \
    }                                                                                              \
    Spectrumf eval_scattering(const MediumInteraction3f &mi) const override {                      \
        ScopedPhase p(EProfilerPhase::ETransmittanceEval);                                         \
        return eval_scattering_impl(mi, true);                                                     \
    }                                                                                              \
    Spectrumf eval_albedo(const Interaction3f &it) const override {                                \
        ScopedPhase p(EProfilerPhase::ETransmittanceEval);                                         \
        return eval_albedo_impl(it, true);                                                         \
    }                                                                                              \
    Float pdf_distance(const Ray3f &ray, Float t) const override {                                 \
        return pdf_distance_impl(ray, t, true);                                                    \
    }                                                                                              \
    Float density(const Interaction3f &it) const override { return density_impl(it, true); }

#define MTS_IMPLEMENT_MEDIUM_PACKET()                                                              \
    std::tuple<SurfaceInteraction3fP, MediumInteraction3fP, SpectrumfP> sample_distance(           \
        const Scene *scene, const Ray3fP &ray, const Point2fP &sample, Sampler *sampler,           \
        UInt32P index, MaskP active = true) const override {                                       \
        ScopedPhase p(EProfilerPhase::EMediumSampleP);                                             \
        return sample_distance_impl(scene, ray, sample, sampler, index, active);                   \
    }                                                                                              \
    std::pair<MediumInteraction3fP, SpectrumfP> sample_distance_with_si(                           \
        const SurfaceInteraction3fP &si, const Ray3fP &ray, const Point2fP &sample,                \
        Sampler *sampler, UInt32P index, MaskP active = true) const override {                     \
        ScopedPhase p(EProfilerPhase::EMediumSampleP);                                             \
        return sample_distance_with_si_impl(si, ray, sample, sampler, index, active);              \
    }                                                                                              \
                                                                                                   \
    SpectrumfP eval_transmittance(const Ray3fP &ray, Sampler *sampler, UInt32P index,              \
                                  MaskP active = true) const override {                            \
        ScopedPhase p(EProfilerPhase::ETransmittanceEvalP);                                        \
        return eval_transmittance_impl(ray, sampler, index, active);                               \
    }                                                                                              \
    SpectrumfP eval_scattering(const MediumInteraction3fP &mi, MaskP active = true)                \
        const override {                                                                           \
        ScopedPhase p(EProfilerPhase::ETransmittanceEvalP);                                        \
        return eval_scattering_impl(mi, active);                                                   \
    }                                                                                              \
    SpectrumfP eval_albedo(const Interaction3fP &it, MaskP active = true) const override {         \
        ScopedPhase p(EProfilerPhase::ETransmittanceEvalP);                                        \
        return eval_albedo_impl(it, active);                                                       \
    }                                                                                              \
    FloatP pdf_distance(const Ray3fP &ray, FloatP t, MaskP active = true) const override {         \
        return pdf_distance_impl(ray, t, active);                                                  \
    }                                                                                              \
    FloatP density(const Interaction3fP &it, MaskP active = true) const override {                 \
        return density_impl(it, active);                                                           \
    }

#if !defined(MTS_ENABLE_AUTODIFF)
#define MTS_IMPLEMENT_MEDIUM_AUTODIFF()
#else
#define MTS_IMPLEMENT_MEDIUM_AUTODIFF()                                                            \
                                                                                                   \
    std::tuple<SurfaceInteraction3fD, MediumInteraction3fD, SpectrumfD> sample_distance(           \
        const Scene *scene, const Ray3fD &ray, const Point2fD &sample, Sampler *sampler,           \
        UInt32D index, MaskD active) const override {                                              \
        ScopedPhase p(EProfilerPhase::EMediumSampleP);                                             \
        return sample_distance_impl(scene, ray, sample, sampler, index, active);                   \
    }                                                                                              \
    std::pair<MediumInteraction3fD, SpectrumfD> sample_distance_with_si(                           \
        const SurfaceInteraction3fD &si, const Ray3fD &ray, const Point2fD &sample,                \
        Sampler *sampler, UInt32D index, MaskD active) const override {                            \
        ScopedPhase p(EProfilerPhase::EMediumSampleP);                                             \
        return sample_distance_with_si_impl(si, ray, sample, sampler, index, active);              \
    }                                                                                              \
                                                                                                   \
    SpectrumfD eval_transmittance(const Ray3fD &ray, Sampler *sampler, UInt32D index,              \
                                  MaskD active = true) const override {                            \
        ScopedPhase p(EProfilerPhase::ETransmittanceEvalP);                                        \
        return eval_transmittance_impl(ray, sampler, index, active);                               \
    }                                                                                              \
    SpectrumfD eval_scattering(const MediumInteraction3fD &mi, MaskD active = true)                \
        const override {                                                                           \
        ScopedPhase p(EProfilerPhase::ETransmittanceEvalP);                                        \
        return eval_scattering_impl(mi, active);                                                   \
    }                                                                                              \
    SpectrumfD eval_albedo(const Interaction3fD &it, MaskD active = true) const override {         \
        ScopedPhase p(EProfilerPhase::ETransmittanceEvalP);                                        \
        return eval_albedo_impl(it, active);                                                       \
    }                                                                                              \
    FloatD pdf_distance(const Ray3fD &ray, FloatD t, MaskD active = true) const override {         \
        return pdf_distance_impl(ray, t, active);                                                  \
    }                                                                                              \
    FloatD density(const Interaction3fD &it, MaskD active = true) const override {                 \
        return density_impl(it, active);                                                           \
    }

#endif

#define MTS_IMPLEMENT_MEDIUM_ALL()                                                                 \
    using Medium::this_pointer;                                                                    \
    MTS_IMPLEMENT_MEDIUM_SCALAR()                                                                  \
    MTS_IMPLEMENT_MEDIUM_PACKET()                                                                  \
    MTS_IMPLEMENT_MEDIUM_AUTODIFF()
