#pragma once

#include <mitsuba/core/fwd.h>
#include <mitsuba/core/transform.h>
#include <mitsuba/core/bsphere.h>
#include <mitsuba/render/common.h>
#include <mitsuba/render/endpoint.h>
#include <mitsuba/render/fwd.h>

NAMESPACE_BEGIN(mitsuba)

class MTS_EXPORT_RENDER Emitter : public Endpoint {
public:
    /// Is this an environment map light emitter?
    virtual bool is_environment() const;

    MTS_DECLARE_CLASS()

protected:
    Emitter(const Properties &props);

    virtual ~Emitter();
};


/**
 * Base class for emitters that are located infinitely far away, such as
 * environment emitters.
 */
class MTS_EXPORT_RENDER InfiniteEmitter : public Emitter {
public:
    bool is_environment() const override;

    /// Creates a bounding sphere around the scene and updates \ref m_bsphere.
    ref<Shape> create_shape(const Scene *scene) override;

    /**
     * A pointer to `this` which is automatically registered as a variable
     * when in CUDA mode, so that it is not baked as a constant in the kernel.
     */
    template <typename Value>
    auto this_pointer() const {
        if constexpr (is_cuda_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
            return EmitterPtrC::from_index_(cuda_var_register_ptr(this));
#endif
        } else {
            return this;
        }
    }

    MTS_DECLARE_CLASS()
protected:
    InfiniteEmitter(const Properties &props);

    BoundingSphere3f m_bsphere;
    /// Inverse surface area of the bounding sphere
    Float m_inv_surface_area;

#if defined(MTS_ENABLE_AUTODIFF)
    using EmitterPtrC = replace_scalar_t<FloatC, Emitter *>;
    const EmitterPtrC m_this_pointer;
#endif
};


NAMESPACE_END(mitsuba)

#include "detail/emitter.inl"
