#pragma once

#include <mitsuba/core/fwd.h>
#include <mitsuba/core/object.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/spectrum.h>
#include <mitsuba/core/timer.h>
#include <mitsuba/core/tls.h>
#include <mitsuba/core/vector.h>
#include <mitsuba/render/fwd.h>
#include <mitsuba/render/imageblock.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/records.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/shape.h>

NAMESPACE_BEGIN(mitsuba)

/**
 * \brief Abstract integrator base-class; does not make any assumptions on
 * how radiance is computed.
 *
 * In Mitsuba, the different rendering techniques are collectively referred to
 * as \a integrators, since they perform integration over a high-dimensional
 * space. Each integrator represents a specific approach for solving the light
 * transport equation---usually favored in certain scenarios, but at the same
 * time affected by its own set of intrinsic limitations. Therefore, it is
 * important to carefully select an integrator based on user-specified accuracy
 * requirements and properties of the scene to be rendered.
 *
 * This is the base class of all integrators; it does not make any assumptions on
 * how radiance is computed, which allows for many different kinds of implementations
 * ranging from software-based path tracing and Markov-Chain based techniques such
 * as Metropolis Light Transport up to hardware-accelerated rasterization.
 */
class MTS_EXPORT_RENDER Integrator : public Object {
public:
    using BSDFPtrP = replace_scalar_t<FloatP, const BSDF *>;
    using BSDFPtrD = replace_scalar_t<FloatD, const BSDF *>;

public:
    /// Perform the main rendering job
    virtual bool render(Scene *scene, bool vectorize) = 0;

    /**
     * \brief Cancel a running render job
     *
     * This function can be called asynchronously to cancel a running render
     * job. In this case, \ref render() will quit with a return value of
     * \c false.
     */
    virtual void cancel();

    /**
     * Indicates whether \ref cancel() or a timeout has occured. Should be
     * checked often in each integrator's main loop so that timeout is
     * enforced accurately.
     *
     * Note that render time measurements rely on \ref m_render_timer, which
     * needs to be reset at the beginning of the rendering phase.
     */
    bool should_stop() const {
        return m_stop || (m_timeout > 0.0f &&
                          m_render_timer.value() > 1000.0f * m_timeout);
    }

    // =========================================================================
    //! @{ \name Progress reporting / callbacks
    // =========================================================================
    /**
     * Signature of callbacks that may be registered. Called with:
     * - Thread ID
     * - Pointer to this thread's bitmap, used to accumulate the results
     *   locally. This should typically point to the same object across calls.
     * - Extra information (e.g. an enumeration value indicating a type of
     *   event).
     */
    using CallbackFunction = std::function<void(size_t, Bitmap*, Float)>;

    /**
     * Adds a callback function, which will be called with the desired
     * frequency.
     * Not thread-safe.
     *
     * @param cb     Callback function to be called.
     * @param period Minimum time elapsed (in seconds) between two calls. May
     *               be zero, in which case the callback is called at each
     *               iteration of the algorithm.
     * @return The index of this callback (used to de-register this callback).
     */
    size_t register_callback(CallbackFunction cb, Float period);

    /**
     * De-registers callback at the specified index.
     * If (size_t) -1 is passed, all callbacks are cleared.
     */
    void remove_callback(size_t cb_index = (size_t) -1);

    size_t callback_count() const {
        return m_callbacks.size();
    }

    /**
     * Maybe trigger a call to the callbacks. This will be rate-limited so that
     * the period specified by the caller is respected *per-thread* (i.e. each
     * thread will count time independently).
     *
     * Callbacks are called in the order the were registered.
     */
    virtual void notify(Bitmap *bitmap, Float extra = 1.0f);

    //! @}
    // =========================================================================


    MTS_DECLARE_CLASS()
protected:
    /// Create an integrator
    Integrator(const Properties &props);

    /// Virtual destructor
    virtual ~Integrator() { }

    /// Integrators should stop computing when this flag is set to true.
    bool m_stop;
    /** Maximum amount of time to spend rendering (excluding scene parsing).
     * Specified in seconds. A negative values indicates no timeout. */
    Float m_timeout;
    /// Timer used to enforce the timeout.
    Timer m_render_timer;

protected:
    struct CallbackInfo {
        CallbackFunction f;
        Float period;
        ThreadLocal<Float> last_called;

        CallbackInfo(const CallbackFunction &f, Float period)
            : f(f), period(period) {}
    };

    /// List of registered callback functions.
    std::vector<CallbackInfo> m_callbacks;
    /// Timer used to enforce the callback rate limits.
    Timer m_cb_timer;
    /// Color handling mode.
    EColorMode m_color_mode = ESpectral;
};

/** \brief Abstract base class, which describes integrators
 * capable of computing samples of the scene's radiance function.
 */
class MTS_EXPORT_RENDER SamplingIntegrator : public Integrator {
public:
    /**
     * \brief Sample the incident radiance along a ray. The record passed is
     * used to store additional information about the result.
     */
    virtual Spectrumf eval(const RayDifferential3f &ray,
                           RadianceSample3f &rs) const;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref eval()
    Spectrumf eval(const RayDifferential3f &ray, RadianceSample3f &rs,
                   bool /* unused */) const {
        return eval(ray, rs);
    }

    /// Vectorized variant of \ref eval.
    virtual SpectrumfP eval(const RayDifferential3fP &ray,
                            RadianceSample3fP &rs,
                            MaskP active = true) const;

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable variant of \ref eval.
    virtual SpectrumfD eval(const RayDifferential3fD &ray,
                            RadianceSample3fD &rs,
                            MaskD active = true) const;
#endif

    // =========================================================================
    //! @{ \name Integrator interface implementation
    // =========================================================================

    bool render(Scene *scene, bool vectorize) override;

    //! @}
    // =========================================================================

    MTS_DECLARE_CLASS()
protected:
    SamplingIntegrator(const Properties &props);
    virtual ~SamplingIntegrator();

    virtual void render_block_scalar(const Scene *scene, Sampler *sampler,
                                     ImageBlock *block,
                                     size_t sample_count = size_t(-1)) const;

    virtual void render_block_vector(const Scene *scene, Sampler *sampler,
                                     ImageBlock *block, Point2fX &points,
                                     size_t sample_count = size_t(-1)) const;

protected:
    /// Size of (square) image blocks to render per core.
    size_t m_block_size;
    /** Number of samples to compute for each pass over the image blocks.
     * Must be a multiple of the total sample count per pixel.
     * If set to (size_t) -1, all the work is done in a single pass (default). */
    size_t m_samples_per_pass;
};

/*
 * \brief Base class of all recursive Monte Carlo integrators, which compute
 * unbiased solutions to the rendering equation (and optionally the radiative
 * transfer equation).
 */
class MTS_EXPORT_RENDER MonteCarloIntegrator : public SamplingIntegrator {
public:
    int max_depth() const { return m_max_depth; }
    int rr_depth() const { return m_rr_depth; }
    bool hide_emitters() const { return m_hide_emitters; }
    void set_hide_emitters(bool hide) { m_hide_emitters = hide; }

    MTS_DECLARE_CLASS()
protected:
    /// Create an integrator
    MonteCarloIntegrator(const Properties &props);

protected:
    int m_max_depth;
    int m_rr_depth;
    bool m_hide_emitters;
};

/**
 * Base class for integrators that start paths from the light source, as opposed
 * to sensor-based integrators.
 */
class MTS_EXPORT_RENDER LightTracerIntegrator : public Integrator {
public:
    explicit LightTracerIntegrator(const Properties &props);

    /// Perform the main rendering job
    bool render(Scene *scene, bool vectorize) override;

    /**
     * Samples an emitter in the scene and connects it directly to the sensor,
     * writing the emitted radiance to the given image block.
     */
    virtual Spectrumf sample_visible_emitters(RadianceSample3f &rs,
                                              ImageBlock *block) const;
    /// Vectorized variant of \ref sample_visible_emitters.
    virtual SpectrumfP sample_visible_emitters(RadianceSample3fP &rs,
                                               ImageBlock *block,
                                               MaskP active = true) const;
    /// Adapter for \ref sample_visible_emitters which discards the boolean.
    Spectrumf sample_visible_emitters(RadianceSample3f &rs, ImageBlock *block,
                                      bool /*unused*/) const {
        return sample_visible_emitters(rs, block);
    }
#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable variant of \ref sample_visible_emitters.
    virtual SpectrumfD
    sample_visible_emitters(RadianceSample3fD &rs, ImageBlock *block,
                            const MaskD &active = true) const;
#endif

    /**
     * Samples a ray from a random emitter in the scene.
     */
    virtual std::pair<Ray3f, Spectrumf>
    prepare_ray(RadianceSample3f &rs) const = 0;
    /// Vectorized variant of \ref prepare_ray.
    virtual std::pair<Ray3fP, SpectrumfP>
    prepare_ray(RadianceSample3fP &rs, const MaskP &active = true) const = 0;
    /// Adapter for \ref prepare_ray which discards the boolean.
    std::pair<Ray3f, Spectrumf> prepare_ray(RadianceSample3f &rs,
                                            bool /*unused*/) const {
        return prepare_ray(rs);
    }
#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable variant of \ref prepare_ray.
    virtual std::pair<Ray3fD, SpectrumfD>
    prepare_ray(RadianceSample3fD &rs, const MaskD &active) const = 0;
#endif

    /**
     * Intersects the given ray with the scene and recursively trace using
     * BSDF sampling. The given `throughput` should account for emitted
     * radiance from the sampled light source, wavelengths sampling weights,
     * etc. At each interaction, we attempt connecting to the sensor and add
     * throughput to the given `block`.
     *
     * Note: this will *not* account for directly visible emitters, since
     * they require a direct connection from the emitter to the sensor. See
     * \ref sample_visible_emitters.
     */
    virtual Spectrumf trace_light_ray(Ray3f ray, RadianceSample3f &rs,
                                      Spectrumf throughput,
                                      ImageBlock *block) const = 0;
    /// Vectorized variant of \ref trace_light_ray.
    virtual SpectrumfP trace_light_ray(Ray3fP ray, RadianceSample3fP &rs,
                                       SpectrumfP throughput, ImageBlock *block,
                                       MaskP active = true) const = 0;
    /// Adapter for \ref trace_light_ray, discarding the boolean parameter.
    Spectrumf trace_light_ray(Ray3f ray, RadianceSample3f &rs,
                              Spectrumf throughput, ImageBlock *block,
                              bool /*active*/) const {
        return trace_light_ray(ray, rs, throughput, block);
    }
#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable variant of \ref trace_light_ray.
    virtual SpectrumfD trace_light_ray(Ray3fD ray, RadianceSample3fD &rs,
                                       SpectrumfD throughput, ImageBlock *block,
                                       const MaskD &active = true) const = 0;
#endif

    /**
     * From the given surface interaction, attempt connecting to the sensor
     */
    virtual Spectrumf connect_sensor(RadianceSample3f &rs,
                                     SurfaceInteraction3f &si, const BSDF *bsdf,
                                     const Spectrumf &weight,
                                     ImageBlock *block) const = 0;
    /// Vectorized variant of \ref connect_sensor.
    virtual SpectrumfP connect_sensor(RadianceSample3fP &rs,
                                      SurfaceInteraction3fP &si, BSDFPtrP bsdf,
                                      const SpectrumfP &weight,
                                      ImageBlock *block,
                                      const MaskP &active = true) const = 0;
    /// Adapter for \ref connect_sensor, discarding the boolean parameter.
    Spectrumf connect_sensor(RadianceSample3f &rs, SurfaceInteraction3f &si,
                             const BSDF *bsdf, const Spectrumf &weight,
                             ImageBlock *block, bool /*active*/) const {
        return connect_sensor(rs, si, bsdf, weight, block);
    }
#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable variant of \ref connect_sensor.
    virtual SpectrumfD connect_sensor(RadianceSample3fD &rs,
                                      SurfaceInteraction3fD &si, BSDFPtrD bsdf,
                                      const SpectrumfD &weight,
                                      ImageBlock *block,
                                      const MaskD &active = true) const = 0;
#endif

    /**
     * Overwrites the accumulated per-pixel normalization weights with a
     * constant weight computed from the number of overall accumulated samples.
     *
     * \return The new normalization weight.
     */
    Float normalize_block(ImageBlock *block, size_t total_samples) const {
        Float new_weight = total_samples / Float(hprod(block->size()));
        block->reweight(new_weight);
        return new_weight;
    }
    Float normalize_film(Film *film, size_t total_samples) const {
        Float new_weight = total_samples / Float(hprod(film->size()));
        film->bitmap()->overwrite(4, new_weight);
        return new_weight;
    }

    MTS_DECLARE_CLASS()

protected:
    int m_max_depth;
    int m_rr_depth;

protected:
    template <typename Value> bool render_impl(Scene *scene);

    template <typename RadianceSample,
              typename Value = typename RadianceSample::Value>
    auto sample_visible_emitters_impl(RadianceSample &rs, ImageBlock *block,
                                      const mask_t<Value> &active) const;
};

NAMESPACE_END(mitsuba)

#include "detail/integrator.inl"
