#pragma once

#include <enoki/stl.h>

#include <mitsuba/core/object.h>
#include <mitsuba/core/spectrum.h>
#include <mitsuba/render/autodiff.h>

#include <mitsuba/render/fwd.h>
#include <mitsuba/render/interaction.h>

NAMESPACE_BEGIN(mitsuba)

class MTS_EXPORT_RENDER Medium : public DifferentiableObject {
public:
    // =============================================================
    //! @{ \name Medium sampling strategy
    // =============================================================

    /**
     * \brief Sample a free-flight distance in the medium.
     *
     * Should ideally importance sample with respect to the transmittance.
     *
     * \param scene    A pointer to the current scene
     * \param ray      Ray, along which a distance should be sampled
     * \param sampler  Sampler to produce random numbers. Note that some media
     * (e.g. heterogenous media) may require an arbitrary number of random
     * numbers to sample a distance.
     *
     * \return         This method returns both a medium and a surface
     *                 interaction. If a medium interaction was sampled, a valid
     *                 MediumInteraction is returned. If the sampled distance
     * results in an intersection with a surface, the corresponding
     * SurfaceInteraction is returned. Further, a spectrum with the sampled
     * throughput value is returned.
     */
    virtual std::tuple<SurfaceInteraction3f, MediumInteraction3f, Spectrumf>
    sample_distance(const Scene *scene, const Ray3f &ray, const Point2f &sample,
                    Sampler *sampler) const = 0;

    /// Vectorized version of \ref sample_distance()
    virtual std::tuple<SurfaceInteraction3fP, MediumInteraction3fP, SpectrumfP>
    sample_distance(const Scene *scene, const Ray3fP &ray, const Point2fP &sample, Sampler *sampler,
                    UInt32P index, MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref
    /// sample_distance()
    std::tuple<SurfaceInteraction3f, MediumInteraction3f, Spectrumf>
    sample_distance(const Scene *scene, const Ray3f &ray, const Point2f &sample, Sampler *sampler,
                    int /* unused */, bool /* unused */) const {
        return sample_distance(scene, ray, sample, sampler);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref sample_distance()
    virtual std::tuple<SurfaceInteraction3fD, MediumInteraction3fD, SpectrumfD>
    sample_distance(const Scene *scene, const Ray3fD &ray, const Point2fD &sample, Sampler *sampler,
                    UInt32D index, MaskD active = true) const = 0;
#endif

    /**
     * Helper to \ref sample_distance.
     * Split off in order to have most of the logic available in one function for PTX extraction.
     */
    virtual std::pair<MediumInteraction3f, Spectrumf>
    sample_distance_with_si(const SurfaceInteraction3f &si, const Ray3f &ray, const Point2f &sample,
                            Sampler *sampler) const = 0;

    /// Vectorized version of \ref sample_distance_with_si()
    virtual std::pair<MediumInteraction3fP, SpectrumfP>
    sample_distance_with_si(const SurfaceInteraction3fP &si, const Ray3fP &ray,
                            const Point2fP &sample, Sampler *sampler, UInt32P index,
                            MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref sample_distance_with_si()
    std::pair<MediumInteraction3f, Spectrumf>
    sample_distance_with_si(const SurfaceInteraction3f &si, const Ray3f &ray, const Point2f &sample, Sampler *sampler,
                    uint32_t /* unused */, bool /* unused */) const {
        return sample_distance_with_si(si, ray, sample, sampler);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref sample_distance_with_si()
    virtual std::pair<MediumInteraction3fD, SpectrumfD>
    sample_distance_with_si(const SurfaceInteraction3fD &si, const Ray3fD &ray,
                            const Point2fD &sample, Sampler *sampler, UInt32D index,
                            MaskD active = true) const = 0;
#endif

    //! @}
    // =============================================================

    // =============================================================
    //! @{ \name Functions for querying the medium
    // =============================================================

    /**
     * \brief Compute the transmittance along a ray segment
     *
     * Computes the transmittance along a ray segment
     * [mint, maxt] associated with the ray. It is assumed
     * that the ray has a normalized direction value.
     */
    virtual Spectrumf eval_transmittance(const Ray3f &ray,
                                         Sampler *sampler) const = 0;

    /// Vectorized version of \ref eval_transmittance()
    virtual SpectrumfP eval_transmittance(const Ray3fP &ray, Sampler *sampler,
                                          UInt32P index, MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref eval_transmittance()
    Spectrumf eval_transmittance(const Ray3f &ray, Sampler *sampler,
                                 int /*unused*/, bool /*unused*/) const {
        return eval_transmittance(ray, sampler);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref eval_transmittance()
    virtual SpectrumfD eval_transmittance(const Ray3fD &ray, Sampler *sampler,
                                          UInt32D index, MaskD active = true) const = 0;
#endif


    /**
     * \brief Looks up the scattering coefficient at a given location.
     *     scattering = albedo * transmittance
     */
    virtual Spectrumf eval_scattering(const MediumInteraction3f &mi) const = 0;

    /// Vectorized version of \ref eval_scattering()
    virtual SpectrumfP eval_scattering(const MediumInteraction3fP &mi,
                                       MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref eval_scattering()
    Spectrumf eval_scattering(const MediumInteraction3f &mi, bool /*unused*/) const {
        return eval_scattering(mi);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref eval_scattering()
    virtual SpectrumfD eval_scattering(const MediumInteraction3fD &mi,
                                       MaskD active = true) const = 0;
#endif

    /// Looks up the albedo at a given location.
    virtual Spectrumf eval_albedo(const Interaction3f &it) const = 0;

    /// Vectorized version of \ref eval_albedo()
    virtual SpectrumfP eval_albedo(const Interaction3fP &it, MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref eval_albedo()
    Spectrumf eval_albedo(const Interaction3f &it, bool /*unused*/) const {
        return eval_albedo(it);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref eval_albedo()
    virtual SpectrumfD eval_albedo(const Interaction3fD &it, MaskD active = true) const = 0;
#endif


    /**
     * \brief Compute the PDF of sampling a given distance along a ray.
     * If t = +inf, then the ray's (mint, maxt) are used as the range to
     * compute the transmittance over.
     * If t < inf, corresponding to sampling a medium interaction,
     * then the range (0, t) is integrated along the ray.
     */
    virtual Float pdf_distance(const Ray3f &ray, Float t) const = 0;

    /// Vectorized version of \ref pdf_distance()
    virtual FloatP pdf_distance(const Ray3fP &ray, FloatP t,
                                    MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref pdf_distance()
    Float pdf_distance(const Ray3f &ray, Float t, bool /*unused*/) const {
        return pdf_distance(ray, t);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref pdf_distance()
    virtual FloatD pdf_distance(const Ray3fD &ray, FloatD t, MaskD active = true) const = 0;
#endif


    /// Looks up the medium's scalar density at the given point.
    virtual Float density(const Interaction3f &it) const = 0;

    /// Vectorized version of \ref density()
    virtual FloatP density(const Interaction3fP &it, MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref density()
    Float density(const Interaction3f &it, bool /*unused*/) const {
        return density(it);
    }

#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref density()
    virtual FloatD density(const Interaction3fD &it, MaskD active = true) const = 0;
#endif


    //! @}
    // =============================================================

    // =============================================================
    //! @{ \name Accessors and helpers
    // =============================================================

    /// Return the phase function of this medium
    MTS_INLINE const PhaseFunction *phase_function() const {
        return m_phase_function.get();
    }

    /// Determine whether the medium is homogeneous
    virtual bool is_homogeneous() const = 0;

    /// Returns whether this specific medium instance uses emitter sampling
    MTS_INLINE bool use_emitter_sampling() const { return m_sample_emitters; }

    /// Return a human-readable representation of the Medium
    std::string to_string() const override = 0;

    /**
     * Extracts the following medium parameters from the given properties:
     * - sigma_a (absoption)
     * - sigma_s (scattering)
     * - sigma_t (transmission)
     * - albedo
     */
    static std::tuple<ref<ContinuousSpectrum>, ref<ContinuousSpectrum>,
                      ref<ContinuousSpectrum>, ref<ContinuousSpectrum>>
    extract_medium_parameters(const Properties &props);

    std::string id() const override;

    /**
     * A pointer to `this` which is automatically registered as a variable
     * when in CUDA mode, so that it is not baked as a constant in the kernel.
     */
    template <typename Value>
    auto this_pointer() const {
        if constexpr (is_cuda_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
            using ValueC =
                std::conditional_t<is_diff_array_v<Value>, typename Value::UnderlyingType, Value>;
            using MediumPtr = replace_scalar_t<ValueC, Medium *>;
            return MediumPtr::from_index_(cuda_var_register_ptr(this));
#endif
        } else {
            return this;
        }
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override;

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override;
#endif

    ENOKI_CALL_SUPPORT_FRIEND()
    MTS_DECLARE_CLASS()
protected:
    Medium();
    Medium(const Properties &props);
    virtual ~Medium();


protected:
    ref<PhaseFunction> m_phase_function;
    bool m_sample_emitters;
    std::string m_id;

#if defined(MTS_ENABLE_AUTODIFF)
    using MediumPtrC = replace_scalar_t<FloatC, Medium *>;
    const MediumPtrC m_this_pointer;
#endif
};

NAMESPACE_END(mitsuba)

#include "detail/medium.inl"
