#pragma once

#include <mitsuba/core/profiler.h>
#include <mitsuba/render/common.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/mueller.h>

NAMESPACE_BEGIN(mitsuba)

/**
 * \brief Abstract phase function base-class.
 *
 * This class provides an abstract interface to all Phase function plugins in
 * Mitsuba. It exposes functions for evaluating and sampling the model.
 */
class MTS_EXPORT_RENDER PhaseFunction : public Object {
public:
    /**
     * \brief Importance sample the phase function model
     *
     * The function returns a sampled direction.
     *
     *
     * \param mi
     *     A medium interaction data structure describing the underlying
     *     medium position. The incident direction is obtained from
     *     the field <tt>mi.wi</tt>.
     *
     * \param sample1
     *     A uniformly distributed sample on \f$[0,1]^2\f$. It is
     *     used to generate the sampled direction.
     *
     * \return A sampled direction wo
     */
    virtual Vector3f sample(const MediumInteraction3f &mi,
                            const Point2f &sample1) const = 0;

    /// Vectorized version of \ref sample()
    virtual Vector3fP sample(const MediumInteraction3fP &mi,
                             const Point2fP &sample1,
                             MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref sample()
    Vector3f sample(const MediumInteraction3f &mi, const Point2f &sample1,
                    bool /*unused*/) const {
        return sample(mi, sample1);
    }


#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref sample()
    virtual Vector3fD sample(const MediumInteraction3fD &mi,
                             const Point2fD &sample1,
                             MaskD active = true) const = 0;
#endif


    /**
     * \brief Evaluates the phase function model
     *
     * The function returns the value (which equals the PDF) of the phase
     * function in the query direction.
     *
     * \param wo
     *      An outgoing direction to evaluate.
     *
     * \return The value of the phase function in direction wo
     */
    virtual Float eval(const Vector3f &wo) const = 0;

    /// Vectorized version of \ref eval()
    virtual FloatP eval(const Vector3fP &wo, MaskP active = true) const = 0;

    /// Compatibility wrapper, which strips the mask argument and invokes \ref eval()
    Float eval(const Vector3f &wo, bool /* unused */) const { return eval(wo); }


#if defined(MTS_ENABLE_AUTODIFF)
    /// Differentiable version of \ref eval()
    virtual FloatD eval(const Vector3fD &wo, MaskD active = true) const = 0;
#endif


    /// Return a human-readable representation of the BSDF
    std::string to_string() const override = 0;

    //! @}
    // -----------------------------------------------------------------------

    ENOKI_CALL_SUPPORT_FRIEND()
    MTS_DECLARE_CLASS()

protected:
    PhaseFunction();
    virtual ~PhaseFunction();
};

//! @}
// -----------------------------------------------------------------------
NAMESPACE_END(mitsuba)

#include "detail/phase.inl"
