#include <mitsuba/core/string.h>
#include <mitsuba/core/fwd.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/ior.h>
#include <mitsuba/render/microfacet.h>
#include <mitsuba/render/spectrum.h>

NAMESPACE_BEGIN(mitsuba)

class RoughConductor final : public BSDF {
public:
    RoughConductor(const Properties &props) : BSDF(props) {
        m_eta = props.spectrum("eta", 0.f);
        m_k   = props.spectrum("k", 1.f);

        MicrofacetDistribution<Float> distr(props);
        m_type = distr.type();
        m_sample_visible = distr.sample_visible();

        m_flags = EGlossyReflection | EFrontSide;

        // m_alpha_u = m_alpha_v = distr.alpha_u();
        // if (distr.is_anisotropic()) {
        //     m_alpha_v = distr.alpha_v();
        //     m_flags |= EAnisotropic;
        // }
        // TODO: less hacky support
        m_alpha = props.spectrum("alpha_texture", 0.1f);

        if (props.has_property("alpha") || props.has_property("alpha_u") || props.has_property("alpha_v")) {
            Throw("`alpha_texture` specified, values of alpha / alpha_u / alpha_v will be ignored");
        }

        m_components.push_back(m_flags);

        if (m_alpha->id().empty())
            m_alpha->set_id(this->id() + "_alpha");
        if (m_eta->id().empty())
            m_eta->set_id(this->id() + "_eta");
        if (m_k->id().empty())
            m_k->set_id(this->id() + "_k");
    }

    template <typename SurfaceInteraction, typename Value, typename Point2,
              typename BSDFSample = BSDFSample<typename SurfaceInteraction::Point3>,
              typename Spectrum   = Spectrum<Value>>
    MTS_INLINE
    std::pair<BSDFSample, Spectrum> sample_impl(const BSDFContext &ctx,
                                                const SurfaceInteraction &si,
                                                Value /* sample1 */,
                                                const Point2 & sample2,
                                                mask_t<Value> active) const {
        using Vector3 = typename BSDFSample::Vector3;
        using Normal3 = typename BSDFSample::Normal3;
        using Frame = mitsuba::Frame<Vector3>;

        BSDFSample bs = zero<BSDFSample>();;
        Value cos_theta_i = Frame::cos_theta(si.wi);
        active &= cos_theta_i > 0.f;

        if (unlikely(!ctx.is_enabled(EGlossyReflection) || none_or<false>(active)))
            return { bs, 0.f };

        /* Construct a microfacet distribution matching the
           roughness values at the current surface position. */
        Value alpha = m_alpha->eval(si, active).x();
        MicrofacetDistribution<Value> distr(m_type, alpha,
                                            m_sample_visible);

        /* Sample M, the microfacet normal */
        Normal3 m;
        std::tie(m, bs.pdf) = distr.sample(si.wi, sample2);

        /* Perfect specular reflection based on the microfacet normal */
        bs.wo = reflect(si.wi, m);
        bs.eta = 1.f;
        bs.sampled_component = 0;
        bs.sampled_type = EGlossyReflection;

        /* Ensure that this is a valid sample */
        active &= neq(bs.pdf, 0.f) && Frame::cos_theta(bs.wo) > 0.f;

        Value weight;
        if (likely(m_sample_visible))
            weight = distr.smith_g1(bs.wo, m);
        else
            weight = distr.G(si.wi, bs.wo, m) * dot(si.wi, m) /
                     (cos_theta_i * Frame::cos_theta(m));

        /* Evaluate the Fresnel factor */
        Complex<Spectrum> eta_c(m_eta->eval(si, active),
                                m_k->eval(si, active));

        Spectrum F = fresnel_conductor(Spectrum(dot(si.wi, m)), eta_c);

        /* Jacobian of the half-direction mapping */
        bs.pdf /= 4.f * dot(bs.wo, m);

        return { bs, (F * weight) & active };
    }

    template <typename SurfaceInteraction, typename Vector3,
              typename Value    = typename SurfaceInteraction::Value,
              typename Spectrum = Spectrum<Value>>
    MTS_INLINE
    Spectrum eval_impl(const BSDFContext &ctx, const SurfaceInteraction &si,
                       const Vector3 &wo, mask_t<Value> active) const {
        using Frame = mitsuba::Frame<Vector3>;

        Value cos_theta_i = Frame::cos_theta(si.wi),
              cos_theta_o = Frame::cos_theta(wo);

        active &= cos_theta_i > 0.f && cos_theta_o > 0.f;

        if (unlikely(!ctx.is_enabled(EGlossyReflection) || none_or<false>(active)))
            return 0.f;

        /* Calculate the half-direction vector */
        Vector3 H = normalize(wo + si.wi);

        /* Construct a microfacet distribution matching the
           roughness values at the current surface position. */
        Value alpha = m_alpha->eval(si, active).x();
        MicrofacetDistribution<Value> distr(m_type, alpha,
                                            m_sample_visible);

        /* Evaluate the microfacet normal distribution */
        Value D = distr.eval(H);

        active &= neq(D, 0.f);

        /* Evaluate the Fresnel factor */
        Complex<Spectrum> eta_c(m_eta->eval(si, active),
                                m_k->eval(si, active));
        Spectrum F = fresnel_conductor(Spectrum(dot(si.wi, H)), eta_c);

        /* Evaluate Smith's shadow-masking function */
        Value G = distr.G(si.wi, wo, H);

        /* Evaluate the full microfacet model (except Fresnel) */
        Value result = D * G / (4.f * Frame::cos_theta(si.wi));

        return (F * result) & active;
    }

    template <typename SurfaceInteraction, typename Vector3,
              typename Value = value_t<Vector3>>
    MTS_INLINE
    Value pdf_impl(const BSDFContext &ctx, const SurfaceInteraction &si,
                   const Vector3 &wo, mask_t<Value> active) const {
        using Frame = mitsuba::Frame<Vector3>;

        Value cos_theta_i = Frame::cos_theta(si.wi),
              cos_theta_o = Frame::cos_theta(wo);

        active &= cos_theta_i > 0.f && cos_theta_o > 0.f;

        if (unlikely(!ctx.is_enabled(EGlossyReflection) || none_or<false>(active)))
            return 0.f;

        /* Calculate the half-direction vector */
        Vector3 H = normalize(wo + si.wi);

        /* Construct a microfacet distribution matching the
           roughness values at the current surface position. */
        Value alpha = m_alpha->eval(si, active).x();
        MicrofacetDistribution<Value> distr(m_type, alpha,
                                            m_sample_visible);

        Value result;
        if (likely(m_sample_visible))
            result = distr.eval(H) * distr.smith_g1(si.wi, H) /
                     (4.f * cos_theta_i);
        else
            result = distr.pdf(si.wi, H) / (4.f * dot(wo, H));

        return select(active, result, 0.f);
    }


    std::string to_string() const override {
        std::ostringstream oss;
        oss << "RoughConductor[" << std::endl
            << "  distribution = ";
        if (m_type == EBeckmann)
            oss << "beckmann";
        else
            oss << "ggx";
        oss << "," << std::endl
            << "  sample_visible = " << m_sample_visible << "," << std::endl
            << "  alpha = " << string::indent(m_alpha) << "," << std::endl
            // << "  alpha_u = " << m_alpha_u << "," << std::endl
            // << "  alpha_v = " << m_alpha_v << "," << std::endl
            << "  eta = " << string::indent(m_eta) << "," << std::endl
            << "  k = " << string::indent(m_k) << std::endl
            << "]";
        return oss.str();
    }


    std::vector<ref<Object>> children() override {
        return { m_alpha.get(), m_eta.get(), m_k.get() };
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        m_alpha->set_labels(prefix + ".m_alpha");
        m_eta->set_labels(prefix + ".m_eta");
        m_k->set_labels(prefix + ".m_k");
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        m_alpha->set_globals(prefix + "_m_alpha", context, material);
        m_eta->set_globals(prefix + "_m_eta", context, material);
        m_k->set_globals(prefix + "_m_k", context, material);
    }
#endif

    MTS_IMPLEMENT_BSDF_ALL()
    MTS_DECLARE_CLASS()

private:
    /// Specifies the type of microfacet distribution
    EType m_type;

    /// TODO: less hacky support for textured alpha
    ref<ContinuousSpectrum> m_alpha;
    // /// Anisotropic roughness values (ignored if m_alpha is specified)
    // Float m_alpha_u, m_alpha_v;

    /// Importance sample the distribution of visible normals?
    bool m_sample_visible;

    /// Relative refractive index (real component)
    ref<ContinuousSpectrum> m_eta;

    /// Relative refractive index (imaginary component).
    ref<ContinuousSpectrum> m_k;
};

MTS_IMPLEMENT_CLASS(RoughConductor, BSDF)
MTS_EXPORT_PLUGIN(RoughConductor, "Rough conductor");

NAMESPACE_END(mitsuba)
