#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/bsphere.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/emitter.h>
#include <mitsuba/render/kdtree.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/core/plugin.h>

NAMESPACE_BEGIN(mitsuba)

class ConstantBackgroundEmitter final : public InfiniteEmitter {
public:
    ConstantBackgroundEmitter(const Properties &props)
        : InfiniteEmitter(props) {
        m_radiance = props.spectrum("radiance", ContinuousSpectrum::D65(1.f, m_color_mode));
    }

    template <typename SurfaceInteraction, typename Mask,
              typename Spectrum = typename SurfaceInteraction::Spectrum>
    MTS_INLINE Spectrum eval_impl(const SurfaceInteraction &si,
                                  Mask active) const {
        return m_radiance->eval(si.wavelengths, active);
    }

    template <typename Point2, typename Value = value_t<Point2>>
    MTS_INLINE auto
    sample_ray_impl(Value time, const Spectrum<Value> &wavelengths,
                    const Point2 &sample1, const Point2 &sample2,
                    const mask_t<Value> & /* active */) const {
        using Point3  = Point<Value, 3>;
        using Ray3    = Ray<Point3>;
        using Vector3 = vector3_t<Point3>;
        using Frame   = Frame<Vector<Value, 3>>;
        using Ray3    = Ray<Point3>;

        // 1. Sample spatial component
        Vector3 v0 = warp::square_to_uniform_sphere(sample1);

        // 2. Sample directional component
        Vector3 v1 = warp::square_to_cosine_hemisphere(sample2);

        Float r2 = m_bsphere.radius * m_bsphere.radius;
        return std::make_pair(Ray3(m_bsphere.center + v0 * m_bsphere.radius,
                                   Frame(-v0).to_world(v1), time, wavelengths),
                              (4 * math::Pi * math::Pi * r2));
    }

    template <typename Interaction,
              typename Point2 = typename Interaction::Point2,
              typename Value  = typename Interaction::Value>
    MTS_INLINE auto sample_direction_impl(const Interaction &it,
                                          const Point2 &sample,
                                          mask_t<Value> active) const {
        using Point3   = typename Interaction::Point3;
        using Vector3  = vector3_t<Point3>;
        using DirectionSample = DirectionSample<Point3>;

        Vector3 d = warp::square_to_uniform_sphere(sample);
        Float dist = 2.f * m_bsphere.radius;

        DirectionSample ds;
        ds.p      = it.p + d * dist;
        ds.n      = -d;
        ds.uv     = Point2(0.f);
        ds.time   = it.time;
        ds.pdf    = warp::square_to_uniform_sphere_pdf(d);
        ds.delta  = false;
        ds.object = this_pointer<Value>();
        ds.d      = d;
        ds.dist   = dist;

        return std::make_pair(
            ds,
            m_radiance->eval(it.wavelengths, active) / ds.pdf
        );
    }

    template <typename Point2, typename Value = value_t<Point2>,
              typename Point3         = mitsuba::Point<Value, 3>,
              typename PositionSample = mitsuba::PositionSample<Point3>>
    MTS_INLINE std::pair<PositionSample, Value>
    sample_position_impl(const Value &time, const Point2 &sample,
                         const mask_t<Value> &active) const {
        /* Note that sampling a position from an infinitely distant light source
         * cannot really make sense. Instead, we return a position from a
         * virtual bounding sphere around the scene. */
        using Vector3 = vector3_t<Point3>;
        Vector3 d     = warp::square_to_uniform_sphere(sample);

        PositionSample ps;
        ps.p      = m_bsphere.center + d * m_bsphere.radius;
        ps.n      = -d;
        ps.uv     = sample;
        ps.time   = time;
        ps.pdf    = select(active, m_inv_surface_area, Value(0.f));
        ps.delta  = false;
        ps.object = this_pointer<Value>();
        return { ps,
                 select(ps.pdf > 0.f, rcp(m_inv_surface_area), Value(0.f)) };
    }

    template <typename Value>
    auto sample_wavelengths_impl(const Value &sample,
                                 const mask_t<Value> &active) const {
        using Spectrum = mitsuba::Spectrum<Value>;
        return m_radiance->sample(
            math::sample_shifted<Spectrum>(sample), active);
    }

    template <typename Interaction, typename DirectionSample,
              typename Value = typename DirectionSample::Value>
    MTS_INLINE Value pdf_direction_impl(const Interaction &,
                                        const DirectionSample &ds,
                                        mask_t<Value>) const {
        return warp::square_to_uniform_sphere_pdf(ds.d);
    }

    template <typename PositionSample,
              typename Value = typename PositionSample::Value>
    Value pdf_position_impl(const PositionSample & /*ps*/,
                            const mask_t<Value> & active) const {
        // Note that in principle, we should check that ps.p lies on the bsphere
        return select(active, m_inv_surface_area, Value(0.f));
    }

    template <typename Spectrum, typename Value = value_t<Spectrum>>
    Spectrum pdf_wavelengths_impl(const Spectrum &wavelengths,
                                  const mask_t<Value> &active) const {
        return m_radiance->pdf(wavelengths, active);
    }

    BoundingBox3f bbox() const override {
        /* This emitter does not occupy any particular region
           of space, return an invalid bounding box */
        return BoundingBox3f();
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "ConstantBackgroundEmitter[" << std::endl
            << "  radiance = " << string::indent(m_radiance) << "," << std::endl
            << "  bsphere = " << m_bsphere << "," << std::endl
            << "]";
        return oss.str();
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        m_radiance->set_labels(prefix + ".m_radiance");
    }
    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        m_radiance->set_globals(prefix + "_m_radiance", context, material);
    }
#endif

    MTS_IMPLEMENT_EMITTER_ALL()
    MTS_DECLARE_CLASS()
protected:
    ref<ContinuousSpectrum> m_radiance;
};

MTS_IMPLEMENT_CLASS(ConstantBackgroundEmitter, InfiniteEmitter)
MTS_EXPORT_PLUGIN(ConstantBackgroundEmitter, "Constant background emitter");
NAMESPACE_END(mitsuba)
