#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/bsphere.h>
#include <mitsuba/core/bitmap.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/emitter.h>
#include <mitsuba/render/kdtree.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/srgb.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#include <optix_cuda_interop.h>
#include <mitsuba/render/scene.h>
#endif

NAMESPACE_BEGIN(mitsuba)

class EnvironmentMapEmitter final : public InfiniteEmitter {
public:
    using Warp = warp::Hierarchical2D<0>;

    EnvironmentMapEmitter(const Properties &props) : InfiniteEmitter(props) {
        auto fs = Thread::thread()->file_resolver();
        fs::path file_path = fs->resolve(props.string("filename"));

        m_bitmap = new Bitmap(file_path);

        /* Convert to linear RGB float bitmap, will be converted
           into spectral profile coefficients below */
        m_bitmap = m_bitmap->convert(Bitmap::ERGBA, Bitmap::EFloat, false);
        m_name = file_path.filename().string();

        std::unique_ptr<Float[]> luminance(new Float[hprod(m_bitmap->size())]);

        Float *ptr = (Float *) m_bitmap->data(),
              *lum_ptr = (Float *) luminance.get();

        for (size_t y = 0; y < m_bitmap->size().y(); ++y) {
            Float sin_theta =
                std::sin(y / Float(m_bitmap->size().y() - 1) * math::Pi);

            for (size_t x = 0; x < m_bitmap->size().x(); ++x) {
                Color3f rgb = load_unaligned<Vector3f>(ptr);
                if (m_color_mode == ERGBColor) {
                    Vector4f coeff = concat(rgb, 0.f);
                    store_unaligned(ptr, coeff);
                } else {
                    Float scale = hmax(rgb) * 2.f;
                    Color3f rgb_norm = rgb / std::max((Float) 1e-8, scale);
                    /* Fetch spectral fit for given sRGB color value */
                    Vector4f coeff = concat(srgb_model_fetch(rgb_norm), scale);
                    /* Overwrite the pixel value with the coefficients */
                    store_unaligned(ptr, coeff);
                }
                *lum_ptr++ = mitsuba::luminance(rgb) * sin_theta;
                ptr += 4;
            }
        }

        m_scale = props.float_("scale", 1.f);
        m_warp = Warp(m_bitmap->size(), luminance.get());
        m_d65 = ContinuousSpectrum::D65(1.f, m_color_mode);

#if defined(MTS_ENABLE_AUTODIFF)
        m_bitmap_d = FloatC::copy(m_bitmap->data(), hprod(m_bitmap->size()) * m_bitmap->channel_count());
#endif
    }

    template <typename SurfaceInteraction, typename Mask,
              typename Spectrum = typename SurfaceInteraction::Spectrum>
    MTS_INLINE Spectrum eval_impl(const SurfaceInteraction &si,
                                  Mask active) const {
        using Value    = typename SurfaceInteraction::Value;
        using Point2f  = Point<Value, 2>;
        using Vector3f = Vector<Value, 3>;

        Vector3f v = m_world_transform->eval(si.time, active)
                         .inverse()
                         .transform_affine(-si.wi);

        /* Convert to latitude-longitude texture coordinates */
        Point2f uv = Point2f(atan2(v.x(), -v.z()) * math::InvTwoPi,
                             safe_acos(v.y()) * math::InvPi);
        uv -= floor(uv);

        return eval_spectral(uv, si.wavelengths, active);
    }

    template <typename Point2, typename Value = value_t<Point2>>
    MTS_INLINE auto
    sample_ray_impl(Value time, const Spectrum<Value> &wavelengths,
                    const Point2 &sample1, const Point2 &sample2,
                    mask_t<Value> active) const {
        using Point3  = Point<Value, 3>;
        using Ray3    = Ray<Point3>;
        using Vector3 = vector3_t<Point3>;
        using Frame   = Frame<Vector<Value, 3>>;

        // 1. Sample spatial component (ray origin)
        Point2 offset = warp::square_to_uniform_disk_concentric(sample1);

        // 2. Sample directional component
        auto [uv, pdf] =
            m_warp.sample(sample2, (const Value *) nullptr, active);

        Value theta = uv.y() * math::Pi;
        Value phi   = uv.x() * (2.f * math::Pi);

        Vector3 d           = math::sphdir(theta, phi);
        d                   = Vector3(d.y(), d.z(), -d.x());
        Value inv_sin_theta = safe_rsqrt(sqr(d.x()) + sqr(d.z()));
        pdf *= inv_sin_theta * math::InvTwoPi;

        d = m_world_transform->eval(time, active).transform_affine(d);

        // Evaluate radiance along ray
        auto value = eval_spectral(uv, wavelengths, active);

        Float r2 = m_bsphere.radius * m_bsphere.radius;
        Vector3 perpendicular_offset =
            Frame(d).to_world(Vector3(offset.x(), offset.y(), 0));
        Ray3 ray(m_bsphere.center +
                     (perpendicular_offset - d) * m_bsphere.radius,
                 d, time, wavelengths);
        return std::make_pair(ray, math::Pi * r2 * value / pdf);
    }

    template <typename Interaction,
              typename Point2 = typename Interaction::Point2,
              typename Value  = typename Interaction::Value>
    MTS_INLINE auto sample_direction_impl(const Interaction &it,
                                          const Point2 &sample,
                                          mask_t<Value> active) const {
        using Point3  = typename Interaction::Point3;
        using Vector3 = vector3_t<Point3>;
        using DirectionSample = DirectionSample<Point3>;

        auto [uv, pdf] = m_warp.sample(sample);

        Value theta = uv.y() * math::Pi,
              phi = uv.x() * (2.f * math::Pi);

        Vector3 d = math::sphdir(theta, phi);
        d = Vector3(d.y(), d.z(), -d.x());

        Float dist = 2.f * m_bsphere.radius;

        Value inv_sin_theta = safe_rsqrt(sqr(d.x()) + sqr(d.z()));

        d = m_world_transform->eval(it.time, active).transform_affine(d);

        DirectionSample ds;
        ds.p      = it.p + d * dist;
        ds.n      = -d;
        ds.uv     = uv;
        ds.time   = it.time;
        ds.pdf    = pdf * inv_sin_theta * (1.f / (2.f * math::Pi * math::Pi));
        masked(ds.pdf, pdf <= 0.f) = 0.f;
        ds.delta  = false;
        ds.object = this_pointer<Value>();
        ds.d      = d;
        ds.dist   = dist;

        if constexpr (is_scalar_v<Value>) {
            if (enoki::isnan(ds.pdf)) {
                Log(EWarn, "ds.pdf = %s, pdf = %s, uv = %s, sample = %s, eval = %s\n"
                           "theta = %s, phi = %s, inv_sin_theta = %s, d = %s",
                    ds.pdf, pdf, uv, sample, eval_spectral(uv, it.wavelengths, active),
                    theta, phi, inv_sin_theta, d);
            }
        }

        return std::make_pair(ds, select(ds.pdf > 0.f,
                                         eval_spectral(uv, it.wavelengths, active) / ds.pdf,
                                         Spectrum<Value>(0.f)));
    }

    template <typename Point2, typename Value = value_t<Point2>,
              typename Point3         = mitsuba::Point<Value, 3>,
              typename PositionSample = mitsuba::PositionSample<Point3>>
    MTS_INLINE std::pair<PositionSample, Value>
    sample_position_impl(const Value &time, const Point2 &sample,
                         const mask_t<Value> &active) const {
        /* Note that sampling a position from an infinitely distant light source
         * cannot really make sense. Instead, we return a position from a
         * virtual bounding sphere around the scene. */
        using Vector3 = vector3_t<Point3>;
        Vector3 d     = warp::square_to_uniform_sphere(sample);

        PositionSample ps;
        ps.p      = m_bsphere.center + d * m_bsphere.radius;
        ps.n      = -d;
        ps.uv     = sample;
        ps.time   = time;
        ps.pdf    = select(active, m_inv_surface_area, Value(0.f));
        ps.delta  = false;
        ps.object = this_pointer<Value>();
        return { ps,
                 select(ps.pdf > 0.f, rcp(m_inv_surface_area), Value(0.f)) };
    }

    template <typename Value>
    auto sample_wavelengths_impl(const Value &sample,
                                 const mask_t<Value> &active) const {
        using Spectrum = mitsuba::Spectrum<Value>;
        // TODO: consider sampling from a more specific continuous spectrum.
        return m_d65->sample(
            math::sample_shifted<Spectrum>(sample), active);
    }

    template <typename Interaction, typename DirectionSample,
              typename Value = typename DirectionSample::Value>
    MTS_INLINE Value pdf_direction_impl(const Interaction &it,
                                        const DirectionSample &ds,
                                        mask_t<Value> active) const {
        using Vector3f = Vector<Value, 3>;
        using Point2f  = Point<Value, 2>;

        Vector3f d = m_world_transform->eval(it.time, active)
                         .inverse()
                         .transform_affine(ds.d);

        /* Convert to latitude-longitude texture coordinates */
        Point2f uv = Point2f(atan2(d.x(), -d.z()) * math::InvTwoPi,
                             safe_acos(d.y()) * math::InvPi);
        uv -= floor(uv);

        Value inv_sin_theta = safe_rsqrt(sqr(d.x()) + sqr(d.z()));
        Value result = m_warp.eval(uv) * inv_sin_theta * (1.f / (2.f * math::Pi * math::Pi));

        return result;
    }

    template <typename PositionSample,
              typename Value = typename PositionSample::Value>
    Value pdf_position_impl(const PositionSample & /*ps*/,
                            const mask_t<Value> & active) const {
        // Note that in principle, we should check that ps.p lies on the bsphere
        return select(active, m_inv_surface_area, Value(0.f));
    }

    template <typename Spectrum, typename Value = value_t<Spectrum>>
    Spectrum pdf_wavelengths_impl(const Spectrum &wavelengths,
                                  const mask_t<Value> &active) const {
        return m_d65->pdf(wavelengths, active);
    }

    BoundingBox3f bbox() const override {
        /* This emitter does not occupy any particular region
           of space, return an invalid bounding box */
        return BoundingBox3f();
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "EnvironmentMapEmitter[" << std::endl
            << "  name = \"" << m_name << "\"," << std::endl
            << "  bitmap = " << string::indent(m_bitmap) << "," << std::endl
            << "  bsphere = " << m_bsphere << std::endl
            << "]";
        return oss.str();
    }

    bool is_environment() const override {
        return true;
    }

    MTS_IMPLEMENT_EMITTER_ALL()
    MTS_DECLARE_CLASS()

protected:
    template <typename Point2f, typename Spectrum, typename Mask>
    Spectrum eval_spectral(Point2f uv, const Spectrum &wavelengths, Mask active) const {
        using Value   = value_t<Point2f>;
        using UInt32  = uint32_array_t<Value>;
        using Point2u = Point<UInt32, 2>;
        using Vector3  = Vector<Value, 3>;
        using Vector4 = Vector<Value, 4>;

        uv *= Vector2f(m_bitmap->size() - 1u);

        Point2u pos = min(Point2u(uv), m_bitmap->size() - 2u);

        Point2f w1 = uv - Point2f(pos),
                w0 = 1.f - w1;

        UInt32 index = pos.x() + pos.y() * (uint32_t) m_bitmap->size().x();

        Vector4 v00, v10, v01, v11;
        if constexpr (is_diff_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
            uint32_t width = (uint32_t) m_bitmap->size().x();
            v00            = gather<Vector4>(m_bitmap_d, index, active);
            v10            = gather<Vector4>(m_bitmap_d, index + 1u, active);
            v01            = gather<Vector4>(m_bitmap_d, index + width, active);
            v11            = gather<Vector4>(m_bitmap_d, index + width + 1u, active);
#endif
        } else {
            uint32_t width   = (uint32_t) m_bitmap->size().x() * 4;
            const Float *ptr = (const Float *) m_bitmap->data();

            v00 = gather<Vector4>(ptr, index, active);
            v10 = gather<Vector4>(ptr + 4, index, active);
            v01 = gather<Vector4>(ptr + width, index, active);
            v11 = gather<Vector4>(ptr + width + 4, index, active);
        }

        if (m_color_mode == ERGBColor) {
            // Note: scale was left multiplied-in, unlike in spectral mode.
            Vector3 v0 = fmadd(w0.x(), head<3>(v00), w1.x() * head<3>(v10)),
                    v1 = fmadd(w0.x(), head<3>(v01), w1.x() * head<3>(v11)),
                    v  = fmadd(w0.y(), v0, w1.y() * v1);
            v = select(any(v > 0.f), v, 0.f);
            return m_scale * Spectrum(v.x(), v.y(), v.z(), 0.f);
        }

        Spectrum s00 = srgb_model_eval(head<3>(v00), wavelengths),
                 s10 = srgb_model_eval(head<3>(v10), wavelengths),
                 s01 = srgb_model_eval(head<3>(v01), wavelengths),
                 s11 = srgb_model_eval(head<3>(v11), wavelengths),
                 s0  = fmadd(w0.x(), s00, w1.x() * s10),
                 s1  = fmadd(w0.x(), s01, w1.x() * s11),
                 f0  = fmadd(w0.x(), v00.w(), w1.x() * v10.w()),
                 f1  = fmadd(w0.x(), v01.w(), w1.x() * v11.w()),
                 s   = fmadd(w0.y(), s0, w1.y() * s1),
                 f   = fmadd(w0.y(), f0, w1.y() * f1);

        return s * f * m_d65->eval(wavelengths, active) * m_scale;
    }


#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        // Note: no need for `m_d65` since it is unused in RGB mode.
        auto var_id = cuda_var_register_ptr(m_bitmap_d.data());
        auto name = prefix + ".m_bitmap_d_data";
        cuda_var_set_label(var_id, name.c_str());

        m_warp.set_labels(prefix + ".m_warp");
    }
    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {

        cuda_eval();

        auto name = prefix + "_m_bitmap_d_data";
        RTvariable var = get_or_create_optix_var(name, context, material);
        const auto *ptr_value = (float *) m_bitmap_d.data();
        rtVariableSetUserData(var, sizeof(ptr_value), &ptr_value);

        m_warp.set_globals(prefix + "_m_warp", context, material);
    }
#endif

protected:
    std::string m_name;
    ref<Bitmap> m_bitmap;
    Warp m_warp;
    ref<ContinuousSpectrum> m_d65;
    Float m_scale;

#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_bitmap_d;
#endif
};

MTS_IMPLEMENT_CLASS(EnvironmentMapEmitter, InfiniteEmitter)
MTS_EXPORT_PLUGIN(EnvironmentMapEmitter, "Environment map emitter");
NAMESPACE_END(mitsuba)
