#include <mitsuba/render/emitter.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/core/bbox.h>

NAMESPACE_BEGIN(mitsuba)

class PointLight final : public Emitter {
public:
    PointLight(const Properties &props) : Emitter(props) {
        if (props.has_property("position")) {
            if (props.has_property("to_world"))
                Throw("Only one of the parameters 'position' and 'to_world' "
                      "can be specified at the same time!'");

            m_world_transform = new AnimatedTransform(
                Transform4f::translate(Vector3f(props.point3f("position"))));
        }

        m_intensity = props.spectrum("intensity", ContinuousSpectrum::D65(1.f, m_color_mode));
        m_needs_sample_1 = false;
    }

    template <typename Point2, typename Value = value_t<Point2>,
              typename Point3   = Point<Value, 3>,
              typename Ray3     = mitsuba::Ray<Point3>,
              typename Spectrum = mitsuba::Spectrum<Value>>
    std::pair<Ray3, Spectrum>
    sample_ray_impl(Value time, const Spectrum &wavelengths,
                    const Point2 &/* pos_sample */, const Point2 &dir_sample,
                    const mask_t<Value> &/* active */) const {

        const auto &trafo = m_world_transform->eval(time);
        Ray3 ray(trafo * Point3(0.f),
                 warp::square_to_uniform_sphere(dir_sample), time, wavelengths);

        return { ray, 4.f * math::Pi };
    }

    template <typename Interaction, typename Mask,
              typename Point2          = typename Interaction::Point2,
              typename Point3          = typename Interaction::Point3,
              typename Value           = typename Interaction::Value,
              typename Spectrum        = typename Interaction::Spectrum,
              typename DirectionSample = DirectionSample<Point3>>
    MTS_INLINE std::pair<DirectionSample, Spectrum>
    sample_direction_impl(const Interaction &it, const Point2 &sample,
                          Mask active) const {
        PositionSample<Point3> ps;
        Spectrum weight;
        std::tie(ps, weight) = sample_position(it.time, sample, active);

        DirectionSample ds(ps);
        ds.d           = ds.p - it.p;
        ds.dist        = norm(ds.d);
        Value inv_dist = rcp(ds.dist);
        ds.d *= inv_dist;

        weight *= m_intensity->eval(it.wavelengths, active) * sqr(inv_dist);

        return { ds, weight };
    }

    template <typename Point2, typename Value = value_t<Point2>,
              typename Point3         = mitsuba::Point<Value, 3>,
              typename PositionSample = mitsuba::PositionSample<Point3>>
    MTS_INLINE std::pair<PositionSample, Value>
    sample_position_impl(const Value &time, const Point2 &/*sample*/,
                         const mask_t<Value> &active) const {
        PositionSample ps;
        ps.p      = m_world_transform->eval(time, active).translation();
        ps.n      = 0.f;
        ps.uv     = 0.f;
        ps.time   = time;
        ps.pdf    = select(active, Value(1.f), Value(0.f));
        ps.delta  = true;
        ps.object = this;
        return { ps, 1.f };
    }

    template <typename Value>
    auto sample_wavelengths_impl(const Value &sample,
                                 const mask_t<Value> &active) const {
        using Spectrum = mitsuba::Spectrum<Value>;
        return m_intensity->sample(
            math::sample_shifted<Spectrum>(sample), active);
    }

    template <typename Interaction, typename DirectionSample, typename Mask,
              typename Value = typename DirectionSample::Value>
    MTS_INLINE Value pdf_direction_impl(const Interaction &, const DirectionSample &, Mask) const {
        return 0.f;
    }

    template <typename PositionSample,
              typename Value = typename PositionSample::Value>
    Value pdf_position_impl(const PositionSample & /*ps*/,
                            const mask_t<Value> & /*active*/) const {
        return 1.f;
    }

    template <typename Spectrum, typename Value = value_t<Spectrum>>
    Spectrum pdf_wavelengths_impl(const Spectrum &wavelengths,
                                  const mask_t<Value> &active) const {
        return m_intensity->pdf(wavelengths, active);
    }

    template <typename SurfaceInteraction, typename Mask,
              typename Spectrum = typename SurfaceInteraction::Spectrum>
    MTS_INLINE Spectrum eval_impl(const SurfaceInteraction &, Mask) const {
        return 0.f;
    }

    template <typename Ray,
              typename Value    = typename Ray::Value,
              typename Spectrum = mitsuba::Spectrum<Value>>
    MTS_INLINE Spectrum eval_environment_impl(const Ray &, mask_t<Value>) const {
        return 0.f;
    }

    BoundingBox3f bbox() const override {
        return m_world_transform->translation_bounds();
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "PointLight[" << std::endl
            << "  world_transform = " << string::indent(m_world_transform->to_string()) << "," << std::endl
            << "  intensity = " << m_intensity << "," << std::endl
            << "  medium = " << (m_medium ? string::indent(m_medium->to_string()) : "")
            << "]";
        return oss.str();
    }

    MTS_IMPLEMENT_EMITTER_ALL()
    MTS_DECLARE_CLASS()
private:
    ref<ContinuousSpectrum> m_intensity;
};


MTS_IMPLEMENT_CLASS(PointLight, Emitter)
MTS_EXPORT_PLUGIN(PointLight, "Point emitter");
NAMESPACE_END(mitsuba)
