#include <enoki/stl.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/ray.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/emitter.h>
#include <mitsuba/render/integrator.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/phase.h>
#include <mitsuba/render/records.h>
#include <random>

NAMESPACE_BEGIN(mitsuba)

class VolumetricPathIntegrator : public MonteCarloIntegrator {
public:
    VolumetricPathIntegrator(const Properties &props) : MonteCarloIntegrator(props) {}

    template <typename RayDifferential, typename Point3 = typename RayDifferential::Point,
              typename RadianceSample = RadianceSample<Point3>,
              typename Value          = typename RayDifferential::Value>
    auto eval_impl(RayDifferential ray_diff, RadianceSample &rs, mask_t<Value> active) const {
        using Mask                = mask_t<Value>;
        using Spectrum            = mitsuba::Spectrum<Value>;
        using SurfaceInteraction3 = SurfaceInteraction<Point3>;
        using MediumInteraction3  = MediumInteraction<Point3>;
        using BSDFPtr             = replace_scalar_t<Value, const BSDF *>;
        using MediumPtr           = replace_scalar_t<Value, const Medium *>;
        using EmitterPtr          = replace_scalar_t<Value, const Emitter *>;
        using Vector3             = vector3_t<Point3>;
        using Index               = uint_array_t<Value>;

        // Create as many indices as we have different rays
        size_t n_elems = slices(ray_diff.d);
        if constexpr (!is_dynamic_array_v<Value>)
            n_elems = array_size_v<Value>;
        Index index = arange<Index>(n_elems);

        const Scene *scene = rs.scene;
        rs.alpha           = 0.0f;

        Ray ray = ray_diff; // For now, ignore ray differentials

        // Tracks radiance scaling due to index of refraction changes
        Value eta(1.f);

        Spectrum throughput(1.f), result(0.f);

        MediumPtr medium = rs.medium;
        MediumInteraction3 mi;

        Mask specular_chain = active && !m_hide_emitters;
        for (int depth = 0;; ++depth) {
            // ----------------- Handle termination of paths ------------------

            /* Russian roulette: try to keep path weights equal to one,
            while accounting for the solid angle compression at refractive
            index boundaries. Stop with at least some probability to avoid
            getting stuck (e.g. due to total internal reflection) */
            active &= any(neq(throughput, 0.f));
            if (depth > m_rr_depth) {
                Value q = min(hmax(throughput) * sqr(eta), .95f);
                active &= rs.next_1d(active) < q;
                throughput *= rcp(detach(q));
            }

            if (none(active) || (uint32_t) depth >= (uint32_t) m_max_depth)
                break;

            // ----------------------- Sampling the RTE -----------------------
            Mask active_medium  = active && neq(medium, nullptr);
            Mask active_surface = active && !active_medium;
            Mask escaped_medium = false;
            SurfaceInteraction3 si_medium;
            if (any_or<true>(active_medium)) {
                Spectrum medium_throughput;
                std::tie(si_medium, mi, medium_throughput) =
                    medium->sample_distance(scene, ray, rs.next_2d(active_medium),
                                            rs.sampler, index, active_medium);
                escaped_medium = active_medium && si_medium.is_valid() && !mi.is_valid();
                masked(throughput, active_medium) *= medium_throughput;
                active_medium &= mi.is_valid();
            }

            if (any_or<true>(active_medium)) {
                // Still inside medium after distance sampling, sample the phase function now
                auto phase = mi.medium->phase_function();
                masked(phase, !active_medium) = nullptr;
                auto wo = phase->sample(mi, rs.next_2d(active_medium), active_medium);
                Ray new_ray  = mi.spawn_ray(mi.to_world(wo));
                new_ray.mint = 0.0f;
                new_ray.maxt = math::Infinity;
                masked(ray, active_medium) = new_ray;

                masked(rs.alpha, active_medium) = 1.f;

                // Emitter sampling (if requested)
                Mask sample_emitters = mi.medium->use_emitter_sampling();
                specular_chain &= !active_medium;
                specular_chain |= active_medium && !sample_emitters;
                Mask active_e = active_medium && sample_emitters;
                if (any_or<true>(active_e)) {
                    auto [ds, emitter_val] = scene->sample_emitter_direction_attenuated(
                        mi, true, medium, rs.sampler, true, index, active_e);
                    active_e &= neq(ds.pdf, 0.f);
                    Value phase_val =
                        mi.medium->phase_function()->eval(mi.to_local(ds.d), active_e);
                    masked(result, active_e) += throughput * phase_val * emitter_val;
                    // masked(result, active_e) = Spectrum(0.5f, 0.f, 0.5f, 0.f);
                }
            }

            // --------------------- Surface Interactions ---------------------
            SurfaceInteraction3 si     = scene->ray_intersect(ray, active_surface);
            masked(si, escaped_medium) = si_medium;
            active_surface |= escaped_medium;

            if (any_or<true>(active_surface)) {
                // ---------------- Intersection with emitters ----------------
                EmitterPtr emitter = si.emitter(scene);
                // Note: we don't actually require a valid surface interaction yet here,
                // e.g. environment emitters are fine.
                Mask use_emitter_contribution =
                    active_surface && specular_chain && neq(emitter, nullptr);
                if (any_or<true>(use_emitter_contribution)) {
                    masked(result, use_emitter_contribution) +=
                        throughput * emitter->eval(si, use_emitter_contribution);

                    // masked(result, use_emitter_contribution) = Spectrum(0.0f, 0.5f, 0.5f, 0.f);
                    // Rays which can see the envmap definitely shouldn't still
                    // have a medium.
                    // masked(result, use_emitter_contribution && neq(medium, nullptr)) = Spectrum(1.f, 0.f, 0.f, 0.f);
                    // masked(result, use_emitter_contribution && neq(medium, nullptr)) = Spectrum(0.f / 0.f);
                }
            }

            active_surface &= si.is_valid();
            if (any_or<true>(active_surface)) {
                // --------------------- Emitter sampling ---------------------
                BSDFContext ctx;
                BSDFPtr bsdf  = si.bsdf(ray);
                Mask active_e = active_surface && neq(bsdf->flags() & BSDF::ESmooth, 0u);
                if (likely(any_or<true>(active_e))) {
                    auto [ds, emitter_val] = scene->sample_emitter_direction_attenuated(
                        si, false, medium, rs.sampler, true, index, active_e);
                    active_e &= neq(ds.pdf, 0.f);

                    // Query the BSDF for that emitter-sampled direction
                    Vector3 wo        = si.to_local(ds.d);
                    Spectrum bsdf_val = bsdf->eval(ctx, si, wo, active_e);

                    // Determine probability of having sampled that same
                    // direction using BSDF sampling.
                    Value bsdf_pdf = bsdf->pdf(ctx, si, wo, active_e);
                    result[active_e] +=
                        throughput * emitter_val * bsdf_val * mis_weight(ds.pdf, bsdf_pdf);
                }

                // ----------------------- BSDF sampling ----------------------
                auto [bs, bsdf_val] = bsdf->sample(ctx, si, rs.next_1d(active_surface),
                                                   rs.next_2d(active_surface), active_surface);
                throughput[active_surface] *= bsdf_val;
                masked(eta, active_surface) *= bs.eta;

                Ray bsdf_ray = si.spawn_ray(si.to_world(bs.wo));
                masked(ray, active_surface) = bsdf_ray;

                Mask add_emitter = active_surface && eq(bs.sampled_type & BSDF::EDelta, 0u) &&
                                   any(neq(throughput, 0.f));
                // Handle the case when m_max_depth is negative
                int max_intersections = m_max_depth < 0
                                            ? std::numeric_limits<int>::max() - depth - 1
                                            : m_max_depth - depth - 1;
                auto [emitted, emitter_pdf] =
                    ray_intersect_and_look_for_emitter(si, scene, rs.sampler, medium, ray,
                                                       max_intersections, index, add_emitter);
                result += select(add_emitter && neq(emitter_pdf, 0),
                                 mis_weight(bs.pdf, emitter_pdf) * throughput * emitted, 0.0f);

                Mask has_medium_trans            = si.is_valid() && si.is_medium_transition();
                masked(medium, has_medium_trans) = si.target_medium(ray.d);

                Mask is_null_bsdf = active_surface && neq(bs.sampled_type & BSDF::ENull, 0u);
                Mask non_null_bsdf = active_surface && !is_null_bsdf;
                masked(rs.alpha, non_null_bsdf) = 1.f;

                specular_chain |= non_null_bsdf && neq(bs.sampled_type & BSDF::EDelta, 0u);
                // TODO: should we allow this type of paths to receive light?
                // specular_chain |= is_null_bsdf && has_medium_trans;
                specular_chain &= !active_surface || eq(bs.sampled_type & BSDF::ESmooth, 0u);

                // // TODO: remove this
                // masked(result, active_surface) = select(
                //     neq(medium, nullptr),
                //     Spectrum(0.0f, 0.0f, (depth + 1) * 0.1f, 0.f),
                //     Spectrum(0.0f, (depth + 1) * 0.1f, 0.0f, 0.f)
                // );
                // masked(result, !active) = 0.f;

                // if constexpr (is_scalar_v<Value>) {
                //     if (active_surface && eq(medium, nullptr) && si.t < 0.1f) {
                //     // if (active_surface && !has_medium_trans) {
                //         Log(EInfo, "depth = %s, bsdf_val = %s, throughput = %s, si.t = %s", depth,
                //             bsdf_val, throughput, si.t);
                //     }
                // }
            }
            active &= (active_surface | active_medium);
        }
        return result;
    }

    template <typename Ray, typename Point3 = typename Ray::Point, typename Value = value_t<Point3>,
              typename Spectrum = mitsuba::Spectrum<Value>, typename Index = uint_array_t<Value>,
              typename Mask                = mask_t<Value>, typename MediumPtr,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>>
    std::pair<Spectrum, Value>
    ray_intersect_and_look_for_emitter(const SurfaceInteraction3 &si_ref, const Scene *scene,
                                       Sampler *sampler, MediumPtr medium, Ray ray,
                                       int maxInteractions, Index index, Mask active) const {
        using EmitterPtr = replace_scalar_t<Value, const Emitter *>;
        Spectrum value(0.0f);
        Spectrum transmittance(1.0f);
        Value emitter_pdf(0.0f);
        int interactions = 0;
        EmitterPtr emitter;
        while (any(active) && interactions < maxInteractions) {

            // Intersect the value ray with the scene
            SurfaceInteraction3 si = scene->ray_intersect(ray, active);

            // If intersection is found: Is it a null bsdf or an occlusion?
            Mask active_surface = active && si.is_valid();
            // Check if we hit an emitter and add illumination if needed
            emitter          = si.emitter(scene, active);
            Mask emitter_hit = neq(emitter, nullptr) && active;
            if (any_or<true>(emitter_hit)) {
                value[emitter_hit] += transmittance * emitter->eval(si, emitter_hit);
                DirectionSample ds(si, si_ref);
                ds.object = emitter;
                masked(emitter_pdf, emitter_hit) =
                    scene->pdf_emitter_direction(si_ref, ds, emitter_hit);
                active &= !emitter_hit; // turn off lanes which already
                                        // found emitter
            }
            if (any_or<true>(active_surface)) {
                auto bsdf         = si.bsdf(ray);
                Spectrum bsdf_val = bsdf->eval_transmission(si, si.to_local(ray.d), active_surface);
                masked(transmittance, active_surface) *= bsdf_val;
            }
            Mask active_medium = neq(medium, nullptr) && active;
            if (any_or<true>(active_medium)) {
                masked(transmittance, active_medium) *=
                    medium->eval_transmittance(Ray(ray, 0, si.t), sampler, index, active_medium);
            }

            active &= si.is_valid() && any(neq(transmittance, 0.f));

            // If a medium transition is taking place: Update the medium pointer
            Mask has_medium_trans = active && si.is_medium_transition();
            if (any_or<true>(has_medium_trans)) {
                masked(medium, has_medium_trans) = si.target_medium(ray.d);
            }
            // Update the ray with new origin & t parameter
            masked(ray.o, active)    = si.p;
            masked(ray.mint, active) = math::Epsilon;
            interactions++;
        }
        return { value, emitter_pdf };
    }

    //! @}
    // =============================================================

    std::string to_string() const override {
        return tfm::format("VolumetricPathIntegrator[\n"
                           "  max_depth = %i,\n"
                           "  rr_depth = %i\n"
                           "]",
                           m_max_depth, m_rr_depth);
    }

    MTS_IMPLEMENT_INTEGRATOR_ALL()
    MTS_DECLARE_CLASS()

protected:
    template <typename Value> Value mis_weight(Value pdf_a, Value pdf_b) const {
        pdf_a *= pdf_a;
        pdf_b *= pdf_b;
        return select(pdf_a > 0.0f, pdf_a / (pdf_a + pdf_b), Value(0.0f));
    };
};

MTS_IMPLEMENT_CLASS(VolumetricPathIntegrator, MonteCarloIntegrator);
MTS_EXPORT_PLUGIN(VolumetricPathIntegrator, "Volumetric Path Tracer integrator");
NAMESPACE_END(mitsuba)
