#include <mitsuba/render/autodiff.h>
#include <mitsuba/core/logger.h>
#include <mitsuba/render/scene.h>
#include <map>
#include <set>

NAMESPACE_BEGIN(mitsuba)

void DifferentiableObject::parameters_changed() { }

#if defined(MTS_ENABLE_AUTODIFF)
struct DifferentiableParameters::Details {
    std::string prefix;
    std::map<std::string, std::tuple<DifferentiableObject *, void *, size_t>> entries;

    Details() {}
    Details(const Details &other)
        : prefix(other.prefix), entries(other.entries) {}
    Details &operator=(const Details &other) = delete;
};

void DifferentiableObject::set_labels(const std::string &) const {}

void DifferentiableObject::set_globals(const std::string &, const RTcontext &,
                                       const RTmaterial &) const {}

void DifferentiableObject::put_parameters(DifferentiableParameters &) { }



DifferentiableParameters::DifferentiableParameters() : d(new Details()) { }

DifferentiableParameters::DifferentiableParameters(const DifferentiableParameters &other)
    : d(new Details(*other.d)) {}

DifferentiableParameters::~DifferentiableParameters() { }

const std::tuple<DifferentiableObject *, void *, size_t> &
DifferentiableParameters::operator[](const std::string &key) const {
    return d->entries.at(key);
}

void DifferentiableParameters::set_prefix(const std::string &prefix) {
    d->prefix = prefix;
}

void DifferentiableParameters::put(DifferentiableObject *obj, const std::string &name, void *ptr, size_t dim) {
    auto qualified = d->prefix + name;
    if (d->entries.find(qualified) != d->entries.end())
        Throw("DifferentiableParameters::put(): parameters \"%s\" is already "
              "registered!", qualified);
    d->entries.emplace(qualified, std::make_tuple(obj, ptr, dim));
}

std::vector<std::string> DifferentiableParameters::keys() const {
    std::vector<std::string> result;
    result.reserve(d->entries.size());
    for (auto &kv: d->entries)
        result.push_back(kv.first);
    return result;
}

void DifferentiableParameters::keep(const std::vector<std::string> &keys) {
    std::map<std::string, std::tuple<DifferentiableObject *, void *, size_t>> entries;
    for (const std::string &k : keys) {
        auto it = d->entries.find(k);
        if (it == d->entries.end()) {
            throw std::runtime_error(tfm::format(
                "DifferentiableParameters::keep(): could not find "
                "parameter \"%s\". Available keys are:\n%s",
                k, this->keys()));
        }
        entries.emplace(*it);
    }
    d->entries = entries;
}

size_t DifferentiableParameters::size() const {
    return d->entries.size();
}

void DifferentiableParameters::clear() {
    return d.reset(new Details());
}

size_t DifferentiableParameters::depth(const std::string &key) const {
    return std::get<2>(d->entries.at(key));
}

size_t DifferentiableParameters::slices(const std::string &key) const {
    auto [parent, ptr, depth] = d->entries.at(key);
    switch (depth) {
        case 1: return enoki::slices(*((FloatD *) ptr));
        case 2: return enoki::slices(*((Vector2fD *) ptr));
        case 3: return enoki::slices(*((Vector3fD *) ptr));
        case 4: return enoki::slices(*((Vector4fD *) ptr));
        case 16: return enoki::slices(*((Matrix4fD *) ptr));
        default: throw std::runtime_error("DifferentiableParameters::slices: unsupported type!");
    }
}

std::string DifferentiableParameters::to_string() const {
    std::ostringstream oss;
    oss << "DifferentiableParameters[" << std::endl;
    size_t i = 0;
    for (auto &kv: d->entries) {
        oss << "  " << kv.first;
        ++i;
        if (i != d->entries.size())
            oss << ",";
        oss << std::endl;
    }
    oss << "]";
    return oss.str();
}

void traverse(DifferentiableParameters *params, std::set<Object *> &visited, std::string path,
              Object *obj) {
    if (visited.find(obj) != visited.end())
        return;  // Avoid cycles

    path += obj->class_()->name();
    std::string identifier = obj->id();
    if (identifier.size() > 0 && identifier.find("[0x") == std::string::npos) {
        path += "[id=\"" + identifier + "\"]";
    }
    path += "/";
    params->set_prefix(path);

    visited.insert(obj);

    DifferentiableObject *diff_obj = dynamic_cast<DifferentiableObject *>(obj);
    if (diff_obj)
        diff_obj->put_parameters(*params);

    for (ref<Object> &child : obj->children())
        traverse(params, visited, path, child.get());
}

MTS_EXPORT_RENDER void get_differentiable_parameters(Object *root,
                                                     DifferentiableParameters *params) {
    auto visited = std::set<Object *>();
    traverse(params, visited, "/", root);
}

MTS_IMPLEMENT_CLASS(DifferentiableParameters, Object)
#endif

MTS_IMPLEMENT_CLASS(DifferentiableObject, Object)
NAMESPACE_END(mitsuba)
