#include <enoki/morton.h>
#include <enoki/stl.h>
#include <mitsuba/core/profiler.h>
#include <mitsuba/core/progress.h>
#include <mitsuba/core/timer.h>
#include <mitsuba/core/util.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/film.h>
#include <mitsuba/render/integrator.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/sensor.h>
#include <mitsuba/render/spiral.h>
#include <thread>

NAMESPACE_BEGIN(mitsuba)

Integrator::Integrator(const Properties & props)
    : m_stop(false) {
    m_color_mode = (EColorMode) props.int_("color_mode");
    m_timeout = props.float_("timeout", -1.0f);
}

void Integrator::cancel() {
    m_stop = true;
}

size_t Integrator::register_callback(CallbackFunction cb, Float period) {
    m_callbacks.push_back(CallbackInfo(cb, period));
    return m_callbacks.size() - 1;
}

void Integrator::remove_callback(size_t cb_index) {
    if (cb_index == (size_t) -1)
        m_callbacks.clear();
    else {
        Assert(cb_index < m_callbacks.size());
        m_callbacks.erase(m_callbacks.begin() + cb_index);
    }
}

void Integrator::notify(Bitmap *bitmap, Float extra) {
    if (m_callbacks.empty())
        return;
    Float current_time = m_cb_timer.value() / 1000.f;
    for (auto &entry : m_callbacks) {
        auto &last_called = (Float &) entry.last_called;
        if (last_called < 0.f || last_called > current_time ||
            current_time - last_called > entry.period) {
            entry.f(Thread::thread()->thread_id(), bitmap, extra);
            last_called = current_time;
        }
    }
}

// -----------------------------------------------------------------------------

SamplingIntegrator::SamplingIntegrator(const Properties &props)
    : Integrator(props) {
    m_block_size = props.size_("block_size", MTS_BLOCK_SIZE);
    size_t b = math::round_to_power_of_two(m_block_size);
    m_samples_per_pass = props.size_("samples_per_pass", (size_t) -1);

    if (b != m_block_size) {
        m_block_size = b;
        Log(EWarn, "Setting block size to next higher power of two: %i", m_block_size);
    }
}

SamplingIntegrator::~SamplingIntegrator() { }

Spectrumf SamplingIntegrator::eval(const RayDifferential3f &,
                                   RadianceSample3f &) const {
    NotImplementedError("eval");
}
SpectrumfP SamplingIntegrator::eval(const RayDifferential3fP &,
                                    RadianceSample3fP &, MaskP) const {
    NotImplementedError("eval_p");
}
#if defined(MTS_ENABLE_AUTODIFF)
SpectrumfD SamplingIntegrator::eval(const RayDifferential3fD &,
                                    RadianceSample3fD &, MaskD) const {
    NotImplementedError("eval_d");
}
#endif

bool SamplingIntegrator::render(Scene *scene, bool vectorize) {
    ScopedPhase sp(EProfilerPhase::ERender);
    m_stop = false;

    ref<Sensor> sensor = scene->sensor();
    ref<Film> film = sensor->film();

    size_t n_threads        = __global_thread_count;
    size_t total_spp        = scene->sampler()->sample_count();
    size_t samples_per_pass = (m_samples_per_pass == (size_t) -1)
                               ? total_spp : std::min(m_samples_per_pass, total_spp);
    if ((total_spp % samples_per_pass) != 0)
        Throw("sample_count (%d) must be a multiple of samples_per_pass (%d).",
              total_spp, samples_per_pass);

    size_t n_passes = ceil(total_spp / (Float) samples_per_pass);
    film->clear();

    Log(EInfo, "Starting render job (%ix%i, %i sample%s,%s %i thread%s)",
        film->crop_size().x(), film->crop_size().y(),
        total_spp, total_spp == 1 ? "" : "s",
        n_passes > 1 ? tfm::format(" %d passes,", n_passes) : "",
        n_threads, n_threads == 1 ? "" : "s");
    if (m_timeout > 0.f)
        Log(EInfo, "Timeout specified: %.2f seconds.", m_timeout);

    Spiral spiral(film, m_block_size, n_passes);

    ThreadEnvironment env;
    ref<ProgressReporter> progress = new ProgressReporter("Rendering");
    tbb::spin_mutex mutex;
    size_t blocks_done = 0;
    // Total number of blocks to be handled, including multiple passes.
    Float total_blocks = (spiral.block_count() * n_passes);
    size_t grain_count = std::max(
        (size_t) 1, (size_t) std::rint(total_blocks / Float(n_threads * 2)));

    m_render_timer.reset();
    tbb::parallel_for(
        tbb::blocked_range<size_t>(0, total_blocks, grain_count),
        [&](const tbb::blocked_range<size_t> &range) {
            ScopedSetThreadEnvironment set_env(env);
            ref<Sampler> sampler = scene->sampler()->clone();
            ref<ImageBlock> block =
                new ImageBlock(Bitmap::EXYZAW, Vector2i(m_block_size),
                               film->reconstruction_filter(), 0, true, m_color_mode);

            scoped_flush_denormals flush_denormals(true);

            Point2fX points;
            // For each block
            for (auto i = range.begin(); i != range.end() && !should_stop(); ++i) {
                auto [offset, size] = spiral.next_block();
                if (hprod(size) == 0)
                    Throw("Internal error -- generated empty image block!");
                if (size != block->size())
                    block = new ImageBlock(Bitmap::EXYZAW, size, film->reconstruction_filter(), 0,
                                                               true, m_color_mode);

                block->set_offset(offset);

                // Ensure that the sample generation is fully deterministic
                size_t seed = (size_t) offset.x() +
                              (size_t) offset.y() * (size_t) film->crop_size().x();
                if (n_passes > 1)
                    seed += i * hprod(film->crop_size());
                sampler->seed(seed);

                if (vectorize)
                    render_block_vector(scene, sampler, block, points,
                                        samples_per_pass);
                else
                    render_block_scalar(scene, sampler, block,
                                        samples_per_pass);

                film->put(block);

                /* locked */ {
                    tbb::spin_mutex::scoped_lock lock(mutex);
                    blocks_done++;
                    progress->update(blocks_done / (Float) total_blocks);
                    notify(film->bitmap());
                }
            }
        }
    );

    if (!m_stop)
        Log(EInfo, "Rendering finished. (took %s)",
            util::time_string(m_render_timer.value(), true));

    return !m_stop;
}

void SamplingIntegrator::render_block_scalar(const Scene *scene,
                                             Sampler *sampler,
                                             ImageBlock *block,
                                             size_t sample_count_) const {
    const Sensor *sensor = scene->sensor();
    block->clear();
    uint32_t pixel_count  = (uint32_t)(m_block_size * m_block_size),
             sample_count = (uint32_t)(sample_count_ == (size_t) -1
                                           ? sampler->sample_count()
                                           : sample_count_);

    RadianceSample3f rs(scene, sampler, sensor->medium());
    bool needs_aperture_sample = sensor->needs_aperture_sample();
    bool needs_time_sample = sensor->shutter_open_time() > 0;

    Float diff_scale_factor = rsqrt((Float) sampler->sample_count());

    Point2f aperture_sample(.5f);
    Vector2f inv_resolution = 1.f / sensor->film()->crop_size();

    for (uint32_t i = 0; i < pixel_count && !should_stop(); ++i) {
        Point2u p = enoki::morton_decode<Point2u>(i);
        if (any(p >= block->size()))
            continue;

        p += block->offset();
        for (uint32_t j = 0; j < sample_count && !should_stop(); ++j) {
            Vector2f position_sample = p + rs.next_2d();

            if (needs_aperture_sample)
                aperture_sample = rs.next_2d();

            Float time = sensor->shutter_open();
            if (needs_time_sample)
                time += rs.next_1d() * sensor->shutter_open_time();

            Spectrumf wavelengths, wav_weight;
            std::tie(wavelengths, wav_weight) =
                sensor->sample_wavelengths(rs.next_1d());

            auto adjusted_position =
                (position_sample - sensor->film()->crop_offset()) *
                inv_resolution;
            auto [ray, ray_weight] = sensor->sample_ray_differential(
                time, wavelengths, adjusted_position, aperture_sample);

            ray.scale_differential(diff_scale_factor);

            Spectrumf result;
            /* Integrator::eval */ {
                ScopedPhase sp(EProfilerPhase::ESamplingIntegratorEval);
                rs.depth = 0;
                result = eval(ray, rs);
            }
            /* ImageBlock::put */ {
                ScopedPhase sp(EProfilerPhase::EImageBlockPut);
                block->put(position_sample, ray.wavelengths,
                           ray_weight * wav_weight * result, rs.alpha);
            }
        }
    }
}

void SamplingIntegrator::render_block_vector(const Scene *scene,
                                             Sampler *sampler,
                                             ImageBlock *block,
                                             Point2fX &points,
                                             size_t sample_count_) const {
    const Sensor *sensor = scene->sensor();
    block->clear();
    uint32_t sample_count = (uint32_t)(
        sample_count_ == (size_t) -1 ? sampler->sample_count() : sample_count_);

    if (block->size().x() != block->size().y()
        || (uint32_t) hprod(block->size()) * sample_count
           != (uint32_t) slices(points)) {

        size_t max_pixel_count = m_block_size * m_block_size,
               n_entries = 0;
        // TODO: this scales poorly for large sample counts and / or CPU counts.
        set_slices(points, max_pixel_count * sample_count);
        for (uint32_t i = 0; i < max_pixel_count; ++i) {
            Point2u p = enoki::morton_decode<Point2u>(i);
            if (any(p >= block->size()))
                continue;
            Point2f pf = Point2f(p);

            for (uint32_t j = 0; j < sample_count; ++j)
                slice(points, n_entries++) = pf;
        }
        set_slices(points, n_entries);
    }

    Vector2f inv_resolution = 1.f / sensor->film()->crop_size();

    Vector2fP offset = block->offset(),
              aperture_sample = .5f;

    RadianceSample3fP rs(scene, sampler, sensor->medium());

    bool needs_aperture_sample = sensor->needs_aperture_sample(),
         needs_time_sample = sensor->shutter_open_time() == 0;

    Float diff_scale_factor = rsqrt((Float) sampler->sample_count());
    UInt32P index = arange<UInt32P>();

    for (size_t i = 0; i < packets(points) && !should_stop(); ++i) {
        MaskP active = index < slices(points);
        Vector2fP position_sample =
            packet(points, i) + offset + rs.next_2d(active);

        if (needs_aperture_sample)
            aperture_sample = rs.next_2d(active);

        FloatP time = sensor->shutter_open();
        if (needs_time_sample)
            time += rs.next_1d(active) * sensor->shutter_open_time();

        SpectrumfP wavelengths, wav_weight;
        std::tie(wavelengths, wav_weight) =
            sensor->sample_wavelengths(rs.next_1d(active), active);

        auto adjusted_position =
            (position_sample - sensor->film()->crop_offset()) * inv_resolution;
        auto [ray, ray_weight] = sensor->sample_ray_differential(
            time, wavelengths, adjusted_position, aperture_sample, active);

        ray.scale_differential(diff_scale_factor);

        SpectrumfP result;
        /* Integrator::eval */ {
            ScopedPhase p(EProfilerPhase::ESamplingIntegratorEvalP);
            rs.depth = 0;
            result = eval(ray, rs);
        }
        /* ImageBlock::put */ {
            ScopedPhase p(EProfilerPhase::EImageBlockPutP);
            block->put(position_sample, ray.wavelengths,
                       ray_weight * wav_weight * result, rs.alpha, active);
        }

        index += (uint32_t) PacketSize;
    }
}

// -----------------------------------------------------------------------------

MonteCarloIntegrator::MonteCarloIntegrator(const Properties &props)
    : SamplingIntegrator(props) {
    /// Depth to begin using russian roulette
    m_rr_depth = props.int_("rr_depth", 5);
    if (m_rr_depth <= 0)
        Throw("\"rr_depth\" must be set to a value greater than zero!");

    m_hide_emitters = props.bool_("hide_emitters", false);

    /** Longest visualized path depth (\c -1 = infinite).
     * A value of \c 1 will visualize only directly visible light sources.
     * \c 2 will lead to single-bounce (direct-only) illumination, and so on.
     */
    m_max_depth = props.int_("max_depth", -1);
    if (m_max_depth < 0 && m_max_depth != -1)
        Throw("\"max_depth\" must be set to -1 (infinite) or a value >= 0");
}

// -----------------------------------------------------------------------------

LightTracerIntegrator::LightTracerIntegrator(const Properties &props)
    : Integrator(props) {
    // TODO: consider moving those parameters to the base class.
    /// Depth to begin using russian roulette
    m_rr_depth = props.int_("rr_depth", 5);
    if (m_rr_depth <= 0)
        Throw("\"rr_depth\" must be set to a value greater than zero!");

    /** Longest visualized path depth (\c -1 = infinite).
     * A value of \c 1 will visualize only directly visible light sources.
     * \c 2 will lead to single-bounce (direct-only) illumination, and so on.
     */
    m_max_depth = props.int_("max_depth", -1);
    if (m_max_depth < 0 && m_max_depth != -1)
        Throw("\"max_depth\" must be set to -1 (infinite) or a value >= 0");
}

bool LightTracerIntegrator::render(Scene *scene, bool vectorize) {
    ScopedPhase sp(EProfilerPhase::ERender);
    m_stop = false;
    m_render_timer.reset();
    bool success;
    if (vectorize)
        success = render_impl<FloatP>(scene);
    else
        success = render_impl<Float>(scene);
    return success;
}

template <typename Value>
bool LightTracerIntegrator::render_impl(Scene *scene) {
    using RadianceSample        = RadianceSample<Point<Value, 3>>;
    constexpr bool IsVectorized = is_array<Value>::value;
    Film *film                  = scene->film();

    if (scene->emitters().empty()) {
        Log(EWarn,
            "Scene does not contain any emitter, returning black image.");
        normalize_film(film, 1);
        return true;
    }

    ThreadEnvironment env;
    tbb::spin_mutex mutex;
    ref<ProgressReporter> progress = new ProgressReporter("Rendering");

    size_t n_cores = util::core_count();
    /* Multiply sample count by pixel count to obtain a similar scale
     * to the standard path tracer. */
    size_t total_samples =
        scene->sampler()->sample_count() * hprod(scene->film()->size());
    size_t iteration_count = IsVectorized
                                 ? std::rint(total_samples / (Float) PacketSize)
                                 : total_samples;
    size_t grain_count = std::max(
        (size_t) 1, (size_t) std::rint(iteration_count / Float(n_cores * 2)));

    Log(EInfo, "Starting render job (%ix%i, %i sample%s, %i core%s)",
        film->crop_size().x(), film->crop_size().y(), total_samples,
        total_samples == 1 ? "" : "s", n_cores, n_cores == 1 ? "" : "s");
    if (m_timeout > 0.0f)
        Log(EInfo, "Timeout specified: %.2f seconds.", m_timeout);

    size_t total_samples_done = 0;
    tbb::parallel_for(
        tbb::blocked_range<size_t>(0, iteration_count, grain_count),
        [&](const tbb::blocked_range<size_t> &range) {
            ScopedSetThreadEnvironment set_env(env);

            // Thread-specific accumulators.
            // Enable warning, border and normalization for the image block.
            ref<ImageBlock> block =
                new ImageBlock(Bitmap::EXYZAW, film->size(), film->reconstruction_filter(), 0, true,
                               m_color_mode, true, true);

            size_t samples_done = 0;

            // Ensure that sample generation is fully deterministic.
            ref<Sampler> sampler = scene->sampler()->clone();
            size_t seed          = (size_t) range.begin();
            sampler->seed(seed);

            RadianceSample rs(scene, sampler);
            for (auto i = range.begin(); i != range.end() && !should_stop();
                 ++i) {
                rs.depth = 0;

                // Account for visible emitters
                sample_visible_emitters(rs, block);

                // Primary & further bounces illumination
                auto [ray, throughput] = prepare_ray(rs);
                trace_light_ray(ray, rs, throughput, block);

                samples_done += (enoki::is_array_v<Value> ? PacketSize : 1);
                if (samples_done > 10000) {
                    tbb::spin_mutex::scoped_lock lock(mutex);
                    total_samples_done += samples_done;
                    samples_done = 0;
                    progress->update(total_samples_done /
                                     (Float) total_samples);
                    notify(block->bitmap());
                }
            }

            /* locked */ {
                tbb::spin_mutex::scoped_lock lock(mutex);
                total_samples_done += samples_done;

                film->put(block);
            }
        });

    // Apply proper normalization.
    auto new_weight = normalize_film(film, total_samples_done);
    Log(EInfo, "Processed %d samples, normalization weight: %f",
        total_samples_done, new_weight);

    return true;
}

template <typename RadianceSample, typename Value>
auto LightTracerIntegrator::sample_visible_emitters_impl(
    RadianceSample &rs, ImageBlock *block, const mask_t<Value> &active) const {
    using Spectrum            = mitsuba::Spectrum<Value>;
    using EmitterPtr          = replace_scalar_t<Value, const Emitter *>;
    using Point2              = Point<Value, 2>;
    using Point3              = Point<Value, 3>;
    using SurfaceInteraction3 = SurfaceInteraction<Point3>;

    const auto *sensor = rs.scene->sensor();
    Value time         = sensor->shutter_open();
    if (sensor->shutter_open_time() > 0)
        time += rs.next_1d(active) * sensor->shutter_open_time();

    Value idx_sample = rs.next_1d(active);
    auto [emitter_idx, emitter_weight] =
        rs.scene->sample_emitter(idx_sample, active);
    EmitterPtr emitter =
        gather<EmitterPtr>(rs.scene->emitters().data(), emitter_idx, active);

    Point2 emitter_sample = rs.next_2d(active);
    auto [ps, pos_weight] =
        emitter->sample_position(time, emitter_sample, active);

    Value wavelength_sample = rs.next_1d(active);
    auto [wavelengths, wav_weight] =
        emitter->sample_wavelengths(wavelength_sample, active);

    rs.si       = SurfaceInteraction3(ps, wavelengths);
    rs.si.shape = emitter->shape();

    Spectrum weight = emitter_weight * pos_weight * wav_weight;

    // No BSDF passed (should not evaluate it since there's no scattering)
    return connect_sensor(rs, rs.si, nullptr, weight, block, active);
}
Spectrumf
LightTracerIntegrator::sample_visible_emitters(RadianceSample3f &rs,
                                               ImageBlock *block) const {
    return sample_visible_emitters_impl(rs, block, true);
}
SpectrumfP LightTracerIntegrator::sample_visible_emitters(RadianceSample3fP &rs,
                                                          ImageBlock *block,
                                                          MaskP active) const {
    return sample_visible_emitters_impl(rs, block, active);
}
#if defined(MTS_ENABLE_AUTODIFF)
SpectrumfD LightTracerIntegrator::sample_visible_emitters(
    RadianceSample3fD &rs, ImageBlock *block, const MaskD &active) const {
    return sample_visible_emitters_impl(rs, block, active);
}
#endif

MTS_IMPLEMENT_CLASS(Integrator, Object)
MTS_IMPLEMENT_CLASS(SamplingIntegrator, Integrator)
MTS_IMPLEMENT_CLASS(MonteCarloIntegrator, SamplingIntegrator)
MTS_IMPLEMENT_CLASS(LightTracerIntegrator, Integrator)

NAMESPACE_END(mitsuba)
