#include <mitsuba/render/medium.h>

#include <mitsuba/core/plugin.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/spectrum.h>
#include <mitsuba/render/phase.h>
#include <mitsuba/render/spectrum.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#include <mitsuba/render/scene.h>
#endif

NAMESPACE_BEGIN(mitsuba)

Medium::Medium() {}

Medium::Medium(const Properties &props)
    : m_id(props.id())
#if defined(MTS_ENABLE_AUTODIFF)
      , m_this_pointer(this_pointer<FloatD>())
#endif
     {

    for (auto &kv : props.objects()) {
        auto *phase =
            dynamic_cast<PhaseFunction *>(kv.second.get());
        if (phase) {
            m_phase_function = phase;
        }
    }
    if (!m_phase_function) {
        /* Create a default diffuse BSDF */
        m_phase_function =
            PluginManager::instance()->create_object<PhaseFunction>(
                Properties("isotropic"));
    }

    m_sample_emitters = props.bool_("sample_emitters", true);
}

Medium::~Medium() {}

std::string Medium::id() const { return m_id; }

std::tuple<ref<ContinuousSpectrum>, ref<ContinuousSpectrum>,
           ref<ContinuousSpectrum>, ref<ContinuousSpectrum>>
Medium::extract_medium_parameters(const Properties &props) {
    bool has_sigma_a_sigma_s =
        props.has_property("sigma_s") || props.has_property("sigma_a");
    bool has_sigma_t_albedo =
        props.has_property("sigma_t") || props.has_property("albedo");

    if ((has_sigma_a_sigma_s && has_sigma_t_albedo) || !(has_sigma_a_sigma_s || has_sigma_t_albedo))
        Log(EError, "You can either specify sigma_s & sigma_a *or* "
                    "sigma_t & albedo, but no other combinations!");

    ref<ContinuousSpectrum> sigma_s, sigma_a, sigma_t, albedo;
    if (has_sigma_a_sigma_s) {
        sigma_s = props.spectrum("sigma_s", 1.0f);
        sigma_a = props.spectrum("sigma_a", 1.0f);
        Properties props2("sum");
        props2.set_object("spectrum1", sigma_s.get());
        props2.set_object("spectrum2", sigma_a.get());
        props2.set_int("color_mode", props.int_("color_mode"));
        auto obj =
            PluginManager::instance()->create_object<ContinuousSpectrum>(props2);
        sigma_t = (ContinuousSpectrum *) obj.get();
        Properties props3("division");
        props3.set_object("spectrum1", sigma_s.get());
        props3.set_object("spectrum2", sigma_t.get());
        props3.set_int("color_mode", props.int_("color_mode"));
        auto obj2 =
            PluginManager::instance()->create_object<ContinuousSpectrum>(
                props3);
        albedo = (ContinuousSpectrum *) obj2.get();
    } else if (has_sigma_t_albedo) {
        sigma_t = props.spectrum("sigma_t", 1.f);
        albedo  = props.spectrum("albedo", 1.f);
        Properties props2("product");
        props2.set_object("spectrum1", sigma_t.get());
        props2.set_object("spectrum2", albedo.get());
        props2.set_int("color_mode", props.int_("color_mode"));
        auto obj =
            PluginManager::instance()->create_object<ContinuousSpectrum>(props2);
        sigma_s = (ContinuousSpectrum *) obj.get();
        Properties props3("difference");
        props3.set_object("spectrum1", sigma_t.get());
        props3.set_object("spectrum2", sigma_s.get());
        props3.set_int("color_mode", props.int_("color_mode"));
        auto obj2 =
            PluginManager::instance()->create_object<ContinuousSpectrum>(
                props3);
        sigma_a = (ContinuousSpectrum *) obj2.get();
    }
    return { sigma_a, sigma_s, sigma_t, albedo };
}


#if defined(MTS_ENABLE_AUTODIFF)
void Medium::set_labels(const std::string &prefix) const {
    auto ptr = this_pointer<FloatD>();
    set_label(ptr, (prefix + ".m_this").c_str());
    cuda_eval();
}

void Medium::set_globals(const std::string &prefix, const RTcontext &context,
                         const RTmaterial &material) const {
    std::string name = prefix + "_m_this";
    RTvariable var = get_or_create_optix_var(name, context, material);
    auto *ptr_value = this;
    rtVariableSetUserData(var, sizeof(ptr_value), &ptr_value);
    cuda_eval();
}
#endif


MTS_IMPLEMENT_CLASS(Medium, Object)
NAMESPACE_END(mitsuba)
