#include <mitsuba/core/thread.h>
#include <mitsuba/core/tls.h>
#include <mitsuba/python/python.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/integrator.h>

// TODO: move all signal-related mechanisms to libcore?
#if defined(__APPLE__) || defined(__linux__)
#  define MTS_HANDLE_SIGINT 1
#else
#  define MTS_HANDLE_SIGINT 0
#endif

#if MTS_HANDLE_SIGINT
#include <signal.h>

/// Integrator currently in use (has to be global)
ThreadLocal<Integrator *> current_integrator;
/// Previously installed signal handler
static void (*sigint_handler_prev)(int) = nullptr;
/// Custom signal handler definition
static void sigint_handler(int sig) {
    Log(EWarn, "Received interrupt signal, winding down..");
    (*current_integrator).cancel();
    signal(sig, sigint_handler_prev);
    raise(sig);
}
#endif

MTS_PY_EXPORT(Integrator) {
    /// Integrator (base class).
    MTS_PY_CLASS(Integrator, Object)
        .def("render",
             [&](Integrator &integrator, Scene *scene, bool vectorize) {
                 py::gil_scoped_release release;

#if MTS_HANDLE_SIGINT
                 // Install new signal handler
                 current_integrator  = &integrator;
                 sigint_handler_prev = signal(SIGINT, sigint_handler);
#endif

                 bool res = integrator.render(scene, vectorize);

#if MTS_HANDLE_SIGINT
                 // Restore previous signal handler
                 signal(SIGINT, sigint_handler_prev);
#endif

                 return res;
             },
             D(Integrator, render), "scene"_a, "vectorize"_a)
        .mdef(Integrator, cancel)
        .mdef(Integrator, should_stop)
        .mdef(Integrator, register_callback, "cb"_a, "period"_a)
        .mdef(Integrator, remove_callback, "index"_a)
        .mdef(Integrator, callback_count)
        .mdef(Integrator, notify, "bitmap"_a, "extra"_a);

    /// SamplingIntegrator.
    MTS_PY_CLASS(SamplingIntegrator, Integrator)
        .def("eval",
             py::overload_cast<const RayDifferential3f &, RadianceSample3f &, bool>(
                 &SamplingIntegrator::eval, py::const_),
             D(SamplingIntegrator, eval), "ray"_a, "rs"_a, "unused"_a = true)
#if 0
         .def("eval",
              enoki::vectorize_wrapper(
                  py::overload_cast<const RayDifferential3fP &,
                                    RadianceSample3fP &, MaskP>(
                      &SamplingIntegrator::eval, py::const_)),
              "ray"_a, "rs"_a, "active"_a = true)
#endif
#if defined(MTS_ENABLE_AUTODIFF)
        .def("eval",
             py::overload_cast<const RayDifferential3fD &, RadianceSample3fD &, MaskD>(
                 &SamplingIntegrator::eval, py::const_),
             "ray"_a, "rs"_a, "active"_a = true)
#endif
        ;

    /// MonteCarloIntegrator.
    MTS_PY_CLASS(MonteCarloIntegrator, SamplingIntegrator)
        .def("max_depth", &MonteCarloIntegrator::max_depth)
        .def("hide_emitters", &MonteCarloIntegrator::hide_emitters)
        .def("set_hide_emitters", &MonteCarloIntegrator::set_hide_emitters, "hide"_a)
        ;

    /// LightTracerIntegrator.
    MTS_PY_CLASS(LightTracerIntegrator, Integrator)
        .def("sample_visible_emitters",
             py::overload_cast<RadianceSample3f &, ImageBlock *, bool>(
                 &LightTracerIntegrator::sample_visible_emitters, py::const_),
             "rs"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, sample_visible_emitters))
        .def("sample_visible_emitters",
             py::overload_cast<RadianceSample3fP &, ImageBlock *, MaskP>(
                 &LightTracerIntegrator::sample_visible_emitters, py::const_),
             "rs"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, sample_visible_emitters, 2))
#if defined(MTS_ENABLE_AUTODIFF)
        .def(
            "sample_visible_emitters",
            py::overload_cast<RadianceSample3fD &, ImageBlock *, const MaskD &>(
                &LightTracerIntegrator::sample_visible_emitters, py::const_),
            "rs"_a, "block"_a, "active"_a = true,
            D(LightTracerIntegrator, sample_visible_emitters, 3))
#endif

        .def("prepare_ray",
             py::overload_cast<RadianceSample3f &, bool>(
                 &LightTracerIntegrator::prepare_ray, py::const_),
             "rs"_a, "active"_a = true, D(LightTracerIntegrator, prepare_ray))
        .def("prepare_ray",
             py::overload_cast<RadianceSample3fP &, const MaskP &>(
                 &LightTracerIntegrator::prepare_ray, py::const_),
             "rs"_a, "active"_a = true,
             D(LightTracerIntegrator, prepare_ray, 2))
#if defined(MTS_ENABLE_AUTODIFF)
        .def("prepare_ray",
             py::overload_cast<RadianceSample3fD &, const MaskD &>(
                 &LightTracerIntegrator::prepare_ray, py::const_),
             "rs"_a, "active"_a = true,
             D(LightTracerIntegrator, prepare_ray, 3))
#endif

        .def("trace_light_ray",
             py::overload_cast<Ray3f, RadianceSample3f &, Spectrumf,
                               ImageBlock *, bool>(
                 &LightTracerIntegrator::trace_light_ray, py::const_),
             "ray"_a, "rs"_a, "throughput"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, trace_light_ray))
        .def("trace_light_ray",
             py::overload_cast<Ray3fP, RadianceSample3fP &,
                               SpectrumfP, ImageBlock *, MaskP>(
                 &LightTracerIntegrator::trace_light_ray, py::const_),
             "ray"_a, "rs"_a, "throughput"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, trace_light_ray, 2))
#if defined(MTS_ENABLE_AUTODIFF)
        .def("trace_light_ray",
             py::overload_cast<Ray3fD, RadianceSample3fD &,
                               SpectrumfD, ImageBlock *, const MaskD &>(
                 &LightTracerIntegrator::trace_light_ray, py::const_),
             "ray"_a, "rs"_a, "throughput"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, trace_light_ray, 3))
#endif

        .def("connect_sensor",
             py::overload_cast<RadianceSample3f &, SurfaceInteraction3f &,
                               const BSDF *, const Spectrumf &, ImageBlock *>(
                 &LightTracerIntegrator::connect_sensor, py::const_),
             "rs"_a, "si"_a, "bsdf"_a, "weight"_a, "block"_a,
             D(LightTracerIntegrator, connect_sensor))
        .def("connect_sensor",
             py::overload_cast<RadianceSample3fP &, SurfaceInteraction3fP &,
                               typename LightTracerIntegrator::BSDFPtrP,
                               const SpectrumfP &, ImageBlock *, const MaskP &>(
                 &LightTracerIntegrator::connect_sensor, py::const_),
             "rs"_a, "si"_a, "bsdf"_a, "weight"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, connect_sensor, 2))
#if defined(MTS_ENABLE_AUTODIFF)
        .def("connect_sensor",
             py::overload_cast<RadianceSample3fD &, SurfaceInteraction3fD &,
                               typename LightTracerIntegrator::BSDFPtrD,
                               const SpectrumfD &, ImageBlock *, const MaskD &>(
                 &LightTracerIntegrator::connect_sensor, py::const_),
             "rs"_a, "si"_a, "bsdf"_a, "weight"_a, "block"_a, "active"_a = true,
             D(LightTracerIntegrator, connect_sensor, 3))
#endif


        .mdef(LightTracerIntegrator, normalize_block, "block"_a, "total_samples"_a)
        .mdef(LightTracerIntegrator, normalize_film, "film"_a, "total_samples"_a)
        ;
}
