#include <mitsuba/core/properties.h>
#include <mitsuba/python/python.h>
#include <mitsuba/render/integrator.h>
#include <mitsuba/render/kdtree.h>
#include <mitsuba/render/mesh.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/sensor.h>

using Index = Shape::Index;

MTS_PY_EXPORT(ShapeKDTree) {
#if !defined(MTS_USE_EMBREE)
    MTS_PY_CLASS(ShapeKDTree, Object)
        .def(py::init<const Properties &>(), D(ShapeKDTree, ShapeKDTree))
        .mdef(ShapeKDTree, add_shape)
        .mdef(ShapeKDTree, primitive_count)
        .mdef(ShapeKDTree, shape_count)
        .def("shape", (Shape *(ShapeKDTree::*)(size_t)) &ShapeKDTree::shape, D(ShapeKDTree, shape))
        .def("__getitem__", [](ShapeKDTree &s, size_t i) -> py::object {
            if (i >= s.primitive_count())
                throw py::index_error();
            Shape *shape = s.shape(i);
            if (shape->class_()->derives_from(MTS_CLASS(Mesh)))
                return py::cast(static_cast<Mesh *>(s.shape(i)));
            else
                return py::cast(s.shape(i));
        })
        .def("__len__", &ShapeKDTree::primitive_count)
        .def("bbox", [] (ShapeKDTree &s) { return s.bbox(); })
        .mdef(ShapeKDTree, build);
#else
    ENOKI_MARK_USED(m);
#endif
}

MTS_PY_EXPORT(Scene) {
    MTS_PY_CLASS(Scene, Object)
        .def(py::init<const Properties>())

        // Full intersection
        .def("ray_intersect",
             py::overload_cast<const Ray3f &, bool>(&Scene::ray_intersect,
                                                    py::const_),
             "ray"_a, "unused"_a = true, D(Scene, ray_intersect))
        .def("ray_intersect",
             enoki::vectorize_wrapper(py::overload_cast<const Ray3fP &, MaskP>(
                 &Scene::ray_intersect, py::const_)),
             "ray"_a, "active"_a = true)

        // Shadow rays
        .def("ray_test",
             py::overload_cast<const Ray3f &, bool>(&Scene::ray_test,
                                                    py::const_),
             "ray"_a, "unused"_a = true, D(Scene, ray_test))
        .def("ray_test",
             enoki::vectorize_wrapper(py::overload_cast<const Ray3fP &, MaskP>(
                 &Scene::ray_test, py::const_)),
             "ray"_a, "active"_a = true)

#if !defined(MTS_USE_EMBREE)
        // Full intersection (brute force, for testing)
        .def("ray_intersect_naive",
             py::overload_cast<const Ray3f &, bool>(&Scene::ray_intersect_naive,
                                                    py::const_),
             "ray"_a, "unused"_a = true)
        .def("ray_intersect_naive",
             enoki::vectorize_wrapper(py::overload_cast<const Ray3fP &, MaskP>(
                 &Scene::ray_intersect_naive, py::const_)),
             "ray"_a, "active"_a = true)
#endif

        // Sampling
        .def("sample_emitter",
             py::overload_cast<Float>(&Scene::sample_emitter, py::const_),
             "sample"_a, D(Scene, sample_emitter))
        .def("sample_emitter",
             py::overload_cast<FloatP, MaskP>(&Scene::sample_emitter, py::const_),
             "sample"_a, "active"_a = true, D(Scene, sample_emitter, 2))
#if defined(MTS_ENABLE_AUTODIFF)
        .def("sample_emitter",
             py::overload_cast<const FloatD&, const MaskD&>(
                &Scene::sample_emitter, py::const_),
             "sample"_a, "active"_a = true, D(Scene, sample_emitter, 3))
#endif

        .def("sample_emitter_direction",
             py::overload_cast<const Interaction3f &, const Point2f &, bool>(
                 &Scene::sample_emitter_direction, py::const_),
             "ref"_a, "sample"_a, "test_visibility"_a = true,
             D(Scene, sample_emitter_direction))
        .def("sample_emitter_direction",
             py::overload_cast<const Interaction3fP &, const Point2fP &, bool, MaskP>(
                 &Scene::sample_emitter_direction, py::const_),
             "ref"_a, "sample"_a, "test_visibility"_a = true, "active"_a = true,
             D(Scene, sample_emitter_direction, 2))

        .def("sample_emitter_ray",
             py::overload_cast<Float, Float, const Point2f &, const Point2f &>(
                 &Scene::sample_emitter_ray, py::const_),
             "time"_a, "sample1"_a, "sample2"_a, "sample3"_a,
             D(Scene, sample_emitter_ray))
        .def("sample_emitter_ray",
             py::overload_cast<FloatP, FloatP, const Point2fP &,
                               const Point2fP &, MaskP>(
                 &Scene::sample_emitter_ray, py::const_),
             "time"_a, "sample1"_a, "sample2"_a, "sample3"_a, "active"_a = true,
             D(Scene, sample_emitter_ray, 2))
#if defined(MTS_ENABLE_AUTODIFF)
        .def("sample_emitter_ray",
             py::overload_cast<const FloatD&, const FloatD&, const Point2fD &,
                               const Point2fD &, const MaskD&>(
                 &Scene::sample_emitter_ray, py::const_),
             "time"_a, "sample1"_a, "sample2"_a, "sample3"_a, "active"_a = true,
             D(Scene, sample_emitter_ray, 3))
#endif

        .def(
            "pdf_emitter_direction",
            py::overload_cast<const Interaction3f &, const DirectionSample3f &>(
                &Scene::pdf_emitter_direction, py::const_),
            "ref"_a, "ds"_a)
        .def("pdf_emitter_direction",
             py::overload_cast<const Interaction3fP &,
                               const DirectionSample3fP &, MaskP>(
                 &Scene::pdf_emitter_direction, py::const_),
             "ref"_a, "ds"_a, "active"_a = true)

        // Accessors
#if !defined(MTS_USE_EMBREE)
        .def("kdtree",     py::overload_cast<>(&Scene::kdtree))
#endif

#if defined(MTS_USE_OPTIX)
        .def("optix_init", [](Scene &scene, size_t n_shapes) {
                scene.optix_init(n_shapes);
            }, "n_shapes"_a)
        .def("optix_release", [](Scene &scene) {
                scene.optix_release();
            })
        .def("optix_register", [](Scene &scene, Shape *shape) {
                scene.optix_register(shape);
            }, "shape"_a)
        .def("optix_build", [](Scene &scene) {
                scene.optix_build();
            })
#endif
        .def("bbox", &Scene::bbox, D(Scene, bbox))
        .def("sensor",     py::overload_cast<>(&Scene::sensor),     D(Scene, sensor))
        .def("sensors",     py::overload_cast<>(&Scene::sensors),     D(Scene, sensors))
        .mdef(Scene, set_current_sensor, "index"_a)
        .def("emitters",     py::overload_cast<>(&Scene::emitters),     D(Scene, emitters))
        .mdef(Scene, environment)
        .def("film",       py::overload_cast<>(&Scene::film),       D(Scene, film))
        .def("sampler",    py::overload_cast<>(&Scene::sampler),    D(Scene, sampler))
        .def("shapes",     py::overload_cast<>(&Scene::shapes),     D(Scene, shapes))
        .def("integrator", [](Scene &scene) {
#define PY_CAST(Name) {                                                        \
        Name *temp = dynamic_cast<Name *>(o);                                  \
        if (temp)                                                              \
            return py::cast(temp);                                             \
    }
            Integrator *o = scene.integrator();
            PY_CAST(LightTracerIntegrator);
            PY_CAST(MonteCarloIntegrator);
            PY_CAST(SamplingIntegrator);
            return py::cast(o);

#undef PY_CAST
        }, D(Scene, integrator))
        .def("color_mode", &Scene::color_mode)
        .def("__repr__", &Scene::to_string);
}
