#include <enoki/stl.h>

#include <mitsuba/core/frame.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/texture3d.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#endif

NAMESPACE_BEGIN(mitsuba)

/**
 * Purely absorptive homogeneous medium.
 */
class HomogeneousAbsorptiveMedium final : public Medium {
public:
    HomogeneousAbsorptiveMedium(const Properties &props) : Medium(props) {
        m_sigma_t = props.spectrum("sigma_t", 1.0f);
        m_density = props.float_("density", 1.0f);

#if defined(MTS_ENABLE_AUTODIFF)
        m_density_d = m_density;
#endif
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override {
        dp.put(this, "density", m_density_d);
    }

    void parameters_changed() override {
        m_needs_reupload = true;
    }
#endif

    MTS_AUTODIFF_GETTER(density_v, m_density, m_density_d)

    template <typename Interaction3, typename Value = typename Interaction3::Value>
    Value density_impl(const Interaction3 & /*it*/, mask_t<Value> /*active*/) const {
        return density_v<Value>();
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value               = typename Ray::Value,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename Spectrum            = Spectrum<Value>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_impl(const Scene *scene,
                                         const Ray &ray,
                                         const Point2 & sample,
                                         Sampler *sampler,
                                         Index index,
                                         mask_t<Value> active) const {
        SurfaceInteraction3 si = scene->ray_intersect(ray, active);
        auto [mi, throughput]  = sample_distance_with_si(si, ray, sample, sampler, index, active);
        // A purely absorptive medium never samples a distance in the medium.
        Assert(!any(mi.is_valid()));
        return std::make_tuple(si, mi, throughput);
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value, typename Spectrum = Spectrum<Value>,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_with_si_impl(const SurfaceInteraction3 &si, const Ray &ray,
                                                 const Point2 & /*sample*/, Sampler *sampler,
                                                 Index index, mask_t<Value> active) const {
        using MediumInteraction3 = MediumInteraction<Point3>;

        // A purely absorptive medium always samples t = +infinity.
        // The throughput is affected by absorption, though.
        auto mi = zero<MediumInteraction3>();
        mi.t = math::Infinity;

        // Note: (mint, maxt) should correspond to the distance travelled inside.
        Ray tr_ray = ray;
        tr_ray.mint = 0;
        tr_ray.maxt = si.t;
        auto throughput = eval_transmittance(tr_ray, sampler, index, active);
        return std::make_pair(mi, throughput);
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value>
    MTS_INLINE auto eval_transmittance_impl(const Ray &ray, Sampler * /*unused*/,
                                            uint_array_t<Value> /*unused*/,
                                            mask_t<Value> /*active*/) const {
        auto density_val = density_v<Value>();
        auto sigma_t     = m_sigma_t->eval(ray.wavelengths);

        // Note: we assume that (mint, maxt) correspond to the bounding box of the volume.
        auto d = ray.maxt - ray.mint;
        return enoki::exp(-d * sigma_t * density_val);
    }

    template <typename Ray3, typename Value = typename Ray3::Value>
    Value pdf_distance_impl(const Ray3 & /*ray*/, Value /*t*/, mask_t<Value> /*active*/) const {
        // Only ever samples t = +inf.
        // TODO: check t against boundary of medium.
        return 1.f;
    }

    template <typename MediumInteraction3, typename Value = typename MediumInteraction3::Value>
    MTS_INLINE auto eval_scattering_impl(const MediumInteraction3 & /*mi*/,
                                         mask_t<Value> /*active*/) const {
        // No scattering
        return Spectrum<Value>(0.f) + 0.f;  // Workaround
    }

    template <typename Interaction3, typename Value = typename Interaction3::Value>
    MTS_INLINE auto eval_albedo_impl(const Interaction3 & /*it*/, mask_t<Value> /*active*/) const {
        // No scattering -> zero albedo
        return Spectrum<Value>(0.f) + 0.f;  // Workaround
    }

    bool is_homogeneous() const override { return true; }

    std::vector<ref<Object>> children() override {
        if (m_sigma_t->id().empty())
            m_sigma_t->set_id(this->id() + "_sigma_t");
        return { m_sigma_t.get() };
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "HomogeneousAbsorptiveMedium[" << std::endl
            << "  m_sigma_t = " << string::indent(m_sigma_t) << std::endl
            << "  m_density = " << string::indent(m_density) << std::endl
            << "]";
        return oss.str();
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        Medium::set_labels(prefix);

        m_sigma_t->set_labels(prefix + ".m_sigma_t");
        set_label(m_density_d, (prefix + ".m_density_d").c_str());
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        Medium::set_globals(prefix, context, material);
        m_sigma_t->set_globals(prefix + "_m_sigma_t", context, material);

        if (!m_needs_reupload)
            return;

        cuda_eval();

        // Get current value from buffer
        // TODO: it would be much nicer to avoid this round-trip
        float m_density;
        cuda_memcpy_from_device(&m_density, m_density_d.data(), 1 * sizeof(float));
        cuda_eval();

        std::string name = prefix + "_m_density_d";
        RTvariable var = get_or_create_optix_var(name, context, material);
        rtVariableSet1f(var, m_density);
        cuda_eval();
        m_needs_reupload = false;
    }
#endif

    MTS_IMPLEMENT_MEDIUM_ALL()
    MTS_DECLARE_CLASS()
private:
    ref<ContinuousSpectrum> m_sigma_t;
    Float m_density;
#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_density_d;
    mutable bool m_needs_reupload = true;
#endif
};

MTS_IMPLEMENT_CLASS(HomogeneousAbsorptiveMedium, Medium)
MTS_EXPORT_PLUGIN(HomogeneousAbsorptiveMedium, "Homogeneous absorptive medium")

NAMESPACE_END(mitsuba)
