#include <enoki/stl.h>

#include <mitsuba/core/frame.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/texture3d.h>

NAMESPACE_BEGIN(mitsuba)

class HeterogeneousMedium final : public Medium {
public:
    HeterogeneousMedium(const Properties &props) : Medium(props) {
        m_sigma_t               = props.spectrum("sigma_t", 1.f);
        m_density_scale         = props.float_("density_scale", 1.f);
        m_use_ratio_tracking    = props.bool_("use_ratio_tracking", false);
        m_use_raymarching       = props.bool_("use_raymarching", false);
        m_step_size_scaling     = props.float_("raymarching_step_scaling", 1.f);
        m_density_tex           = props.texture3d("density", 1.f);
        m_albedo_tex            = props.texture3d("albedo", 1.f);

        // TODO: Should also get the maximum of sigmaT
        m_max_density     = m_density_scale * m_density_tex->max();
        m_inv_max_density = 1.f / m_max_density;
        m_density_aabb    = m_density_tex->bbox();

        // This is just for information purposes only
        Float step_size = get_step_size();
        Float diag_norm = norm(m_density_aabb.max - m_density_aabb.min);
        Log(EInfo, "Volume resolution %s", m_density_tex->resolution());
        Log(EInfo, "Step size: %f", step_size);
        Log(EInfo, "Max steps: %d", int(diag_norm / step_size + 1));
    }


    template <typename Interaction3, typename Value = typename Interaction3::Value>
    Value density_impl(const Interaction3 &it, mask_t<Value> active) const {
        // TODO: Support spectrally varying density here
        return m_density_scale * m_density_tex->eval(it, active).coeff(0);
    }

    template <typename Value, typename Index, typename Mask>
    Value next_1d(Sampler *sampler, Index index, Mask active) const {
        ENOKI_MARK_USED(index);

#if defined(MTS_ENABLE_AUTODIFF)
        if constexpr (is_diff_array_v<Value>)
            return sampler->next_1d_d(index, active);
        else
#endif
            return sampler->next<Value>(active);
    }

    MTS_INLINE Float get_step_size() const {
        Vector3f diag = m_density_aabb.max - m_density_aabb.min;
        Vector3f res  = enoki::max(3, Vector3f(m_density_tex->resolution()));
        return m_step_size_scaling * hmin(diag * rcp(res));
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value, typename Spectrum = Spectrum<Value>,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_impl(const Scene *scene, const Ray &ray, const Point2 &sample,
                                         Sampler *sampler, Index index,
                                         mask_t<Value> active) const {
        // TODO: remove index argument, just pass random variable

        // Intersect the ray with the scene. Note: We do not use short ray
        // optimization here since it would increase noise in combination with
        // delta tracking
        SurfaceInteraction3 si = scene->ray_intersect(ray, active);
        auto [mi, throughput]  = sample_distance_with_si(si, ray, sample, sampler, index, active);
        masked(si.t, mi.is_valid()) = math::Infinity;
        return std::make_tuple(si, mi, throughput);
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value, typename Spectrum = Spectrum<Value>,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_with_si_impl(const SurfaceInteraction3 &si, const Ray &ray,
                                                 const Point2 &sample, Sampler *sampler,
                                                 Index index, mask_t<Value> active) const {
        using Vector3            = mitsuba::Vector<Value, 3>;
        using Frame              = Frame<Vector3>;
        using Mask               = mask_t<Value>;
        using Interaction3       = mitsuba::Interaction<Point3>;
        using MediumInteraction3 = mitsuba::MediumInteraction<Point3>;

        Float step_size = get_step_size();
        Float half_step_size = 0.5f * step_size;

        Value mean_sigma_t          = enoki::mean(m_sigma_t->eval(ray.wavelengths));
        auto [aabb_its, aabb_mint, aabb_maxt] = m_density_aabb.ray_intersect(ray);
        Value mint                            = enoki::max(ray.mint, aabb_mint);
        Value maxt                            = enoki::min(ray.maxt, aabb_maxt);
        active &= aabb_its;

        Interaction3 lookup_it;
        lookup_it.p           = ray(mint);
        lookup_it.wavelengths = ray.wavelengths;
        lookup_it.time        = ray.time;

        MediumInteraction3 mi;
        mi.p           = Point3(0.f, 0.f, 0.f);
        mi.sh_frame    = Frame(ray.d);
        mi.wi          = Vector3(0.f, 0.f, 1.f);
        mi.time        = ray.time;
        mi.wavelengths = ray.wavelengths;
        mi.medium      = nullptr;
        mi.t           = math::Infinity;

        if (m_use_raymarching) {
            Vector3f diag   = m_density_aabb.max - m_density_aabb.min;
            Float diag_norm = norm(diag);

            Value desired_density = -enoki::log(1 - sample.x());

            int max_steps = diag_norm / step_size + 1;

            Value t_a                = mint;
            lookup_it.p              = ray(t_a);
            Value f_a                = density(lookup_it, active) * mean_sigma_t;
            Value max_t              = enoki::min(maxt, si.t);
            Value integrated_density = zero<Value>();
            Mask reached_density     = false;

            // Jitter: Perform first step outside of loop
            Value t_b         = t_a + sample.y() * step_size;
            lookup_it.p       = ray(t_b);
            Value f_b         = density(lookup_it, active) * mean_sigma_t;
            Value new_density = 0.5f * (t_b - t_a) * (f_a + f_b);
            reached_density |= active && (new_density >= desired_density);
            reached_density |= (new_density >= desired_density);
            active &= !(active && (reached_density || (t_b > max_t)));
            masked(integrated_density, active) = new_density;
            masked(t_a, active)                = t_b;
            masked(f_a, active)                = f_b;
            mint                               = t_b;

            for (int i = 1; i < max_steps; ++i) {
                masked(t_b, active) = fmadd(i, step_size, mint);
                lookup_it.p         = ray(t_b);
                masked(f_b, active) = density(lookup_it, active) * mean_sigma_t;

                Value new_density = fmadd(half_step_size, f_a + f_b, integrated_density);

                reached_density |= active && (new_density >= desired_density);
                reached_density |= (new_density >= desired_density);
                active &= !reached_density && (t_b <= max_t);

                masked(integrated_density, active) = new_density;
                masked(t_a, active)                = t_b;
                masked(f_a, active)                = f_b;
            }

            // Solve quadratic equation to get exact intersection location
            Value a = 0.5f * (f_b - f_a);
            Value b = f_a;
            Value c = (integrated_density - desired_density) * rcp(t_b - t_a);
            auto [has_solution, solution1, solution2] = math::solve_quadratic(a, b, c);
            // Assert(none(active && reached_density && !has_solution));
            Value interp = select(solution1 >= 0.f && solution1 <= 1.f, solution1, solution2);
            Value sampled_t = t_a + (t_b - t_a) * interp;
            // Assert(none(active && reached_density && (interp < 0.f || interp > 1.f)));
            Value f_c = (1 - interp) * f_a + interp * f_b;

            // Update integrated_density for rays which do not interact with the medium
            Mask escaped =
                (!reached_density && (t_b > max_t)) || (reached_density && (sampled_t > max_t));
            lookup_it.p = ray(max_t - math::Epsilon);
            masked(integrated_density, escaped) +=
                0.5f * (f_a + density(lookup_it, active) * mean_sigma_t);

            // Record medium interaction if the generated "t" is within the range of valid "t"
            Mask valid_mi = reached_density && (sampled_t <= max_t);
            // Assert(none(valid_mi && f_c <= 0.f));

            // Compute transmittance
            Value tr                             = exp(-integrated_density);
            masked(integrated_density, valid_mi) = desired_density;
            masked(mi.t, valid_mi) = detach(sampled_t);
            masked(mi.p, valid_mi) = ray(mi.t);
            masked(mi.medium, valid_mi) = this_pointer<Value>();

            Spectrum throughput = 1.f;
            Value pdf_tr = f_c * tr;
            masked(throughput, valid_mi) *= m_albedo_tex->eval(mi, valid_mi) * select(pdf_tr > 0.f, f_c * tr / detach(pdf_tr), 1.f);
            masked(throughput, si.is_valid()) *= select(tr > 0.f, tr / detach(tr), 1.f);
            return std::make_pair(mi, throughput);
        } // ------ done ray marching

        // Delta tracking
        Spectrum throughput = 1.f;
        Value t = mint;
        while (any(active)) {

            // 1. Sample tentative free-flight distance using max. density
            Value u = next_1d<Value>(sampler, index, active);
            t += -enoki::log(1 - u) * m_inv_max_density;

            // 2. If sampled distance exceeds distance to surface:
            active &= (t <= si.t) && (t <= maxt);

            // 3. Check whether we sampled a real or virtual particle interaction
            lookup_it.p = ray(t);
            Value d     = density(lookup_it, active) * mean_sigma_t;

            Value p_s     = d * m_inv_max_density;
            Value p_n     = 1.f - p_s;
            Mask valid_mi = active && (next_1d<Value>(sampler, index, active) < p_s);

            masked(mi.t, valid_mi) = t;
            masked(mi.p, valid_mi) = ray(mi.t);
            masked(mi.medium, valid_mi) = this_pointer<Value>();
            masked(throughput, valid_mi) *= p_s / detach(p_s) * m_albedo_tex->eval(mi, valid_mi);
            masked(throughput, active && !valid_mi) *= p_n / detach(p_n);
            active &= !valid_mi;
        }
        return std::make_pair(mi, throughput);
    }


    template <typename Ray, typename Value = typename Ray::Value>
    MTS_INLINE auto integrate_density(const Ray &ray, mask_t<Value> active) const {
        using Mask         = mask_t<Value>;
        using Interaction3 = mitsuba::Interaction<Point<Value, 3>>;
        Assert(m_use_raymarching);

        Value mean_sigma_t = enoki::mean(m_sigma_t->eval(ray.wavelengths));

        auto [aabb_its, aabb_mint, aabb_maxt] = m_density_aabb.ray_intersect(ray);
        Value maxt = enoki::min(aabb_maxt - math::Epsilon, ray.maxt);
        Value mint = enoki::max(ray.mint, aabb_mint + math::Epsilon);
        active &= aabb_its && maxt > mint;

        Interaction3 lookup_it;
        lookup_it.p           = ray(mint);
        lookup_it.wavelengths = ray.wavelengths;
        lookup_it.time        = ray.time;

        Float step_size = get_step_size();
        Vector3f diag   = m_density_aabb.max - m_density_aabb.min;
        Float diag_norm = norm(diag);
        int max_steps   = 1 + diag_norm / step_size;
        Value t_a       = mint;
        lookup_it.p     = ray(t_a);
        Value f_a       = density(lookup_it, active) * mean_sigma_t;

        Value result = zero<Value>();
        Value last_interpolant = f_a;
        for (int i = 1; i < max_steps; ++i) {
            Value t_b                 = fmadd(i, step_size, mint);
            Mask reached_maxt         = t_b >= maxt;
            masked(t_b, reached_maxt) = maxt;
            lookup_it.p               = ray(t_b);
            Value f_b                 = density(lookup_it, active) * mean_sigma_t;

            masked(last_interpolant, active) = 0.5f * (f_a + f_b);
            result += select(active, last_interpolant * (t_b - t_a), 0.f);
            // result += 0.5f * (f_a + f_b) * (t_b - t_a);
            active &= !reached_maxt;
            t_a = t_b;
            f_a = f_b;
        }
        return std::make_pair(result, last_interpolant);
    }


    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value, typename Index = uint_array_t<Value>,
              typename Spectrum = Spectrum<Value>>
    MTS_INLINE auto eval_transmittance_impl(const Ray &ray, Sampler *sampler, Index index,
                                            mask_t<Value> active) const {
        using Mask         = mask_t<Value>;
        using Interaction3 = mitsuba::Interaction<Point3>;

        if (m_use_raymarching) {
            // ----- Ray marching
            // For verification against a homogeneous medium with some known density:
            // auto density_val      = Value(5.f);
            // auto d                = maxt - mint;
            // auto expected_density = d * mean_sigma_t * density_val;
            // Log(EInfo, "integrated_density = %s, expected_density = %s", integrated_density,
            //     expected_density);

            auto [integrated_density, interpolant] = integrate_density(ray, active);
            return Spectrum(enoki::exp(-integrated_density));

        } else {
            // ----- Delta tracking
            Value mean_sigma_t = enoki::mean(m_sigma_t->eval(ray.wavelengths));

            auto [aabb_its, aabb_mint, aabb_maxt] = m_density_aabb.ray_intersect(ray);
            active &= aabb_its;
            Value maxt = enoki::min(aabb_maxt - math::Epsilon, ray.maxt);
            Value mint = enoki::max(ray.mint, aabb_mint + math::Epsilon);
            Value t    = mint;

            Interaction3 lookup_it;
            lookup_it.p           = ray(mint);
            lookup_it.wavelengths = ray.wavelengths;
            lookup_it.time        = ray.time;

            Spectrum tr(1.f);
            while (any(active)) {
                // 1. Sample tentative free-flight distance using max. density
                Value u = next_1d<Value>(sampler, index, active);
                t += -enoki::log(1 - u) * m_inv_max_density;

                // 2. If sampled distance exceeds max distance: everything good
                active &= t <= maxt;

                lookup_it.p = ray(t);
                Value d     = density(lookup_it, active) * mean_sigma_t;
                Value p_s = d * m_inv_max_density;
                if (m_use_ratio_tracking) {
                    masked(tr, active) *= 1.f - p_s;
                } else {
                    Value p_n     = 1.f - p_s;
                    Mask valid_mi = active && (p_s > next_1d<Value>(sampler, index, active));
                    masked(tr, valid_mi) *= 0.f;
                    masked(tr, active && !valid_mi) *= p_n / detach(p_n);
                    active &= !valid_mi;
                }
            }
            return tr;
        }
    }


    template <typename Ray3, typename Value = typename Ray3::Value>
    Value pdf_distance_impl(Ray3 ray, Value t, mask_t<Value> active) const {
        Assert(m_use_raymarching);  // Not implemented for delta tracking

        ray.mint                                    = 0;
        ray.maxt                                    = select(t < math::Infinity, t, ray.maxt);
        auto [integrated_density, last_interpolant] = integrate_density(ray, active);
        Value tr                                    = exp(-integrated_density);

        // Compute sigma_t at ray(t), assuming it is within the medium
        auto sigma_t = enoki::mean(m_sigma_t->eval(ray.wavelengths)) * last_interpolant;
        return select(t < math::Infinity, sigma_t * tr, tr);
    }


    template <typename MediumInteraction3, typename Value = typename MediumInteraction3::Value>
    MTS_INLINE auto eval_scattering_impl(const MediumInteraction3 &mi, mask_t<Value> active) const {
        auto density_value = density(mi, active);
        auto sigma_t = m_sigma_t->eval(mi.wavelengths) * density_value;
        // Value avg_sigma_t = enoki::mean(sigma_t);
        // return m_albedo_tex->eval(mi, active) * sigma_t / detach(avg_sigma_t);
        // TODO: shouldn't the scattering coefficient be sigma_t * density * albedo?
        return m_albedo_tex->eval(mi, active) * sigma_t;
    }

    template <typename Interaction3, typename Value = typename Interaction3::Value>
    MTS_INLINE auto eval_albedo_impl(const Interaction3 &mi, mask_t<Value> active) const {
        return m_albedo_tex->eval(mi, active);
    }

    bool is_homogeneous() const override { return false; }

    std::vector<ref<Object>> children() override {
        if (m_sigma_t->id().empty())
            m_sigma_t->set_id(this->id() + "_sigma_t");
        if (m_density_tex->id().empty())
            m_density_tex->set_id(this->id() + "_density_tex");
        if (m_albedo_tex->id().empty())
            m_albedo_tex->set_id(this->id() + "_albedo_tex");

        return { m_sigma_t.get(), m_density_tex.get(), m_albedo_tex.get() };
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "HeterogeneousMedium[" << std::endl
            << "  m_sigma_t = " << string::indent(m_sigma_t) << std::endl
            << "]";
        return oss.str();
    }


#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        Medium::set_labels(prefix);
        m_sigma_t->set_labels(prefix + ".m_sigma_t");
        m_density_tex->set_labels(prefix + ".m_density_tex");
        m_albedo_tex->set_labels(prefix + ".m_albedo_tex");
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        Medium::set_globals(prefix, context, material);
        m_sigma_t->set_globals(prefix + "_m_sigma_t", context, material);
        m_density_tex->set_globals(prefix + "_m_density_tex", context, material);
        m_albedo_tex->set_globals(prefix + "_m_albedo_tex", context, material);

        // Extra variables to help with re-implementation
        RTvariable var = get_or_create_optix_var(prefix + "_step_size", context, material);
        rtVariableSet1f(var, get_step_size());

        auto rgb_sigma_t = m_sigma_t->eval(Spectrumf(1.f));
        var = get_or_create_optix_var(prefix + "_rgb_sigma_t", context, material);
        rtVariableSet3f(var, rgb_sigma_t.x(), rgb_sigma_t.y(), rgb_sigma_t.z());
    }
#endif


    MTS_IMPLEMENT_MEDIUM_ALL()
    MTS_DECLARE_CLASS()
private:
    ref<ContinuousSpectrum> m_sigma_t;
    ref<Texture3D> m_density_tex, m_albedo_tex;
    BoundingBox3f m_density_aabb;
    Float m_density_scale, m_max_density, m_inv_max_density, m_step_size_scaling;
    bool m_use_ratio_tracking, m_use_raymarching;
};

MTS_IMPLEMENT_CLASS(HeterogeneousMedium, Medium)
MTS_EXPORT_PLUGIN(HeterogeneousMedium, "Heterogeneous Medium")

NAMESPACE_END(mitsuba)
