#include <enoki/stl.h>

#include <mitsuba/core/frame.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/texture3d.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#endif

/**
 * It's generally better for performance to use the short rays optimization,
 * but it makes PTX extraction more difficult (since it requires tracing a
 * ray during distance sampling).
 */
#define MTS_DISABLE_SHORT_RAYS


NAMESPACE_BEGIN(mitsuba)

class HomogeneousMedium final : public Medium {
public:
    HomogeneousMedium(const Properties &props) : Medium(props) {
        ref<ContinuousSpectrum> sigma_a, sigma_s;
        std::tie(sigma_a, sigma_s, m_sigma_t, m_albedo) = extract_medium_parameters(props);
        m_density = props.float_("density", 1.0f);
        m_use_albedo_texture = false;

        if (props.has_property("albedotex")) {
            m_albedo_tex         = props.texture3d("albedotex", 1.0f);
            m_use_albedo_texture = true;

            Log(EInfo, "Albedo texture resolution: %s", m_albedo_tex->resolution());
        }

#if defined(MTS_ENABLE_AUTODIFF)
        m_density_d = m_density;
#endif
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override {
        dp.put(this, "density", m_density_d);
    }

    void parameters_changed() override {
        m_needs_reupload = true;
    }
#endif

    MTS_AUTODIFF_GETTER(density_v, m_density, m_density_d)


    template <typename Interaction3, typename Value = typename Interaction3::Value>
    Value density_impl(const Interaction3 & /*it*/, mask_t<Value> /*active*/) const {
        return density_v<Value>();
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value               = typename Ray::Value,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename Spectrum            = Spectrum<Value>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_impl(const Scene *scene,
                                         const Ray &ray,
                                         const Point2 & sample,
                                         Sampler *sampler,
                                         Index index,
                                         mask_t<Value> active) const {

#if defined(MTS_DISABLE_SHORT_RAYS)
        // No short rays optimization
        SurfaceInteraction3 si = scene->ray_intersect(ray, active);
        auto [mi, throughput]  = sample_distance_with_si(si, ray, sample, sampler, index, active);
        masked(si.t, mi.is_valid()) = math::Infinity;
        return std::make_tuple(si, mi, throughput);

#else
        using MediumInteraction3 = MediumInteraction<Point3>;
        using Vector3            = mitsuba::Vector<Value, 3>;
        using Frame              = Frame<Vector3>;
        ENOKI_MARK_USED(index);

        // Using the short rays optimization
        auto density_val = density_v<Value>();

        auto sigma_t = m_sigma_t->eval(ray.wavelengths) * density_val;
        MediumInteraction3 mi;
        Value avg_sigma_t = enoki::mean(sigma_t);

        Value u = sample.x();

        Value sampled_t       = -enoki::log(1 - u) / avg_sigma_t;
        mi.t                  = detach(sampled_t);
        mi.p                  = ray(mi.t);

        mi.sh_frame           = Frame(ray.d);
        mi.wi                 = Vector3(0.0f, 0.0f, 1.0f);
        mi.time               = ray.time;
        mi.wavelengths        = ray.wavelengths;
        mi.medium             = nullptr;
        masked(mi.t, !active) = math::Infinity;

        // Intersect the generated (short) ray with the scene
        Ray ray2(ray, ray.mint, mi.t);
        SurfaceInteraction3 si = scene->ray_intersect(ray2, active);
        Spectrum throughput(1.0f);

        Value t = mi.t;
        masked(t, si.is_valid() && active) = si.t;
        Value pdf_tr = exp(-t * avg_sigma_t);
        masked(throughput, active) =
             exp(-t * m_sigma_t->eval(ray.wavelengths) * density_val) / detach(pdf_tr);

        if (m_use_albedo_texture) {
            masked(throughput, !si.is_valid() && active) *=
               m_albedo_tex->eval(mi, !si.is_valid() && active) * sigma_t / detach(avg_sigma_t);
        } else {
            masked(throughput, !si.is_valid() && active) *=
                m_albedo->eval(ray.wavelengths) * sigma_t / detach(avg_sigma_t);
        }

        masked(si.t, !active) = math::Infinity;
        // Medium interactions are invalid if the surface was intersected
        masked(mi.t, si.is_valid() && active) = math::Infinity;
        masked(mi.medium, mi.is_valid() && active) = this_pointer<Value>();
        return std::make_tuple(si, mi, throughput);
#endif
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value, typename Spectrum = Spectrum<Value>,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_with_si_impl(const SurfaceInteraction3 &si, const Ray &ray,
                                                 const Point2 &sample, Sampler * /*sampler*/,
                                                 Index /*index*/, mask_t<Value> active) const {
        using MediumInteraction3 = MediumInteraction<Point3>;

#if defined(MTS_DISABLE_SHORT_RAYS)
        using Vector3 = mitsuba::Vector<Value, 3>;
        using Frame   = Frame<Vector3>;

        auto density_val = density_v<Value>();

        auto sigma_t      = m_sigma_t->eval(ray.wavelengths) * density_val;
        Value avg_sigma_t = enoki::mean(sigma_t);
        Value u           = sample.x();
        Value sampled_t   = -enoki::log(1 - u) / avg_sigma_t;

        MediumInteraction3 mi;
        mi.t                  = select(active, detach(sampled_t), math::Infinity);
        mi.p                  = ray(mi.t);
        mi.sh_frame           = Frame(ray.d);
        mi.wi                 = Vector3(0.0f, 0.0f, 1.0f);
        mi.time               = ray.time;
        mi.wavelengths        = ray.wavelengths;
        mi.medium             = nullptr;

        Spectrum throughput(1.0f);
        mask_t<Value> valid_si     = si.is_valid() && (si.t <= mi.t) && active;
        mask_t<Value> not_valid_si = (!si.is_valid() || (si.t > mi.t)) && active;

        Value t             = mi.t;
        masked(t, valid_si) = si.t;
        Value pdf_tr        = exp(-t * avg_sigma_t);
        masked(throughput, active) =
            exp(- density_val * t * m_sigma_t->eval(ray.wavelengths)) / detach(pdf_tr);

        Spectrum albedo;
        if (m_use_albedo_texture)
            albedo = m_albedo_tex->eval(mi, not_valid_si);
        else
            albedo = m_albedo->eval(ray.wavelengths);
        masked(throughput, not_valid_si) *= albedo * sigma_t / detach(avg_sigma_t);

        // Medium interactions are invalid if the surface was intersected
        masked(mi.t, valid_si)                     = math::Infinity;
        masked(mi.medium, mi.is_valid() && active) = this_pointer<Value>();
        return std::make_pair(mi, throughput);
#else
        NotImplementedError("sample_distance_with_si() when short rays are enabled");
        return std::make_pair(zero<MediumInteraction3>(1), zero<Spectrum>(1));
#endif
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value>
    MTS_INLINE auto eval_transmittance_impl(const Ray &ray,
                                            Sampler * /*unused*/,
                                            uint_array_t<Value> /*unused*/,
                                            mask_t<Value> active) const {
        ENOKI_MARK_USED(active);
        auto density_val = density_v<Value>();
        auto sigma_t     = m_sigma_t->eval(ray.wavelengths);
        auto d           = ray.maxt - ray.mint;
        return enoki::exp(-d * sigma_t * density_val);
    }

    template <typename Ray3, typename Value = typename Ray3::Value>
    Value pdf_distance_impl(const Ray3 &ray, Value t, mask_t<Value> /*active*/) const {
        auto sigma_t      = m_sigma_t->eval(ray.wavelengths) * density_v<Value>();
        Value avg_sigma_t = enoki::mean(sigma_t);
        Value tr          = exp(-(ray.maxt - ray.mint) * avg_sigma_t);
        return select(t < math::Infinity, avg_sigma_t * tr, tr);
    }

    template <typename MediumInteraction3, typename Value = typename MediumInteraction3::Value>
    MTS_INLINE auto eval_scattering_impl(const MediumInteraction3 &mi,
                                         mask_t<Value> active) const {
        Spectrum<Value> albedo;
        if (m_use_albedo_texture)
            albedo = m_albedo_tex->eval(mi, active);
        else
            albedo = m_albedo->eval(mi.wavelengths);

        // // TODO: shouldn't the scattering coefficient be sigma_t * density * albedo?
        // auto sigma_t      = m_sigma_t->eval(mi.wavelengths) * density_v<Value>();
        // Value avg_sigma_t = enoki::mean(sigma_t);
        // return albedo * sigma_t / detach(avg_sigma_t);

        auto sigma_t = m_sigma_t->eval(mi.wavelengths) * density_v<Value>();
        return albedo * sigma_t;
    }

    template <typename Interaction3, typename Value = typename Interaction3::Value>
    MTS_INLINE auto eval_albedo_impl(const Interaction3 &it, mask_t<Value> active) const {
        if (m_use_albedo_texture)
            return m_albedo_tex->eval(it, active);
        else
            return m_albedo->eval(it.wavelengths);
    }

    bool is_homogeneous() const override { return true; }

    std::vector<ref<Object>> children() override {
        if (m_sigma_t->id().empty())
            m_sigma_t->set_id(this->id() + "_sigma_t");

        if (m_use_albedo_texture) {
            if (m_albedo_tex->id().empty())
                m_albedo_tex->set_id(this->id() + "_albedo_tex");
            return { m_albedo_tex.get(), m_sigma_t.get() };
        } else {
            if (m_albedo->id().empty())
                m_albedo->set_id(this->id() + "_albedo");
            return { m_albedo.get(), m_sigma_t.get() };
        }
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "HomogeneousMedium[" << std::endl
            << "  m_albedo = " << string::indent(m_albedo) << std::endl
            << "  m_sigma_t = " << string::indent(m_sigma_t) << std::endl
            << "]";
        return oss.str();
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        Medium::set_labels(prefix);

        m_sigma_t->set_labels(prefix + ".m_sigma_t");
        if (m_use_albedo_texture)
            m_albedo_tex->set_labels(prefix + ".m_albedo_tex");
        else
            m_albedo->set_labels(prefix + ".m_albedo");
        set_label(m_density_d, (prefix + ".m_density_d").c_str());
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        Medium::set_globals(prefix, context, material);
        m_sigma_t->set_globals(prefix + "_m_sigma_t", context, material);
        if (m_use_albedo_texture)
            m_albedo_tex->set_globals(prefix + "_m_albedo_tex", context, material);
        else
            m_albedo->set_globals(prefix + "_m_albedo", context, material);

        if (!m_needs_reupload)
            return;

        cuda_eval();

        // Get current value from buffer
        // TODO: it would be much nicer to avoid this round-trip
        float m_density;
        cuda_memcpy_from_device(&m_density, m_density_d.data(), 1 * sizeof(float));
        cuda_eval();

        std::string name = prefix + "_m_density_d";
        RTvariable var = get_or_create_optix_var(name, context, material);
        rtVariableSet1f(var, m_density);
        cuda_eval();
        m_needs_reupload = false;
    }
#endif

    MTS_IMPLEMENT_MEDIUM_ALL()
    MTS_DECLARE_CLASS()
private:
    ref<ContinuousSpectrum> m_sigma_t;
    ref<ContinuousSpectrum> m_albedo;
    ref<Texture3D>  m_albedo_tex;
    Float m_density;
    bool m_use_albedo_texture;
#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_density_d;
    mutable bool m_needs_reupload = true;
#endif
};

MTS_IMPLEMENT_CLASS(HomogeneousMedium, Medium)
MTS_EXPORT_PLUGIN(HomogeneousMedium, "Homogeneous Medium")

NAMESPACE_END(mitsuba)
