#include <enoki/stl.h>

#include <mitsuba/core/frame.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/texture3d.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#endif


NAMESPACE_BEGIN(mitsuba)

class GrayscaleHomogeneousMedium final : public Medium {
public:
    GrayscaleHomogeneousMedium(const Properties &props) : Medium(props) {
        m_albedo = props.spectrum("albedo", 1.0f);
        m_density = props.float_("density", 1.0f);

#if defined(MTS_ENABLE_AUTODIFF)
        m_density_d = m_density;
#endif
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override {
        dp.put(this, "density", m_density_d);
    }

    void parameters_changed() override {
        m_needs_reupload = true;
    }
#endif

    MTS_AUTODIFF_GETTER(density_v, m_density, m_density_d)

    template <typename Interaction3, typename Value = typename Interaction3::Value>
    Value density_impl(const Interaction3 & /*it*/, mask_t<Value> /*active*/) const {
        return density_v<Value>();
    }

    template <
        typename Ray, typename Point3 = typename Ray::Point, typename Value = typename Ray::Value,
        typename Point2 = typename mitsuba::Point<Value, 2>, typename Spectrum = Spectrum<Value>,
        typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
        typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_impl(const Scene *scene, const Ray &ray, const Point2 &sample,
                                         Sampler *sampler, Index index,
                                         mask_t<Value> active) const {
        // No short rays optimization
        SurfaceInteraction3 si = scene->ray_intersect(ray, active);
        auto [mi, throughput]  = sample_distance_with_si(si, ray, sample, sampler, index, active);
        masked(si.t, mi.is_valid()) = math::Infinity;
        return std::make_tuple(si, mi, throughput);
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value, typename Spectrum = Spectrum<Value>,
              typename Point2              = typename mitsuba::Point<Value, 2>,
              typename SurfaceInteraction3 = SurfaceInteraction<Point3>,
              typename Index               = uint_array_t<Value>>
    MTS_INLINE auto sample_distance_with_si_impl(const SurfaceInteraction3 &si, const Ray &ray,
                                                 const Point2 &sample, Sampler * /*sampler*/,
                                                 Index /*index*/, mask_t<Value> active) const {
        using MediumInteraction3 = MediumInteraction<Point3>;
        using Vector3            = mitsuba::Vector<Value, 3>;
        using Frame              = Frame<Vector3>;

        auto density_val = density_v<Value>();

        Value u         = sample.x();
        Value sampled_t = -enoki::log(1 - u) / density_val;

        MediumInteraction3 mi;
        mi.t           = select(active, detach(sampled_t), math::Infinity);
        mi.p           = ray(mi.t);
        mi.sh_frame    = Frame(ray.d);
        mi.wi          = Vector3(0.0f, 0.0f, 1.0f);
        mi.time        = ray.time;
        mi.wavelengths = ray.wavelengths;
        mi.medium      = nullptr;

        mask_t<Value> valid_si     = si.is_valid() && (si.t <= mi.t) && active;
        mask_t<Value> not_valid_si = (!si.is_valid() || (si.t > mi.t)) && active;

        Value t             = mi.t;
        masked(t, valid_si) = si.t;
        Value pdf_tr        = exp(-t * density_val);
        Spectrum throughput(1.0f);
        masked(throughput, active) = exp(-t * density_val) / detach(pdf_tr);

        Spectrum albedo = m_albedo->eval(ray.wavelengths);
        masked(throughput, not_valid_si) *= albedo * density_val / detach(density_val);

        // No medium interaction sampled
        masked(mi.t, valid_si)                     = math::Infinity;
        masked(mi.medium, mi.is_valid() && active) = this_pointer<Value>();
        return std::make_pair(mi, throughput);
    }

    template <typename Ray, typename Point3 = typename Ray::Point,
              typename Value = typename Ray::Value>
    MTS_INLINE auto eval_transmittance_impl(const Ray &ray, Sampler * /*unused*/,
                                            uint_array_t<Value> /*unused*/,
                                            mask_t<Value> /*active*/) const {
        auto sigma_t = density_v<Value>();
        auto d       = ray.maxt - ray.mint;
        return enoki::exp(-d * sigma_t);
    }

    template <typename Ray3, typename Value = typename Ray3::Value>
    Value pdf_distance_impl(const Ray3 &ray, Value t, mask_t<Value> /*active*/) const {
        auto sigma_t = density_v<Value>();
        Value tr     = exp(-(ray.maxt - ray.mint) * sigma_t);
        return select(t < math::Infinity, sigma_t * tr, tr);
    }

    template <typename MediumInteraction3, typename Value = typename MediumInteraction3::Value>
    MTS_INLINE auto eval_scattering_impl(const MediumInteraction3 &mi,
                                         mask_t<Value> /*active*/) const {
        Spectrum<Value> albedo = m_albedo->eval(mi.wavelengths);
        return albedo * density_v<Value>();
    }

    template <typename Interaction3, typename Value = typename Interaction3::Value>
    MTS_INLINE auto eval_albedo_impl(const Interaction3 &it, mask_t<Value> /*active*/) const {
        return m_albedo->eval(it.wavelengths);
    }

    bool is_homogeneous() const override { return true; }

    std::vector<ref<Object>> children() override {
        if (m_albedo->id().empty())
            m_albedo->set_id(this->id() + "_albedo");
        return { m_albedo.get() };
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "GrayscaleHomogeneousMedium[" << std::endl
            << "  m_albedo = " << string::indent(m_albedo) << std::endl
            << "  m_density = " << string::indent(m_density) << std::endl
            << "]";
        return oss.str();
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        Medium::set_labels(prefix);
        m_albedo->set_labels(prefix + ".m_albedo");
        set_label(m_density_d, (prefix + ".m_density_d").c_str());
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        Medium::set_globals(prefix, context, material);
        m_albedo->set_globals(prefix + "_m_albedo", context, material);

        if (!m_needs_reupload)
            return;
        cuda_eval();

        // Get current value from buffer
        // TODO: it would be much nicer to avoid this round-trip
        float m_density;
        cuda_memcpy_from_device(&m_density, m_density_d.data(), 1 * sizeof(float));
        cuda_eval();

        std::string name = prefix + "_m_density_d";
        RTvariable var = get_or_create_optix_var(name, context, material);
        rtVariableSet1f(var, m_density);
        cuda_eval();
        m_needs_reupload = false;
    }
#endif

    MTS_IMPLEMENT_MEDIUM_ALL()
    MTS_DECLARE_CLASS()
private:
    ref<ContinuousSpectrum> m_albedo;
    Float m_density;
#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_density_d;
    mutable bool m_needs_reupload = true;
#endif
};

MTS_IMPLEMENT_CLASS(GrayscaleHomogeneousMedium, Medium)
MTS_EXPORT_PLUGIN(GrayscaleHomogeneousMedium, "Grayscale Homogeneous Medium")

NAMESPACE_END(mitsuba)
