#include <iostream>
#include <fstream>
#include <vector>

#include <enoki/cuda.h>
#include <mitsuba/core/argparser.h>
#include <mitsuba/core/class.h>
#include <mitsuba/core/filesystem.h>
#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/jit.h>
#include <mitsuba/core/logger.h>
#include <mitsuba/core/object.h>
#include <mitsuba/core/thread.h>
#include <mitsuba/core/util.h>
#include <mitsuba/core/xml.h>
#include <mitsuba/mitsuba.h>
#include <mitsuba/optix/renderer.h>
#include <mitsuba/render/scene.h>

using namespace mitsuba;

int main(int argc, char **argv) {
    Jit::static_initialization();
    Class::static_initialization();
    Thread::static_initialization();
    Logger::static_initialization();
    Bitmap::static_initialization();
    Profiler::static_initialization();

    // Ensure that the mitsuba-optix shared library is loaded
    // liboptix_nop();

    // CUDA backend only really works single-threaded
    // __global_thread_count = 1;

    // Argument parser
    ArgParser parser;
    using StringVec  = std::vector<std::string>;
    auto arg_verbose = parser.add(StringVec{ "-v", "--verbose" }, false);
    auto arg_volume  = parser.add(StringVec{ "--vol", "--volume" }, false);
    auto arg_passes  = parser.add(StringVec{ "-p", "--passes" }, true);  // Multiplies the spp
    auto arg_define  = parser.add(StringVec{ "-D", "--define" }, true);
    auto arg_output  = parser.add(StringVec{ "-o", "--output" }, true);
    auto arg_extra   = parser.add("", true);
    xml::ParameterList params;

    // Parse all command line options
    parser.parse(argc, argv);

    if (*arg_verbose) {
        auto logger = Thread::thread()->logger();
        if (arg_verbose->next())
            logger->set_log_level(ETrace);
        else
            logger->set_log_level(EDebug);
    }

    while (arg_define && *arg_define) {
        std::string value = arg_define->as_string();
        auto sep = value.find('=');
        if (sep == std::string::npos)
            Throw("-D/--define: expect key=value pair!");
        params.push_back(std::make_pair(value.substr(0, sep), value.substr(sep+1)));
        arg_define = arg_define->next();
    }

    if (!*arg_extra)
        Log(EWarn, "No scene file specified!");
    size_t pass_count = 1;
    if (*arg_passes)
        pass_count = arg_passes->as_int();

    EColorMode color_mode = ERGBColor;
    OptiXRenderer::EIntegratorType integrator_type =
        *arg_volume ? OptiXRenderer::EVolPath : OptiXRenderer::EPath;

    // Append the mitsuba directory to the FileResolver search path list
    ref<FileResolver> fr = Thread::thread()->file_resolver();
    filesystem::path base_path = util::library_path().parent_path();
    if (!fr->contains(base_path))
        fr->append(base_path);

    // For each scene path specified
    bool overall_success = true;
    while (arg_extra && *arg_extra) {
        filesystem::path filename(arg_extra->as_string());
        // Add the scene file's directory to the search path.
        auto scene_dir = filename.parent_path();
        if (!fr->contains(scene_dir))
            fr->append(scene_dir);

        filesystem::path out_fname = filename;
        if (*arg_output)
            out_fname = arg_output->as_string();
        out_fname.replace_extension("exr");

        // Try and parse a scene from the passed file.
        ref<Object> parsed = xml::load_file(arg_extra->as_string(), params, color_mode);

        auto *scene = dynamic_cast<Scene *>(parsed.get());
        if (!scene)
            Throw("Could not parse scene from file: %s", filename);

        OptiXRenderer renderer(scene, integrator_type);
        auto b = renderer.render(1, (size_t) -1, pass_count);
        b->write(out_fname);
        Log(EInfo, "Saved result to: %s", out_fname);

        bool success = true;
        overall_success = overall_success && success;

        arg_extra = arg_extra->next();
    }

    Profiler::static_shutdown();
    Bitmap::static_shutdown();
    Logger::static_shutdown();
    Thread::static_shutdown();
    Class::static_shutdown();
    Jit::static_shutdown();
    return overall_success ? 0 : -1;
}
