#pragma once

#include <optix_world.h>
#include <math_constants.h>
#include "common.cuh"
#include "types.cuh"

/**
 * Area emitter (OptiX callable program)
 */
class AreaEmitter {
public:
    static RT_FUNCTION float3 sample(const EmitterProperties &light, DirectionSample &ds, const SurfaceInteraction &ref, RayPayload &prd) {

        // TODO: actually sample on the mesh
        float u = rnd(prd.seed);
        float v = rnd(prd.seed);
        ds.p = light.p + u * light.v0 + v * light.v1;

        // Note: light_direction is returned in world coordinates,
        // and points toward the light.
        ds.light_direction = ds.p - ref.p;
        float light_dist2 = squared_norm(ds.light_direction);
        ds.light_dist = sqrt(light_dist2);
        ds.light_direction /= ds.light_dist;

        // TODO: get actual normal on the surface of this emitter
        ds.normal = light.normal;

        // Foreshortening on the surface of the emitter
        float cos_theta = optix::dot(ds.normal, -ds.light_direction);
        if (cos_theta <= 0.f) {
            ds.pdf = 0.f;
            return make_float3(0.f);
        }

        // Geometric term (conversion from position sampling to direction sampling)
        ds.pdf = light_dist2 / (light.area * cos_theta);

        // Sampling weight = eval / pdf
        return light.radiance / ds.pdf;
    }


    static RT_FUNCTION float3 eval(const EmitterProperties &light, const SurfaceInteraction &si) {
        return light.radiance * (si.wi.z > 0.f);
    }

    static RT_FUNCTION float pdf(const EmitterProperties &light, const DirectionSample &ds) {
        // Note: light_direction is expected in world coordinates, pointing toward the light.
        float cos_theta = optix::dot(ds.normal, -ds.light_direction);
        if (cos_theta <= 0.f)
            return 0.f;

        // pdf_area_light        = pdf_surface_direction
        // pdf_surface_direction = pdf_position * jacobian to convert to angular domain
        // pdf_position          = 1. / surface_area
        // jacobian              = ds.dist * ds.dist / abs_dot(ds.d, ds.n)
        return (ds.light_dist * ds.light_dist) / (light.area * cos_theta);
    }

};
