#include <optix_world.h>

struct BoundingBox3f {
    optix::float3 pmin, pmax;
};
struct BoundingBoxIntersection {
    bool missed;
    float mint, maxt;
};

RT_FUNCTION BoundingBoxIntersection intersect_bbox(const optix::float3 o, const optix::float3 d,
                                                   const BoundingBox3f bbox) {
    BoundingBoxIntersection it;
    it.missed = true;

    // First, ensure that the ray either has a nonzero slope on each axis,
    // or that its origin on a zero-valued axis is within the box bounds
    bool active = (d.x != 0.f || (o.x > bbox.pmin.x || o.x < bbox.pmax.x))
               && (d.y != 0.f || (o.y > bbox.pmin.y || o.y < bbox.pmax.y))
               && (d.z != 0.f || (o.z > bbox.pmin.z || o.z < bbox.pmax.z));
    if (!active) {
        return it;
    }

    // Compute intersection intervals for each axis
    auto d_rcp = optix::make_float3(1.f) / d;
    optix::float3 t1 = (bbox.pmin - o) * d_rcp,
                  t2 = (bbox.pmax - o) * d_rcp;

    // Ensure proper ordering
    optix::float3 t1p = optix::make_float3(
        optix::min(t1.x, t2.x), optix::min(t1.y, t2.y), optix::min(t1.z, t2.z)
    );
    optix::float3 t2p = optix::make_float3(
        optix::max(t1.x, t2.x), optix::max(t1.y, t2.y), optix::max(t1.z, t2.z)
    );

    // Intersect intervals
    it.mint = optix::max(t1p.x, optix::max(t1p.y, t1p.z));
    it.maxt = optix::min(t2p.x, optix::min(t2p.y, t2p.z));

    if (it.maxt < it.mint) {
        return it;  // Missed
    }

    it.missed = false;
    return it;
}

/// Note: this ignores (mint, maxt)
RT_FUNCTION BoundingBoxIntersection intersect_bbox(const optix::Ray &ray,
                                                   const BoundingBox3f bbox) {
    return intersect_bbox(ray.origin, ray.direction, bbox);
}


RT_FUNCTION bool bbox_contains(const optix::float3 &p, const BoundingBox3f bbox) {
    return (p.x >= bbox.pmin.x && p.y >= bbox.pmin.y && p.z >= bbox.pmin.z)
        && (p.x <= bbox.pmax.x && p.y <= bbox.pmax.y && p.z <= bbox.pmax.z);
}
