#include <optix_world.h>
#include <math_constants.h>
#include "types.cuh"
#include "common.cuh"

using namespace optix;

rtDeclareVariable(uint2, launch_index, rtLaunchIndex, );
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(uint, kNx, , "Image width");
rtDeclareVariable(uint, kNy, , "Image height");
rtDeclareVariable(uint, kSpp, , "Samples / pixel");
rtDeclareVariable(uint, kSampleIndex, , "Index of this sample, used to seed the RNG");
rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(uint, kMaxDepth, , "Maximum path depth (0: visible emitters, 1: direct illumination, etc.)");

rtDeclareVariable(rtObject, top_object, , );
rtDeclareVariable(float3, camera_pos, , );
rtDeclareVariable(float3, camera_u, , );
rtDeclareVariable(float3, camera_v, , );
rtDeclareVariable(float3, camera_w, , );
rtDeclareVariable(float,  camera_near, , );
rtDeclareVariable(float,  camera_far, , );
rtDeclareVariable(Matrix4x4,  sample_to_camera, , );
rtDeclareVariable(float,  scene_scale, , );

rtDeclareVariable(float3, p, attribute p, );
rtDeclareVariable(float3, ns, attribute ns, );
rtDeclareVariable(float3, ng, attribute ng, );

/// Output image buffer
rtBuffer<float> out_image;

RT_PROGRAM void ray_gen() {
    uint2 pixel_pos = make_uint2(launch_index.x, launch_index.y);

    // Perspective camera (using homogeneous coordinates)
    float4 sample_pos = make_float4(pixel_pos.x / float(kNx), pixel_pos.y / float(kNy), 0.f, 1.f);
    float4 near_p = sample_to_camera * sample_pos;
    float3 d = normalize(make_float3(near_p.x, near_p.y, near_p.z));
    float3 direction = normalize(d.x * camera_u + d.y * camera_v + d.z * camera_w);
    // Ray extents
    float inv_z = 1.f / near_p.z;
    float min_t = camera_near * inv_z;
    float max_t = camera_far * inv_z;

    Ray ray = make_Ray(camera_pos, direction, 0, min_t, max_t);

    RayPayload payload(0, 1);
    rtTrace(top_object, ray, payload, RT_RAY_FLAG_TERMINATE_ON_FIRST_HIT);

    // Accumulate ray's contribution to its pixel
    uint idx = 3 * (pixel_pos.x * kNx + pixel_pos.y);
    out_image[idx + 0] = payload.result.x;
    out_image[idx + 1] = payload.result.y;
    out_image[idx + 2] = payload.result.z;
}


RT_PROGRAM void ray_hit_closest() {
    float t = sqrt(squared_norm(p - ray.origin) / squared_norm(ray.direction));
    t /= 2.5f * scene_scale;
    payload.result.x = t;
    payload.result.y = t;
    payload.result.z = t;
}

/// Miss program for standard rays
RT_PROGRAM void ray_miss() {
    payload.result.x = launch_index.x / (1.f * kNx);
    payload.result.y = 0.5f * (launch_index.y + launch_index.x) / (1.f * kNy);
    payload.result.z = 0.f;
}

/// Hit program for shadow rays
RT_PROGRAM void shadow_closest_hit() {
    // Nothing to do for the depth integrator.
}
