/**
 * @file Wrappers for calls to PTX-extracted autodiff Media methods.
 */

#pragma once

#include <optix_world.h>
#include "types.cuh"

namespace detail {
extern "C" {

__device__ void medium_density_d(
    float,        // it.t
    float,        // it.time
    float,        // it.wavelengths.0
    float,        // it.wavelengths.1
    float,        // it.wavelengths.2
    float,        // it.wavelengths.3
    float,        // it.p.0
    float,        // it.p.1
    float,        // it.p.2
    float         // grad_of_result
);

__device__ void medium_eval_transmittance_d(
    float,        // ray.o.0
    float,        // ray.o.1
    float,        // ray.o.2
    float,        // ray.d.0
    float,        // ray.d.1
    float,        // ray.d.2
    float,        // ray.d_rcp.0
    float,        // ray.d_rcp.1
    float,        // ray.d_rcp.2
    float,        // ray.mint
    float,        // ray.maxt
    float,        // ray.time
    float,        // ray.wavelengths.0
    float,        // ray.wavelengths.1
    float,        // ray.wavelengths.2
    float,        // ray.wavelengths.3
    uint32_t,     // index
    float,        // grad_of_result.0
    float,        // grad_of_result.1
    float,        // grad_of_result.2
    float         // grad_of_result.3
);

__device__ void medium_eval_scattering_d(
    float,        // mi.t
    float,        // mi.time
    float,        // mi.wavelengths.0
    float,        // mi.wavelengths.1
    float,        // mi.wavelengths.2
    float,        // mi.wavelengths.3
    float,        // mi.p.0
    float,        // mi.p.1
    float,        // mi.p.2
    void *,       // mi.medium
    float,        // mi.sh_frame.s.0
    float,        // mi.sh_frame.s.1
    float,        // mi.sh_frame.s.2
    float,        // mi.sh_frame.t.0
    float,        // mi.sh_frame.t.1
    float,        // mi.sh_frame.t.2
    float,        // mi.sh_frame.n.0
    float,        // mi.sh_frame.n.1
    float,        // mi.sh_frame.n.2
    float,        // mi.wi.0
    float,        // mi.wi.1
    float,        // mi.wi.2
    float,        // grad_of_result.0
    float,        // grad_of_result.1
    float,        // grad_of_result.2
    float         // grad_of_result.3
);

#if defined(MTS_PTX_USE_COMBINED_SCATTERING_AND_TRANSMITTANCE)
__device__ void medium_eval_scattering_times_transmittance_d(
    float,        // mi.t
    float,        // mi.time
    float,        // mi.wavelengths.0
    float,        // mi.wavelengths.1
    float,        // mi.wavelengths.2
    float,        // mi.wavelengths.3
    float,        // mi.p.0
    float,        // mi.p.1
    float,        // mi.p.2
    void *,       // mi.medium
    float,        // mi.sh_frame.s.0
    float,        // mi.sh_frame.s.1
    float,        // mi.sh_frame.s.2
    float,        // mi.sh_frame.t.0
    float,        // mi.sh_frame.t.1
    float,        // mi.sh_frame.t.2
    float,        // mi.sh_frame.n.0
    float,        // mi.sh_frame.n.1
    float,        // mi.sh_frame.n.2
    float,        // mi.wi.0
    float,        // mi.wi.1
    float,        // mi.wi.2
    float,        // ray.o.0
    float,        // ray.o.1
    float,        // ray.o.2
    float,        // ray.d.0
    float,        // ray.d.1
    float,        // ray.d.2
    float,        // ray.d_rcp.0
    float,        // ray.d_rcp.1
    float,        // ray.d_rcp.2
    float,        // ray.mint
    float,        // ray.maxt
    float,        // ray.time
    float,        // ray.wavelengths.0
    float,        // ray.wavelengths.1
    float,        // ray.wavelengths.2
    float,        // ray.wavelengths.3
    uint32_t,     // index
    float,        // grad_of_result.0
    float,        // grad_of_result.1
    float,        // grad_of_result.2
    float         // grad_of_result.3
);
#endif

}  // end extern
}  // namespace detail


RT_FUNCTION void call_medium_density_d(float t, const optix::float3 &p,
                                       float grad_of_result) {
    detail::medium_density_d(
        t,
        0.f,  // it.time,
        0.f, 0.f, 0.f, 0.f,  // it.wavelengths
        p.x, p.y, p.z,
        grad_of_result
    );
}

RT_FUNCTION void call_medium_eval_transmittance_d(const optix::Ray &ray,
                                                  const optix::float3 &grad_of_result) {
    detail::medium_eval_transmittance_d(
        ray.origin.x, ray.origin.y, ray.origin.z,
        ray.direction.x, ray.direction.y, ray.direction.z,
        // ray.d_rcp
        1.f / ray.direction.x, 1.f / ray.direction.y, 1.f / ray.direction.z,
        ray.tmin, ray.tmax,
        // ray.time
        0.f,
        // ray.wavelengths
        0.f, 0.f, 0.f, 0.f,
        // index (assumed unused!)
        0,
        grad_of_result.x, grad_of_result.y, grad_of_result.z, 0.f
    );
}


RT_FUNCTION void call_medium_eval_scattering_d(const MediumSample &mi,
                                               const optix::float3 &grad_of_result) {
    detail::medium_eval_scattering_d(
        mi.t,
        0.f,  // mi.time,
        0.f, 0.f, 0.f, 0.f,  // mi.wavelengths
        mi.p.x, mi.p.y, mi.p.z,
        nullptr,  // mi.medium
        mi.sh_frame.s.x, mi.sh_frame.s.y, mi.sh_frame.s.z,
        mi.sh_frame.t.x, mi.sh_frame.t.y, mi.sh_frame.t.z,
        mi.sh_frame.n.x, mi.sh_frame.n.y, mi.sh_frame.n.z,
        mi.wi.x, mi.wi.y, mi.wi.z,
        grad_of_result.x, grad_of_result.y, grad_of_result.z, 0.f
    );
}

#if defined(MTS_PTX_USE_COMBINED_SCATTERING_AND_TRANSMITTANCE)
RT_FUNCTION void call_medium_eval_scattering_times_transmittance_d(const MediumSample &mi,
                                                                   const optix::Ray &ray,
                                                                   const optix::float3 &grad_of_result) {
    detail::medium_eval_scattering_times_transmittance_d(
        mi.t,
        0.f,  // mi.time,
        0.f, 0.f, 0.f, 0.f,  // mi.wavelengths
        mi.p.x, mi.p.y, mi.p.z,
        nullptr,  // mi.medium
        mi.sh_frame.s.x, mi.sh_frame.s.y, mi.sh_frame.s.z,
        mi.sh_frame.t.x, mi.sh_frame.t.y, mi.sh_frame.t.z,
        mi.sh_frame.n.x, mi.sh_frame.n.y, mi.sh_frame.n.z,
        mi.wi.x, mi.wi.y, mi.wi.z,

        ray.origin.x, ray.origin.y, ray.origin.z,
        ray.direction.x, ray.direction.y, ray.direction.z,
        // ray.d_rcp
        1.f / ray.direction.x, 1.f / ray.direction.y, 1.f / ray.direction.z,
        ray.tmin, ray.tmax,
        // ray.time
        0.f,
        // ray.wavelengths
        0.f, 0.f, 0.f, 0.f,
        // index (assumed unused!)
        0,
        grad_of_result.x, grad_of_result.y, grad_of_result.z, 0.f
    );
}
#endif
