#include <optix_world.h>
#include "common.cuh"
#include "media.cuh"
#include "medium_heterogeneous.cuh"

// Note: already declared in the header
// rtDeclareVariable(float, medium_step_size, , );
// rtDeclareVariable(optix::float3, medium_rgb_sigma_t, , );


namespace detail {

/**
 * This function assumes that ray.mint corresponds to either a point inside the medium
 * or on the boundary, and ray.maxt should be on the exit boundary in the direction of the ray.
 */
RT_FUNCTION MediumSample custom_sample_distance(const SurfaceInteraction &si,
                                                const optix::Ray &ray,
                                                const optix::float2 &sample) {
    float step_size      = medium_step_size;
    float half_step_size = 0.5f * step_size;
    float mean_sigma_t   = (medium_rgb_sigma_t.x + medium_rgb_sigma_t.y
                            + medium_rgb_sigma_t.z) / 3.f;
    float mint     = ray.tmin;
    float maxt     = min(ray.tmax, si.t);
    int step_count = 1 + (int) ceilf((maxt - mint) / step_size);

    MediumSample mi;
    mi.p           = optix::make_float3(0.f);
    mi.sh_frame    = Onb(ray.direction);
    mi.wi          = optix::make_float3(0.f, 0.f, 1.f);
    mi.t           = math::Infinity;

    float desired_density = -logf(1 - sample.x);

    float t_a                = mint;
    optix::float3 lookup_p   = ray.origin + t_a * ray.direction;
    float f_a                = call_medium_density(0.f, lookup_p) * mean_sigma_t;
    float integrated_density = 0.f;

    // Jitter: Perform first step outside of loop
    float t_b         = fmaf(sample.y, step_size, t_a);
    lookup_p          = ray.origin + t_b * ray.direction;
    float f_b         = call_medium_density(0.f, lookup_p) * mean_sigma_t;
    float new_density = 0.5f * (t_b - t_a) * (f_a + f_b);

    bool reached_density = (new_density >= desired_density);
    if (!reached_density && t_b <= maxt) {
        integrated_density = new_density;
        t_a                = t_b;
        f_a                = f_b;
        mint               = t_b;

        for (int i = 1; i < step_count; ++i) {
            t_b      = fmaf(i, step_size, mint);
            lookup_p = ray.origin + t_b * ray.direction;
            f_b      = call_medium_density(0.f, lookup_p) * mean_sigma_t;
            float new_density = fmaf(half_step_size, f_a + f_b, integrated_density);

            reached_density |= (new_density >= desired_density);
            if (reached_density || t_b > maxt)
                break;

            integrated_density = new_density;
            t_a                = t_b;
            f_a                = f_b;
        }
    }

    // Solve quadratic equation to get exact intersection location
    float a = 0.5f * (f_b - f_a);
    float b = f_a;
    float c = (integrated_density - desired_density) * (1.f / (t_b - t_a));
    float solution1, solution2;
    bool has_solution = call_math_solve_quadratic(a, b, c, &solution1, &solution2);
    float interp      = (solution1 >= 0.f && solution1 <= 1.f ? solution1 : solution2);
    float sampled_t   = t_a + (t_b - t_a) * interp;
    float f_c         = (1 - interp) * f_a + interp * f_b;

    // Update integrated_density for rays which do not interact with the medium
    if ((!reached_density && (t_b > maxt)) || (reached_density && (sampled_t > maxt))) {
        lookup_p = ray.origin + (maxt - math::Epsilon) * ray.direction;
        integrated_density += 0.5f * (f_a + call_medium_density(0.f, lookup_p) * mean_sigma_t);
    }

    // Record medium interaction if the generated "t" is within the range of valid "t"
    mi.throughput = make_float3(1.0f);
    if (reached_density && sampled_t <= maxt) {
        integrated_density = desired_density;
        mi.t               = sampled_t;
        mi.p               = ray.origin + mi.t * ray.direction;

        float tr = expf(-integrated_density);
        float pdf_tr = f_c * tr;
        optix::float3 albedo_value = call_medium_eval_albedo(mi);
        mi.throughput *= albedo_value * (pdf_tr > 0.f ? f_c * tr / pdf_tr : 1.f);
    }

    // No need for the following, we're not doing autodiff
    // if (si.is_valid())
    //     mi.throughput *= (tr > 0.f ? tr / detach(tr) : 1.0f);

    return mi;
}


/**
 * This function assumes that ray.(mint, maxt) correspond to the bounding box of the medium.
 */
RT_FUNCTION optix::float3 custom_eval_transmittance(const optix::Ray &tr_ray) {
    float integrated_density = medium_integrate_density(tr_ray);
    return make_float3(expf(-integrated_density));
}

/**
 * This function ignores the ray's (mint, maxt) and uses `t` as the distance.
 * The integration must start at `ray.origin`.
 */
RT_FUNCTION float custom_pdf_distance(optix::Ray tr_ray, float t) {
    tr_ray.tmin = 0.f;
    tr_ray.tmax = fminf(tr_ray.tmax, t);

    float last_interpolant;
    float integrated_density = medium_integrate_density(tr_ray, &last_interpolant);
    float tr = expf(-integrated_density);
    if (t < math::Infinity) {
        float mean_sigma_t     = (medium_rgb_sigma_t.x + medium_rgb_sigma_t.y
                                  + medium_rgb_sigma_t.z) / 3.f;
        float sigma_t          = mean_sigma_t * last_interpolant;
        return sigma_t * tr;
    } else {
        return tr;
    }
}

}  // end namespace detail


// External interface
extern "C" {


// __device__ void my_fun_function(float a, float *result) {
//     *result = 43.f * a;
// }

__device__ void medium_eval_transmittance(
    float ray_o_0, float ray_o_1, float ray_o_2,
    float ray_d_0, float ray_d_1, float ray_d_2,
    float /*ray_d_rcp_0*/, float /*ray_d_rcp_1*/, float /*ray_d_rcp_2*/,
    float ray_mint, float ray_maxt, float /*ray_time*/,
    float /*ray_wavelengths_0*/, float /*ray_wavelengths_1*/,
    float /*ray_wavelengths_2*/, float /*ray_wavelengths_3*/,
    uint32_t /*index*/,
    float *result_0, float *result_1, float *result_2, float * /*result_3*/) {
    auto ray = optix::make_Ray(
        optix::make_float3(ray_o_0, ray_o_1, ray_o_2),
        optix::make_float3(ray_d_0, ray_d_1, ray_d_2),
        1,
        ray_mint, ray_maxt
    );

    auto result = detail::custom_eval_transmittance(ray);
    *result_0 = result.x;
    *result_1 = result.y;
    *result_2 = result.z;
}

__device__ void medium_sample_distance(
    float si_t, float /*si_time*/,
    float /*si_wavelengths_0*/, float /*si_wavelengths_1*/,
    float /*si_wavelengths_2*/, float /*si_wavelengths_3*/,
    float si_p_0, float si_p_1, float si_p_2,
    void * /*si_shape*/,
    float si_uv_0, float si_uv_1,
    float si_n_0, float si_n_1, float si_n_2,
    float si_sh_frame_s_0, float si_sh_frame_s_1, float si_sh_frame_s_2,
    float si_sh_frame_t_0, float si_sh_frame_t_1, float si_sh_frame_t_2,
    float si_sh_frame_n_0, float si_sh_frame_n_1, float si_sh_frame_n_2,
    float /*si_dp_du_0*/, float /*si_dp_du_1*/, float /*si_dp_du_2*/,
    float /*si_dp_dv_0*/, float /*si_dp_dv_1*/, float /*si_dp_dv_2*/,
    float /*si_duv_dx_0*/, float /*si_duv_dx_1*/,
    float /*si_duv_dy_0*/, float /*si_duv_dy_1*/,
    float si_wi_0, float si_wi_1, float si_wi_2,
    uint32_t /*si_prim_index*/,
    void * /*si_instance*/,
    float ray_o_0, float ray_o_1, float ray_o_2,
    float ray_d_0, float ray_d_1, float ray_d_2,
    float /*ray_d_rcp_0*/, float /*ray_d_rcp_1*/, float /*ray_d_rcp_2*/,
    float ray_mint, float ray_maxt,
    float /*ray_time*/,
    float /*ray_wavelengths_0*/, float /*ray_wavelengths_1*/,
    float /*ray_wavelengths_2*/, float /*ray_wavelengths_3*/,
    float sample_0, float sample_1,
    uint32_t /*index*/,
    float *mi_t, float * /*mi_time*/,
    float * /*mi_wavelengths_0*/, float * /*mi_wavelengths_1*/,
    float * /*mi_wavelengths_2*/, float * /*mi_wavelengths_3*/,
    float *mi_p_0, float *mi_p_1, float *mi_p_2,
    void ** /*mi_medium*/,
    float *mi_sh_frame_s_0, float *mi_sh_frame_s_1, float *mi_sh_frame_s_2,
    float *mi_sh_frame_t_0, float *mi_sh_frame_t_1, float *mi_sh_frame_t_2,
    float *mi_sh_frame_n_0, float *mi_sh_frame_n_1, float *mi_sh_frame_n_2,
    float *mi_wi_0, float *mi_wi_1, float *mi_wi_2,
    float *throughput_0, float *throughput_1, float *throughput_2, float * /*throughput_3*/) {

    SurfaceInteraction si;
    si.t = si_t;
    si.p.x = si_p_0;
    si.p.y = si_p_1;
    si.p.z = si_p_2;
    si.uv.x = si_uv_0;
    si.uv.y = si_uv_1;
    si.normal.x = si_n_0;
    si.normal.y = si_n_1;
    si.normal.z = si_n_2;
    si.sh_frame.s.x = si_sh_frame_s_0;
    si.sh_frame.s.y = si_sh_frame_s_1;
    si.sh_frame.s.z = si_sh_frame_s_2;
    si.sh_frame.t.x = si_sh_frame_t_0;
    si.sh_frame.t.y = si_sh_frame_t_1;
    si.sh_frame.t.z = si_sh_frame_t_2;
    si.sh_frame.n.x = si_sh_frame_n_0;
    si.sh_frame.n.y = si_sh_frame_n_1;
    si.sh_frame.n.z = si_sh_frame_n_2;
    si.wi.x = si_wi_0;
    si.wi.y = si_wi_1;
    si.wi.z = si_wi_2;

    auto ray = optix::make_Ray(
        optix::make_float3(ray_o_0, ray_o_1, ray_o_2),
        optix::make_float3(ray_d_0, ray_d_1, ray_d_2),
        1,
        ray_mint, ray_maxt
    );

    MediumSample mi = detail::custom_sample_distance(si, ray,
                                                     optix::make_float2(sample_0, sample_1));
    *mi_t = mi.t;
    *mi_p_0 = mi.p.x;
    *mi_p_1 = mi.p.y;
    *mi_p_2 = mi.p.z;
    *mi_sh_frame_s_0 = mi.sh_frame.s.x;
    *mi_sh_frame_s_1 = mi.sh_frame.s.y;
    *mi_sh_frame_s_2 = mi.sh_frame.s.z;
    *mi_sh_frame_t_0 = mi.sh_frame.t.x;
    *mi_sh_frame_t_1 = mi.sh_frame.t.y;
    *mi_sh_frame_t_2 = mi.sh_frame.t.z;
    *mi_sh_frame_n_0 = mi.sh_frame.n.x;
    *mi_sh_frame_n_1 = mi.sh_frame.n.y;
    *mi_sh_frame_n_2 = mi.sh_frame.n.z;
    *mi_wi_0 = mi.wi.x;
    *mi_wi_1 = mi.wi.y;
    *mi_wi_2 = mi.wi.z;
    *throughput_0 = mi.throughput.x;
    *throughput_1 = mi.throughput.y;
    *throughput_2 = mi.throughput.z;
}

__device__ void medium_pdf_distance(
    float ray_o_0, float ray_o_1, float ray_o_2,
    float ray_d_0, float ray_d_1, float ray_d_2,
    float /*ray_d_rcp_0*/, float /*ray_d_rcp_1*/, float /*ray_d_rcp_2*/,
    float ray_mint, float ray_maxt,
    float /*ray_time*/,
    float /*ray_wavelengths_0*/, float /*ray_wavelengths_1*/,
    float /*ray_wavelengths_2*/, float /*ray_wavelengths_3*/,
    float t,
    float *result) {

    auto ray = optix::make_Ray(
        optix::make_float3(ray_o_0, ray_o_1, ray_o_2),
        optix::make_float3(ray_d_0, ray_d_1, ray_d_2),
        1,
        ray_mint, ray_maxt
    );
    *result = detail::custom_pdf_distance(ray, t);
}



}  // end extern "C"
