/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include <math_constants.h>
#include "area.cuh"
#include "bsdf.cuh"
#include "common.cuh"
#include "random.cuh"
#include "types.cuh"
#include "envmap.cuh"

using namespace optix;

/// Main constants
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(ShadowPayload, payload_shadow, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(int, kHasEnvmap, , );

rtDeclareVariable(rtObject, top_object, , );
/// Atributes (on hit)
rtDeclareVariable(float3, p, attribute p, );
rtDeclareVariable(float2, uv, attribute uv, );
rtDeclareVariable(float3, ns, attribute ns, );
rtDeclareVariable(float3, ng, attribute ng, );

/// List of all material parameters in the scene
rtBuffer<MaterialProperties> material_parameters;
/// Index of the curent material parameters
rtDeclareVariable(uint, material_id, , );
/// List of all emitter parameters in the scene
rtBuffer<EmitterProperties> emitter_parameters;


RT_FUNCTION float3 sample_emitters(SurfaceInteraction &ref, RayPayload &prd) {
    // Select a light source uniformly (including envmap if there's one)
    uint index = static_cast<uint>(floorf(rnd(prd.seed) * kNumberOfLights));
    index = optix::clamp(index, 0u, kNumberOfLights - 1);

    DirectionSample ds;
    // Convention: envmap (if present) is the emitter with index 0
    if (kHasEnvmap && index == 0) {
        float2 sample = make_float2(rnd(prd.seed), rnd(prd.seed));
        ds = call_envmap_sample_direction(ref, sample);
    } else {
        EmitterProperties light = emitter_parameters[kHasEnvmap ? index - 1 : index];
        // Sample a direction toward the light source
        ds.radiance = AreaEmitter::sample(light, ds, ref, prd);
    }

    if (ds.pdf <= 0.f)
        return make_float3(0.f);
    ds.pdf /= kNumberOfLights;

    /* Check for visibility with a shadow ray (ray type = 2) */ {
        ShadowPayload prd_shadow;
        prd_shadow.shadowed = false;
        Ray shadow_ray = spawn_shadow_ray(ref.p, ds.light_direction, ds.light_dist);
        rtTrace(top_object, shadow_ray, prd_shadow, RT_RAY_FLAG_TERMINATE_ON_FIRST_HIT);

        if (prd_shadow.shadowed)
            return make_float3(0.f);
    }

    // Light direction in the frame of the surface being queried (not the emitter's surface)
    float3 to_light_local = ref.to_local(ds.light_direction);
    // BSDF value toward that emitter
    float3 bsdf_val = call_bsdf_eval(ref, to_light_local);
    ds.radiance *= bsdf_val * kNumberOfLights;

    // Probability of sampling that direction from the BSDF (for MIS)
    float bsdf_pdf = call_bsdf_pdf(ref, to_light_local);
    float mis      = power_heuristic(ds.pdf, bsdf_pdf);

    return mis * ds.radiance;
}

RT_PROGRAM void ray_hit_normals() {
    // Visualize normal vectors
    float3 norm;
    norm = rtTransformNormal(RT_OBJECT_TO_WORLD, ns);
    norm = normalize(norm);
    payload.result = make_float4(abs(norm.x), abs(norm.y), abs(norm.z), 0.f);
}

RT_PROGRAM void ray_hit_closest() {
    MaterialProperties props = material_parameters[material_id];
    SurfaceInteraction si(ray, p, normalize(ns));
    si.uv = uv;

    // Emission at this object (not an envmap, an actual shape)
    if (props.emitter >= 0) {
        EmitterProperties light = emitter_parameters[props.emitter];
        float3 Le = AreaEmitter::eval(light, si);

        float mis_weight = 1.f;
        // MIS: only if the previous interaction wasn't a delta BSDF
        if (payload.depth > 1 && payload.bsdf_sampled_type != 1) {
            // Query the PDF of sampling this emitter (for MIS)
            // Note: this doesn't make sense for camera rays.
            DirectionSample ds;
            // World coordinates, points toward the light from the reference point
            ds.light_direction = ray.direction;
            ds.normal = light.normal;  // TODO: si.normal
            ds.light_dist = sqrtf(squared_norm(ray.origin - si.p));
            float light_pdf = AreaEmitter::pdf(light, ds) / kNumberOfLights;
            mis_weight = power_heuristic(payload.bsdf_pdf, light_pdf);
        }

        payload.result += make_float4(mis_weight * payload.throughput * Le, 0.f);
    }

    // Skip sampling the next direction if max_depth is already reached
    if (payload.depth >= payload.max_depth)
        return;

    // ----- Emitter sampling
    if (!props.is_delta) {
        float3 emitter_weight = sample_emitters(si, payload);
        payload.result += make_float4(payload.throughput * emitter_weight, 0.f);
    }

    // ----- BSDF sampling
    // Sample next direction from the BSDF (result stored in the payload).
    // Also sets the pdf of sampling that direction.
    float sample1 = rnd(payload.seed);
    float2 sample2 = make_float2(rnd(payload.seed), rnd(payload.seed));
    auto bs = call_bsdf_sample(si, sample1, sample2);

    // Transform sampled direction to world coordinates
    payload.bsdf_dir = si.to_world(bs.wo);
    payload.bsdf_pdf = bs.pdf;
    payload.throughput *= bs.sampling_weight;
    payload.bsdf_origin = si.p;
    payload.eta *= bs.eta;
    // TODO use bs.sampled_type
    payload.bsdf_sampled_type = props.is_delta ? 1 : 0;
}
