#pragma once

#include <optix_world.h>
#include <math_constants.h>
#include "common.cuh"
#include "random.cuh"
#include "types.cuh"
#include "envmap.cuh"

rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(int, kHasEnvmap, , );
rtDeclareVariable(uint, kRRDepth, , "Maximum path depth (0: visible emitters, 1: direct illumination, etc.)");
rtDeclareVariable(int, kHideEmitters, , "Whether to hide emitters from the final image (e.g. envmap)");
rtDeclareVariable(rtObject, top_object, , );


RT_FUNCTION void eval(optix::Ray ray, RayPayload& payload) {
    payload.bsdf_origin = ray.origin;
    payload.bsdf_dir = ray.direction;

    while (payload.depth < payload.max_depth && !payload.escaped) {
        payload.depth += 1;

        /* Russian roulette: try to keep path weights equal to one.
           Stop with at least some probability to avoid getting stuck
           (e.g. due to total internal reflection) */
        if (payload.depth > kRRDepth) {
            float q = min(fmaxf(payload.throughput) * payload.eta * payload.eta, .95f);
            if (rnd(payload.seed) >= q)
                break;
            payload.throughput *= 1.f / q;
        }

        // Trace the next ray, accumulate emission there
        rtTrace(top_object, ray, payload);
        // Prepare next ray
        ray = spawn_ray(payload.bsdf_origin, payload.bsdf_dir);
    }

    // TODO: for recursive rays, how to handle kHideEmitters?
    if (payload.escaped && !kHideEmitters && kHasEnvmap) {
        float3 Le = call_envmap_eval(payload.bsdf_dir);

        float mis_weight = 1.f;
        if (payload.depth > 1 && payload.bsdf_sampled_type != 1) {
            // Query the PDF of sampling this emitter (for MIS)
            float light_pdf = call_envmap_pdf_direction(payload.bsdf_dir) / kNumberOfLights;
            mis_weight = power_heuristic(payload.bsdf_pdf, light_pdf);
        }

        payload.result += make_float4(mis_weight * payload.throughput * Le, 0.f);
    }
}
