/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include <math_constants.h>
#include "area.cuh"
#include "bsdf.cuh"
#include "bsdf_d.cuh"
#include "envmap.cuh"
#include "common.cuh"
#include "random.cuh"
#include "types.cuh"
#include "radiative_backprop.cuh"

using namespace optix;

#define RB_USE_EMITTER_SAMPLING


/// Main constants
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(int, kBackpropEnabled, , "");
rtDeclareVariable(int, kMakeImage, , "");

// Already declared in header
// rtDeclareVariable(uint, kNumberOfLights, , );
// rtDeclareVariable(uint, kHasEnvmap, , );
// rtDeclareVariable(rtObject, top_object, , );

/// Atributes (on hit)
rtDeclareVariable(float3, p, attribute p, );
rtDeclareVariable(float2, uv, attribute uv, );
rtDeclareVariable(float3, ns, attribute ns, );
rtDeclareVariable(float3, ng, attribute ng, );

/// List of all material parameters in the scene
rtBuffer<MaterialProperties> material_parameters;
/// Index of the curent material parameters
rtDeclareVariable(uint, material_id, , );
/// List of all emitter parameters in the scene
rtBuffer<EmitterProperties> emitter_parameters;

// --- Optimization flags
rtDeclareVariable(int, kUseApproxLi1, , );


RT_FUNCTION float3 sample_emitters(SurfaceInteraction &ref, RayPayload &prd) {
    // Select a light source uniformly (including envmap if there's one)
    uint index = static_cast<uint>(floorf(rnd(prd.seed) * kNumberOfLights));
    index = optix::clamp(index, 0u, kNumberOfLights - 1);

    DirectionSample ds;
    // Convention: envmap (if present) is the emitter with index 0
    if (kHasEnvmap && index == 0) {
        float2 sample = make_float2(rnd(prd.seed), rnd(prd.seed));
        ds = call_envmap_sample_direction(ref, sample);
    } else {
        EmitterProperties light = emitter_parameters[kHasEnvmap ? index - 1 : index];
        // Sample a direction toward the light source
        ds.radiance = AreaEmitter::sample(light, ds, ref, prd);
    }

    if (ds.pdf <= 0.f)
        return make_float3(0.f);
    // Adjust PDF to include choice of that emitter
    ds.pdf /= kNumberOfLights;
    ds.radiance *= kNumberOfLights;

    /* Check for visibility with a shadow ray (ray type = 2) */ {
        ShadowPayload prd_shadow;
        prd_shadow.shadowed = false;
        Ray shadow_ray = spawn_shadow_ray(ref.p, ds.light_direction, ds.light_dist);
        rtTrace(top_object, shadow_ray, prd_shadow, RT_RAY_FLAG_TERMINATE_ON_FIRST_HIT);

        if (prd_shadow.shadowed)
            return make_float3(0.f);
    }

    // Light direction in the frame of the surface we're querying from (not the emitter's surface)
    float3 to_light_local = ref.to_local(ds.light_direction);

    // Probability of sampling that direction from the BSDF (for MIS)
    float bsdf_pdf = call_bsdf_pdf(ref, to_light_local);
    float mis      = power_heuristic(ds.pdf, bsdf_pdf);

#if defined(RB_USE_EMITTER_SAMPLING)
    // static_assert(false, "this is most likely broken, check carefully with envmaps");

    const MaterialProperties &props = material_parameters[material_id];
    bool is_differentiable = kBackpropEnabled && props.is_differentiable && prd.carries_derivatives();

    if (prd.max_depth <= prd.depth)
        printf("RD assert: shouldn't call emitter sampling, max_depth reached! \n");

    if (is_differentiable && !props.is_delta) {

        float3 Li = make_float3(0.f);
        if (kUseApproxLi1) {
            Li = make_float3(1.f);
        } else {
            // TODO: why do we match AD better when advancing the seed here?
            // size_t secondary_seed = tea<16>(prd.seed, 0);
            rnd(payload.seed);  // Advance seed

            // TODO: should we decorrelate?
            // Note: ds.radiance already includes the 1. / light_pdf factor
            Li = ds.radiance;

            // Ray weight_ray = spawn_ray(ref.p, ds.light_direction);
            // auto weight_payload = eval(secondary_seed, weight_ray, make_float3(0.f),
            //                            prd.max_depth - prd.depth);
            // Li = make_float3(weight_payload.result.x, weight_payload.result.y,
            //                  weight_payload.result.z);

        }
        float3 edge_weight = payload.throughput * Li * mis * prd.delta;
        call_bsdf_eval_d(ref, to_light_local, edge_weight);
    }
#endif

    // BSDF value toward that emitter
    float3 bsdf_val = call_bsdf_eval(ref, to_light_local);

    return mis * bsdf_val * ds.radiance;
}

RT_PROGRAM void ray_hit_closest() {
    const MaterialProperties &props = material_parameters[material_id];
    SurfaceInteraction si(ray, p, normalize(ns));
    si.uv = uv;
    bool is_differentiable = kBackpropEnabled && props.is_differentiable
                             && payload.carries_derivatives();
    bool is_weight_ray = kBackpropEnabled && !payload.carries_derivatives();
    bool gather_light = !kBackpropEnabled || kMakeImage || is_differentiable || is_weight_ray;

    // Emission at this object
    if (gather_light && props.emitter >= 0) {
        const EmitterProperties &light = emitter_parameters[props.emitter];
        float3 Le = AreaEmitter::eval(light, si);

        float mis_weight = 1.f;
        // MIS: only if the previous interaction wasn't a delta BSDF
        // Note: since we are on a surface, we know that it couldn't have been
        // the result of envmap sampling.
        // Note: this doesn't make sense for camera rays, since we
        // don't do emitter sampling from the camera, only from surfaces.
        if (payload.depth > 1 && payload.bsdf_sampled_type != 1) {
            // Query the PDF of sampling this emitter directly with emitter sampling from the
            // previous surface (MIS).
            DirectionSample ds;
            ds.light_direction = ray.direction;
            ds.normal = light.normal;  // TODO: si.normal at the light source
            ds.light_dist = sqrtf(squared_norm(ray.origin - si.p));
            float light_pdf = AreaEmitter::pdf(light, ds) / kNumberOfLights;
            mis_weight = power_heuristic(payload.bsdf_pdf, light_pdf);
        }

        // TODO: accumulate gradients for emitters
        // if (is_differentiable) {
        //     float3 d = payload.throughput * payload.delta;
        // }

        payload.result += make_float4(mis_weight * payload.throughput * Le, 0.f);
    }

    // Skip sampling the next direction if max_depth is already reached
    if (payload.depth >= payload.max_depth)
        return;

    // ----- Emitter sampling
    if (gather_light && !props.is_delta) {
        // Note: `emitter_weight` includes the MIS weight
        float3 emitter_weight = sample_emitters(si, payload);
        payload.result += make_float4(payload.throughput * emitter_weight, 0.f);
    }

    // ----- BSDF sampling
    // Sample next direction from the BSDF (result stored in the payload).
    // Also sets the pdf of sampling that direction.
    float sample1 = rnd(payload.seed);
    float2 sample2 = make_float2(rnd(payload.seed), rnd(payload.seed));
    auto bs = call_bsdf_sample(si, sample1, sample2);

    // ----- BSDF derivatives
    // Note: here we know that `payload.depth < payload.max_depth`, so the
    // derivative edge weight can still be > 0 by at least hitting a light source
    // directly (bringing depth to max_depth).
    if (is_differentiable) {
        // Weight of this edge: non-differentiable incoming Li
        // Estimate it with 1 non-differentiable sample (= 1 ray sampled from this BSDF).
        Ray weight_ray = spawn_ray(si.p, si.to_world(bs.wo));
        RayPayload weight_payload(tea<16>(payload.seed, 0),
                                  payload.max_depth - payload.depth + 1);

        if (kUseApproxLi1) {
            weight_payload.result = make_float4(1.f);
        } else {
            rnd(payload.seed);  // Advance seed
            weight_payload.depth = 1;
            weight_payload.bsdf_pdf = bs.pdf;
            weight_payload.eta *= bs.eta;
            weight_payload.bsdf_sampled_type = props.is_delta ? 1 : 0;

            eval(weight_ray, weight_payload);
        }

        float3 Li = make_float3(weight_payload.result.x, weight_payload.result.y,
                                weight_payload.result.z);
        float3 edge_weight = payload.throughput * Li * payload.delta / bs.pdf;
        call_bsdf_eval_d(si, bs.wo, edge_weight);
    }

    // Transform sampled direction to world coordinates
    payload.bsdf_dir = si.to_world(bs.wo);
    payload.bsdf_pdf = bs.pdf;
    payload.throughput *= bs.sampling_weight;
    payload.bsdf_origin = si.p;
    payload.eta *= bs.eta;
    // TODO use bs.sampled_type
    payload.bsdf_sampled_type = props.is_delta ? 1 : 0;
}
