/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include <math_constants.h>
#include "radiative_backprop_vol.cuh"
#include "types.cuh"
#include "common.cuh"

using namespace optix;

/// Main constants
rtDeclareVariable(uint3, launch_index, rtLaunchIndex, );
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(ShadowPayload, payload_shadow, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(uint, kNx, , "Image width");
rtDeclareVariable(uint, kNy, , "Image height");
rtDeclareVariable(uint, kSpp, , "Samples per pixel == number of launches");
rtDeclareVariable(uint, kBaseSeed, , "Base seed (e.g. changing frame to frame)");
rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(uint, kMaxDepth, , "Maximum path depth (0: visible emitters, 1: direct illumination, etc.)");
rtDeclareVariable(int, kBackpropEnabled, , "");
rtDeclareVariable(int, kMakeImage, , "");

// rtDeclareVariable(rtObject, top_object, , );  // Already declared in the header
rtDeclareVariable(float3, camera_pos, , "Position of the camera in world space");
rtDeclareVariable(float3, camera_u, , );
rtDeclareVariable(float3, camera_v, , );
rtDeclareVariable(float3, camera_w, , "Vector pointing in the eye direction in world space");
rtDeclareVariable(float,  camera_near, , "Near clip plane");
rtDeclareVariable(float,  camera_far, , "Far clip plane");
rtDeclareVariable(Matrix4x4,  sample_to_camera, , "Transformation from a 2D point on the film to camera space");
rtDeclareVariable(float,  scene_scale, , "Maximal extent of the scene's bounding box");

/**
 * Input derivatives buffer: partial derivative of the loss
 * with respect to each image pixel (RGB).
 */
rtBuffer<float> in_delta;
/// Output buffers
rtBuffer<float> out_image;


RT_PROGRAM void ray_gen() {
    const float kInvSpp = 1.f / kSpp;

    uint seed_seed = kSpp * (launch_index.y * kNx + launch_index.x) + launch_index.z;
    seed_seed = tea<16>(seed_seed, 0) + kBaseSeed;

    /// Corner of the pixel from which to shoot this ray
    uint idx = 3 * (launch_index.y * kNx + launch_index.x);

    /// Position of the ray origin on the film (jittered)
    float4 sample_pos = make_float4(
        (launch_index.x + rnd(seed_seed)) / float(kNx),
        (launch_index.y + rnd(seed_seed)) / float(kNy),
        0.f, 1.f);
    // Ray generation from a perspective camera (using homogeneous coordinates)
    float4 near_p = sample_to_camera * sample_pos;
    float3 d = normalize(make_float3(near_p.x, near_p.y, near_p.z));
    float3 primary_dir = normalize(d.x * camera_v + d.y * camera_u + d.z * camera_w);
    // Lookup derivatives carried by this ray
    // TODO: use bilinear interpolation? (should match the reconstruction filter)
    float3 primary_delta;
    if (kBackpropEnabled)
        primary_delta = make_float3(in_delta[idx + 0], in_delta[idx + 1],
                                    in_delta[idx + 2]) / kSpp;
    else
        primary_delta = make_float3(0.f);
    float inv_z = 1.f / d.z;

    Ray primary_ray = make_Ray(camera_pos, primary_dir, 1,
                               // min_t, max_t
                               camera_near * inv_z, camera_far * inv_z);
    RayPayload prd = eval(seed_seed, primary_ray, primary_delta, kMaxDepth);

    if (kMakeImage) {
        // Accumulate ray's contribution to its pixel
        atomicAdd(&out_image[idx + 0], kInvSpp * prd.result.x);
        atomicAdd(&out_image[idx + 1], kInvSpp * prd.result.y);
        atomicAdd(&out_image[idx + 2], kInvSpp * prd.result.z);
    }
}

/// Miss program for standard rays
RT_PROGRAM void ray_miss() {
    payload.escaped = true;
}

/// Hit program for shadow rays
RT_PROGRAM void shadow_closest_hit() {
    payload_shadow.shadowed = true;
}
