/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include <math_constants.h>
#include "radiative_backprop_vol.cuh"
#include "area.cuh"
#include "bsdf.cuh"
#include "envmap.cuh"
#include "media.cuh"
#include "media_d.cuh"
#include "common.cuh"
#include "random.cuh"
#include "types.cuh"
#include "bbox.cuh"

using namespace optix;

/// Main constants
rtDeclareVariable(uint3, launch_index, rtLaunchIndex, );
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(Ray, global_ray, rtCurrentRay,);
rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(uint, kMaxDepth, , "Maximum path depth (0: visible emitters, 1: direct illumination, etc.)");
rtDeclareVariable(int, kBackpropEnabled, , "");

// rtDeclareVariable(rtObject, top_object, , );  // Already declared in the header

/// Atributes (on hit)
rtDeclareVariable(float3, p, attribute p, );
rtDeclareVariable(float2, uv, attribute uv, );
rtDeclareVariable(float3, ns, attribute ns, );
rtDeclareVariable(float3, ng, attribute ng, );

/// List of all material parameters in the scene
rtBuffer<MaterialProperties> material_parameters;
/// Index of the curent material parameters
rtDeclareVariable(uint, material_id, , );
/// List of all emitter parameters in the scene
rtBuffer<EmitterProperties> emitter_parameters;

/// Index of the curent interior medium, if any
rtDeclareVariable(int, medium_id, , );
/// Bounding box corners for the medium
rtDeclareVariable(float3, medium_bbox_min, , );
rtDeclareVariable(float3, medium_bbox_max, , );

// --- Optimization flags
rtDeclareVariable(int, kUseApproxLi1, , );

// TODO: replace by a global variable, taken from the scene
constexpr size_t kRRDepth = 16;


/**
 * Sample a position on a light source and evaluate how much light makes it
 * through the medium to the current location.
 *
 * Note: we assume that we are sampling from within a medium, so there's attenuation.
 * Note: we assume there's *no surface* and *no other medium* in the scene.
 * Note: we ignore all types of emitters except for the environment map.
 */
RT_FUNCTION DirectionSample sample_emitter_direction_attenuated(const RayPayload &prd,
                                                                const MediumSample &mi,
                                                                const optix::float2 &sample,
                                                                const BoundingBox3f &medium_bbox) {
    // TODO: also support sampling from area emitters (and multiply by prob. of picking emitter)
    if (emitter_parameters.size() > 0) {
        printf("Not implemented yet: sample_emitter_direction_attenuated with area emitters.\n");
    }

    // Sample a direction toward the emitter.
    // Since we ignore surfaces, no need to test for visibility in that direction.
    DirectionSample ds = call_envmap_sample_direction(mi, sample);

    ds.radiance = call_envmap_eval(ds.light_direction) / ds.pdf;

    if (prd.in_medium) {
        // Attenuation by the medium.
        auto it = intersect_bbox(mi.p, ds.light_direction, medium_bbox);

        Ray tr_ray = make_Ray(mi.p, ds.light_direction, 1, optix::max(0.f, it.mint), it.maxt);
        ds.radiance *= call_medium_eval_transmittance(tr_ray);
    } else {
        printf("Weird, shouldn't do emitter sampling from outside medium!\n");
    }

    return ds;
}


RT_PROGRAM void ray_hit_closest() {
    payload.escaped = false;

    // ----- Assumptions
    const BoundingBox3f medium_bbox = { medium_bbox_min, medium_bbox_max };
    if (medium_id < 0) {
        // TODO: proper surface case with BSDF sampling, etc
        printf("Not implemented yet: surfaces with volumes.\n");
    }

    MaterialProperties props = material_parameters[material_id];
    bool is_differentiable = kBackpropEnabled && props.is_differentiable
                             && payload.carries_derivatives();

    // ----- Traversing the medium
    bool active = !payload.escaped;
    Ray inner_ray = global_ray;
    do {
        /* Russian roulette: try to keep path weights equal to one.
           Stop with at least some probability to avoid getting stuck
           (e.g. due to total internal reflection) */
        // TODO: check that this doesn't break gradients
        if (payload.depth > kRRDepth) {
            float q = min(fmaxf(payload.throughput), 0.95f);
            if (rnd(payload.seed) >= q) {
                payload.escaped = false;
                break;
            }
            payload.throughput *= 1.f / q;
        }

        // ----- Intersecting the medium's bounding box
        auto it = intersect_bbox(payload.bsdf_origin, payload.bsdf_dir, medium_bbox);
        if (it.missed) {
            // This should rarely happen (just numerical errors)
            payload.escaped = true;
            payload.in_medium = false;
            printf("Accidentally missed:"
                   " inner_ray.o = %f, %f, %f     inner_ray.d = %f, %f, %f. \n"
                   "        it.mint = %f, it.maxt = %f\n",
                   inner_ray.origin.x, inner_ray.origin.y, inner_ray.origin.z,
                   inner_ray.direction.x, inner_ray.direction.y, inner_ray.direction.z,
                   it.mint, it.maxt);
            return;
        }
        if (!payload.in_medium) {
            // Implicitly traverse the null BSDF, this adds 1 interaction
            payload.depth += 1;

            // We might have exceeded the max depth by traversing the null interface
            if (payload.depth >= payload.max_depth) {
                payload.escaped = false;
                payload.in_medium = true;
                break;
            }
        }

        // For the current interaction
        payload.depth += 1;
        payload.escaped = false;
        payload.in_medium = true;

        // Clip ray to extents of the medium (important for transmittance evaluation)
        float safe_mint = optix::max(0.f, it.mint);
        inner_ray = optix::make_Ray(payload.bsdf_origin + safe_mint * payload.bsdf_dir,
                                    payload.bsdf_dir, 1,
                                    // Starts at the current location in the medium and goes
                                    // to the medium's boundary on the other side.
                                    0.f, it.maxt - safe_mint);

        // Note: there may be precision issues here when setting the ray origin.
        // if (!bbox_contains(inner_ray.origin, medium_bbox)) {
        //     printf("Doesn't contain before sampling distance:\n"
        //            "    payload.depth = %d,\n"
        //            "    payload.bsdf_origin = %f, %f, %f\n"
        //            "    payload.bsdf_dir    = %f. %f, %f\n"
        //            "    inner_ray.origin    = %f, %f, %f\n"
        //            "    inner_ray.direction = %f. %f, %f\n"
        //            "    it.mint = %f, it.maxt = %f\n",
        //            payload.depth,
        //            payload.bsdf_origin.x, payload.bsdf_origin.y, payload.bsdf_origin.z,
        //            payload.bsdf_dir.x, payload.bsdf_dir.y, payload.bsdf_dir.z,
        //            inner_ray.origin.x, inner_ray.origin.y, inner_ray.origin.z,
        //            inner_ray.direction.x, inner_ray.direction.y, inner_ray.direction.z,
        //            it.mint, it.maxt
        //         );
        // }

        // Surface interaction with the medium boundary in the current direction
        SurfaceInteraction si(inner_ray.tmax,
                              inner_ray.origin + inner_ray.tmax * inner_ray.direction,
                              /* normal */ make_float3(0.f / 0.f),
                              -inner_ray.direction);

        float2 dist_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
        MediumSample mi = call_medium_sample_distance(si, inner_ray, dist_sample);
        if (mi.is_valid()) {
            // ----- Scattering event
            payload.escaped = false;

            // Sample a new direction from the phase function
            auto phase_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
            auto wo_local = sample_isotropic_phase(phase_sample);

            // Prepare origin & direction for next iteration
            payload.bsdf_origin = mi.p;
            payload.bsdf_dir = mi.to_world(wo_local);

            // TODO: make sure this doesn't happen too often.
#if 0
            if (!bbox_contains(payload.bsdf_origin, medium_bbox)) {
                // Accidentally escaped, even though we shouldn't have to
                payload.escaped = true;

                printf("Doesn't contain when preparing next ray:\n"
                        "    payload.depth = %d,\n"
                        "    payload.bsdf_origin = %f, %f, %f\n"
                        "    payload.bsdf_dir    = %f. %f, %f\n"
                        "    inner_ray.origin    = %f, %f, %f\n"
                        "    inner_ray.direction = %f. %f, %f\n"
                        "    wo_local            = %f, %f, %f\n"
                        "    mi.t = %f, it.mint = %f, it.maxt = %f, inner_ray.tmax = %f\n",
                        payload.depth,
                        payload.bsdf_origin.x, payload.bsdf_origin.y, payload.bsdf_origin.z,
                        payload.bsdf_dir.x, payload.bsdf_dir.y, payload.bsdf_dir.z,
                        inner_ray.origin.x, inner_ray.origin.y, inner_ray.origin.z,
                        inner_ray.direction.x, inner_ray.direction.y, inner_ray.direction.z,
                        wo_local.x, wo_local.y, wo_local.z,
                        mi.t, it.mint, it.maxt, inner_ray.tmax
                    );
                // TODO: remove this
                payload.escaped = false;
                payload.throughput = make_float3(0.f);
                payload.result = make_float4(0.f);
                return;
            }
#endif

            // ----- Emitter sampling
            auto emitter_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
            auto ds = sample_emitter_direction_attenuated(payload, mi, emitter_sample, medium_bbox);
            payload.did_emitter_sampling = true;
            if (ds.pdf > 0.f) {
                float phase_val = eval_isotropic_phase();
                payload.result += make_float4(mi.throughput * payload.throughput
                                              * phase_val * ds.radiance, 0.f);
            }

            // // TODO: remove this
            // // This is to check that recursive rays work as expected
            // Ray recursive_ray = optix::make_Ray(payload.bsdf_origin, payload.bsdf_dir,
            //                                     1, math::Epsilon, RT_DEFAULT_MAX);
            // RayPayload recursive_payload = eval(payload.seed, recursive_ray,
            //                                     make_float3(0.f),
            //                                     payload.max_depth - payload.depth, true,
            //                                     payload.did_emitter_sampling);
            // payload.escaped = false;
            // // Still need this payload's throughput if it receives envmap light
            // payload.throughput *= mi.throughput;
            // payload.result += make_float4(payload.throughput, 0.f) * recursive_payload.result;
            // return;

        } else {
            // Implicitly traverse the null BSDF to go out, this adds 1 interaction
            // We might have exceeded the max depth by traversing the null interface
            bool did_escape = payload.depth < payload.max_depth;
            payload.escaped = did_escape;
            payload.in_medium = !did_escape;
            if (did_escape)
                payload.depth += 1;
        }

        // ----- Radiative backprop
        if (is_differentiable) {
            Ray tr_ray = inner_ray;
            tr_ray.tmin = 0.f;
            tr_ray.tmax = mi.is_valid() ? mi.t : si.t;
            // `mi.t` is +inf if there's no medium interaction, which is the right convention.
            float tr_pdf = call_medium_pdf_distance(tr_ray, mi.t);


            // /* Term: Q_2 * L_e */ {
            //     // Note: we accumulate here regardless of whether we sampled a medium interaction
            //     // or not, since it will be less noisy. If we only accumulated it when !mi.is_valid(),
            //     // we would need an additional 1 / tr_pdf term.
            //     // Here, inner_ray goes from the current position (perhaps the
            //     // boundary of the medium) to the next boundary along the ray.
            //     auto Le = call_envmap_eval(inner_ray.direction);

            //     // Ray exit_ray = inner_ray;
            //     // exit_ray.tmin = 0.f;
            //     // exit_ray.tmax = si.t;
            //     // float exit_pdf = call_medium_pdf_distance(exit_ray, math::Infinity);
            //     // auto edge_weight = Le * payload.throughput * payload.delta / exit_pdf;

            //     auto edge_weight = Le * payload.throughput * payload.delta;
            //     call_medium_eval_transmittance_d(inner_ray, edge_weight);
            // }


            // // TODO: check why this happens, make sure it doesn't happen too often
            // if (mi.is_valid() && tr_pdf <= 0.f) {
            //     float t = tr_ray.tmax - tr_ray.tmin;
            //     float density_val = 5.f;
            //     float recomputed = density_val * std::exp(-t * density_val);

            //     optix::float3 lookup_p = tr_ray.origin + mi.t * tr_ray.direction;
            //     float density_lookup   = call_medium_density(0.f, lookup_p);

            //     printf("\nWeird: payload.depth = %d, payload.max_depth = %d, mi.t = %f, si.t = %f\n"
            //            "mi.throughput = %f, %f, %f, payload.throughput = %f, %f, %f,\n"
            //            "inner_ray.tmin = %f, inner_ray.tmax = %f,\n"
            //            "tr_ray.tmin = %f, tr_ray.tmax = %f,\n"
            //            "tr_pdf = %f, recomputed for d = 5: %f\n"
            //            "medium lookup = %f  at p = %f, %f, %f\n",
            //            payload.depth, payload.max_depth, mi.t, si.t,
            //            mi.throughput.x, mi.throughput.y, mi.throughput.z,
            //            payload.throughput.x, payload.throughput.y, payload.throughput.z,
            //            inner_ray.tmin, inner_ray.tmax,
            //            tr_ray.tmin, tr_ray.tmax, tr_pdf, recomputed, density_lookup,
            //            lookup_p.x, lookup_p.y, lookup_p.z);
            // }

            if (mi.is_valid() && tr_pdf > 0.f) {
                // Term: G * Q_1 * L_i + Q_2 * K * L_i


                // Estimate Li at the newly sampled position & direction
                bool reached_end = payload.depth >= payload.max_depth;
                float3 Li = make_float3(0.f);
                if (!reached_end && kUseApproxLi1) {
                    Li = make_float3(1.f);
                } else if (!reached_end) {
                    Ray weight_ray = optix::make_Ray(payload.bsdf_origin, payload.bsdf_dir,
                                                     1, 0.f, RT_DEFAULT_MAX);

                    // TODO: figure out what's appropriate here
                    // bool recursive_ray_already_did_emitter_sampling = payload.did_emitter_sampling;
                    bool recursive_ray_already_did_emitter_sampling = false;
                    size_t secondary_seed = tea<16>(payload.seed, 0);
                    rnd(payload.seed);  // Advance seed
                    RayPayload weight_payload = eval(secondary_seed, weight_ray,
                                                     make_float3(0.f),
                                                     payload.max_depth - payload.depth, true,
                                                     recursive_ray_already_did_emitter_sampling);
                    // Perfect importance sampling of the phase function, so end contribution is 1.
                    float phase_weight = 1.f;
                    float mis_weight = 0.5f;
                    Li = make_float3(weight_payload.result.x, weight_payload.result.y,
                                     weight_payload.result.z) * phase_weight * mis_weight;
                }

                /* Second strategy: emitter sampling */ {
                    auto emitter_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
                    auto ds = sample_emitter_direction_attenuated(payload, mi, emitter_sample,
                                                                  medium_bbox);
                    if (ds.pdf > 0.f) {
                        float phase_val = eval_isotropic_phase();
                        // Here we didn't sample the phase function, so we multiply by its value.
                        // Note that `ds.radiance` includes the pdf and (non-differentiable)
                        // attenuation by the transmittance.
                        float mis_weight = 0.5f;
                        Li += phase_val * mis_weight * ds.radiance;
                    }
                }


#if defined(MTS_PTX_USE_COMBINED_SCATTERING_AND_TRANSMITTANCE)
                // -- Combined
                float3 edge_weight = payload.throughput * payload.delta * Li / tr_pdf;

                // /* TODO: remove this */ {
                //
                //     float pdf_re = 1.f / (4.f * math::Pi);
                //     // Attenuation
                //     auto its = intersect_bbox(mi.p, ds_leak.light_direction, medium_bbox);
                //     float tr = std::exp(-its.maxt * density_val);

                //     float3 radiance_re = make_float3(tr / pdf_re);
                //     float phase_val = eval_isotropic_phase();
                //     float3 li_recomputed = phase_val * radiance_re;
                //     printf("Li = %f, %f, %f,  li_recomputed = %f, %f, %f\n"
                //            "ds.pdf = %f   vs  pdf_re = %f\n"
                //            "ds.radiance = %f, %f, %f   vs  radiance_re = %f, %f, %f\n",
                //            Li.x, Li.y, Li.z, li_recomputed.x, li_recomputed.y, li_recomputed.z,
                //            ds_leak.pdf, pdf_re,
                //            ds_leak.radiance.x, ds_leak.radiance.y, ds_leak.radiance.z,
                //            radiance_re.x, radiance_re.y, radiance_re.z
                //         );
                // }

                call_medium_eval_scattering_times_transmittance_d(mi, tr_ray, edge_weight);

#else
                // printf("Accidental nope\n");

                // -- Separate
                // Note: mi.throughput is the transmittance sampling weight:
                //    scattering * Tr / pdf(Tr)
                // So we need to divide out a scattering term.
                float3 scattering = call_medium_eval_scattering(mi);
                float3 sigma_s_edge_weight = mi.throughput * payload.throughput * Li
                                             * payload.delta / scattering;
                // Recover from division by 0
                if (scattering.x == 0.f) sigma_s_edge_weight.x = 0.f;
                if (scattering.y == 0.f) sigma_s_edge_weight.y = 0.f;
                if (scattering.z == 0.f) sigma_s_edge_weight.z = 0.f;
                // Note: scattering coefficient will be evaluated mi.p, which is the new location.
                call_medium_eval_scattering_d(mi, sigma_s_edge_weight);

                // Here we can't use `mi.throughput` because it would include an unwanted Tr term.
                float3 tr_edge_weight = scattering * payload.throughput * Li
                                        * payload.delta / tr_pdf;
                call_medium_eval_transmittance_d(tr_ray, tr_edge_weight);
#endif

                // // Verification
                // float pdf_recomputed = exp(-mi.t * 0.1f);
                // printf("mi.t = %f, si.t = %f,\n"
                //        "recomputed = %f  vs hardcoded = %f\n\n",
                //        mi.t, si.t, tr_pdf, pdf_recomputed);

                // {  // Verification
                //     Ray tr_ray = inner_ray;
                //     tr_ray.tmin = 0.f;
                //     tr_ray.tmax = mi.t;

                //     float3 scattering = call_medium_eval_scattering(mi);
                //     float3 transmittance = call_medium_eval_transmittance(tr_ray);
                //     float3 recomputed = scattering * transmittance / tr_pdf;
                //     printf("mi.throughput = %f, %f, %f  vs  recomputed = %f, %f, %f\n"
                //            "scattering = %f, %f, %f;  transmittance = %f, %f, %f;  tr_pdf = %f \n",
                //            mi.throughput.x, mi.throughput.y, mi.throughput.z,
                //            recomputed.x, recomputed.y, recomputed.z,
                //            scattering.x, scattering.y, scattering.z,
                //            transmittance.x, transmittance.y, transmittance.z,
                //            tr_pdf
                //         );
                // }
            } else if (tr_pdf > 0.f) {
                // Term: Q_2 * L_e
                // Note that we still have to cross the null interface to see the light source.
                if (payload.escaped) {
                    // // TODO: remove this
                    // float t = tr_ray.tmax - tr_ray.tmin;
                    // float density_val = 5.f;
                    // float recomputed = std::exp(-t * density_val);

                    // printf("HEY payload.depth = %d, payload.max_depth = %d, mi.t = %f, si.t = %f\n"
                    //        "inner_ray.tmin = %f, inner_ray.tmax = %f,\n"
                    //        "tr_ray.tmin = %f, tr_ray.tmax = %f,\n"
                    //        "tr_pdf = %f, recomputed = %f\n",
                    //        payload.depth, payload.max_depth, mi.t, si.t,
                    //        inner_ray.tmin, inner_ray.tmax,
                    //        tr_ray.tmin, tr_ray.tmax, tr_pdf, recomputed);

                    auto Le = call_envmap_eval(inner_ray.direction);
                    auto edge_weight = Le * payload.throughput * payload.delta / tr_pdf;
                    call_medium_eval_transmittance_d(inner_ray, edge_weight);
                } else {
                    // printf("WHO payload.depth = %d, payload.max_depth = %d, mi.t = %f, si.t = %f\n"
                    //        "inner_ray.tmin = %f, inner_ray.tmax = %f,\n"
                    //        "tr_ray.tmin = %f, tr_ray.tmax = %f,\n"
                    //        "tr_pdf = %f\n",
                    //        payload.depth, payload.max_depth, mi.t, si.t,
                    //        inner_ray.tmin, inner_ray.tmax,
                    //        tr_ray.tmin, tr_ray.tmax, tr_pdf);
                }
            }
        }

        // Attenuation by the medium
        payload.throughput *= mi.throughput;
        active = !payload.escaped && (payload.throughput.x > 0.f || payload.throughput.y > 0.f
                                      || payload.throughput.z > 0.f);

    } while (payload.depth < payload.max_depth && active);
}
