/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include "area.cuh"
#include "bsdf.cuh"
#include "envmap.cuh"
#include "media.cuh"
#include "media_d.cuh"
#include "common.cuh"
#include "random.cuh"
#include "types.cuh"
#include "bbox.cuh"

using namespace optix;


/// Main constants
rtDeclareVariable(uint3, launch_index, rtLaunchIndex, );
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(uint, kMaxDepth, , "Maximum path depth (0: visible emitters, 1: direct illumination, etc.)");
rtDeclareVariable(int, kBackpropEnabled, , "");

/// Atributes (on hit)
rtDeclareVariable(float3, p, attribute p, );
rtDeclareVariable(float2, uv, attribute uv, );
rtDeclareVariable(float3, ns, attribute ns, );
rtDeclareVariable(float3, ng, attribute ng, );

/// List of all material parameters in the scene
rtBuffer<MaterialProperties> material_parameters;
/// Index of the curent material parameters
rtDeclareVariable(uint, material_id, , );

/// Index of the curent interior medium, if any
rtDeclareVariable(int, medium_id, , );


RT_PROGRAM void ray_hit_closest() {
    // ----- Assumptions
    const BoundingBox3f medium_bbox = { make_float3(0.f), make_float3(1.f) };

    // No scattering: we'll always go all the way through
    payload.escaped = true;

    // ----- Intersecting the medium
    auto it = intersect_bbox(ray, medium_bbox);
    if (it.missed) {
        // This should rarely happen (just numerical errors)
        return;
    }

    // ----- Traversing the medium
    // Clip ray to extents travelled in the medium (important for transmittance evaluation)
    Ray tr_ray = optix::make_Ray(ray.origin + it.mint * ray.direction, ray.direction,
                                 1, 0.f, it.maxt - it.mint);
    SurfaceInteraction si(tr_ray.tmax, tr_ray.origin + tr_ray.tmax * tr_ray.direction,
                          /* normal */ make_float3(0.f / 0.f),
                          -tr_ray.direction);

    // // TODO: remove this
    // float3 tr = call_medium_eval_transmittance(tr_ray);
    // float integrated_density = -std::log(tr.x);
    // payload.escaped = false;
    // // payload.result = make_float4(tr, 0.f);
    // payload.result = make_float4(integrated_density);
    // return;

    float2 dist_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
    MediumSample mi = call_medium_sample_distance(si, tr_ray, dist_sample);
    if (mi.is_valid()) {
        // Ray died (absorption)
        payload.escaped = false;
    }

    // ----- Radiative backprop
    MaterialProperties props = material_parameters[material_id];
    bool is_differentiable = kBackpropEnabled && props.is_differentiable
                            && payload.carries_derivatives();
    if (is_differentiable) {
        // \bar G \bar Q_1 * L_i
        auto Le = call_envmap_eval(ray.direction);
        auto edge_weight = Le * payload.delta;
        call_medium_eval_transmittance_d(tr_ray, edge_weight);
    }

    // Attenuation by the medium
    payload.throughput *= mi.throughput;
}
