#pragma once

#include <limits>
#include <optix.h>
#include <optix_world.h>


// #define MTS_PTX_USE_COMBINED_SCATTERING_AND_TRANSMITTANCE


#if !defined(RT_FUNCTION)
#define RT_FUNCTION __forceinline__ __device__
#endif

namespace math_detail {
    constexpr float kInfinity = std::numeric_limits<float>::infinity();
};

struct RayPayload {
    RT_FUNCTION RayPayload() {}

    RT_FUNCTION RayPayload(size_t seed_, size_t max_depth_)
        : seed(seed_), throughput(optix::make_float3(1.f)),
          result(optix::make_float4(0.f)), depth(0), max_depth(max_depth_),
          bsdf_origin(optix::make_float3(0.f)), bsdf_dir(optix::make_float3(0.f)),
          bsdf_pdf(0.f), bsdf_sampled_type(0), eta(1.f), escaped(false), in_medium(false),
          did_emitter_sampling(false),
          delta(optix::make_float3(0.f)) { }

    uint seed;
    optix::float3 throughput;
    optix::float4 result;
    uint depth;
    uint max_depth;
    /// Origin of the next ray
    optix::float3 bsdf_origin;
    /// Direction sampled by the BSDF
    optix::float3 bsdf_dir;
    float bsdf_pdf;
    uint32_t bsdf_sampled_type;

    float eta;

    /// Marks whether the ray escaped
    bool escaped;
    /// Whether the ray started from inside a medium
    bool in_medium;
    /// Whether the path included emitter sampling
    bool did_emitter_sampling;

    /// Amount of derivatives carried by this ray
    optix::float3 delta;

    RT_FUNCTION bool carries_derivatives() const {
        return (delta.x != 0) || (delta.y != 0) || (delta.z != 0);
    }
};

struct ShadowPayload {
    bool shadowed;
};

RT_FUNCTION void coordinate_system(optix::float3 n, optix::float3 &x, optix::float3 &y) {
    /* Based on "Building an Orthonormal Basis, Revisited" by
       Tom Duff, James Burgess, Per Christensen,
       Christophe Hery, Andrew Kensler, Max Liani,
       and Ryusuke Villemin (JCGT Vol 6, No 1, 2017) */

    float s = optix::copysignf(1.f, n.z),
          a = -1.f / (s + n.z),
          b = n.x * n.y * a;

    x = optix::make_float3(n.x * n.x * a * s + 1.f, b * s, -n.x * s);
    y = optix::make_float3(b, s + n.y * n.y * a, -n.y);
}

struct Onb {
    optix::float3 s, t, n;

    RT_FUNCTION Onb(const optix::float3 &n) : n(n) {
        coordinate_system(n, s, t);
    }

    RT_FUNCTION optix::float3 to_local(const optix::float3 &v) {
        return optix::make_float3(
            optix::dot(v, s),
            optix::dot(v, t),
            optix::dot(v, n)
        );
    }

    RT_FUNCTION optix::float3 to_world(const optix::float3 &v) {
        return s * v.x + t * v.y + n * v.z;
    }
};

struct Interaction {
    RT_FUNCTION Interaction()
        : t(0.f), p(optix::make_float3(0.f)) {}

    RT_FUNCTION Interaction(float t_, const optix::float3 &point_)
        : t(t_), p(point_) {}

    /// Distance traveled along the ray
    float t;
    /// Position on the surface in world coordinates
    optix::float3 p;
};

struct SurfaceInteraction {
    RT_FUNCTION SurfaceInteraction()
        : sh_frame(optix::make_float3(0.f, 0.f, 1.f)) {}

    RT_FUNCTION SurfaceInteraction(float t, const optix::float3 &point, const optix::float3 &n,
                                   const optix::float3 &wi_global)
        : t(t), p(point), normal(n), sh_frame(n), wi(to_local(wi_global)) {}

    RT_FUNCTION SurfaceInteraction(const optix::Ray &ray, const optix::float3 &it_point, const optix::float3 &n)
        : p(it_point), normal(n), sh_frame(n), wi(to_local(-ray.direction)) {
        optix::float3 d = it_point - ray.origin;
        t = std::sqrt(optix::dot(d, d));
    }

    /// Progress along the ray
    float t;
    /// Position on the surface in world coordinates
    optix::float3 p;
    /// Surface normal in world coordinates
    optix::float3 normal;
    /// Local shading frame
    Onb sh_frame;
    /// Incident direction in local coordinates, pointing away from the surface
    optix::float3 wi;
    /// Texture coordinates
    optix::float2 uv;


    RT_FUNCTION optix::float3 to_local(const optix::float3 &global_vector) {
        return sh_frame.to_local(global_vector);
    }

    RT_FUNCTION optix::float3 to_world(const optix::float3 &local_vector) {
        return sh_frame.to_world(local_vector);
    }
};

struct DirectionSample {
    optix::float3 p;
    // Points toward the light source, in world coordinates
    optix::float3 light_direction;
    optix::float3 normal;
    optix::float3 radiance;
    float pdf;
    float light_dist;
};

struct MediumSample {
    RT_FUNCTION MediumSample()
        : t(math_detail::kInfinity), sh_frame(optix::make_float3(0.f, 0.f, 1.f)) {}

    RT_FUNCTION bool is_valid() const {
        return t != math_detail::kInfinity;
    }

    float t;
    optix::float3 p;
    Onb sh_frame;
    optix::float3 wi;
    optix::float3 throughput;

    RT_FUNCTION optix::float3 to_local(const optix::float3 &global_vector) {
        return sh_frame.to_local(global_vector);
    }

    RT_FUNCTION optix::float3 to_world(const optix::float3 &local_vector) {
        return sh_frame.to_world(local_vector);
    }
};

struct alignas(16) MaterialProperties {
    RT_FUNCTION MaterialProperties() {
        emitter = -1;  // No emitter by default
        is_differentiable = false;
        is_delta = false;
    }

    /// Index into the array of emitter properties, or -1 if none.
    int32_t emitter;
    /// Whether this material should receive derivatives
    bool is_differentiable;
    /// If true, we can't do NEE from this material
    bool is_delta;
};

enum EEmitterType : uint32_t {
    EAreaMesh = 0,
    EEmitterTypeCount
};

struct alignas(16) EmitterProperties {
    RT_FUNCTION EmitterProperties() {
        sampling_table = nullptr;
        radiance = optix::make_float3(0.f);
        p = v0 = v1 = normal = optix::make_float3(0.f);
        area = 0.f;
        is_differentiable = false;
        emitter_type = EAreaMesh;
    }

    // optix::rtBuffer<float> sampling_table;
    float *sampling_table;
    optix::float3 radiance;
    /// Corner of the emitter's bounding box.
    optix::float3 p;
    /**
     * Sides of the emitter's bounding box.
     * Interpreted as a rectangle, as a temporary workaround for proper mesh sampling.
     */
    // TODO: should be replaced by proper sampling on the mesh
    optix::float3 v0, v1;
    float area;

    /// Normal direction of the emitter (assumed a plane)
    optix::float3 normal;

    /// Whether this material should receive derivatives
    bool is_differentiable;

    EEmitterType emitter_type;
};
