/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include <math_constants.h>
#include "types.cuh"
#include "common.cuh"
#include "envmap.cuh"

using namespace optix;

/// Main constants
rtDeclareVariable(uint3, launch_index, rtLaunchIndex, );
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(ShadowPayload, payload_shadow, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(uint, kNx, , "Image width");
rtDeclareVariable(uint, kNy, , "Image height");
rtDeclareVariable(uint, kSpp, , "Samples per pixel == number of launches");
rtDeclareVariable(uint, kBaseSeed, , "Base seed (e.g. changing frame to frame)");
rtDeclareVariable(uint, kNumberOfLights, , );
rtDeclareVariable(uint, kMaxDepth, , "Maximum path depth (0: visible emitters, 1: direct illumination, etc.)");
rtDeclareVariable(int, kHideEmitters, , "Whether to hide emitters from the final image (e.g. envmap)");

rtDeclareVariable(rtObject, top_object, , );
rtDeclareVariable(float3, camera_pos, , "Position of the camera in world space");
rtDeclareVariable(float3, camera_u, , );
rtDeclareVariable(float3, camera_v, , );
rtDeclareVariable(float3, camera_w, , "Vector pointing in the eye direction in world space");
rtDeclareVariable(float,  camera_near, , "Near clip plane");
rtDeclareVariable(float,  camera_far, , "Far clip plane");
rtDeclareVariable(Matrix4x4,  sample_to_camera, , "Transformation from a 2D point on the film to camera space");
rtDeclareVariable(float,  scene_scale, , "Maximal extent of the scene's bounding box");

/// Output buffers
rtBuffer<float> out_image;

// TODO: replace by a global variable, taken from the scene
constexpr size_t kRRDepth = 16;


RT_PROGRAM void ray_gen() {
    const float kInvSpp = 1.f / kSpp;

    uint seed_seed = kSpp * (launch_index.y * kNx + launch_index.x) + launch_index.z;
    uint seed = tea<16>(seed_seed, 0) + kBaseSeed;
    RayPayload payload(seed, kMaxDepth);

    /// Position of the ray origin on the film (jittered)
    float4 sample_pos = make_float4(
        (launch_index.x + rnd(payload.seed)) / float(kNx),
        (launch_index.y + rnd(payload.seed)) / float(kNy),
        0.f, 1.f);
    // Ray generation from a perspective camera (using homogeneous coordinates)
    float4 near_p = sample_to_camera * sample_pos;
    float3 d = normalize(make_float3(near_p.x, near_p.y, near_p.z));
    payload.bsdf_dir = normalize(d.x * camera_v + d.y * camera_u + d.z * camera_w);
    payload.bsdf_origin = camera_pos;
    // Ray extents
    float inv_z = 1.f / d.z;
    float min_t = camera_near * inv_z;
    float max_t = camera_far * inv_z;

    while (payload.depth < payload.max_depth && !payload.escaped) {
        /* Russian roulette: try to keep path weights equal to one.
           Stop with at least some probability to avoid getting stuck
           (e.g. due to total internal reflection) */
        if (payload.depth > kRRDepth) {
            float q = min(fmaxf(payload.throughput), 0.95f);
            if (rnd(payload.seed) >= q) {
                payload.escaped = false;
                break;
            }
            payload.throughput *= 1.f / q;
        }

        // Since the ray_miss seems to never trigger, we assume escape
        // and undo it if there's an intersection.
        payload.escaped = true;

        // Trace the next ray
        Ray next_ray = make_Ray(payload.bsdf_origin, payload.bsdf_dir, 1, min_t, max_t);
        rtTrace(top_object, next_ray, payload);
        min_t = math::Epsilon * (
         1.f + max(max(abs(payload.bsdf_origin.x), abs(payload.bsdf_origin.y)),
                   abs(payload.bsdf_origin.z)));
        max_t = RT_DEFAULT_MAX;
    }

    if (payload.escaped && !payload.did_emitter_sampling && !kHideEmitters) {
        // Envmap is visible now
        payload.result += make_float4(payload.throughput, 0.f)
                          * make_float4(call_envmap_eval(payload.bsdf_dir), 0.f);
    } else {
        // Ray stayed trapped (probably in medium)
        // payload.result = make_float4(0.f / 0.f);
    }

    // Accumulate ray's contribution to its pixel
    uint idx = 3 * (launch_index.y * kNx + launch_index.x);
    atomicAdd(&out_image[idx + 0], kInvSpp * payload.result.x);
    atomicAdd(&out_image[idx + 1], kInvSpp * payload.result.y);
    atomicAdd(&out_image[idx + 2], kInvSpp * payload.result.z);
 }

/// Miss program for standard rays
RT_PROGRAM void ray_miss() {
    payload.escaped = true;
}

/// Hit program for shadow rays
RT_PROGRAM void shadow_closest_hit() {
    payload_shadow.shadowed = true;
}
