/**
 * @file OptiX path tracer
 */
#include <optix_world.h>
#include <math_constants.h>
#include "area.cuh"
#include "bsdf.cuh"
#include "common.cuh"
#include "random.cuh"
#include "types.cuh"
#include "media.cuh"
#include "envmap.cuh"
#include "bbox.cuh"

using namespace optix;

/// State variables
rtDeclareVariable(RayPayload, payload, rtPayload, );
rtDeclareVariable(Ray, ray, rtCurrentRay,);
rtDeclareVariable(uint, kNumberOfLights, , );

/// Atributes (on hit)
rtDeclareVariable(float3, p, attribute p, );
rtDeclareVariable(float2, uv, attribute uv, );
rtDeclareVariable(float3, ns, attribute ns, );
rtDeclareVariable(float3, ng, attribute ng, );

/// List of all material parameters in the scene
rtBuffer<MaterialProperties> material_parameters;
/// Index of the curent material parameters
rtDeclareVariable(uint, material_id, , );
/// List of all emitter parameters in the scene
rtBuffer<EmitterProperties> emitter_parameters;

/// Index of the curent interior medium, if any
rtDeclareVariable(int, medium_id, , );

/// Bounding box corners for the only medium in the scene
rtDeclareVariable(float3, medium_bbox_min, , );
rtDeclareVariable(float3, medium_bbox_max, , );

/**
* Sample a position on a light source and evaluate how much light makes it
* through the medium to the current location.
*
* Note: we assume there's *no surface* and *no other medium* in the scene.
* Note: we ignore all types of emitters except for the environment map.
*/
RT_FUNCTION DirectionSample sample_emitter_direction_attenuated(const RayPayload &prd,
                                                                const MediumSample &mi,
                                                                const optix::float2 &sample,
                                                                const BoundingBox3f &medium_bbox) {
    // TODO: also support sampling from area emitters (and multiply by prob. of picking emitter)
    if (emitter_parameters.size() > 0) {
        printf("Not implemented yet: sample_emitter_direction_attenuated with area emitters.\n");
    }

    // Sample a direction toward the emitter.
    // Since we ignore surfaces, no need to test for visibility in that direction.
    DirectionSample ds = call_envmap_sample_direction(mi, sample);

    ds.radiance = call_envmap_eval(ds.light_direction) / ds.pdf;

    if (prd.in_medium) {
        // Attenuation by the medium.
        auto it = intersect_bbox(mi.p, ds.light_direction, medium_bbox);

        Ray tr_ray = make_Ray(mi.p, ds.light_direction, 1, optix::max(0.f, it.mint), it.maxt);
        ds.radiance *= call_medium_eval_transmittance(tr_ray);
    }

    return ds;
}


RT_PROGRAM void ray_hit_closest() {
    payload.depth += 1;
    payload.escaped = false;

    if (medium_id >= 0) {
        if (!payload.in_medium) {
            // Just hit the volume's bounding box, perform null interaction
            // to continue inside.
            payload.bsdf_dir = ray.direction;
            payload.bsdf_pdf = 1.f;
            payload.bsdf_origin = p;
            payload.in_medium = true;
            return;
        }


        // Sample a distance in the medium. Assumes that the given ray has `mint` at
        // the start of the bbox, and `maxt` at the end.
        const BoundingBox3f medium_bbox = { medium_bbox_min, medium_bbox_max };
        SurfaceInteraction si(ray, p, normalize(ns));
        float2 dist_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
        MediumSample mi = call_medium_sample_distance(si, ray, dist_sample);
        payload.throughput *= mi.throughput;

        if (mi.is_valid()) {
            // We're still in the medium.
            // Sample a new direction from the phase function.
            float2 phase_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
            float3 wo_local = sample_isotropic_phase(phase_sample);

            payload.bsdf_dir = mi.to_world(wo_local);
            payload.bsdf_pdf = 1.f;
            payload.bsdf_origin = mi.p;

            // Emitter sampling
            float2 emitter_sample = make_float2(rnd(payload.seed), rnd(payload.seed));
            DirectionSample ds = sample_emitter_direction_attenuated(payload, mi, emitter_sample, medium_bbox);
            payload.did_emitter_sampling = true;
            if (ds.pdf > 0.f) {
                float phase_val = eval_isotropic_phase();
                payload.result += make_float4(payload.throughput * phase_val * ds.radiance, 0.f);
            }

        } else {
            // Escaped, do another null surface interaction
            payload.bsdf_dir = ray.direction;
            payload.bsdf_pdf = 1.f;
            payload.bsdf_origin = p;
            payload.in_medium = false;
            return;
        }

    } else {
        // TODO: proper surface case with BSDF sampling, etc
        MaterialProperties props = material_parameters[material_id];
        SurfaceInteraction si(ray, p, ns);
        si.uv = uv;

        // Emission at this object
        if (props.emitter >= 0) {
            EmitterProperties light = emitter_parameters[props.emitter];
            float3 Le = AreaEmitter::eval(light, si);
            float mis_weight = 1.f;
            payload.result += make_float4(mis_weight * payload.throughput * Le, 0.f);
        }
        // TODO: emitter sampling (attenuated)

        // ----- BSDF sampling
        // Sample next direction from the BSDF (result stored in the payload).
        // Also sets the pdf of sampling that direction.
        float sample1 = rnd(payload.seed);
        float2 sample2 = make_float2(rnd(payload.seed), rnd(payload.seed));
        auto bs = call_bsdf_sample(si, sample1, sample2);

        // Transform sampled direction to world coordinates
        payload.bsdf_dir = si.to_world(bs.wo);
        payload.bsdf_pdf = bs.pdf;
        payload.throughput *= bs.sampling_weight;
        payload.bsdf_origin = si.p;
    }
}
