#include <cmath>

#include <optix.h>
#include <optix_cuda_interop.h>
#include "optix_ptx.h"
#include "ptx/types.cuh"

#include <mitsuba/core/bitmap.h>
#include <mitsuba/core/filesystem.h>
#include <mitsuba/core/timer.h>
#include <mitsuba/core/util.h>
#include <mitsuba/optix/common.h>
#include <mitsuba/optix/extraction.h>
#include <mitsuba/optix/renderer.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/common.h>
#include <mitsuba/render/integrator.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/scene.h>

#define MTS_WRITE_PTX

#if defined(MTS_WRITE_PTX)
#include <fstream>
#endif

using namespace mitsuba;

constexpr int kDeviceID = 0;
static const char *globals_names[kOptixGlobalsCount] = {
    "kNx",
    "kNy",
    "kSpp",
    "kBaseSeed",
    "kNumberOfLights",
    "kHasEnvmap",
    "kMaxDepth",
    "kRRDepth",
    "kHideEmitters",
    "kBackpropEnabled",
    "kMakeImage",
    "camera_pos",
    "camera_u",
    "camera_v",
    "camera_w",
    "camera_near",
    "camera_far",
    "sample_to_camera",
    "scene_scale",
    "medium_bbox_min",
    "medium_bbox_max",
    "kUseApproxLi1",
};
static const char *buffer_names[kOptixBuffersCount] = {
    "out_image",
};
static const char *rb_buffer_names[kOptixRbBuffersCount] = {
    "in_delta",
};
static const char *kBSDFGlobalsPrefix = "bsdf";
static const char *kMediumGlobalsPrefix = "medium";
static const char *kEnvmapGlobalsPrefix = "envmap";

// Never use re-implemented heterogeneous medium functions
constexpr bool kDoNotUseMediumReimplementation = false;
// constexpr bool kDoNotUseMediumReimplementation = true;

static void __rt_check(RTcontext context, RTresult errval, const char *file,
                       const int line) {
    if (errval != RT_SUCCESS) {
        const char *message;
        rtContextGetErrorString(context, errval, &message);
        fprintf(stderr,
                "rt_check(): OptiX API error = %04d (%s) in "
                "%s:%i.\n", (int) errval, message, file, line);
        exit(EXIT_FAILURE);
    }
}

// TODO: avoid this duplication
struct Scene::OptixState {
    RTcontext context;
    RTbuffer var_buf[kOptixStandardVariableCount];
    RTmaterial material;
    RTprogram attr_prog;
    RTacceleration accel;
    RTgeometrygroup group;
    uint32_t shape_index = 0, n_shapes = 0;
};


std::string sanitize_label(std::string label) {
    std::replace_if(
        label.begin(), label.end(),
        [&](char c) {
            return !((c >= '0' && c <= '9') || (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'));
        },
        '_');
    return label;
}

OptiXRenderer::OptiXRenderer(Scene *scene, EIntegratorType integrator,
                             DifferentiableParameters *params)
    : m_scene(scene), m_integrator_type(integrator), m_params(params) {

    switch (m_integrator_type) {
        case EDepth:
            m_program_source = { (const char *) depth_integrator_hit_ptx,
                                 (const char *) depth_integrator_ptx };
            Log(EInfo, "Integrator type: depth integrator");
            break;
        case EDirect:
            Throw("Not implemented: direct integrator");
            break;
        case EPath:
            m_program_source = {
                (const char *) path_integrator_hit_ptx,
                (const char *) path_integrator_ptx,
            };
            Log(EInfo, "Integrator type: path tracer");
            break;
        case EVolPath:
            m_program_source = { (const char *) volpath_integrator_hit_ptx,
                                 (const char *) volpath_integrator_ptx };
            Log(EInfo, "Integrator type: volumetric path tracer");
            break;
        case ERadiativeBackprop:
            m_program_source = { (const char *) radiative_backprop_hit_ptx,
                                 (const char *) radiative_backprop_ptx };
            Log(EInfo, "Integrator type: radiative backprop");
            break;
        case ERadiativeBackpropVol:
            m_program_source = { (const char *) radiative_backprop_vol_hit_ptx,
                                 (const char *) radiative_backprop_vol_ptx };
            Log(EInfo, "Integrator type: radiative backprop with volumes");
            break;
        case ERadiativeBackpropVolSimple:
            m_program_source = { (const char *) rb_simple_hit_ptx,
                                 (const char *) rb_simple_ptx };
            Log(EInfo, "Integrator type: simple radiative backprop with volumes");
            break;
        default:
            Throw("Unsupported integrator type: %s", m_integrator_type);
    }

    if (supports_gradients(m_integrator_type) && params == nullptr)
        Throw("Must specify `params` for radiative backprop rendering.");

    init();
}
OptiXRenderer::~OptiXRenderer() {
    // TODO: there are GPU resources being leaked!
    if (m_optix_state)
        delete m_optix_state;
    if (m_image_buffer)
        cuda_host_free(m_image_buffer);
}

void OptiXRenderer::init() {
    if (m_optix_state != nullptr)
        return;

    m_optix_state = new MyOptixState(m_scene->optix_state());
    auto &s = *m_optix_state->state;
    auto &s2 = *m_optix_state;

    // JIT-compile functions from all BSDFs & emitters
    extract_programs();

    // Create & declare all buffers (0-sized for now)
    for (size_t i = 0; i < kOptixBuffersCount; ++i) {
        RTvariable var;
        rt_check(rtBufferCreate(s.context, RT_BUFFER_INPUT, &s2.buffer_vars[i]));
        rt_check(rtBufferSetSize1D(s2.buffer_vars[i], 0));
        rt_check(rtContextDeclareVariable(s.context, buffer_names[i], &var));
        rt_check(rtVariableSetObject(var, s2.buffer_vars[i]));
        // Standard float buffers, storage will be in pinned memory
        rt_check(rtBufferSetFormat(s2.buffer_vars[i], RT_FORMAT_FLOAT));
        rt_check(rtBufferSetDevicePointer(s2.buffer_vars[i], kDeviceID, (void *) 8));
    }
    // Buffers specific to radiative backprop
    if (supports_gradients(m_integrator_type)) {
        for (size_t i = 0; i < kOptixRbBuffersCount; ++i) {
            RTvariable var;
            rt_check(rtBufferCreate(s.context, RT_BUFFER_INPUT, &s2.rb_buffer_vars[i]));
            rt_check(rtBufferSetSize1D(s2.rb_buffer_vars[i], 0));
            rt_check(rtContextDeclareVariable(s.context, rb_buffer_names[i], &var));
            rt_check(rtVariableSetObject(var, s2.rb_buffer_vars[i]));
            // Standard float buffers, storage will be in pinned memory
            rt_check(rtBufferSetFormat(s2.rb_buffer_vars[i], RT_FORMAT_FLOAT));
            rt_check(rtBufferSetDevicePointer(s2.rb_buffer_vars[i], kDeviceID, (void *) 8));
        }
    }

    // Declare global variables
    for (size_t i = 0; i < kOptixGlobalsCount; ++i) {
        rt_check(rtContextDeclareVariable(s.context, globals_names[i], &s2.global_vars[i]));
    }

    // Create materials & emitter properties
    create_bsdfs_and_emitters_props();

    // Create programs for the chosen integrator
    std::vector<const char *> common_sources = {
        m_program_source.rest,
        s2.extra_ptx.c_str()
    };

    // Main entry point
    RTprogram ray_gen_program;
    rt_check(rtProgramCreateFromPTXStrings(s.context, common_sources.size(), common_sources.data(),
                                           "ray_gen", &ray_gen_program));
    rt_check(rtContextSetRayGenerationProgram(s.context, 2, ray_gen_program));
    // Ray miss program
    RTprogram ray_miss_program;
    rt_check(rtProgramCreateFromPTXString(s.context, m_program_source.rest, "ray_miss", &ray_miss_program));
    rt_check(rtContextSetMissProgram(s.context, 1, ray_miss_program));
    // Shadow ray closest hit
    RTprogram shadow_ray_hit_program;
    rt_check(rtProgramCreateFromPTXString(s.context, m_program_source.rest, "shadow_closest_hit", &shadow_ray_hit_program));

    // For each material created, create and associate ray hit programs
    std::unordered_map<size_t, RTprogram> program_per_hash;
    for (const auto &kv : s2.bsdf_programs) {
        Log(EDebug, "Creating program for hash: %s", kv.first);
        std::vector<const char *> extended_sources = {
            m_program_source.hit,
            s2.extra_ptx.c_str()
        };
        for (const auto &ptx : kv.second.ptx)
            extended_sources.push_back(ptx.c_str());

#if defined(MTS_WRITE_PTX)
        std::ofstream f(tfm::format("/tmp/program_%s.ptx", kv.first));
        for (const auto &ptx : kv.second.ptx) {
            f << "// ----- Program" << std::endl;
            f << ptx << std::endl;
        }
        f.close();
#endif

        RTprogram program;
        rt_check(rtProgramCreateFromPTXStrings(s.context, extended_sources.size(),
                                               extended_sources.data(), "ray_hit_closest",
                                               &program));
        program_per_hash[kv.first] = program;
    }

    for (const auto &kv : s2.materials) {
        if (dynamic_cast<const BSDF *>((const Object *) kv.first) == nullptr)
            continue;

        Log(EDebug, "Registering program for BSDF: %p, hash: %s",
            kv.first, s2.bsdf_program_hashes[kv.first]);
        const auto &program = program_per_hash[s2.bsdf_program_hashes[kv.first]];
        rt_check(rtMaterialSetClosestHitProgram(kv.second, 1, program));
        rt_check(rtMaterialSetClosestHitProgram(kv.second, 2, shadow_ray_hit_program));
    }

    upload_bsdf_globals();
}

void OptiXRenderer::extract_programs() {
    auto &s2 = *m_optix_state;
    // bool used_graph_simpl = FloatD::graph_simplification_enabled_();
    // FloatD::set_graph_simplification_(false);

    // Shape / BSDF / Medium specific functions
    for (auto &shape : m_scene->shapes()) {
        auto bsdf_ptr = static_cast<const void *>(shape->bsdf());
        if (s2.bsdf_program_hashes.find(bsdf_ptr) != s2.bsdf_program_hashes.end()) {
            // Case of multiple shapes sharing the exact same BSDF instance.
            if (shape->interior_medium() || shape->exterior_medium())
                Throw("BSDF sharing is not supported for shapes with a medium");

            continue;
        }

        BSDFPrograms programs = extract_bsdf_programs(*shape->bsdf(), *shape);
        s2.bsdf_program_hashes[bsdf_ptr] = programs.hash;
        if (s2.bsdf_programs.find(programs.hash) != s2.bsdf_programs.end()) {
            // Case of multiple shapes sharing similar-enough BSDF instances that they
            // can reuse the BSDF programs.
            continue;
        }
        // Log(EInfo, "programs.ptx = %s", programs.ptx);
        s2.bsdf_programs[programs.hash] = programs;
    }

    // Functions shared across the whole scene (e.g. envmaps)
    s2.extra_ptx = extract_common_programs();

    // Restore value
    // FloatD::set_graph_simplification_(used_graph_simpl);
}

BSDFPrograms OptiXRenderer::extract_bsdf_programs(BSDF &bsdf, Shape &shape) const {
    auto &s2 = *m_optix_state;
    BSDFPrograms programs;
    programs.hash = 0;
    programs.extracted_from = &bsdf;

    std::hash<std::string> hasher;

    // Extract BSDF methods
    cuda_eval();
    cuda_create_ptx_module_context();
    using Point2fC              = typename BSDF::Point2fC;
    using SurfaceInteraction3fC = typename BSDF::SurfaceInteraction3fC;

    BSDFContext ctx;
    auto si = zero<SurfaceInteraction3fC>(1);
    auto wo = zero<Vector<FloatC, 3>>(1);

    /* ----- eval */ {
        cuda_start_recording_ptx_function("bsdf_eval");
        auto result = bsdf.eval(ctx, si, wo);

        // Workaround for "duplicate definition of variable"
        if (result.x().index_() == result.y().index_())
            result += 0.f;

        // Set label for nicer ptx and globals naming
        bsdf.set_labels(kBSDFGlobalsPrefix);
        set_label(si, "si");
        set_label(wo, "wo");
        set_label(result, "result");

        cuda_set_inputs(si, wo);
        cuda_set_outputs(result);
        cuda_stop_recording_ptx_function();
    }

    /* ----- pdf */ {
        cuda_start_recording_ptx_function("bsdf_pdf");
        auto result = bsdf.pdf(ctx, si, wo);
        set_label(result, "result");
        cuda_set_inputs(si, wo);
        cuda_set_outputs(result);
        cuda_stop_recording_ptx_function();
    }

    /* ----- sample */ {
        FloatC sample1   = zero<FloatC>(1);
        Point2fC sample2 = zero<Point2fC>(1);
        cuda_start_recording_ptx_function("bsdf_sample");
        auto [bs, weight] = bsdf.sample(ctx, si, sample1, sample2);

        // Workaround for "duplicate definition of variable"
        if (weight.x().index_() == weight.y().index_())
            weight += 0.f;

        set_label(sample1, "sample1");
        set_label(sample2, "sample2");
        set_label(bs, "bs");
        set_label(weight, "weight");

        cuda_set_inputs(si, sample1, sample2);
        cuda_set_outputs(bs, weight);
        cuda_stop_recording_ptx_function();
    }

    // ----- eval (autodiff)
    if (supports_gradients(m_integrator_type)) {
        using SpectrumfC = Spectrum<FloatC>;

        auto si2 = zero<SurfaceInteraction3fD>(1);
        auto wo2 = zero<Vector3fD>(1);
        auto grad_of_result = zero<SpectrumfC>(1);
        set_label(si2, "si");
        set_label(wo2, "wo");
        set_label(grad_of_result, "grad_of_result");
        cuda_eval();

        // Extract differentiable parameters from that BSDF
        DifferentiableParameters params = *m_params;
        std::vector<std::string> keys;
        for (const auto &k : m_params->keys()) {
            if (k.find(bsdf.id()) != std::string::npos) {
                keys.push_back(k);
                s2.enabled_differentiable_objects.insert(&bsdf);
            }
        }

        if (keys.empty()) {
            // Extract a dummy empty function since this is not enabled
            cuda_start_recording_ptx_function("bsdf_eval_d");
            auto result3 = zero<SpectrumfD>(1);
            cuda_set_inputs(si2, wo2, grad_of_result);
            set_label(result3, "result");
            cuda_stop_recording_ptx_function();
        } else {
            // Where the parameter gradients will be output
            std::vector<FloatC> grad_buffers;
            for (const auto &key : keys) {
                grad_buffers.push_back(zero<FloatC>(1));
                set_label(grad_buffers.back(), "grad_buffer");

                set_requires_gradients_for_param(params, key);
                // Set gradient storage to a buffer we control.
                // TODO: this is a hacky
                if (params.depth(key) == 1)
                    add_gradient_for_param(params, key, grad_buffers.back(), true);
            }

            cuda_start_recording_ptx_function("bsdf_eval_d");
            auto result3 = bsdf.eval(ctx, si2, wo2);
            set_label(result3, "result");

            set_gradient(result3, grad_of_result);
            auto var_id = cuda_var_register_ptr(grad_of_result.data());
            cuda_var_set_label(var_id, "grad_of_result_ptr");

            // Propagate `grad_of_result` to `input` through the computation.
            backward<FloatD>(/* free_graph = */ true);

            for (size_t i = 0; i < keys.size(); ++i) {
                const auto &key = keys[i];
                // Atomic accumulation to output buffers
                void *grad_buffer_ptr = accumulate_param_gradient_to_buffer(
                    params, key, "bsdf_eval_d.", &grad_buffers[i]);

                // Note that `grad_buffer_ptr` may or may not be `grad_buffer.data()`
                auto var_id2               = cuda_var_register_ptr(grad_buffer_ptr);
                auto grad_buffer_ptr_label = sanitize_label(key + "/grad_buffer_ptr");
                cuda_var_set_label(var_id2, grad_buffer_ptr_label.c_str());

                cuda_set_inputs(si2, wo2, grad_of_result);
                // Note: for this function we actually don't care about the result
                // of the main computation, so there's no output.
                cuda_stop_recording_ptx_function();

                // Already create the actual buffer that will hold gradients for that parameter
                size_t grad_entries      = params.depth(key) * params.slices(key);
                s2.gradient_buffers[key] = std::make_tuple(
                    (const void *) &bsdf, grad_buffer_ptr_label, zero<FloatC>(grad_entries));
            }
        }
    }

    // --- Medium
    if (supports_volumes(m_integrator_type)) {
        using SpectrumfC           = mitsuba::Spectrum<FloatC>;

        auto ray    = zero<Ray3fD>(1);
        auto sample = zero<Point2fD>(1);
        auto index  = zero<UInt32D>(1);
        auto si     = zero<SurfaceInteraction3fD>(1);
        auto it     = zero<Interaction3fD>(1);
        auto t      = zero<FloatD>(1);
        set_label(ray, "ray");
        set_label(sample, "sample");
        set_label(index, "index");
        set_label(si, "si");
        set_label(it, "it");
        set_label(t, "t");

        if (shape.exterior_medium()) {
            NotImplementedError("Shapes with exterior_medium");
        } else if (shape.interior_medium()) {
            const Medium *medium = shape.interior_medium();
            medium->set_labels(kMediumGlobalsPrefix);

            if (medium->is_homogeneous() || kDoNotUseMediumReimplementation) {
                /* medium_eval_transmittance */ {
                    cuda_start_recording_ptx_function("medium_eval_transmittance");

                    auto result = medium->eval_transmittance(ray, nullptr, index);
                    // Workaround for variables initialized from a single value
                    result += 0.f;
                    set_label(result, "result");

                    cuda_set_inputs(ray, index);
                    cuda_set_outputs(result);
                    cuda_stop_recording_ptx_function();
                }

                /* medium_sample_distance */ {
                    cuda_start_recording_ptx_function("medium_sample_distance");

                    auto [mi, throughput] =
                        medium->sample_distance_with_si(si, ray, sample, nullptr, index);

                    // Workaround for "variable cannot be set as output with an initial value"
                    mi.time += 0.f;
                    mi.wavelengths += 0.f;
                    mi.sh_frame.n += 0.f;
                    set_label(mi, "mi");
                    set_label(throughput, "throughput");

                    cuda_set_inputs(si, ray, sample, index);
                    cuda_set_outputs(mi, throughput);
                    cuda_stop_recording_ptx_function();
                }

                /* medium_pdf_distance */ {
                    cuda_start_recording_ptx_function("medium_pdf_distance");
                    auto pdf = medium->pdf_distance(ray, t);
                    set_label(pdf, "pdf");

                    cuda_set_inputs(ray, t);
                    cuda_set_outputs(pdf);
                    cuda_stop_recording_ptx_function();
                }
            } else {
                // For heterogeneous media, we provide a CUDA implementation.
                // Otherwise, the kernels get huge and compilation never finishes.
                programs.ptx.push_back(std::string((const char *) medium_heterogeneous_ptx));
            }

            /* medium_eval_scattering */ {
                auto mi = zero<MediumInteraction3fD>(1);
                set_label(mi, "mi");
                cuda_start_recording_ptx_function("medium_eval_scattering");
                auto result = medium->eval_scattering(mi);
                set_label(result, "result");
                cuda_set_inputs(mi);
                cuda_set_outputs(result);
                cuda_stop_recording_ptx_function();
            }

            /* medium_eval_albedo */ {
                cuda_start_recording_ptx_function("medium_eval_albedo");
                auto result = medium->eval_albedo(it);
                result += 0.f;
                set_label(result, "result");
                cuda_set_inputs(it);
                cuda_set_outputs(result);
                cuda_stop_recording_ptx_function();
            }

            /* medium_density */ {
                cuda_start_recording_ptx_function("medium_density");
                FloatD result = medium->density(it);
                result += 0.f;
                set_label(result, "result");
                cuda_set_inputs(it);
                cuda_set_outputs(result);
                cuda_stop_recording_ptx_function();
            }

        } else {
            // Extract dummy empty functions since this is not enabled
            cuda_start_recording_ptx_function("medium_eval_transmittance");
            auto result = zero<SpectrumfD>(1);
            set_label(result, "result_unused");
            cuda_set_inputs(ray, index);
            cuda_set_outputs(result);
            cuda_stop_recording_ptx_function();

            cuda_start_recording_ptx_function("medium_sample_distance");
            auto [mi, throughput] = std::make_pair(zero<MediumInteraction3fD>(1), zero<SpectrumfD>(1));
            set_label(mi, "mi_unused");
            set_label(throughput, "throughput_unused");
            cuda_set_inputs(si, ray, sample, index);
            cuda_set_outputs(mi, throughput);
            cuda_stop_recording_ptx_function();

            cuda_start_recording_ptx_function("medium_pdf_distance");
            auto pdf = zero<FloatD>();
            set_label(pdf, "pdf");
            cuda_set_inputs(ray, t);
            cuda_set_outputs(pdf);
            cuda_stop_recording_ptx_function();

            auto mi2 = zero<MediumInteraction3fD>(1);
            set_label(mi2, "mi");
            cuda_start_recording_ptx_function("medium_eval_scattering");
            auto scattering = zero<FloatD>();
            set_label(scattering, "scattering");
            cuda_set_inputs(mi2);
            cuda_set_outputs(scattering);
            cuda_stop_recording_ptx_function();

            cuda_start_recording_ptx_function("medium_eval_albedo");
            auto albedo = zero<SpectrumfD>();
            set_label(albedo, "albedo_unused");
            cuda_set_inputs(it);
            cuda_set_outputs(albedo);
            cuda_stop_recording_ptx_function();

            cuda_start_recording_ptx_function("medium_density");
            auto density = zero<FloatD>(1);
            set_label(density, "density_unused");
            cuda_set_inputs(it);
            cuda_set_outputs(density);
            cuda_stop_recording_ptx_function();
        }


        /* math::solve_quadratic */ {
            auto a = zero<FloatD>(1), b = zero<FloatD>(1), c = zero<FloatD>(1);
            set_label(a, "a");
            set_label(b, "b");
            set_label(c, "c");

            cuda_start_recording_ptx_function("math_solve_quadratic");
            auto [has_solution, solution1, solution2] = math::solve_quadratic(a, b, c);
            set_label(has_solution, "has_solution");
            set_label(solution1, "solution1");
            set_label(solution2, "solution2");
            cuda_set_inputs(a, b, c);
            cuda_set_outputs(has_solution, solution1, solution2);
            cuda_stop_recording_ptx_function();
        }


        // --- Differentiable medium
        if (supports_gradients(m_integrator_type)) {
            // Extract differentiable parameters from that medium
            std::vector<std::string> keys;
            const Medium *medium = nullptr;
            DifferentiableParameters params = *m_params;
            if (shape.interior_medium()) {
                medium = shape.interior_medium();
                // Find out if this medium has a differentiable parameter
                for (const auto &k : m_params->keys()) {
                    if (k.find(medium->id()) != std::string::npos) {
                        keys.push_back(k);
                        s2.enabled_differentiable_objects.insert(medium);
                    }
                }
            }
            // Make sure that `this` pointer is registered as a variable.
            medium->this_pointer<FloatD>();

            if (!keys.empty()) {
                // Where the parameter gradients will be output, unless the parameter
                // is a buffer itself (then it will come with its own grad output buffer).
                std::vector<FloatC> grad_buffers;
                std::vector<void *> actual_grad_buffer_ptrs;
                std::vector<std::string> grad_buffer_ptr_labels;
                for (const auto &key : keys) {
                    grad_buffers.push_back(zero<FloatC>(1));
                    set_label(grad_buffers.back(), sanitize_label(key + "/grad_buffer").c_str());
                    actual_grad_buffer_ptrs.push_back(nullptr);
                    grad_buffer_ptr_labels.push_back(sanitize_label(key + "/grad_buffer_ptr"));
                }

                auto ensured_controlled_gradients = [&]() {
                    for (size_t i = 0; i < keys.size(); ++i) {
                        set_requires_gradients_for_param(params, keys[i]);
                        // Set gradient storage to a buffer we control (mostly a hack).
                        if (params.depth(keys[i]) == 1)
                            add_gradient_for_param(params, keys[i], grad_buffers[i], true);
                    }
                };

                /* recording */ {
                    ensured_controlled_gradients();
                    auto it = zero<Interaction3fD>(1);
                    set_label(it, "it");
                    auto grad_of_result = zero<FloatC>(1);
                    set_label(grad_of_result, "grad_of_result");

                    cuda_start_recording_ptx_function("medium_density_d");

                    auto result = medium->density(it);
                    result += 0.f;
                    set_label(result, "result_medium_density");

                    // It may be that the differentiable quantity is not involved in `density()`.
                    if (enoki::requires_gradient(result)) {
                        set_gradient(result, grad_of_result);

                        // Propagate `grad_of_result` to `input` through the computation.
                        backward<FloatD>(/* free_graph = */ true);

                        for (size_t i = 0; i < keys.size(); ++i) {
                            // Atomic accumulation to an output buffer
                            actual_grad_buffer_ptrs[i] = accumulate_param_gradient_to_buffer(
                                params, keys[i], "medium_density_d.", &grad_buffers[i]);

                            auto var_id = cuda_var_register_ptr(actual_grad_buffer_ptrs[i]);
                            cuda_var_set_label(var_id, grad_buffer_ptr_labels[i].c_str());
                        }
                    }
                    cuda_set_inputs(it, grad_of_result);
                    cuda_stop_recording_ptx_function();
                }

                if (medium->is_homogeneous() || kDoNotUseMediumReimplementation) {
                    ensured_controlled_gradients();
                    auto grad_of_result = zero<SpectrumfC>(1);
                    set_label(grad_of_result, "grad_of_result");

                    cuda_start_recording_ptx_function("medium_eval_transmittance_d");
                    auto result = medium->eval_transmittance(ray, nullptr, index);
                    // Workaround
                    result += 0.f;
                    // It may be that the differentiable quantity is not involved
                    // in `eval_transmittance`.
                    if (enoki::requires_gradient(result)) {
                        set_gradient(result, grad_of_result);

                        // Propagate `grad_of_result` to `input` through the computation.
                        backward<FloatD>(/* free_graph = */ true);

                        for (size_t i = 0; i < keys.size(); ++i) {
                            if (actual_grad_buffer_ptrs[i] != nullptr) {
                                // Note: we keep the same label as above since it's the same buffer
                                void *other_grad_buffer_ptr = accumulate_param_gradient_to_buffer(
                                    params, keys[i], "medium_density_d.",
                                    &grad_buffers[i]);
                                // Output pointer should be the same, it's already registered
                                if (actual_grad_buffer_ptrs[i] != other_grad_buffer_ptr)
                                    Throw("Gradient output buffer was different! Expected %p  "
                                          "found  %p, grad_buffer.data() = %p",
                                          actual_grad_buffer_ptrs[i], other_grad_buffer_ptr,
                                          grad_buffers[i].data());
                            } else {
                                // Register that fresh output buffer
                                actual_grad_buffer_ptrs[i] = accumulate_param_gradient_to_buffer(
                                    params, keys[i], "medium_eval_transmittance_d.", &grad_buffers[i]);
                                auto var_id = cuda_var_register_ptr(actual_grad_buffer_ptrs[i]);
                                cuda_var_set_label(var_id, grad_buffer_ptr_labels[i].c_str());
                            }
                        }
                    }

                    cuda_set_inputs(ray, index, grad_of_result);
                    cuda_stop_recording_ptx_function();
                } else {
                    // For heterogeneous media, we re-implement this function ourselves in CUDA.
                    // Otherwise, the kernel size becomes huge for large volumes and compilation
                    // of the extracted kernel in OptiX never finishes.
                    programs.ptx.push_back(std::string((const char *) medium_heterogeneous_d_ptx));
                }

                /* recording */ {
                    auto mi = zero<MediumInteraction3fD>(1);
                    set_label(mi, "mi");
                    for (size_t i = 0; i < keys.size(); ++i) {
                        set_requires_gradients_for_param(params, keys[i]);
                        // Set gradient storage to a buffer we control (mostly a hack).
                        if (params.depth(keys[i]) == 1)
                            add_gradient_for_param(params, keys[i], grad_buffers[i], true);
                    }
                    auto grad_of_result = zero<SpectrumfC>(1);
                    set_label(grad_of_result, "grad_of_result");

                    ensured_controlled_gradients();

                    cuda_start_recording_ptx_function("medium_eval_scattering_d");

                    auto result = medium->eval_scattering(mi);
                    set_label(result, "result_eval_scattering");

                    // It may be that the differentiable quantity is not involved in
                    // `eval_scattering()`.
                    if (enoki::requires_gradient(result)) {
                        set_gradient(result, grad_of_result);

                        // Propagate `grad_of_result` to `input` through the computation.
                        backward<FloatD>(/* free_graph = */ true);

                        for (size_t i = 0; i < keys.size(); ++i) {
                            // Atomic accumulation to an output buffer
                            if (actual_grad_buffer_ptrs[i] != nullptr) {
                                // Note: we keep the same label as above since it's the same buffer
                                void *other_grad_buffer_ptr = accumulate_param_gradient_to_buffer(
                                    params, keys[i], "medium_eval_transmittance_d.",
                                    &grad_buffers[i]);
                                // Output pointer should be the same as before, it's already
                                // registered
                                if (actual_grad_buffer_ptrs[i] != other_grad_buffer_ptr)
                                    Throw("Gradient output buffer was different! Expected %p  "
                                          "found  %p, grad_buffer.data() = %p",
                                          actual_grad_buffer_ptrs[i], other_grad_buffer_ptr,
                                          grad_buffers[i].data());
                            } else {
                                // Register that fresh output buffer
                                actual_grad_buffer_ptrs[i] = accumulate_param_gradient_to_buffer(
                                    params, keys[i], "medium_eval_scattering_d.", &grad_buffers[i]);
                                auto var_id = cuda_var_register_ptr(actual_grad_buffer_ptrs[i]);
                                cuda_var_set_label(var_id, grad_buffer_ptr_labels[i].c_str());
                            }
                        }
                    }
                    cuda_set_inputs(mi, grad_of_result);
                    cuda_stop_recording_ptx_function();
                }

                /* recording */ {
#if defined(MTS_PTX_USE_COMBINED_SCATTERING_AND_TRANSMITTANCE)
                    // TODO: special handling of heterogeneous (large kernel)
                    auto mi = zero<MediumInteraction3fD>(1);
                    set_label(mi, "mi");
                    auto grad_of_result = zero<SpectrumfC>(1);
                    set_label(grad_of_result, "grad_of_result");

                    ensured_controlled_gradients();
                    cuda_start_recording_ptx_function(
                        "medium_eval_scattering_times_transmittance_d");

                    auto scattering = medium->eval_scattering(mi);
                    auto transmittance = medium->eval_transmittance(ray, nullptr, index);
                    auto result = scattering * transmittance;
                    set_label(result, "result");

                    // It may be that the differentiable quantity is not involved in this
                    if (enoki::requires_gradient(result)) {
                        set_gradient(result, grad_of_result);

                        // Propagate `grad_of_result` to `input` through the computation.
                        backward<FloatD>(/* free_graph = */ true);

                        for (size_t i = 0; i < keys.size(); ++i) {
                            // Atomic accumulation to an output buffer
                            if (actual_grad_buffer_ptrs[i] != nullptr) {
                                // Note: we keep the same label as above since it's the same buffer
                                void *other_grad_buffer_ptr = accumulate_param_gradient_to_buffer(
                                    params, keys[i], "medium_eval_transmittance_d.",
                                    &grad_buffers[i]);
                                // Output pointer should be the same, it's already registered
                                if (actual_grad_buffer_ptrs[i] != other_grad_buffer_ptr)
                                    Throw("Gradient output buffer was different! Expected %p  "
                                          "found  %p, grad_buffer.data() = %p",
                                          actual_grad_buffer_ptrs[i], other_grad_buffer_ptr,
                                          grad_buffers[i].data());
                            } else {
                                // Register that fresh output buffer
                                actual_grad_buffer_ptrs[i] = accumulate_param_gradient_to_buffer(
                                    params, keys[i],
                                    "medium_eval_scattering_times_transmittance_d.",
                                    &grad_buffers[i]);
                                auto var_id = cuda_var_register_ptr(actual_grad_buffer_ptrs[i]);
                                cuda_var_set_label(var_id, grad_buffer_ptr_labels[i].c_str());
                            }
                        }
                    }
                    cuda_set_inputs(mi, ray, index, grad_of_result);
                    cuda_stop_recording_ptx_function();
#endif
                }

                // Already prepare the actual buffers that will hold gradients for those parameters
                for (size_t i = 0; i < keys.size(); ++i) {
                    size_t grad_entries = params.depth(keys[i]) * params.slices(keys[i]);
                    s2.gradient_buffers[keys[i]] =
                        std::make_tuple((const void *) medium, grad_buffer_ptr_labels[i],
                                        zero<FloatC>(grad_entries));
                }
            } else {
                // No medium to extract, or medium's param not enabled: extract empty function
                auto it               = zero<Interaction3fD>(1);
                auto mi               = zero<MediumInteraction3fD>(1);
                auto grad_of_result_f = zero<FloatC>(1);
                auto grad_of_result   = zero<SpectrumfC>(1);
                set_label(it, "it");
                set_label(mi, "mi");
                set_label(grad_of_result, "grad_of_result");

                cuda_start_recording_ptx_function("medium_eval_transmittance_d");
                auto result = zero<SpectrumfD>(1);
                set_label(result, "result_unused");
                cuda_set_inputs(ray, index, grad_of_result);
                cuda_stop_recording_ptx_function();

                cuda_start_recording_ptx_function("medium_density_d");
                auto result_f = zero<FloatD>(1);
                set_label(result_f, "result_unused");
                cuda_set_inputs(it, grad_of_result_f);
                cuda_stop_recording_ptx_function();

                cuda_start_recording_ptx_function("medium_eval_scattering_d");
                result = zero<SpectrumfD>(1);
                set_label(result, "result_unused");
                cuda_set_inputs(mi, grad_of_result);
                cuda_stop_recording_ptx_function();

#if defined(MTS_PTX_USE_COMBINED_SCATTERING_AND_TRANSMITTANCE)
                cuda_start_recording_ptx_function("medium_eval_scattering_times_transmittance_d");
                auto result2 = zero<SpectrumfD>(1);
                set_label(result2, "result_unused");
                cuda_set_inputs(mi, ray, index, grad_of_result);
                cuda_stop_recording_ptx_function();
#endif
            }

        }  // end differentiable volumes extraction
    }  // end volume-related extraction

    // Finalize extraction
    const char *ptx_c = cuda_get_ptx_module();
    if (ptx_c == nullptr)
        Throw("Could not extract ptx module for BSDF: %s", bsdf.to_string());
    programs.ptx.push_back(std::string(ptx_c));
    free((void *) ptx_c);

    // Get hash of complete program
    std::string complete = "";
    for (const auto &s : programs.ptx)
        complete += s;
    programs.hash = hasher(complete);

    // if (supports_gradients(m_integrator_type)) {
    //     std::cout << programs.ptx << std::endl;
    //     exit(1);
    // }

    cuda_destroy_ptx_module_context();
    return programs;
}

std::string OptiXRenderer::extract_common_programs() const {
    using SpectrumfC = mitsuba::Spectrum<FloatC>;

    cuda_create_ptx_module_context();

    // Extract InfiniteEmitter methods
    auto si     = zero<SurfaceInteraction3fD>(1);
    auto ref    = zero<Interaction3fD>(1);
    auto in_ds  = zero<DirectionSample3fD>(1);
    auto sample = zero<Point2fD>(1);
    set_label(si, "si");
    set_label(ref, "ref");
    set_label(in_ds, "in_ds");
    set_label(sample, "sample");
    if (m_scene->environment()) {
        auto emitter = m_scene->environment();
        emitter->set_labels(kEnvmapGlobalsPrefix);

        /* InfiniteEmitter::eval */ {
            cuda_start_recording_ptx_function("envmap_eval");
            auto result = emitter->eval(si);

            // Workaround for "variable cannot be set as output with an initial value"
            if (result.x().index_() == result.y().index_())
                result += 0.f;
            set_label(result, "result");

            cuda_set_inputs(si);
            cuda_set_outputs(result);
            cuda_stop_recording_ptx_function();
        }

        /* InfiniteEmitter::sample_direction */ {
            cuda_start_recording_ptx_function("envmap_sample_direction");
            auto [ds, result] = emitter->sample_direction(ref, sample);

            // Workaround for "variable cannot be set as output with an initial value"
            if (result.x().index_() == result.y().index_())
                result += 0.f;
            if (ds.uv.x().index_() == ds.uv.y().index_())
                ds.uv += 0.f;
            ds.time += 0.f;
            set_label(result, "result");
            set_label(ds, "ds");

            cuda_set_inputs(ref, sample);
            cuda_set_outputs(ds, result);
            cuda_stop_recording_ptx_function();
        }

        /* InfiniteEmitter::pdf_direction */ {
            cuda_start_recording_ptx_function("envmap_pdf_direction");
            auto pdf = emitter->pdf_direction(ref, in_ds);

            set_label(pdf, "pdf");

            cuda_set_inputs(ref, in_ds);
            cuda_set_outputs(pdf);
            cuda_stop_recording_ptx_function();
        }
    } else {
        cuda_start_recording_ptx_function("envmap_eval");
        auto result = zero<SpectrumfC>(1);
        set_label(result, "result_unused");
        cuda_set_inputs(si);
        cuda_set_outputs(result);
        cuda_stop_recording_ptx_function();

        cuda_start_recording_ptx_function("envmap_sample_direction");
        auto [ds, result2] = std::make_pair(zero<DirectionSample3fD>(1), zero<SpectrumfC>(1));

        set_label(ds, "ds_unused");
        set_label(result2, "result_unused");

        cuda_set_inputs(ref, sample);
        cuda_set_outputs(ds, result2);
        cuda_stop_recording_ptx_function();

        cuda_start_recording_ptx_function("envmap_pdf_direction");
        auto pdf = zero<FloatD>(1);
        set_label(pdf, "pdf_unused");
        cuda_set_inputs(ref, in_ds);
        cuda_set_outputs(pdf);
        cuda_stop_recording_ptx_function();
    }

    char *ptx_c = cuda_get_ptx_module();
    if (ptx_c == nullptr)
        Throw("Could not extract ptx module for common functions");
    std::string ptx = ptx_c;

    free(ptx_c);
    cuda_destroy_ptx_module_context();

    // // TODO: remove this
    // std::cout << ptx << std::endl;
    // exit(1);

    return ptx;
}

void OptiXRenderer::upload_bsdf_globals() const {
    auto &s = *m_optix_state->state;
    auto &s2 = *m_optix_state;

    // For each material created (BSDF, Medium, etc), upload its globals to the device
    for (const auto &kv : s2.materials) {
        auto obj           = dynamic_cast<const DifferentiableObject *>((const Object *) kv.first);
        std::string prefix;
        if (dynamic_cast<const BSDF *>(obj))
            prefix = kBSDFGlobalsPrefix;
        else if (dynamic_cast<const Medium *>(obj))
            prefix = kMediumGlobalsPrefix;
        else
            Throw("Unrecognized object type registered with a material: %s", obj);

        obj->set_globals(prefix, s.context, kv.second);
        Log(EDebug, "Setting globals for object %p, prefix = %s", obj, prefix);
    }

    // For each differentiable parameter, upload a gradient output buffer pointer
    for (const auto &kv : s2.gradient_buffers) {
        const void *obj          = std::get<0>(kv.second);
        const std::string &label = std::get<1>(kv.second);
        const FloatC &buffer     = std::get<2>(kv.second);

        // TODO: avoid re-uploading at each iteration
        const RTmaterial &material = s2.materials.at(obj);
        RTvariable var = get_or_create_optix_var(label, s.context, material);
        const auto *ptr_value = buffer.data();
        rt_check(rtVariableSetUserData(var, sizeof(ptr_value), &ptr_value));
    }

    // Common globals
    if (m_scene->environment()) {
        m_scene->environment()->set_globals("envmap", s.context, nullptr);
    }
}

/// Creates and register a BSDF for each shape in the scene
void OptiXRenderer::create_bsdfs_and_emitters_props() {
    auto &s = *m_optix_state->state;
    auto &s2 = *m_optix_state;
    auto &buffer_mats = s2.material_params_buffer;
    auto &buffer_lights = s2.emitter_params_buffer;
    size_t n_materials = m_scene->shapes().size();
    size_t n_emitters = m_scene->emitters().size();
    // Note that we don't create emitter properties for the envmap
    if (m_scene->environment())
        n_emitters -= 1;

    RTvariable var;
    // Create the BSDF parameters buffer
    rt_check(rtBufferCreate(s.context, RT_BUFFER_INPUT, &buffer_mats));
    rt_check(rtBufferSetFormat(buffer_mats, RT_FORMAT_USER));
    rt_check(rtBufferSetElementSize(buffer_mats, sizeof(MaterialProperties)));
    rt_check(rtBufferSetSize1D(buffer_mats, n_materials));
    rt_check(rtContextDeclareVariable(s.context, "material_parameters", &var));
    rt_check(rtVariableSetObject(var, buffer_mats));
    // Create the emitter parameters buffer
    rt_check(rtBufferCreate(s.context, RT_BUFFER_INPUT, &buffer_lights));
    rt_check(rtBufferSetFormat(buffer_lights, RT_FORMAT_USER));
    rt_check(rtBufferSetElementSize(buffer_lights, sizeof(EmitterProperties)));
    rt_check(rtBufferSetSize1D(buffer_lights, n_emitters));
    rt_check(rtContextDeclareVariable(s.context, "emitter_parameters", &var));
    rt_check(rtVariableSetObject(var, buffer_lights));

    // For each shape, create material properties (and emitter, if associated)
    MaterialProperties *mats_data;
    EmitterProperties *emitters_data;
    rt_check(rtBufferMap(buffer_mats, (void **) &mats_data));
    rt_check(rtBufferMap(buffer_lights, (void **) &emitters_data));
    size_t light_i = 0;
    size_t medium_i = 0;
    for (size_t shape_i = 0; shape_i < n_materials; ++shape_i) {
        const auto &shape = *m_scene->shapes()[shape_i];
        void *bsdf_ptr = (void *) shape.bsdf();

        MaterialProperties props;
        props.emitter = -1;
        if (shape.emitter()) {
            emitters_data[light_i] = create_emitter_props(shape, *shape.emitter());
            props.emitter = light_i;
            ++light_i;
        }

        // TODO: which category really indicates a delta interaction? (pdf == 0)
        props.is_delta = eq(shape.bsdf()->flags() & BSDF::ESmooth, 0u)
                         || neq(shape.bsdf()->flags() & BSDF::EDelta, 0u);

        if (supports_gradients(m_integrator_type)) {
            props.is_differentiable = s2.enabled_differentiable_objects.count(bsdf_ptr) > 0;
            if (shape.interior_medium())
                props.is_differentiable |=
                    s2.enabled_differentiable_objects.count(shape.interior_medium()) > 0;
        }

        mats_data[shape_i] = props;

        // Set the material_id variable for this shape (= geometry group)
        RTgeometryinstance geometry;
        rt_check(rtGeometryGroupGetChild(s.group, shape_i, &geometry));
        RTvariable mat_id_var;
        rt_check(rtGeometryInstanceDeclareVariable(geometry, "material_id", &mat_id_var));
        rt_check(rtVariableSet1ui(mat_id_var, shape_i));

        // Create an OptiX material for this shape's BSDF
        if (s2.materials.find(bsdf_ptr) == s2.materials.end()) {
            RTmaterial material;
            rt_check(rtMaterialCreate(s.context, &material));
            s2.materials[bsdf_ptr] = material;

            if (shape.interior_medium())
                s2.materials[(const void *) shape.interior_medium()] = material;
        }
        rt_check(rtGeometryInstanceSetMaterial(geometry, 0, s2.materials[bsdf_ptr]));

        if (shape.exterior_medium())
            NotImplementedError("shape with exterior medium");

        RTvariable medium_id_var;
        rt_check(rtGeometryInstanceDeclareVariable(geometry, "medium_id", &medium_id_var));
        if (shape.interior_medium()) {
            if (!supports_volumes(m_integrator_type))
                Throw("Found medium in scene while trying to render with non-volumetric path tracer");
            rt_check(rtVariableSet1i(medium_id_var, medium_i));
            ++medium_i;
        } else {
            rt_check(rtVariableSet1i(medium_id_var, -1));
        }
    }

    rt_check(rtBufferUnmap(buffer_mats));
    rt_check(rtBufferUnmap(buffer_lights));

    cuda_eval();
}

EmitterProperties OptiXRenderer::create_emitter_props(const Shape &shape,
                                                      const Emitter &emitter) {
    EmitterProperties props;
    // TODO: actual emitter properties

    /* Hacky */ {
        SurfaceInteraction3f si;
        si.wi = Vector3f(0.f, 0.f, 1.f);
        si.wavelengths = 0.f;
        si.uv = 0.5f;
        auto radiance = emitter.eval(si);
        props.radiance.x = radiance.x();
        props.radiance.y = radiance.y();
        props.radiance.z = radiance.z();
        // Hardcoded value for cbox:
        // props.radiance.x = 18.387f;
        // props.radiance.y = 10.9873f;
        // props.radiance.z = 2.75357f;
    }

    auto p = shape.bbox().min;

    // Squish the bounding box, pretending we have a perfect plane
    auto extents = shape.bbox().extents();
    auto minor_axis = shape.bbox().minor_axis();
    auto v0 = extents, v1 = extents;
    v0[(minor_axis + 1) % 3] = 0.f;
    v1[(minor_axis + 2) % 3] = 0.f;
    props.p.x = p.x(); props.p.y = p.y(); props.p.z = p.z();
    props.v0.x = v0.x(); props.v0.y = v0.y(); props.v0.z = v0.z();
    props.v1.x = v1.x(); props.v1.y = v1.y(); props.v1.z = v1.z();

    // Set surface area and normal as plausibly as we can
    props.area = shape.surface_area();
    props.emitter_type = EAreaMesh;
    // TODO: get actual emitter normals
    Normal3f fake_n = 0.f;
    fake_n[minor_axis] = -1.f;
    props.normal.x = fake_n.x(); props.normal.y = fake_n.y(); props.normal.z = fake_n.z();

    // TODO: actually enable only the differentiable objects
    if (supports_gradients(m_integrator_type)) {
        props.is_differentiable = false;
    }

    return props;
}

ref<Bitmap> OptiXRenderer::render(size_t seed, size_t _spp, size_t pass_count) {
    size_t nx = m_scene->film()->size().x();
    size_t ny = m_scene->film()->size().y();
    size_t spp = (_spp != (size_t) -1) ? _spp : m_scene->sampler()->sample_count();
    Log(EDebug, "Rendering (%s x %s @ %s spp, %s passes)...", nx, ny, spp, pass_count);
    Timer render_timer;
    float *image_buffer = trace(nx, ny, spp, pass_count, seed);
    Log(EDebug, "Done! Took %s.", util::time_string(render_timer.value(), true));

    // With this constructor, the Bitmap doesn't own the memory, so we create a copy.
    ref<Bitmap> b =
        new Bitmap(Bitmap::ERGB, Struct::EFloat32, { nx, ny }, 0, (uint8_t *) image_buffer);
    return new Bitmap(*b);
}

ref<Bitmap> OptiXRenderer::delta_render(size_t seed, const Bitmap *image_derivatives, size_t _spp,
                                        size_t pass_count, bool return_image) {
    size_t nx = m_scene->film()->size().x();
    size_t ny = m_scene->film()->size().y();
    const auto &b = *image_derivatives;
    if (b.component_format() != Struct::EFloat32 || b.pixel_format() != Bitmap::ERGB)
        Throw("Wrong component format (expected float32) or pixel format (expected RGB)");
    if (b.width() != nx || b.height() != ny)
        Throw("Incorrect image dimensions for input gradients! Expected (%s, %s), "
              "found (%s, %s)",
              nx, ny, b.width(), b.height());

    auto image_derivatives_c = FloatC::copy((void *) b.data(), hprod(b.size()) * b.channel_count());
    return delta_render(seed, image_derivatives_c, _spp, pass_count, return_image);
}

ref<Bitmap> OptiXRenderer::delta_render(size_t seed, const FloatC &image_derivatives, size_t _spp,
                                        size_t pass_count, bool return_image) {
    if (!supports_gradients(m_integrator_type))
        Throw("`delta_render` is only available for the RadiativeBackprop integrators");

    cuda_eval();
    size_t nx  = m_scene->film()->size().x();
    size_t ny  = m_scene->film()->size().y();
    size_t spp = (_spp != (size_t) -1) ? _spp : m_scene->sampler()->sample_count();

    Log(EDebug, "Rendering (%s x %s @ %s spp)...", nx, ny, spp);
    Timer render_timer;
    float *image_buffer = trace(nx, ny, spp, pass_count, seed, return_image, &image_derivatives);
    Log(EDebug, "Done! Took %s.", util::time_string(render_timer.value(), true));

    // Copy gradients back to their Enoki variable
    unpack_gradients(m_params);

    if (!return_image)
        return nullptr;
    // With this constructor, the Bitmap doesn't own the memory, so we create a copy.
    ref<Bitmap> b =
        new Bitmap(Bitmap::ERGB, Struct::EFloat32, { nx, ny }, 0, (uint8_t *) image_buffer);
    return new Bitmap(*b);
}

float *OptiXRenderer::trace(size_t nx, size_t ny, size_t spp, size_t pass_count, size_t base_seed,
                            bool return_image, const FloatC *image_derivatives) {
    auto &s           = *m_optix_state->state;
    auto &s2          = *m_optix_state;
    size_t image_size = 3 * nx * ny;

    bool backprop_enabled =
        supports_gradients(m_integrator_type) && image_derivatives != nullptr;

    // Output image buffer.
    // Note that radiative backprop doesn't produce an image when computing gradients,
    // unless it has been explicitly enable.
    if (return_image) {
        if (m_image_buffer_size != image_size) {
            if (m_image_buffer != nullptr)
                enoki::cuda_free(m_image_buffer);
            m_image_buffer      = (float *) enoki::cuda_host_malloc(image_size * sizeof(float));
            m_image_buffer_size = image_size;
        }
        for (size_t i = 0; i < image_size; ++i)
            m_image_buffer[i] = 0.f;
        rt_check(rtBufferSetSize1D(s2.buffer_vars[0], image_size));
        rt_check(rtBufferSetDevicePointer(s2.buffer_vars[0], kDeviceID, m_image_buffer));
    } else {
        rt_check(rtBufferSetSize1D(s2.buffer_vars[0], 0));
        rt_check(rtBufferSetDevicePointer(s2.buffer_vars[0], kDeviceID, (void *) 8));
    }

    // Set radiative backprop-specific buffers, if relevant
    if (supports_gradients(m_integrator_type)) {
        if (image_derivatives != nullptr) {
            if (image_derivatives->size() != image_size)
                Throw("Incorrect size for image_derivatives: expected %d, found %d", image_size,
                      image_derivatives->size());

            // in_delta: partial derivative with respect to pixel values
            rt_check(rtBufferSetSize1D(s2.rb_buffer_vars[0], image_size));
            rt_check(rtBufferSetDevicePointer(s2.rb_buffer_vars[0], kDeviceID,
                                              (void *) image_derivatives->data()));
        } else {
            rt_check(rtBufferSetSize1D(s2.rb_buffer_vars[0], 0));
            rt_check(rtBufferSetDevicePointer(s2.rb_buffer_vars[0], kDeviceID, (void *) 8));
        }
    }


    /* Set global variables value (image dimensions, sensor, etc) */ {
        rt_check(rtVariableSet1ui(s2.global_vars[0], nx));
        rt_check(rtVariableSet1ui(s2.global_vars[1], ny));
        rt_check(rtVariableSet1ui(s2.global_vars[2], pass_count*spp));
        // Variable 3: Base seed (per frame)
        rt_check(rtVariableSet1ui(s2.global_vars[3], base_seed));

        rt_check(rtVariableSet1ui(s2.global_vars[4], m_scene->emitters().size()));
        bool has_envmap = m_scene->environment();
        rt_check(rtVariableSet1i(s2.global_vars[5], has_envmap));

        auto *integrator = dynamic_cast<MonteCarloIntegrator *>(m_scene->integrator());
        Assert(integrator != nullptr);
        rt_check(rtVariableSet1ui(s2.global_vars[6], integrator->max_depth()));
        rt_check(rtVariableSet1ui(s2.global_vars[7], integrator->rr_depth()));
        rt_check(rtVariableSet1i(s2.global_vars[8], integrator->hide_emitters()));
        rt_check(rtVariableSet1i(s2.global_vars[9], backprop_enabled));
        rt_check(rtVariableSet1i(s2.global_vars[10], return_image));

        auto *sensor = dynamic_cast<ProjectiveCamera *>(m_scene->sensor());
        Assert(sensor != nullptr);
        auto sensor_transform = sensor->world_transform()->eval(0);
        auto sensor_pos = sensor_transform.translation();
        auto frame = sensor_transform.to_frame<Vector3f>();
        rt_check(rtVariableSet3f(s2.global_vars[11], sensor_pos.x(), sensor_pos.y(), sensor_pos.z()));
        rt_check(rtVariableSet3f(s2.global_vars[12], frame.t.x(), frame.t.y(), frame.t.z()));
        rt_check(rtVariableSet3f(s2.global_vars[13], frame.s.x(), frame.s.y(), frame.s.z()));
        rt_check(rtVariableSet3f(s2.global_vars[14], frame.n.x(), frame.n.y(), frame.n.z()));
        // Near and far plane
        rt_check(rtVariableSet1f(s2.global_vars[15], sensor->near_clip()));
        rt_check(rtVariableSet1f(s2.global_vars[16], sensor->far_clip()));
        // Sensor to camera transform
        auto sample_to_camera = sensor->sample_to_camera();
        rt_check(rtVariableSetMatrix4x4fv(s2.global_vars[17], 1, sample_to_camera.matrix.x().data()));
        auto scene_bbox = m_scene->bbox();
        auto scene_scale = scene_bbox.extents()[scene_bbox.major_axis()];
        rt_check(rtVariableSet1f(s2.global_vars[18], scene_scale));

        if (supports_volumes(m_integrator_type)) {
            // We assume that the shape's bbox is the containing medium's bbox
            const Shape *shape = nullptr;
            for (const auto sh : m_scene->shapes()) {
                if (sh->interior_medium()) {
                    if (shape != nullptr)
                        NotImplementedError("More than a single medium in a scene");
                    shape = sh.get();
                }
            }
            auto medium_bbox = shape->bbox();
            rt_check(rtVariableSet3f(s2.global_vars[19], medium_bbox.min.x(), medium_bbox.min.y(),
                                     medium_bbox.min.z()));
            rt_check(rtVariableSet3f(s2.global_vars[20], medium_bbox.max.x(), medium_bbox.max.y(),
                                     medium_bbox.max.z()));
        }
    }

    /* Optimization flags and other parameters */ {
        rt_check(rtVariableSet1i(s2.global_vars[21], m_use_approx_li_1));
    }

    // Run once with zero-size for initialization
    // TODO: do we need to do this every single time, or just once?
    cuda_eval();
    rt_check(rtContextValidate(s.context));
    rt_check(rtContextLaunch3D(s.context, 2, 0, 0, 0));
    cuda_eval();

    // Launch OptiX programs (starting from ray generation)
    for (size_t i = 0; i < pass_count; i++) {
        rt_check(rtVariableSet1ui(s2.global_vars[3], base_seed + i));
        Log(EDebug, "Launching render kernel with %d spp.", spp);
        rt_check(rtContextLaunch3D(s.context, 2, nx, ny, spp));
        cuda_eval();
    }

    return return_image ? m_image_buffer : nullptr;
}

void OptiXRenderer::unpack_gradients(DifferentiableParameters *params) const {
    auto &s2 = *m_optix_state;
    for (auto &kv : s2.gradient_buffers) {
        auto &grad = std::get<2>(kv.second);
        add_gradient_for_param(*params, kv.first, grad, /* add_to_graph = */ false);
        // TODO: try and keep the same storage in place
        grad = zero<FloatC>(grad.size());
    }
}

void liboptix_nop() { }

MTS_IMPLEMENT_CLASS(OptiXRenderer, Object)
