import os
from os.path import join, dirname, realpath

import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
import numpy as np

from test_types import KEY_FIELDS


def load_runs_metadata(key, run_files):
    import pandas as pd

    run_fields = ['iteration', 'runtime', 'timings_forward', 'timings_backward', 'loss', 'pass_count']
    df = pd.DataFrame(columns=['method', 'time'] + KEY_FIELDS + run_fields)

    for method, fname in run_files.items():
        run = pd.read_csv(fname, header=0)
        # Strip whitespace in column names
        run = run.rename(columns=lambda x: x.strip())
        run = run.rename(columns={'timing': 'runtime', 'Iteration': 'iteration',
                                  'Error': 'loss', 'Elapsed': 'runtime'})

        if 'loss (not clean)' in run.columns:
            run = run.drop(columns='loss (not clean)')
        if 'iteration' not in run.columns:
            run['iteration'] = range(len(run))
        if 'pass_count' not in run.columns:
            run['pass_count'] = 1
        if 'timings_forward' not in run.columns:
            run['timings_forward'] = 0.
        if 'timings_backward' not in run.columns:
            run['timings_backward'] = run['runtime']

        assert set(run.columns) == set(run_fields), \
               '{} vs {} for method {}'.format(str(run.columns), str(run_fields), method)
        run['method'] = method
        for k in KEY_FIELDS:
            run[k] = key._asdict()[k]
        run['time'] = np.cumsum(run['runtime'])

        assert set(run.columns) == set(df.columns)
        df = df.append(run, sort=False)

    return df



def select_equal_time_from_runs(key, run_files, **kwargs):
    df = load_runs_metadata(key, run_files)
    return select_equal_time_from_df(df, **kwargs), df


def select_equal_time_from_df(df, minimum_iteration=0):
    max_it = df['iteration'].max()
    earliest_done = df[df['iteration'] == max_it]['time'].min()

    later = df[(df['time'] >= earliest_done) & (df['iteration'] >= minimum_iteration)]
    selected_its = later.groupby('method').agg({'iteration': 'min'})
    return selected_its.to_dict()['iteration']



def estimate_learning_rate_scaling(config, prepare_integrator,
                                   compute_loss, spp_large=512, output_dir=None):
    """Computes gradients at the first iteration for the given config,
    both with and without the Li = 1 approximation applied.
    From the ratio of the average norm of the gradients, estimates a
    step size factor which must be applied to the optimizer when
    using the approximation so that it is roughly equivalent
    to the learning rate with approximation disabled.
    """
    import enoki as ek
    from test_rb_gradients import gradients_to_df

    # Hardcoded value
    if 'approx_li_lr_scale' in config:
        return config['approx_li_lr_scale']

    # Compute the loss (high quality)
    print('[i] Starting learning rate scaling estimate')
    print('[ ] Rendering initial state at {}spp'.format(spp_large))
    renderer, params, _ = prepare_integrator(allow_approx=False)
    current_b = renderer.render(seed=13, spp=spp_large)
    if output_dir is not None:
        fname = join(output_dir, 'ratio_estimation_initial_{:04d}.exr'.format(0))
        current_b.write(fname)
        print('[+] Saved initial state to: {}'.format(fname))

    image_gradients, _, _ = compute_loss(current_b, cache=None)

    # Get high-quality gradients from RB without approximation
    print('[ ] Getting high-quality gradients ({}spp)'.format(spp_large))
    seed_backward = 14
    renderer.delta_render(seed_backward, image_gradients, spp=spp_large)
    gradients_clean = {k: ek.gradient(v) for k, v in params.items()}
    if output_dir is not None:
        grad_df = gradients_to_df({'rb': gradients_clean})
        fname = join(output_dir, 'gradients_clean.csv')
        with open(fname, 'w') as f:
            f.write(grad_df.to_csv())
        print('[+] Saved clean gradients to: {}'.format(fname))

    # Get high-quality gradients from RB with approximation
    # Note: scene should be reloaded in `prepare_integrator`, so gradients start over
    print('[ ] Getting high-quality approximate gradients ({}spp)'.format(spp_large))
    renderer, params, _ = prepare_integrator(allow_approx=True)
    renderer.delta_render(seed_backward, image_gradients, spp=spp_large)
    gradients_approx = {k: ek.gradient(v) for k, v in params.items()}
    if output_dir is not None:
        grad_df = gradients_to_df({'rb_bias': gradients_approx})
        fname = join(output_dir, 'gradients_approx.csv')
        with open(fname, 'w') as f:
            f.write(grad_df.to_csv())
        print('[+] Saved approximate gradients to: {}'.format(fname))

    # Estimate the ratio to be applied to the approx LR
    norm_clean, norm_approx = [np.mean([
        np.linalg.norm(v.numpy().ravel())
        for _, v in source.items()
    ]) for source in (gradients_clean, gradients_approx)]

    # TODO: detect if ratio will be noisy, perhaps averaging over a few seeds?
    ratio = norm_clean / norm_approx

    print('[i] Estimated ratio: {:.3f},\n    Clean gradients norm: {}, approx gradients norm: {}'.format(ratio, norm_clean, norm_approx))

    return ratio, gradients_clean, gradients_approx
