import threading
import time

import enoki as ek
import numpy as np


class GPUUsageSampler(threading.Thread):
    def __init__(self, sampling_frequency=10, max_entries=int(1e6)):
        """Sampling frequency in hertz."""
        super().__init__()
        self.sampling_frequency = sampling_frequency
        self.sleep_period = 1. / sampling_frequency
        self.max_entries = max_entries
        self.is_done = False
        self.times, self.used = [], []

        self.sample_used_gpu_ram()

    def sample_used_gpu_ram(self):
        if len(self.times) >= self.max_entries:
            return None

        free, total = ek.cuda_mem_get_info()
        used = total - free
        self.times.append(time.time())
        self.used.append(used)
        return used

    def run(self):
        while not self.is_done:
            self.sample_used_gpu_ram()
            time.sleep(self.sleep_period)

    def done(self):
        self.is_done = True
        self.join()
        return (self.times, self.used)

    def median_usage(self):
        baseline = self.used[0]
        return np.median(np.array(self.used) - baseline)

    def max_usage(self):
        baseline = self.used[0]
        return np.max(np.array(self.used) - baseline)


def parameters_expand_wildcard(keys):
    """Expands parameters that are None with all combinations
    of parameters. Returns a set of new keys."""
    from test_performance_radiative_backprop import make_key, SELECTED_CONFIGS, SPP, MAX_DEPTH, KEY_FIELDS

    if not isinstance(keys, set):
        if isinstance(keys, (list, tuple)):
            keys = set(keys)
        else:
            keys = set([keys])

    def expand_entry(keys, field, options):
        results = set()
        for key in keys:
            assert (None not in (key.resolution_x, key.resolution_y)), 'Not supported'

            current_values = key._asdict()
            if current_values[field] is not None:
                results.add(key)
                continue
            for o in options:
                new_values = dict(**current_values)
                new_values[field] = o
                results.add(make_key(**new_values))
        return results

    all_results = set(keys)
    values_for_field = {
        'scene_name': SELECTED_CONFIGS,
        'spp': SPP,
        'max_depth': MAX_DEPTH,
    }
    for field in KEY_FIELDS:
        if field not in values_for_field:
            continue
        # print('\nGoing to expand {}, results so far:'.format(field))
        # for k in all_results:
        #     print(k)
        all_results = expand_entry(all_results, field, values_for_field[field])

    return all_results
