import os
from os.path import join, dirname, realpath
import time
import sys

import enoki as ek
import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
from mitsuba.optix import OptiXRenderer
from mitsuba.render.autodiff import Adam
import numpy as np

from render_helpers import load_scene, save_info_csv, integrator_type_for_scene, \
                           reference_differentiable_render, compute_image_gradients, \
                           ensure_valid_values
from config import CONFIGS, OUTPUT_DIR

differentiable_rendering_dir = realpath(join(dirname(__file__), '../../../../differentiable-rendering'))
sys.path.append(join(differentiable_rendering_dir, 'python/optimize'))
from differentiable import optimize_scene
from autodiff_utils import image_to_float_d, combine_callbacks
import losses

def optimize_ad(config_name, pass_count=1, callback=None, output_dir=None):
    """Optimization of RGB colors or textures in the cbox scene."""
    if isinstance(config_name, str):
        config = CONFIGS[config_name]
    else:
        config = config_name
        config_name = config['name']
    output_dir = output_dir or join(OUTPUT_DIR, config_name + '_ad')
    os.makedirs(output_dir, exist_ok=True)

    # Render reference image, if missing
    ref_fname = config['ref']
    if not os.path.isfile(ref_fname):
        print('[ ] Rendering reference image with OptiX on the GPU...')
        scene = load_scene(config['scene'], parameters=config['ref_scene_params'])
        renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
        # current_b = renderer.render(seed=1234, spp=1024)
        current_b = renderer.render(seed=1234, spp=8192)
        current_b.write(ref_fname)
        print('[+] Saved reference image: {}'.format(ref_fname))
    reference_b = Bitmap(ref_fname)
    reference_b.write(join(output_dir, 'ref.exr'))

    scene_params = config.get('scene_params', [])
    scene = load_scene(config['scene'], parameters=scene_params)
    param_names = config['params']

    loss_function = losses.l2
    learning_rate = config.get('learning_rate', 2e-2)
    make_opt = (lambda params: Adam(params, lr=learning_rate))
    spp = config.get('forward_spp', 32)
    max_iterations = config.get('max_iterations', 150)

    save_intermediate = True
    block_size = config.get('block_size', None)
    spp_per_pass = spp // pass_count
    # TODO: handle non-perfect case
    assert np.abs(spp_per_pass * pass_count - spp) < 0.1 * spp, \
           '{} * {} != {}'.format(spp_per_pass, pass_count, spp)

    def clamp_callback(i, params, *args, **kwargs):
        ensure_valid_values(params)
        # for k, v in params.items():
        #     vv = v.numpy()
        #     print('[{:04d}] {}: min: {}, max: {}'.format(i, k, np.min(vv), np.max(vv)))

    if callback is not None:
        callback = combine_callbacks(clamp_callback, callback)
    else:
        callback = clamp_callback


    _, loss, images_np, metadata = optimize_scene(
        scene,
        param_names,
        image_to_float_d(reference_b, Bitmap.ERGB),
        optimizer=make_opt,
        max_iterations=max_iterations,
        loss_function=loss_function,
        pass_count=pass_count,
        block_size=block_size,
        image_spp=spp_per_pass,
        gradient_spp=spp_per_pass,
        output_dir=output_dir,
        write_image_per_iteration=save_intermediate,
        callback=callback,
        return_metadata=True,
        )


    # Save metadata about the run
    fname = join(output_dir, 'metadata.csv')
    save_info_csv(fname, iteration=list(range(max_iterations)), **metadata)
    print('[+] Saved run metadata to: {}'.format(fname))



if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(prog='test_radiative_backprop.py')
    parser.add_argument('config_name', type=str, help='Configuration (scene) to optimize')
    parser.add_argument('--passes', type=int, default=1, help='Pass count')
    args = parser.parse_args()

    optimize_ad(args.config_name, args.passes)
