from collections import namedtuple
from functools import partial
import os
from os.path import join, dirname, realpath

from mitsuba.core import Bitmap, Struct

from approx_utils import select_equal_time_from_runs
from config import OUTPUT_DIR, CBOX_PARAM_NAMES, RES_SCENE_DIR
from test_types import ResultKey
from optimize_smoke import re_render_initial, re_render_optimized

from notify import SlackNotifier


resolution = 256
max_depth = 8
grid_res = (256, 64, 256)
config = {
    'name': 'cbox-rgb',
    'scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb_initial.xml'),
    'ref_scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb.xml'),
    'ref': join(RES_SCENE_DIR, 'cbox/cbox_rgb_ref.exr'),
    # 'opt_scene': join(DIFF_SCENE_DIR, 'cbox/cbox_rgb_optimized.xml'),
    'params': CBOX_PARAM_NAMES,
    'has_volumes': False,
    'forward_spp': 32,
    'backward_spp': 32,
    'learning_rate': 2e-2,
    'max_iterations': 50,

    'scene_params': [
        ('res', str(resolution)),
        ('max_depth', str(max_depth)),
        ('state', 'initial'),
    ],
    'ref_scene_params': [
        ('res', str(resolution)),
        ('max_depth', str(max_depth)),
        ('state', 'target'),
    ],
}
key = ResultKey(config['name'], resolution, resolution, config['forward_spp'], max_depth)


def main():
    from optimize_rb import optimize_rb
    from optimize_ad import optimize_ad

    notifier = SlackNotifier()

    methods = {
        'rb':                    optimize_rb,
        'rb_biased':             partial(optimize_rb, use_approx_li_1=True),
        'rb_interleaved':        partial(optimize_rb, use_interleaving=True),
        'rb_biased_interleaved': partial(optimize_rb, use_approx_li_1=True,
                                                      use_interleaving=True),
        'ad':                    partial(optimize_ad, pass_count=4),
    }

    def callback(it_i, params, _1, _2, _3, _4, images_np, *args):
        import enoki as ek
        nonlocal current_output_dir

        # Not counted in runtime, it's okay to spend time here
        assert len(images_np) == 1
        current_b = Bitmap(images_np[0], pixel_format=Bitmap.ERGB)
        fname = join(current_output_dir, 'out_{:04d}.exr'.format(it_i))
        current_b.write(fname)
        # TODO: save optimized parameter values


    run_files = {}
    output_dir = join(OUTPUT_DIR, 'rb-optimizations', config['name'])
    for method, fun in methods.items():
        current_output_dir = join(output_dir, method)
        expected_output = join(current_output_dir, 'metadata.csv')
        run_files[method] = expected_output
        if os.path.isfile(expected_output):
            print('[.] Skipping existing method: {}'.format(method))
            continue

        print('[ ] Starting optimization for method: {}'.format(method))
        print('[ ] Output directory is: {}'.format(current_output_dir))
        # glhf
        fun(config, callback=callback, output_dir=current_output_dir)
        message = '[+] Done optimizing with method {}'.format(method)
        print(message)
        notifier.notify(message)


    # def opt_scene_params(method, it):
    #     params = [p for p in config['scene_params']
    #               if p[0] != 'initial_value']
    #     opt_fname = join(output_dir, method, 'albedo_{:04d}_grid-optimized.vol'.format(it))
    #     params.append(('initial_value', opt_fname))
    #     return params

    # selected_its, _ = select_equal_time_from_runs(key, run_files)
    # re_render_optimized(config, selected_its, opt_scene_params, output_dir)

    # # Also re-render the initial state at higher spp
    # re_render_initial(config, join(output_dir, 'rb'))


if __name__ == '__main__':
    main()
