from collections import namedtuple
from functools import partial
import os
from os.path import join, dirname, realpath

from mitsuba.core import Bitmap, Struct

from approx_utils import select_equal_time_from_runs
from config import OUTPUT_DIR, DIFF_SCENE_DIR, HETEROGENEOUS_DENSITY_PARAM_NAMES
from test_types import ResultKey

from notify import SlackNotifier

# --------------------------------------

resolution = 256
config = {
    'name': 'smoke-white',
    'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
    'ref': join(DIFF_SCENE_DIR, 'medium-opt/smoke-plume_ref-white.exr'),
    'opt_scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
    'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
    'has_volumes': True,
    'is_heterogeneous': True,
    'forward_spp': 256,
    'backward_spp': 256,
    'learning_rate': 9e-3,  # 2e-2
    'max_iterations': 100,

    # Estimated ratio: 3.739,
    # Clean gradients norm: 0.005953248590230942, approx gradients norm: 0.0015920683508738875
    # 'approx_lr_ratio': 3.739,
    'approx_lr_ratio': 1,  # TODO

    'block_size': 128,
    'scene_params': [
        ('res', str(resolution)),
        ('volume_filename', 'textures/density-grid-res-111-222-112-value-0.1.vol'),
        # ('volume_filename', 'textures/density-grid-res-111-222-112-value-0.018.vol'),
        ('smoke_color', '1.0, 1.0, 1.0'),
    ],
    'ref_scene_params': [
        ('res', str(resolution)),
        ('volume_filename', 'textures/smoke.vol'),
        ('smoke_color', '1.0, 1.0, 1.0'),
    ],
}
key = ResultKey(config['name'], resolution, resolution, config['forward_spp'], 8)

# --------------------------------------


def re_render_initial(config, output_dir, spp=8196):
    from mitsuba.optix import OptiXRenderer
    from mitsuba.render import DifferentiableParameters
    from render_helpers import load_scene, integrator_type_for_scene

    init_fname = join(output_dir, 'initial_{}spp.exr'.format(spp))
    if os.path.isfile(init_fname):
        print('[.] Skipping existing initial state render: {}'.format(init_fname))
        return init_fname

    scene_params = config.get('scene_params', [])
    scene = load_scene(config['scene'], parameters=scene_params)

    integrator_type = integrator_type_for_scene(scene, with_gradients=False)
    renderer = OptiXRenderer(scene, integrator_type)

    print('[ ] Re-rendering initial state at {} spp, params are: {}'.format(spp, scene_params))
    initial_b = renderer.render(seed=1234, spp=spp)
    initial_b.write(init_fname)
    print('[+] Saved initial image: {}'.format(init_fname))
    return init_fname


def re_render_optimized(config, selected_its, optimized_params_fn, output_dir, spp=8196):
    from mitsuba.optix import OptiXRenderer
    from render_helpers import load_scene, integrator_type_for_scene

    for method, selected_it in selected_its.items():
        opt_fname = join(output_dir, method, '{}_optimized_{:04d}.exr'.format(
            method, selected_it))
        if os.path.isfile(opt_fname):
            print('[.] Skipping optimized render: {} at iteration {}'.format(method, selected_it))
            continue

        opt_scene = config['opt_scene']
        opt_scene_params = optimized_params_fn(method, selected_it)

        scene = load_scene(opt_scene, parameters=opt_scene_params)
        print('[ ] Re-rendering method {} at iteration {}, params are: {}'.format(
            method, selected_it, opt_scene_params))
        renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
        opt_b = renderer.render(seed=1234, spp=spp)
        opt_b.write(opt_fname)
        print('[+] Saved reference image: {}'.format(opt_fname))


def compute_initial_loss(config, output_dir, loss_function, spp=None):
    from render_helpers import compute_image_gradients
    spp = spp or config['forward_spp']
    init_fname = re_render_initial(config, output_dir, spp=spp)
    current_b = Bitmap(init_fname)
    reference_b = Bitmap(config['ref'])
    _, loss, _ = compute_image_gradients(reference_b, current_b, loss_function)
    return loss.numpy().squeeze()


def main():
    from optimize_rb import optimize_rb
    from optimize_ad import optimize_ad

    notifier = SlackNotifier()
    methods = {
        'rb':                    optimize_rb,
        'rb_biased':             partial(optimize_rb, use_approx_li_1=True),
        'rb_interleaved':        partial(optimize_rb, use_interleaving=True),
        'rb_biased_interleaved': partial(optimize_rb, use_approx_li_1=True,
                                                      use_interleaving=True),
        'ad':                    partial(optimize_ad, pass_count=64),
    }

    def callback(it_i, params, _1, _2, _3, _4, images_np, *args):
        # from autodiff_utils import get_medium_resolution
        from io_utils import save_optimized_parameters
        nonlocal current_output_dir

        # Not counted in runtime, it's okay to spend time here
        assert len(images_np) == 1
        current_b = Bitmap(images_np[0], pixel_format=Bitmap.ERGB)
        fname = join(current_output_dir, 'out_{:04d}.exr'.format(it_i))
        current_b.write(fname)

        ADConfig = namedtuple('ADConfig', ['has_volume_data', 'param'])
        grid_res = (111, 222, 112)
        # if (it_i % 10) == 0 or it_i >= config['max_iterations'] - 1:
        if True:
            prefix = join(current_output_dir, 'density_{:04d}_'.format(it_i))
            special_config = ADConfig(True, config['params'][0])
            save_optimized_parameters(special_config, params, prefix=prefix,
                                      volume_resolution=grid_res)


    run_files = {}
    output_dir = join(OUTPUT_DIR, 'rb-optimizations', config['name'])
    for method, fun in methods.items():
        current_output_dir = join(output_dir, method)
        expected_output = join(current_output_dir, 'metadata.csv')
        run_files[method] = expected_output
        if os.path.isfile(expected_output):
            print('[.] Skipping existing method: {}'.format(method))
            continue
        print('Using output dir: {}'.format(current_output_dir))

        # TODO: remove this
        if method == 'ad':
            print('[!] Skipping AD, it\'s too slow')
            continue

        print('[ ] Starting optimization for method: {}'.format(method))
        print('[ ] Output directory is: {}'.format(current_output_dir))
        # glhf
        fun(config, callback=callback, output_dir=current_output_dir)
        message = '[+] Done optimizing with method {}'.format(method)
        print(message)
        notifier.notify(message)


    # Since AD won't finish in time, use intermediate results instead
    run_files['ad'] = join(dirname(run_files['ad']), 'loss.csv')

    def opt_scene_params(method, it):
        params = [p for p in config['scene_params']
                  if p[0] != 'volume_filename']
        opt_fname = join(output_dir, method,
                         'density_{:04d}_grid-optimized.vol'.format(it))
        params.append(('volume_filename', opt_fname))
        return params

    selected_its, _ = select_equal_time_from_runs(key, run_files)
    re_render_optimized(config, selected_its, opt_scene_params, output_dir)

    # Also re-render the initial state at higher spp
    re_render_initial(config, join(output_dir, 'rb'))

    # Compute loss at initial state (same spp to have same noise level)
    from render_helpers import l2_loss
    initial_loss = compute_initial_loss(config, join(output_dir, 'rb'), l2_loss)
    print('[+] Loss value at initial state: {}'.format(initial_loss))

if __name__ == '__main__':
    main()
