import os
from os.path import join, dirname, realpath
from functools import partial

import enoki as ek
import numpy as np

from mitsuba.core import Bitmap, Struct, Thread, EColorMode, Log, EInfo
from mitsuba.core.xml import load_file
from mitsuba.optix import OptiXRenderer
from mitsuba.render.autodiff import render as mts_autodiff_render
from mitsuba.render.autodiff_utils import image_to_float_d, float_d_to_bitmap
from mitsuba.test.util import fresolver_append_path


@fresolver_append_path
def load_scene(filename, *args, **kwargs):
    """Prepares the file resolver and loads a Mitsuba scene from the given path."""
    fr = Thread.thread().file_resolver()
    here = os.path.dirname(__file__)
    fr.append(here)
    fr.append(os.path.join(here, filename))
    fr.append(os.path.dirname(filename))

    kwargs.setdefault('color_mode', EColorMode.ERGBColor)
    scene = load_file(filename, *args, **kwargs)
    assert scene is not None
    return scene


def l1_loss(x, y):
    assert ek.slices(x) == ek.slices(y)
    assert isinstance(x, ek.Vector3fD)

    return ek.hsum(ek.hsum(
        ek.abs(x - y)
    )) / ek.slices(x)


def l2_loss(x, y):
    assert ek.slices(x) == ek.slices(y)
    assert isinstance(x, ek.Vector3fD)

    return ek.hsum(ek.hsum(
        ek.sqr(x - y)
    )) / ek.slices(x)


def ensure_valid_values(params):
    # TODO: more principle way
    for k, v in params.items():
        if k.endswith('/rgb') or k.endswith('/bitmap') or k.endswith('/data_r')\
           or k.endswith('/data_g') or k.endswith('/data_b')\
           or ('alpha' in k and (k.endswith('/value') or k.endswith('/bitmap'))):
            params[k] = ek.clamp(ek.detach(v), 0, 1)
            # params[k] = type(v)(ek.clamp(ek.detach(v), 0, 1))
        elif k.endswith('/density') or ('texture3d' in k.lower() and k.endswith('/data')):
            # This is to avoid numerical precision issues
            params[k] = ek.clamp(ek.detach(v), 0., 6.)
            # params[k] = type(v)(ek.clamp(ek.detach(v), 0., 6.))
        else:
            continue
        ek.set_requires_gradient(params[k])


def save_info_csv(fname, **kwargs):
    headers = kwargs.keys()
    values = zip(*kwargs.values())
    with open(fname, 'w') as f:
        f.write(', '.join(headers) + '\n')
        for r in values:
            f.write(', '.join([str(v) for v in r]) + '\n')

def integrator_type_for_scene(scene, with_gradients, config=None):
    if config is not None and 'integrator' in config:
        name = config['integrator']
        return OptiXRenderer.EIntegratorType.__entries[name][0]

    has_volumes = scene.sensor().medium() is not None
    for shape in scene.shapes():
        has_volumes = has_volumes or shape.is_medium_transition()

    if has_volumes:
        return OptiXRenderer.ERadiativeBackpropVol if with_gradients else OptiXRenderer.EVolPath
    else:
        return OptiXRenderer.ERadiativeBackprop if with_gradients else OptiXRenderer.EPath

# -----------------------------------------------------------------------------


def compute_image_gradients(reference_b, current_b, loss_function,
                            cache=None, return_3d=False):
    """ Computes test inputs for radiative backropagation given a target image and current image.

    The input to radiative backprop is the gradient of the loss with respect
    to each image pixel.
    """
    assert current_b.pixel_format() == Bitmap.ERGB

    # Convert to differentiable Enoki arrays
    if cache is None:
        assert reference_b.pixel_format() == Bitmap.ERGB
        arr = np.array(reference_b, copy=False)
        ref = ek.Vector3fC(*[ek.FloatC(arr[:, :, i].flatten()) for i in range(3)])
        cache = {
            'reference': ref,
            'n_entries': ek.slices(ref),
            'index': 3 * ek.UInt32D.arange(ek.slices(ref)),
        }

    reference = cache['reference']
    arr = np.array(current_b, copy=False)
    current = ek.Vector3fD(*[ek.FloatD(arr[:, :, i].flatten()) for i in range(3)])

    ek.set_requires_gradient(current, True)
    l = loss_function(current, reference)
    ek.backward(l)

    grad3 = ek.gradient(current)
    if return_3d:
        grad_return = grad3
    else:
        # Flatten gradient vector
        grad_return = ek.FloatC.zero(len(grad3) * ek.slices(grad3))
        ek.scatter(grad_return, grad3[0], cache['index'])
        ek.scatter(grad_return, grad3[1], cache['index'] + 1)
        ek.scatter(grad_return, grad3[2], cache['index'] + 2)

    return grad_return, l, cache


def reference_differentiable_render(scene, reference_b, opt, loss_function,
                                    pass_count=10, spp=8, base_seed=12345,
                                    gradient_spp=None,
                                    output_dir=None):
    """Computes gradients of the loss with respect to some tracked parameters
    using automatic differentiation."""
    assert pass_count >= 1
    pixel_format = Bitmap.ERGB
    use_reattach_trick = gradient_spp is not None

    ref_flat = image_to_float_d(reference_b, pixel_format)
    gradients = {}

    for pass_i in range(pass_count):
        seed_i = base_seed + pass_i

        if use_reattach_trick:
            assert gradient_spp > 0
            seed2 = (pass_count + 1) + seed_i
            # Forward pass: high-quality image reconstruction, no gradients
            with opt.no_gradients():
                scene.sampler().seed(seed_i)
                image_block = mts_autodiff_render(scene, pixel_format=pixel_format, spp=spp)

            # Forward pass: low-spp gradient estimate
            scene.sampler().seed(seed2)
            gradients_block = mts_autodiff_render(
                scene, pixel_format=pixel_format, spp=gradient_spp)

            # Reattach trick
            block_values = image_to_float_d(image_block, pixel_format)
            gradient_block_values = image_to_float_d(gradients_block, pixel_format)
            ek.reattach(block_values, gradient_block_values)

        else:
            # Single render for both image values and gradient values (may be biased)
            scene.sampler().seed(seed_i)
            image_block = mts_autodiff_render(scene, pixel_format=pixel_format, spp=spp)
            block_values = image_to_float_d(image_block, pixel_format)

        if output_dir is not None:
            pass_fname = join(output_dir, 'autodiff_{:02d}.exr').format(pass_i)
            shape = (reference_b.width(), reference_b.height(), reference_b.channel_count())
            float_d_to_bitmap(block_values, shape, pixel_format).write(pass_fname)
            Log(EInfo, '[+] Saved pass debug image to: {}'.format(pass_fname))

        loss = loss_function(block_values, ref_flat)
        ek.backward(loss)

        for k, v in opt.params.items():
            if k not in gradients:
                gradients[k] = ek.gradient(v) / float(pass_count)
            else:
                gradients[k] += ek.gradient(v) / float(pass_count)
    return image_block, gradients


def finite_differences_render(scene, reference_b, opt, loss_function,
                              epsilon=1e-3,
                              pass_count=10, spp=8, base_seed=12345,
                              output_dir=None,
                              render_function=None, method_name='fd'):
    """Computes gradients of the loss with respect to some tracked parameters
    using finite differences."""
    assert pass_count >= 1
    pixel_format = Bitmap.ERGB

    # We don't need autodiff here
    for k, _ in opt.params.items():
        ek.set_requires_gradient(opt.params[k], False)


    render_function = render_function or mts_autodiff_render
    ref_flat = image_to_float_d(reference_b, pixel_format)
    gradients = {}

    render_counter = 0
    def finite_differences_single(update_param, epsilon):
        nonlocal render_counter, seed_i, pass_i
        loss_values = []
        for change_i, change in enumerate([-0.5 * epsilon, 0.5 * epsilon]):
            update_param(change)
            scene.sampler().seed(seed_i)  # Use the same seed for all change_i
            image_block = render_function(scene, pixel_format=pixel_format, spp=spp)
            block_values = image_to_float_d(image_block, pixel_format)
            loss = loss_function(block_values, ref_flat)
            loss_values.append(loss.numpy().squeeze())

            if output_dir is not None:
                pass_fname = join(output_dir, '{}_{:04d}_pass_{:02d}_{:d}.exr').format(method_name, render_counter, pass_i, change_i)
                shape = (reference_b.width(), reference_b.height(), reference_b.channel_count())
                float_d_to_bitmap(block_values, shape, pixel_format).write(pass_fname)
                # print('[+] Saved pass debug image to: {}'.format(pass_fname))
            render_counter += 1
        # Reset param to original value
        update_param(0)

        return (loss_values[1] - loss_values[0]) / epsilon

    def finite_differences(param_name, epsilon):
        v = opt.params[param_name]
        if isinstance(v, ek.FloatD):
            original_value = v.numpy().copy()
            result = np.zeros(len(v))
            for idx in range(len(v)):
                def update_param(diff, index):
                    value = original_value.copy()
                    value[index] += diff
                    opt.params[param_name] = ek.FloatD(value)
                result[idx] = finite_differences_single(
                    partial(update_param, index=idx), epsilon)
            return result

        elif isinstance(v, ek.Vector3fD):
            grad3 = np.zeros((3,))
            for k in range(3):
                assert len(
                    v[k]) == 1, "Not implemented yet: Vector3fD parameters with more than 1 entry"
                original_value = v.numpy()

                def update_param(diff):
                    diff3 = [0, 0, 0]
                    diff3[k] = diff
                    opt.params[param_name] = ek.Vector3fD(original_value + diff3)

                grad3[k] = finite_differences_single(update_param, epsilon)
            return grad3

        else:
            raise RuntimeError("Unsupported Enoki type in finite differences: {}".format(type(v)))

    # For every single parameter (including e.g. individual component of a color),
    # render the image twice to compute finite differences gradient.
    for pass_i in range(pass_count):
        seed_i = base_seed + pass_i
        scene.sampler().seed(seed_i)

        for k, _ in opt.params.items():
            # Compute gradients with finite differences
            grad_k = finite_differences(k, epsilon) / float(pass_count)

            if k not in gradients:
                gradients[k] = grad_k
            else:
                gradients[k] += grad_k

    # Convert gradients to Enoki types for consistency with other methods
    for k, v in opt.params.items():
        tp = type(ek.detach(v))
        gradients[k] = tp(gradients[k])

    return gradients
