import os
from os.path import join, dirname, realpath
import subprocess
import pytest
import re
import time
import sys

import enoki as ek
import mitsuba
from mitsuba import set_thread_count
from mitsuba.core import Bitmap, Struct, Thread, EColorMode, EInfo, EError, EWarn, EDebug, Log
from mitsuba.optix import OptiXRenderer
from mitsuba.render.autodiff import get_differentiable_parameters, Adam, SGD, render as render_gpu

# Thread.thread().logger().set_log_level(EDebug)

import numpy as np
import scipy.misc
from scipy import ndimage

from render_helpers import load_scene, l2_loss, save_info_csv, integrator_type_for_scene, \
                           reference_differentiable_render, compute_image_gradients, \
                           ensure_valid_values

differentiable_rendering_dir = realpath(join(dirname(__file__), '../../../../differentiable-rendering'))
sys.path.append(join(differentiable_rendering_dir, 'python/optimize'))
from differentiable import simple_render, optimize_scene
from autodiff_utils import image_to_float_d, combine_callbacks, clamp_parameters
from losses import l2

from config import CONFIGS, OUTPUT_DIR

from approx_utils import estimate_learning_rate_scaling, select_equal_time_from_runs


###################################################################################################


# TODO make it work with other pixel format
pixel_format = Bitmap.ERGB

# Apply a Gaussian filter onto the optimized texture in the callback (on top of regularization)
ENABLE_TEXTURE_BLURRING = False

# Enable/disable opening images in TEV viewer
SEND_RENDERS_TO_TEV = False
SEND_TEXTURE_TO_TEV = False


###################################################################################################


ENABLE_TEV = True
def open_image_tev(image_fname, force=False):
    if force or ENABLE_TEV:
        subprocess.call(["tev", image_fname])

#--------------------------------------------------------------------------------------------------

class Scheduler:
    def __init__(self, schedule = [(0, 0)]):
        self.schedule = schedule

    def get(self, iteration):
        assert iteration >= 0
        for i in range(len(self.schedule)):
            if iteration < self.schedule[i][0]:
                return self.schedule[i-1][1]

        return self.schedule[-1][1]

    def eval(self, iteration):
        assert iteration >= 0

        if iteration <= self.schedule[0][0]:
            return self.schedule[0][1]

        for i in range(len(self.schedule)):
            if iteration < self.schedule[i][0]:
                w = (iteration - self.schedule[i-1][0]) / float(self.schedule[i][0] - self.schedule[i-1][0])
                return (1.0 - w) * self.schedule[i-1][1] + w * self.schedule[i][1]

        return self.schedule[-1][1]

    def has_schedule_changed(self, iteration):
        for k, v in self.schedule:
            if iteration == k:
                return True
        return False

#--------------------------------------------------------------------------------------------------

# NOTE currently only work for FloatD representing flatten images

def laplacian(x, shape, channels):
    size = channels
    strides = [0] * len(shape)
    for i, entry in reversed(list(enumerate(shape))):
        strides[i] = size
        size *= entry
    if len(x) != size:
        raise Exception('laplacian(): Input array "x"'
                        'has size %i, expected %i!', len(x), size)

    # indices = ek.UInt32D.arange(int(size / channels)) * channels
    indices = ek.UInt32D.arange(size)
    tmp = indices

    value = x * (-2 * len(shape))

    for dim in range(len(shape)):
        coordinate     = tmp / strides[dim]
        tmp           -= coordinate * strides[dim]

        source_index_1 = ek.select(
            coordinate > 0,
            indices - strides[dim], indices)

        source_index_2 = ek.select(
            coordinate < shape[dim] - 1,
            indices + strides[dim], indices)

        value += ek.gather(x, source_index_1)
        value += ek.gather(x, source_index_2)

    return ek.hsum(ek.hsum(ek.sqr(value)))


def total_variation(x, shape, channels):
    size = channels
    strides = [0] * len(shape)
    for i, entry in reversed(list(enumerate(shape))):
        strides[i] = size
        size *= entry
    if len(x) != size:
        raise Exception('total_variation(): Input array "x"'
                        'has size %i, expected %i!', len(x), size)

    # indices = ek.UInt32D.arange(int(size / channels)) * channels
    indices = ek.UInt32D.arange(size)
    tmp = indices
    value = 0

    for dim in range(len(shape)):
        coordinate     = tmp / strides[dim]
        tmp           -= coordinate * strides[dim]

        source_index = ek.select(
            coordinate < shape[dim] - 1,
            indices + strides[dim], indices)

        value += ek.abs(x - ek.gather(x, source_index))

    return ek.hsum(ek.hsum(value))


#--------------------------------------------------------------------------------------------------

def l2_clamped(y, x, low, high):
    print('max abs loss: {}'.format(ek.hmax(ek.hmax(ek.detach(ek.abs(y-x))))))
    return  ek.hsum(
              ek.hsum(
                ek.sqr(
                  ek.clamp(ek.abs(y - x), low, high)
            ))) / ek.slices(x)

#--------------------------------------------------------------------------------------------------

def save_optimized_texture(output_dir, prefix, it_i, data, texture_res):
    if it_i == -1:
        texture_fname = join(output_dir, prefix + '_opt_texture_ref.exr')
    else:
        texture_fname = join(output_dir, prefix + '_opt_texture_{:04d}.exr'.format(it_i))
    texture_b = Bitmap(data.numpy().reshape((texture_res[0], texture_res[1], -1)).astype(np.float32))
    texture_b.write(texture_fname)
    return texture_fname

#--------------------------------------------------------------------------------------------------

def blur_optimized_texture(data, texture_res, sigma, alpha=0.5):
    from skimage.filters import gaussian

    texture_np = data.numpy().reshape((texture_res[0], texture_res[1], -1)).astype(np.float32)
    # blurred_texture_np = ndimage.gaussian_filter(texture_np, sigma=sigma)
    # blurred_texture_np = ndimage.uniform_filter(texture_np, size=(2, 2, 1))
    blurred_texture_np = gaussian(texture_np, sigma=sigma, multichannel=True)
    result = (1.0 - alpha) * texture_np + alpha * blurred_texture_np
    data = ek.FloatD(result.flatten())
    ek.set_requires_gradient(data, True)
    return data

#--------------------------------------------------------------------------------------------------

def load_reference_image(config, output_dir):
    ref_fname = config['ref']
    if not os.path.isfile(ref_fname):
        print('[ ] Rendering reference image with OptiX on the GPU...')
        scene = load_scene(config['scene'], parameters=config['ref_scene_params'])
        renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
        current_b = renderer.render(seed=1234, spp=config['ref_spp'], pass_count=4)
        current_b.write(ref_fname)
        print('[+] Saved reference image: {}'.format(ref_fname))

    if SEND_RENDERS_TO_TEV:
        open_image_tev(ref_fname)

    b = Bitmap(ref_fname)
    a = np.array(b, copy=False)
    nan_entries = np.isnan(a)
    if np.any(nan_entries):
        Log(EWarn, "NaN entries found in reference: {}! Replacing with zeros.".format(np.sum(nan_entries)))
        a[nan_entries] = 0.

    b.write(join(output_dir, 'ref.exr'))

    return b


def my_callback(output_dir, prefix):
    def callback(it_i, params, optimized_params, opt, max_iterations, loss, image, image_ref):
        # print('callback it = {}'.format(it_i))
        # Ensure valid gradients
        ensure_valid_values(params)
        # print('ensured valid values')

        # Blur the texture image with a Gaussian filter
        if ENABLE_TEXTURE_BLURRING:
            params[optimized_params[0]] = blur_optimized_texture(params[optimized_params[0]], texture_res, sigma=0.38)

        tex_fname = save_optimized_texture(output_dir, prefix, it_i, params[optimized_params[0]], texture_res)

        # Save the rendered image to a file
        image_fname = join(output_dir, prefix + '_out_{:04d}.exr'.format(it_i))
        Bitmap(image[0], pixel_format).write(image_fname)

        # Open the rendered image in TEV
        if SEND_RENDERS_TO_TEV:
            open_image_tev(image_fname)

        # Open the optimized texture in TEV
        if SEND_TEXTURE_TO_TEV:
            open_image_tev(tex_fname)

    return callback


###################################################################################################


def gen_teaser_RB(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
                  prefix, fwrd_spp_mult=1, bckwrd_spp_mult=1, use_approx_li_1=False,
                  approx_lr_scale=1.0,  use_interleaving=False):

    #----------------------------------------------
    #   Load config and create output directory

    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)
    os.makedirs(output_dir, exist_ok=True)

    #----------------------------------------------
    #    Load or generate reference image

    reference_b = load_reference_image(config, output_dir)

    # Send target texture to TEV
    if SEND_TEXTURE_TO_TEV:
        scene = load_scene(config['scene'], parameters=config['ref_scene_params'])
        params = get_differentiable_parameters(scene)
        fname = save_optimized_texture(output_dir, '', -1, params[config['params'][0]], texture_res)
        open_image_tev(fname)

    #----------------------------------------------
    #    Load scene and optimizer configuration

    scene = load_scene(config['scene'], parameters=config.get('scene_params', []))
    params = get_differentiable_parameters(scene)
    params.keep(config['params'])

    opt = opt_constr(params)

    max_iterations = config.get('max_iterations', 150)

    integrator_type = integrator_type_for_scene(scene, with_gradients=True, config=config)
    renderer = OptiXRenderer(scene, integrator_type, params=params)
    if use_approx_li_1:
        print('[i] Using approximate Li = 1')
        renderer.set_use_approx_li_1(True)
        print('[i] Multiplying lr by {}'.format(approx_lr_scale))
        opt.lr *= approx_lr_scale

    if use_interleaving:
        print('[i] Using interleaving (intentional off-by-one)')

    callback = my_callback(output_dir, prefix)

    #----------------------------------------------
    #    Optimization loop

    cache = None
    metadata = { k: [] for k in ['timing', 'timings_forward', 'timings_backward', 'loss'] }

    if use_interleaving:
        # Render current state for initial image gradients
        current_b = renderer.render(0, spp=spp_scheduler.get(0) * fwrd_spp_mult)

    for it_i in range(max_iterations):
        t0 = time.time()

        # set seeds
        seed_forward = it_i
        seed_backward = max_iterations + seed_forward + 1

        # Query schedulers
        forward_spp   = spp_scheduler.get(it_i) * fwrd_spp_mult
        backward_spp  = spp_scheduler.get(it_i) * bckwrd_spp_mult

        if use_interleaving:
            backward_spp = max(backward_spp, forward_spp)

        # Reset optimizer when SPP schedule changes
        # if spp_scheduler.has_schedule_changed(it_i):
            # opt.reset(config['params'][0])

        # Render current state (forward pass)
        if not use_interleaving:
            current_b = renderer.render(seed_forward, spp=forward_spp)
        t_end_forward = time.time()

        # Apply regularization term
        reg = regularisation(params[config['params'][0]], texture_res) / approx_lr_scale
        if not reg == 0.0:
            print('reg: {}'.format(reg))
            ek.backward(reg)
            # We pass None to avoid for that weight to be weighted per-pass
            opt.accumulate_gradients(None)

        # Compute image gradients
        image_gradients, loss, cache = compute_image_gradients(
            reference_b, current_b, loss_function, cache=cache)

        img_or_none = renderer.delta_render(seed_backward, image_gradients, spp=backward_spp,
                                            return_image=use_interleaving)
        opt.accumulate_gradients(None)

        if use_interleaving:
            current_b = img_or_none

        # Apply gradient step
        opt.step()

        # Re-upload globals to the correct OptiX constants
        renderer.upload_bsdf_globals()

        timings.append(time.time() - t0)
        loss_values.append(loss.numpy()[0])

        images_np = [np.array(current_b, copy=True)]
        callback(it_i, params, config['params'], opt, max_iterations, loss, images_np, None)

        print("[{:03d}]: loss = {:.6f}, took {:.2f}s".format(
            it_i, loss_values[-1], timings[-1]), end='\n')

    # Save metadata about the run
    fname = join(output_dir, prefix + '_metadata.csv')
    save_info_csv(fname, iteration=list(range(len(timings))),
                  runtime=timings, loss=loss_values)
    print('[+] Saved run metadata to: {}'.format(fname))


###################################################################################################


def gen_teaser_AD(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
                  fwrd_spp_mult=1, bckwrd_spp_mult=1, pass_count=1):

    #----------------------------------------------
    #   Load config and create output directory

    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)
    os.makedirs(output_dir, exist_ok=True)

    #----------------------------------------------
    #    Load or generate reference image

    reference_b = load_reference_image(config, output_dir)
    reference_d = image_to_float_d(reference_b, pixel_format)

    #----------------------------------------------
    #    Load scene and optimizer configuration

    scene = load_scene(config['scene'], parameters=config.get('scene_params', []))

    gradient_spp = int(spp_scheduler.get(0) * bckwrd_spp_mult / pass_count)
    image_spp = int(spp_scheduler.get(0) * fwrd_spp_mult / pass_count)

    print('[+] Initial forward SPP: {}'.format(image_spp if image_spp else gradient_spp))
    print('[+] Initial backward SPP: {}'.format(gradient_spp))

    # Combine loss and regularization function
    def loss_reg_function(x, y, i, max_iterations, params):
        return loss_function(x, y) + regularisation(params[config['params'][0]], texture_res)

    #----------------------------------------------
    #    Run optimization

    _, _, _, metadata = optimize_scene(
        scene = scene,
        optimized_params=config['params'],
        image_ref_=reference_d,
        max_iterations=config.get('max_iterations', 150)+1,
        log_level=0,
        pixel_format=pixel_format,
        optimizer=opt_constr,
        loss_function=loss_reg_function,
        callback=my_callback(output_dir, 'ad'),
        block_size=1280,
        pass_count=pass_count,
        image_spp=image_spp,
        gradient_spp=gradient_spp,
        enabled_sensors=None,
        output_dir=output_dir,
        # verbose=True,
        verbose=False,
        spp_multiplier= lambda i: int(spp_scheduler.get(i) / spp_scheduler.get(0)),
        batch_size=None,
        timer_exclude_callback=True
    )

    # Save metadata about the run
    fname = join(output_dir, 'ad' + '_metadata.csv')
    save_info_csv(fname, iteration=list(range(len(metadata['timing']))),
                  runtime=metadata['timing'], loss=metadata['loss'])
    print('[+] Saved run metadata to: {}'.format(fname))


###################################################################################################


def gen_clean_loss(config_name, loss_function, spp, prefix, crop=None, low_res=False):
    # Load config
    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)

    # Load reference
    if not crop is None: # crop
        ref_fname = join(output_dir, 'ref_crop.exr')
        if not os.path.isfile(ref_fname):
            print('[ ] Rendering reference image with OptiX on the GPU...')
            if low_res:
                scene = load_scene(config['scene_lowres'], parameters=config['ref_scene_params'])
            else:
                scene = load_scene(config['scene'], parameters=config['ref_scene_params'])

            crop(scene)

            renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
            current_b = renderer.render(seed=4321, spp=1024, pass_count=1)
            current_b.write(ref_fname)
            print('[+] Saved reference image: {}'.format(ref_fname))

        if SEND_RENDERS_TO_TEV:
            open_image_tev(ref_fname)

        reference_b = Bitmap(ref_fname)
        a = np.array(reference_b, copy=False)
        nan_entries = np.isnan(a)
        if np.any(nan_entries):
            Log(EWarn, "NaN entries found in reference: {}! Replacing with zeros.".format(np.sum(nan_entries)))
            a[nan_entries] = 0.

        reference_b.write(join(output_dir, 'ref_crop.exr'))
    else:
        reference_b = load_reference_image(config, output_dir)

    reference_d = image_to_float_d(reference_b, pixel_format)

    # Load scene
    if low_res:
        scene = load_scene(config['scene_lowres'], parameters=config['scene_params'])
    else:
        scene = load_scene(config['scene'], parameters=config['scene_params'])
    params = get_differentiable_parameters(scene)
    renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))

    max_iterations = 50# config.get('max_iterations', 150)

    # Set crop
    if not crop is None:
        crop(scene)

    loss_values = []
    for it_i in range(max_iterations):
        # Skip if clean render already exists
        render_fname = join(output_dir, prefix + '_clean_loss_{:04d}.exr'.format(it_i))
        if not os.path.isfile(render_fname):
            # Replace optimized texture
            texture_fname = join(output_dir, prefix + '_opt_texture_{:04d}.exr'.format(it_i))
            if not os.path.isfile(texture_fname):
                break
            texture_b = Bitmap(texture_fname)
            params[config['params'][0]] = ek.FloatD(np.array(texture_b).flatten())

            # Render image
            current_b = renderer.render(seed=4321, spp=spp, pass_count=1)

            current_b.write(render_fname)
        else:
            current_b = Bitmap(render_fname)

        # Compute and store loss value
        current_d = image_to_float_d(current_b, pixel_format)
        loss = loss_function(current_d, reference_d)
        loss_values.append(loss.numpy()[0])

        # if SEND_RENDERS_TO_TEV:
            # open_image_tev(render_fname)

        print("[{:03d}]: prefix = {}, loss = {:.12f}".format(it_i, prefix, loss_values[-1]), end='\n')

    # from matplotlib import pyplot as plt
    # plt.plot(loss_values)
    # plt.plot(np.log(loss_values))
    # plt.semilogy(loss_values)
    # plt.show()

    # Save metadata about the loss
    fname = join(output_dir, prefix + '_metadata_losses.csv')
    save_info_csv(fname, iteration=list(range(len(loss_values))), loss=loss_values)
    print('[+] Saved run metadata to: {}'.format(fname))


###################################################################################################


def compute_texture_loss(config_name, loss_function, prefix):
    # Load config
    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)

    scene = load_scene(config['scene'], parameters=config['ref_scene_params'])
    params = get_differentiable_parameters(scene)

    reference_texture_np = np.array(params[config['params'][0]], copy=True)

    max_iterations = config.get('max_iterations', 150)

    loss_values = []
    for it_i in range(max_iterations):
        # Replace optimized texture
        texture_fname = join(output_dir, prefix + '_opt_texture_{:04d}.exr'.format(it_i))
        if not os.path.isfile(texture_fname):
            break

        texture_b = Bitmap(texture_fname)
        current_texture_np = np.array(texture_b).flatten()

        # Compute and store loss value
        loss = np.sum(np.power(current_texture_np - reference_texture_np, 2))
        # loss = np.sum(np.abs(current_texture_np - reference_texture_np))
        loss_values.append(loss)

        print("[{:03d}]: prefix = {}, loss = {:.12f}".format(it_i, prefix, loss_values[-1]), end='\n')

    from matplotlib import pyplot as plt
    # plt.plot(loss_values)
    # plt.plot(np.log(loss_values))
    plt.semilogy(loss_values)
    plt.show()

    # Save metadata about the loss
    fname = join(output_dir, prefix + '_metadata_losses.csv')
    save_info_csv(fname, iteration=list(range(len(loss_values))), loss=loss_values)
    print('[+] Saved run metadata to: {}'.format(fname))


###################################################################################################


def render_config(config_name, spp, prefix='ref', it_i = -1, render_mode='rb', crop=None):
    # Load config
    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)

    # Load scene
    if prefix == 'ref':
        scene = load_scene(config['scene'], parameters=config['ref_scene_params'])
        render_fname = join(output_dir, 'ref_clean_render_ref.exr')
    else:
        scene = load_scene(config['scene'], parameters=config.get('scene_params', []))
        if it_i >= 0:
            render_fname = join(output_dir, prefix + '_clean_render_{:04d}.exr'.format(it_i))
        else:
            render_fname = join(output_dir, prefix + '_clean_render_initial.exr')

    params = get_differentiable_parameters(scene)
    params.keep(config['params'])

    if not crop is None:
        crop(scene)

    if it_i >= 0:
        # Replace optimized texture
        texture_fname = join(output_dir, prefix + '_opt_texture_{:04d}.exr'.format(it_i))
        if not os.path.isfile(texture_fname):
            print("error: texture file doesn't exist: {}".format(texture_fname))
            exit(0)

        texture_b = Bitmap(texture_fname)
        params[config['params'][0]] = ek.FloatD(np.array(texture_b).flatten())

    # Render image
    if render_mode == 'rb':
        # integrator_type = integrator_type_for_scene(scene, with_gradients=False, config=config)
        integrator_type = integrator_type_for_scene(scene, with_gradients=True, config=config)
        renderer = OptiXRenderer(scene, integrator_type, params=params)
        current_b = renderer.render(seed=4321, spp=spp, pass_count=1)
    elif render_mode == 'ad_cpu':
        current_b= simple_render(scene).bitmap().convert(pixel_format, Struct.EFloat32, False)
    elif render_mode == 'ad':
        current_b = render_gpu(scene, pixel_format=pixel_format, spp=spp).bitmap().convert(pixel_format, Struct.EFloat32, False)
    else:
        print('[render_config]: Invalid prefix, should be rb, ad or ad_cpu.')
        exit(0)

    current_b.write(render_fname)

    if True:
        open_image_tev(render_fname)


###################################################################################################

def calculate_approx_lr_scale(config_name, loss_function):
    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)

    def prepare_integrator(allow_approx=True):
        scene_params = config.get('scene_params', [])
        scene = load_scene(config['scene'], parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])

        # Optimization setup
        opt = opt_constr(params)
        integrator_type = integrator_type_for_scene(scene, with_gradients=True, config=config)
        renderer = OptiXRenderer(scene, integrator_type, params=params)
        if allow_approx:
            print('[i] Using approximate Li = 1')
            renderer.set_use_approx_li_1(True)
        return renderer, params, opt

    def compute_loss(current_b, cache=None):
        reference_b = load_reference_image(config, output_dir)
        return compute_image_gradients(reference_b, current_b, loss_function, cache=cache)

    approx_lr_scale, _, _ = estimate_learning_rate_scaling(config, prepare_integrator,
                                                           compute_loss, spp_large=128)
    return approx_lr_scale


###################################################################################################


if __name__ == '__main__':

    #----------------------------------------------
    #   Configure pass

    # config_name = 'cbox-textured'
    config_name = 'bell'
    # config_name = 'bell_no_ceiling'

    # [RB]: doesn't do anything
    # [AD]: defines in how many passes to split the schedule_spp[0] (use to fit GPU memory)
    # pass_count = 8
    pass_count = 4  # TODO: tweak

    # Other optimization parameters
    # spp_scheduler = Scheduler([(0, 16), (15, 32), (25, 64)])
    # spp_scheduler = Scheduler([(0, 16), (10, 32), (25, 64), (40, 128)])
    # spp_scheduler = Scheduler([(0, 8), (20, 16), (50, 32), (80, 64)])
    spp_scheduler = Scheduler([(0, 4), (40, 8), (80, 16), (120, 32)])
    # spp_scheduler = Scheduler([(0, 16)])
    fwrd_spp_mult = 2
    bckwrd_spp_mult = 1

    # texture_res = [512, 1024]
    texture_res = [256, 512]
    # texture_res = [128, 256]

    #----------------------------------------------
    #   Define loss and regularisation function

    loss_function = l2_loss
    # loss_function = lambda x,y: l2_clamped(x, y, 0.0, 0.025)

    def regularisation(data, texture_res):
        # return 0.0
        return 3.5e-11 * total_variation(data, texture_res, 3)
        # return 3.5e-11 * laplacian(data, texture_res, 3)
        # return 1e-9 * ek.hsum(ek.hsum(data))


    # opt_constr = lambda p: Adam(p, lr=CONFIGS[config_name].get('learning_rate', 2e-2), clamp_val=1e-9)
    # opt_constr = lambda p: Adam(p, lr=CONFIGS[config_name].get('learning_rate', 2e-2), clamp_val=6e-9)
    opt_constr = lambda p: Adam(p, lr=CONFIGS[config_name].get('learning_rate', 2e-2), clamp_val=3e-9)
    # opt_constr = lambda p: Adam(p, lr=CONFIGS[config_name].get('learning_rate', 2e-2), beta_1=0.05, beta_2=0.04)
    # opt_constr = lambda p: SGD(p, lr=CONFIGS[config_name].get('learning_rate', 2e-2), momentum=0.5)

    #----------------------------------------------
    #   Calculate approx lr scaling factor

    # approx_lr_scale = calculate_approx_lr_scale(config_name, loss_function)
    approx_lr_scale = 0.05

    #----------------------------------------------
    #   Run optimization

    # AD
    gen_teaser_AD(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
                  fwrd_spp_mult, bckwrd_spp_mult, pass_count)

    # # RB: normal mode
    # gen_teaser_RB(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
    #               'rb', fwrd_spp_mult, bckwrd_spp_mult, False, 1.0, False)

    # # RB: biased
    # gen_teaser_RB(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
    #               'rb_biased', fwrd_spp_mult, bckwrd_spp_mult, True, approx_lr_scale, False)



    # # RB: interleaved
    # gen_teaser_RB(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
    #               'rb_interleaved', fwrd_spp_mult, bckwrd_spp_mult, False, 1.0, True)

    # # RB: biased + interleaved
    # gen_teaser_RB(config_name, opt_constr, texture_res, loss_function, regularisation, spp_scheduler,
    #               'rb_biased_interleaved', fwrd_spp_mult, bckwrd_spp_mult, True, approx_lr_scale, True)


    #----------------------------------------------
    #   Compute clean loss

    # methods = ['rb_biased']
    methods = ['rb', 'rb_biased', 'ad']

    zoom = 1.2
    crop = lambda s: s.sensor().set_crop_window([int(zoom*622), int(zoom*350)], [390, 93])
    # crop = lambda s: s.sensor().set_crop_window([int(zoom*311), int(zoom*175)], [195, 46])
    # crop = None

    if False:
        for mname in methods:
            # gen_clean_loss(config_name, loss_function, spp=128, prefix=mname, crop=crop, low_res=True)
            compute_texture_loss(config_name, loss_function, prefix=mname)

    #----------------------------------------------
    #   Render clean images

    if False:
        config = CONFIGS[config_name]
        output_dir = join(OUTPUT_DIR, config_name)
        run_files = {prefix: join(output_dir, prefix + '_metadata.csv') for prefix in methods}
        selected_its, _ = select_equal_time_from_runs(None, run_files)

        spp_ = 512

        for prefix, it_i in selected_its.items():
            render_config(config_name, spp=spp_, prefix=prefix, it_i=int(it_i), crop=crop)


    #----------------------------------------------
    #   Render initial and ref crop images

    if False:
        # render_config(config_name, spp=1024, prefix='ref', it_i=-1, crop=crop)
        render_config(config_name, spp=1024, prefix='initial', it_i=-1, crop=crop)


    #----------------------------------------------
    #   DEBUG

    # render_config(config_name, spp=4, prefix='ref', render_mode='ad')
    # render_config(config_name, spp=4, prefix='ref', render_mode='rb')
    # render_config(config_name, spp=4, prefix='ref', render_mode='ad_cpu')
