#include <iostream>
#include <fstream>
#include <enoki/cuda.h>
#include <enoki/autodiff.h>

#include <optix.h>
#include <optix_cuda_interop.h>
#include "optix_ptx.h"

using namespace enoki;
using FloatC = CUDAArray<float>;
using FloatD = DiffArray<FloatC>;
using UInt64C = replace_scalar_t<FloatC, uint64_t>;

#define rt_check(err)  __rt_check (context, err, __FILE__, __LINE__)

static void __rt_check(RTcontext context, RTresult errval, const char *file,
                       const int line) {
    if (errval != RT_SUCCESS) {
        const char *message;
        rtContextGetErrorString(context, errval, &message);
        fprintf(stderr,
                "rt_check(): OptiX API error = %04d (%s) in "
                "%s:%i.\n", (int) errval, message, file, line);
        exit(EXIT_FAILURE);
    }
}

class MyClass {
public:
    MyClass() : m_factor(1.f) {
        set_label(m_factor, "m_factor");
        set_requires_gradient(m_factor);
    }

    FloatD test_computation(const FloatD &x) const {
        // return m_factor * enoki::sin(x);
        // return m_factor * (x * x);
        return 2.f * m_factor * (x * x);
    }

    void set_globals(RTcontext &context) const {
        // Get current value from buffer
        // TODO: it would be much nicer to avoid this round-trip
        float factor_value;
        cuda_memcpy_from_device(&factor_value, m_factor.data(), 1 * sizeof(float));
        cuda_eval();
        std::cout << "factor_value = " << factor_value << std::endl;

        RTvariable var;
        rt_check(rtContextQueryVariable(context, "m_factor", &var));
        if (var == nullptr)
            rt_check(rtContextDeclareVariable(context, "m_factor", &var));
        rtVariableSet1f(var, factor_value);
    }

    FloatD m_factor;
};


/// Extraction of the reverse pass of the `test_computation` function.
std::string extract_autodiff_function(const MyClass &object) {
    cuda_create_ptx_module_context();

    /* test_computation */ {
        auto input = zero<FloatD>(1);
        set_label(input, "input");

        cuda_start_recording_ptx_function("test_computation");
        auto out = object.test_computation(input);
        set_label(out, "out");

        cuda_set_inputs(detach(input));
        cuda_set_outputs(out);
        cuda_stop_recording_ptx_function();
    }

    /* test_computation_d */ {
        auto input = zero<FloatD>(1);
        // Gradient computed up until the result of the computation we're
        // extracting now, i.e. gradient of the gradient of that computation.
        auto grad_of_output = zero<FloatC>(1);
        auto out_grad = zero<FloatC>(1);

        cuda_start_recording_ptx_function("test_computation_d");
        auto out = object.test_computation(input);
        set_gradient(out, grad_of_output);

        // Propagate `grad_of_output` to `input` through the computation.
        backward<FloatD>(true);

        auto grad_of_param = gradient(object.m_factor);
        // Atomic accumulation to an output buffer
        auto index = arange<UInt64C>(1);
        scatter_add(out_grad, grad_of_param, index);

        set_label(input, "input");
        set_label(grad_of_output, "grad_of_output");
        set_label(out, "out");
        set_label(grad_of_param, "grad_of_param");
        set_label(index, "index");

        set_label(out_grad, "out_grad");
        auto var_id = cuda_var_register_ptr(out_grad.data());
        cuda_var_set_label(var_id, "out_grad.data");

        cuda_set_inputs(detach(input), grad_of_output);
        cuda_set_outputs(out);
        cuda_stop_recording_ptx_function();
    }

    char* ptx_c = cuda_get_ptx_module();
    if (ptx_c == nullptr)
        throw std::runtime_error("Could not extract function");
    std::string ptx = ptx_c;

    free(ptx_c);
    cuda_destroy_ptx_module_context();

    return ptx;
}

void print_ptx(const std::string &ptx) {
    std::cout << "////////////////////////////*/"    << std::endl;
    std::cout << "///     ____ _______  __    */"    << std::endl;
    std::cout << "///    |  _ \\_   _\\ \\/ /    */" << std::endl;
    std::cout << "///    | |_) || |  \\  /     */"   << std::endl;
    std::cout << "///    |  __/ | |  /  \\     */"   << std::endl;
    std::cout << "///    |_|    |_| /_/\\_\\    */"  << std::endl;
    std::cout << "///                         */"    << std::endl;
    std::cout << "////////////////////////////*/"    << std::endl << std::endl;
    std::cout << ptx << std::endl;
}

void check_graph() {
    FloatD::set_graph_simplification_(0);
    cuda_eval();

    MyClass object;

    // FloatD x = enoki::full<FloatD>(1.f, 100);
    FloatD x = enoki::arange<FloatD>(100) / 100.f;
    FloatD values = object.test_computation(x);
    FloatD loss = 0.5f * enoki::hsum(enoki::sqr(values));
    enoki::set_label(x,  "x");
    enoki::set_label(values,  "values");
    enoki::set_label(loss,  "loss");
    enoki::set_label(object.m_factor, "m_factor");
    set_requires_gradient(values);

    auto fname = "/tmp/test_computation.dot";
    std::ofstream f(fname);
    f << enoki::graphviz(loss);
    std::cout << std::endl << "Graph written to: " << fname << std::endl;

    // Backpropagation (several ways to do it, should all be equivalent)
    // enoki::set_gradient(loss, 1.f);
    // enoki::set_gradient(values, detach(values));
    // enoki::backward<FloatD>(true);
    enoki::backward(loss);
    std::cout << "Loss value: " << loss << std::endl;
    std::cout << "Gradient of factor: " << enoki::gradient(object.m_factor) << std::endl;
    cuda_eval();
}

void run_kernel(MyClass &object, const std::string &extra_ptx) {
    constexpr int kDeviceID = 0;
    const size_t ray_count = 100;
    const size_t output_size = ray_count;

    RTcontext context = nullptr;
    rt_check(rtContextCreate(&context));
    rt_check(rtContextSetRayTypeCount(context, 1));
    rt_check(rtContextSetEntryPointCount(context, 1));
    rt_check(rtContextSetStackSize(context, 0));
    rt_check(rtContextSetMaxTraceDepth(context, 15));
    rt_check(rtContextSetMaxCallableProgramDepth(context, 0));

    float *output_buffer = nullptr;
    auto out_grad_data = zero<FloatC>(1);
    /* Create and set output buffers */ {
        // output buffer
        RTvariable var;
        RTbuffer buffer_var;
        rt_check(rtBufferCreate(context, RT_BUFFER_INPUT, &buffer_var));
        rt_check(rtContextDeclareVariable(context, "output", &var));
        rt_check(rtVariableSetObject(var, buffer_var));
        // Standard float buffers, storage will be in pinned memory
        rt_check(rtBufferSetFormat(buffer_var, RT_FORMAT_FLOAT));

        void *cuda_ptr = enoki::cuda_host_malloc(output_size * sizeof(float));
        for (size_t i = 0; i < output_size; ++i)
            ((float *) cuda_ptr)[i] = 0.f;
        rt_check(rtBufferSetSize1D(buffer_var, output_size));
        rt_check(rtBufferSetDevicePointer(buffer_var, kDeviceID, cuda_ptr));
        output_buffer = (float *) cuda_ptr;

        // Globals for the extracted function
        object.set_globals(context);

        // Gradient output (pointer to a memory region)
        rt_check(rtContextDeclareVariable(context, "out_grad_data", &var));
        const auto *ptr_value = out_grad_data.data();
        rt_check(rtVariableSetUserData(var, sizeof(ptr_value), &ptr_value));
    }

    /* Create and bind programs */ {
        std::vector<const char *> extended_sources {
            (const char *) test_autodiff_extraction_ptx,
            extra_ptx.c_str()
        };

        RTprogram program;
        rt_check(rtProgramCreateFromPTXStrings(context, extended_sources.size(),
                                               extended_sources.data(), "ray_gen",
                                               &program));
        rt_check(rtContextSetRayGenerationProgram(context, 0, program));
    }

    /* Build and prepare context */ {
        enoki::cuda_eval();
        rt_check(rtContextValidate(context));
        // Zero-size launch is required
        rt_check(rtContextLaunch1D(context, 0, 0));
        enoki::cuda_eval();
    }

    /* Actual run */ {
        rt_check(rtContextLaunch1D(context, 0, ray_count));
        enoki::cuda_eval();
    }

    // Extract results
    std::cout << "Survived! Output buffer:" << std::endl;
    double loss = 0.;
    for (size_t i = 0; i < output_size; ++i) {
        std::cout << output_buffer[i] << ", ";
        loss += 0.5 * enoki::sqr((double) output_buffer[i]);
    }
    std::cout << std::endl;
    std::cout << "Loss: " << loss << std::endl;
    std::cout << "Gradient output buffer: " << out_grad_data[0] << std::endl;

}


int main() {
    MyClass object;
    auto ptx = extract_autodiff_function(object);
    print_ptx(ptx);
    run_kernel(object, ptx);

    check_graph();
    std::cout << "Test complete." << std::endl;
}
