from collections import namedtuple
from os import path
from os.path import join, dirname, realpath
import csv
import os
import pytest
import re
import sys
import time

import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
from mitsuba.core.util import time_string, mem_string
from mitsuba.optix import OptiXRenderer
from mitsuba.render.autodiff import get_differentiable_parameters, Adam
import numpy as np

from autodiff_utils import image_to_float_d

from render_helpers import load_scene, l2_loss, reference_differentiable_render, compute_image_gradients, integrator_type_for_scene, ensure_valid_values

from config import CONFIGS, OUTPUT_DIR, TEXTURED_PARAM_NAMES
from benchmark_utils import GPUUsageSampler, parameters_expand_wildcard
from test_types import KEY_FIELDS, RESULT_FIELDS, FIELD_TYPES, ResultKey, Result, make_key, key_to_string, scene_params_for_key

# Whether we should save intermediate images
SAVE_INTERMEDIATE_RESULTS = False

SELECTED_CONFIGS = [
    'cbox-rgb',
    'cbox-rgb-approx-1',
    'cbox-microfacet',
    'cbox-textured',
    'rb-homogeneous-noise-opt',
    'rb-heterogeneous-noise-opt',
    'living-room-01',
    'albedo-opt-null',
]
RESOLUTIONS = [(256, 256)]
# RESOLUTIONS = [(128, 128), (256, 256), (512, 512)]
# SPP         = [1, 2, 3, 4]
# SPP         = [1, 2, 4, 8, 16, 32]
SPP = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]
# MAX_DEPTH   = [2, 3]
MAX_DEPTH = [2, 3, 5]

# DEFAULT_ITERATION_COUNT = 5
DEFAULT_ITERATION_COUNT = 10

KEY_FIELDS = ['scene_name', 'resolution_x', 'resolution_y', 'spp', 'max_depth']
RESULT_FIELDS = ['loss', 'time_per_it', 'total_time', 'iteration_count', 'memory_usage', 'gpu_usage']
FIELD_TYPES = {
    'scene_name': str,
    'resolution_x': int,
    'resolution_y': int,
    'spp': int,
    'max_depth': int,
    'loss': float,
    'time_per_it': float,   # In seconds
    'total_time': float,    # In seconds
    'iteration_count': int,
    'memory_usage': float,  # In bytes
    'gpu_usage': float,     # In percentage
}
assert set(FIELD_TYPES.keys()) == set(KEY_FIELDS + RESULT_FIELDS)

ResultKey = namedtuple('ResultKey', KEY_FIELDS)
Result = namedtuple('Result', ['key'] + RESULT_FIELDS)


# Hand-specified pass counts, chosen so that GPU memory is fully used but not exceeded
# Pass counts specified (so far) so that it fits on a GPU with 5611MB of free RAM.
ad_pass_counts = {
    ResultKey('cbox-rgb', 256, 256, 32, 5):           2,
    ResultKey('cbox-rgb', 256, 256, 64, 2):           2,
    ResultKey('cbox-rgb', 256, 256, 64, 3):           2,
    ResultKey('cbox-rgb', 256, 256, 64, 5):           4,
    ResultKey('cbox-rgb', 256, 256, 128, 2):          3,
    ResultKey('cbox-rgb', 256, 256, 128, 3):          4,
    ResultKey('cbox-rgb', 256, 256, 128, 5):          8,
    ResultKey('cbox-rgb', 256, 256, 256, 2):          5,
    ResultKey('cbox-rgb', 256, 256, 256, 3):          8,
    ResultKey('cbox-rgb', 256, 256, 256, 5):         16,
    ResultKey('cbox-rgb', 256, 256, 512, 2):         10,
    ResultKey('cbox-rgb', 256, 256, 512, 3):         16,
    ResultKey('cbox-rgb', 256, 256, 512, 5):         31,
    ResultKey('cbox-rgb-approx-1', 256, 256, 32, 5):  2,
    ResultKey('cbox-rgb-approx-1', 256, 256, 64, 5):  4,
    ResultKey('cbox-rgb-approx-1', 256, 256, 128, 5): 8,
    ResultKey('cbox-rgb-approx-1', 256, 256, 256, 5):16,
    ResultKey('cbox-rgb-approx-1', 256, 256, 512, 5):31,

    ResultKey('cbox-microfacet', 256, 256, 16, 3):    2,
    ResultKey('cbox-microfacet', 256, 256, 16, 5):    2,
    ResultKey('cbox-microfacet', 256, 256, 32, 3):    2,
    ResultKey('cbox-microfacet', 256, 256, 32, 5):    3,
    ResultKey('cbox-microfacet', 256, 256, 64, 2):    2,
    ResultKey('cbox-microfacet', 256, 256, 64, 3):    4,
    ResultKey('cbox-microfacet', 256, 256, 64, 5):    7,
    ResultKey('cbox-microfacet', 256, 256, 128, 2):   4,
    ResultKey('cbox-microfacet', 256, 256, 128, 3):   8,
    ResultKey('cbox-microfacet', 256, 256, 128, 5):  14,
    ResultKey('cbox-microfacet', 256, 256, 256, 2):   8,
    ResultKey('cbox-microfacet', 256, 256, 256, 3):  16,
    ResultKey('cbox-microfacet', 256, 256, 256, 5):  24,
    ResultKey('cbox-microfacet', 256, 256, 512, 2):  16,
    ResultKey('cbox-microfacet', 256, 256, 512, 3):  31,
    ResultKey('cbox-microfacet', 256, 256, 512, 5):  42,

    ResultKey('cbox-textured', 256, 256, 64, 2):      2,
    ResultKey('cbox-textured', 256, 256, 128, 2):     2,
    ResultKey('cbox-textured', 256, 256, 128, 3):     3,
    ResultKey('cbox-textured', 256, 256, 128, 5):     4,

    ResultKey('living-room-01', 512, 256, 32, 12):    1,
    ResultKey('living-room-01', 1280, 720, 32, 12):   6,

    ResultKey('bell', 640, 360,  32, 24):   2,
    ResultKey('bell', 640, 360,  64, 24):   4,
    ResultKey('bell', 640, 360,  128, 24):  8,
    ResultKey('bell', 640, 360,  256, 24):  16,
    ResultKey('bell', 640, 360,  512, 24):  32, #

    ResultKey('albedo-opt-null', 256, 256,  16, 64):   2,
    ResultKey('albedo-opt-null', 256, 256,  32, 64):   4,
    ResultKey('albedo-opt-null', 256, 256,  64, 64):   8,
    ResultKey('albedo-opt-null', 256, 256, 128, 64):  16,
    ResultKey('albedo-opt-null', 256, 256, 256, 64):  29,
    ResultKey('albedo-opt-null', 256, 256, 512, 64):  58,

    ResultKey('smoke-plume-res-111-white', 128, 128,   8, 8):   2,
    ResultKey('smoke-plume-res-111-white', 128, 128,  16, 8):   4,
    ResultKey('smoke-plume-res-111-white', 128, 128,  32, 8):   8,
    ResultKey('smoke-plume-res-111-white', 128, 128,  64, 8):   15,
    ResultKey('smoke-plume-res-111-white', 128, 128, 128, 8):   29, # A bit large, but didn't divide well
    ResultKey('smoke-plume-res-111-white', 128, 128, 256, 8):   58,
    ResultKey('smoke-plume-res-111-white', 128, 128, 512, 8):   120,
}

# Specific configurations to run, on top of the cross product of resolution, spp, etc
EXTRA_COMBINATIONS = [
    ResultKey('cbox-rgb', 256, 256, None, 5),
    # ResultKey('bell', 1280, 720, None, 24),
    ResultKey('bell', 640, 360, None, 24),
    ResultKey('albedo-opt-null', 256, 256, None, 64),

    # Note: need to go down to 128 for AD, with 256 not even a single spp fits in memory
    # ResultKey('smoke-plume-res-111-white', 256, 256, None, 8),
    ResultKey('smoke-plume-res-111-white', 128, 128, None, 8),
]
ONLY_EXTRAS = True
# Specific combinations that should not be run
COMBINATIONS_BLACKLIST = [
    ResultKey('cbox-rgb-approx-1', 256, 256, None, 2),
    ResultKey('cbox-rgb-approx-1', 256, 256, None, 3),
]

def optimize_scene_RB_with_approx_l1(key, scene, params, reference_b, config, iteration_count):
    return optimize_scene_RB(key, scene, params, reference_b, config, iteration_count,
                             use_approx_li_1=True)

def optimize_scene_RB_with_interleaving(key, scene, params, reference_b, config, iteration_count):
    return optimize_scene_RB(key, scene, params, reference_b, config, iteration_count,
                             use_interleaving=True)

def optimize_scene_RB_with_interleaving_and_approx_l1(key, scene, params, reference_b,
                                                      config, iteration_count):
    return optimize_scene_RB(key, scene, params, reference_b, config, iteration_count,
                             use_approx_li_1=True, use_interleaving=True)

def optimize_scene_RB(key, scene, params, reference_b, config, iteration_count,
                      use_approx_li_1=False, use_interleaving=False):
    loss_function = l2_loss
    opt = Adam(params, lr=config.get('learning_rate', 2e-2))
    integrator_tp = integrator_type_for_scene(scene, with_gradients=True, config=config)
    renderer = OptiXRenderer(scene, integrator_tp, params=params)

    if 'use_approx_li_1' in config:
        raise ValueError('Error: approx L1 is forced in this benchmarking script, config value'
                         ' is not honored.')

    if use_approx_li_1:
        print('[i] Using approximate Li = 1')
        renderer.set_use_approx_li_1(True)
    if use_interleaving:
        print('[i] Using interleaving (intentional off-by-one)')

    cache = None
    metadata = { k: [] for k in ['timing', 'timings_forward', 'timings_backward', 'loss'] }

    if use_interleaving:
        # Render current state for initial image gradients
        current_b = renderer.render(0, spp=key.spp)

    for it_i in range(iteration_count):
        t0 = time.time()

        # Render current state (forward pass)
        seed_forward = it_i
        if not use_interleaving:
            current_b = renderer.render(seed_forward, spp=key.spp)
        t_end_forward = time.time()

        # Compute image gradients
        image_gradients, loss, cache = compute_image_gradients(
            reference_b, current_b, loss_function, cache=cache)

        seed_backward = iteration_count + seed_forward + 1
        img_or_none = renderer.delta_render(seed_backward, image_gradients, spp=key.spp,
                                            return_image=use_interleaving)
        if use_interleaving:
            current_b = img_or_none

        # Apply gradient step
        opt.step()

        ensure_valid_values(params)

        # Re-upload globals to the correct OptiX constants
        renderer.upload_bsdf_globals()

        t_end_backward = time.time()
        elapsed = t_end_backward - t0
        loss_v = loss.numpy()[0]
        metadata['loss'].append(loss_v)
        metadata['timing'].append(elapsed)
        metadata['timings_forward'].append(t_end_forward - t0)
        metadata['timings_backward'].append(t_end_backward - t_end_forward)
        print("[{:03d}]: loss = {:.6f}, took {:.2f}s".format(
              it_i, loss_v, elapsed), end='\r')

    del renderer, opt
    return metadata['loss'][-1], current_b, metadata


def optimize_scene_AD(key, scene, params, reference_b, config, iteration_count):
    differentiable_rendering_dir = realpath(join(dirname(__file__), '../../../../differentiable-rendering'))
    sys.path.append(join(differentiable_rendering_dir, 'python/optimize'))
    from differentiable import optimize_scene
    from losses import l2 as l2_loss_from_dr

    learning_rate = config.get('learning_rate', 2e-2)
    output_dir = join(OUTPUT_DIR, 'benchmark-ad-scratch')
    os.makedirs(output_dir, exist_ok=True)

    param_names = [k for k, _ in params.items()]
    make_opt = (lambda params: Adam(params, lr=learning_rate))
    save_intermediate = config.get('save_intermediate_results', SAVE_INTERMEDIATE_RESULTS)
    pass_count = ad_pass_counts.get(key, 1)
    block_size = config.get('block_size', None)
    spp_per_pass = key.spp // pass_count
    # TODO: handle non-perfect case
    assert np.abs(spp_per_pass * pass_count - key.spp) < 0.1 * key.spp, \
           '{} * {} != {}'.format(spp_per_pass, pass_count, key.spp)

    _, loss, images_np, metadata = optimize_scene(
        scene,
        param_names,
        image_to_float_d(reference_b, Bitmap.ERGB),
        optimizer=make_opt,
        max_iterations=iteration_count,
        loss_function=l2_loss_from_dr,
        pass_count=pass_count,
        block_size=block_size,
        image_spp=spp_per_pass,
        gradient_spp=spp_per_pass,
        output_dir=output_dir,
        write_image_per_iteration=save_intermediate,
        return_metadata=True,
        )

    return loss, Bitmap(images_np[0]), metadata


def create_or_open_results_file(result_file_path):
    restored_results = {}
    if path.exists(result_file_path) and open(result_file_path, 'r').read().strip():
        result_file = open(result_file_path, 'r+', newline='')
        print('[ ] Recovering cached results...')
        csv_reader = csv.reader(result_file)

        headers = next(csv_reader, None)
        all_fields = KEY_FIELDS + RESULT_FIELDS
        assert headers == all_fields, \
               'Headers do not match:\nexpected {},\n found:   {}'.format(all_fields, headers)

        for row in csv_reader:
            assert len(row) == len(all_fields), \
                   'Expected fields {}, found: {} in file {}'.format(row, all_fields, result_file_path)
            key = ResultKey(**{ h: FIELD_TYPES[h](row[i])
                                for i, h in enumerate(KEY_FIELDS) })
            values = Result(key=key, **{ h: FIELD_TYPES[h](row[len(KEY_FIELDS) + i])
                                         for i, h in enumerate(RESULT_FIELDS) })
            restored_results[key] = values
        print('[+] Recovered {} cached results, resuming.'.format(len(restored_results)))
        csv_writer = csv.writer(result_file)
    else:
        # Create file
        result_file = open(result_file_path, 'w+', newline='')
        csv_writer = csv.writer(result_file)
        # Add header line
        assert Result._fields[0] == 'key'
        field_names = KEY_FIELDS + list(Result._fields)[1:]
        csv_writer.writerow(field_names)
        result_file.flush()

    return result_file, csv_writer, restored_results


def add_result_to_file(csv_writer, result):
    csv_writer.writerow(list(result.key) + list(result)[1:])


def write_metadata(method, key, config, result, metadata, output_dir):
    fname = join(output_dir, key_to_string(method, key) + '.csv')
    columns = sorted(list(metadata.keys()))
    rows = list(zip(*[metadata[k] for k in columns]))
    with open(fname, 'w') as f:
        csv_writer = csv.writer(f)
        # Headers
        csv_writer.writerow(['iteration'] + columns)
        # Lines
        for i, row in enumerate(rows):
            csv_writer.writerow([i] + list(row))



def load_or_render_ref(directory, method_name, key, config):
    ref_fname = realpath(join(directory, key_to_string(method_name, key, exclude_spp=True) + '_ref.exr'))
    if os.path.isfile(ref_fname):
        reference_b = Bitmap(ref_fname)
        assert reference_b.pixel_format() == Bitmap.ERGB
        assert (reference_b.width() == key.resolution_x
                and reference_b.height() == key.resolution_y), \
               '{}: {}x{}  vs found {}'.format(ref_fname, key.resolution_x,
                                               key.resolution_y, reference_b)
        print('[+] Loaded reference image from: {}'.format(ref_fname))
    else:
        # Note: we assume there are corresponding variables defined in the scene
        scene_fname = config.get('ref_scene', config['scene'])
        load_scene_params = scene_params_for_key(key, config, is_reference=True)
        scene = load_scene(scene_fname, parameters=load_scene_params)

        ref_spp = 16384 if config.get('has_volumes', False) else 4096
        ref_spp = config.get('ref_spp', ref_spp)
        print('[ ] Rendering reference image with OptiX on the GPU: '
              '{} - {}x{} @ {}spp, max_depth {}...'.format(
              key.scene_name, key.resolution_x, key.resolution_y, ref_spp, key.max_depth))
        integrator_tp = integrator_type_for_scene(scene, with_gradients=False, config=config)
        renderer = OptiXRenderer(scene, integrator_tp)
        reference_b = renderer.render(seed=1234, spp=ref_spp)

        reference_b.write(ref_fname)
        print('[+] Saved reference image: {}'.format(ref_fname))

    a = np.array(reference_b, copy=False)
    nan_entries = np.isnan(a)
    if np.any(nan_entries):
        print('[!] NaN entries found in reference: {}! Replacing with zeros.'.format(np.sum(nan_entries)))
        a[nan_entries] = 0.
    return reference_b


def generate_parameter_combinations():
    extra_keys = sorted(list(
        parameters_expand_wildcard(EXTRA_COMBINATIONS)), key=tuple)
    if ONLY_EXTRAS:
        return extra_keys

    blacklist = parameters_expand_wildcard(COMBINATIONS_BLACKLIST)
    keys = set()
    for scene_name in SELECTED_CONFIGS:
        for resolution in RESOLUTIONS:
            for spp in SPP:
                for max_depth in MAX_DEPTH:
                    key = make_key(scene_name=scene_name,
                                   resolution_x=resolution[0], resolution_y=resolution[1],
                                   spp=spp, max_depth=max_depth)
                    if key in blacklist:
                        continue
                    keys.add(key)
    # Make sure that extra keys will be run first
    return extra_keys + sorted(list(keys), key=tuple)


def benchmark_method(method_name, optimize_method):
    from enoki import cuda_mem_get_info

    if SAVE_INTERMEDIATE_RESULTS:
        raise RuntimeError('Not implemented yet: saving intermediate results')

    output_dir = realpath(join(OUTPUT_DIR, 'benchmark_' + method_name))
    os.makedirs(output_dir, exist_ok=True)

    # Figure out where we left last time
    results_file_path = join(output_dir, method_name.lower() + '_results.csv')
    result_file, csv_writer, results_dict = create_or_open_results_file(results_file_path)

    all_keys = generate_parameter_combinations()
    for current_key in all_keys:
        scene_name = current_key.scene_name
        config = CONFIGS[scene_name]
        resolution = (current_key.resolution_x, current_key.resolution_y)
        spp = current_key.spp
        max_depth = current_key.max_depth
        # iteration_count = config.get('max_iterations', DEFAULT_ITERATION_COUNT)
        iteration_count = DEFAULT_ITERATION_COUNT

        # Skip if this combination of values has already been computed
        if current_key in results_dict:
            print('[-] Skipping: {}, {}x{}, {}spp, max_depth {}'.format(
                  scene_name, resolution[0], resolution[1], spp, max_depth))
            continue

        # Load or render reference image
        reference_b = load_or_render_ref(output_dir, method_name, current_key, config)
        print('[ ] Profiling scene "{}" with resolution {}x{} @ {}spp, max_depth {}.'.format(
              scene_name, resolution[0], resolution[1], spp, max_depth))

        scene = load_scene(config['scene'],
                           scene_params_for_key(current_key, config, is_reference=False))
        # Check parameters have been applied (scene loading is a bit brittle)
        assert np.all(scene.film().size() == resolution), str(scene.film())
        assert scene.integrator().max_depth() == max_depth
        # Extract differentiable parameters of interest
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])

        sampler = GPUUsageSampler()
        sampler.start()
        t_start = time.time()
        try:
            loss, final_b, metadata = optimize_method(current_key, scene, params, reference_b,
                                                      config, iteration_count)
        except Exception as e:
            raise e
        finally:
            sampler.done()  # Need to join the thread, otherwise it will hang
        elapsed = time.time() - t_start

        # TODO: compute gpu usage
        # TODO: need to get memory usage before and during (probably in another thread)
        free, total = cuda_mem_get_info()

        final_image_fname = join(output_dir,
                                 key_to_string(method_name, current_key) + '_final.exr')
        final_b.write(final_image_fname)

        # TODO: consider saving all per-iteration metadata to a separate file as well
        time_per_it = np.median(metadata['timing'])
        gpu_usage = sampler.max_usage()
        result = Result(key=current_key, loss=loss, time_per_it=time_per_it,
                        total_time=elapsed, iteration_count=iteration_count,
                        memory_usage=total-free, gpu_usage=gpu_usage)
        print('[+] Optimized scene {} ({}x{} @ {}spp, max_depth {}): took {} and {}'.format(
            scene_name, resolution[0], resolution[1], spp, max_depth,
            time_string(1000 * result.total_time), mem_string(result.memory_usage)))

        # Add result to file
        write_metadata(method_name, current_key, config, result, metadata, output_dir)
        add_result_to_file(csv_writer, result)
        result_file.flush()

        # Cleanup
        # scene.optix_release()
        # del scene, final_b

    result_file.close()
    print('[+] Benchmarking complete.')

if __name__ == '__main__':
    logger = mitsuba.core.Thread.thread().logger()
    logger.set_log_level(mitsuba.core.EWarn)
    # logger.set_log_level(mitsuba.core.EDebug)

    benchmark_method('rb', optimize_scene_RB)
    benchmark_method('rb_biased', optimize_scene_RB_with_approx_l1)
    benchmark_method('rb_interleaved', optimize_scene_RB_with_interleaving)
    benchmark_method('rb_biased_interleaved', optimize_scene_RB_with_interleaving_and_approx_l1)
    benchmark_method('ad', optimize_scene_AD)
