import os
from os.path import join, dirname, realpath
import pytest
import time

import enoki as ek
import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
from mitsuba.optix import OptiXRenderer
from mitsuba.render.autodiff import get_differentiable_parameters, Adam
import numpy as np

from render_helpers import load_scene, l2_loss, save_info_csv, integrator_type_for_scene, \
                           reference_differentiable_render, compute_image_gradients, ensure_valid_values
from config import CONFIGS, OUTPUT_DIR


def test01_plain_render():
    config = CONFIGS['cbox-rgb']
    scene = load_scene(config['scene'])
    renderer = OptiXRenderer(scene, OptiXRenderer.EPath)
    b = renderer.render(seed=1234)
    assert np.all(b.size() == scene.film().size())


def test02_differentiable_render():
    config = CONFIGS['cbox-rgb']
    output_dir = join(OUTPUT_DIR, 'cbox-rgb')
    os.makedirs(output_dir, exist_ok=True)
    scene = load_scene(config['scene'])
    params = get_differentiable_parameters(scene)
    params.keep(config['params'])

    print('[ ] Rendering initial state with OptiX on the GPU...')
    renderer = OptiXRenderer(scene, OptiXRenderer.ERadiativeBackprop, params)
    current_b = renderer.render(seed=1234)

    initial_state_fname = join(output_dir, 'cbox_rgb_initial_rb.exr')
    current_b.write(initial_state_fname)
    print('[+] Saved initial state: {}'.format(initial_state_fname))

    # TODO: create reference bitmaps inside the test
    reference_b = Bitmap(config['ref'])

    loss_function = l2_loss
    image_gradients, _, _ = compute_image_gradients(reference_b, current_b, loss_function)

    if True:
        input_gradients_fname = join(output_dir, 'cbox_rgb_input_gradients.exr')
        image_gradients_arr = image_gradients.numpy().reshape(
            reference_b.width(), reference_b.height(), -1)
        Bitmap(image_gradients_arr).write(input_gradients_fname)
        print('[+] Saved input image gradients for RB to: {}'.format(input_gradients_fname))
        image_gradients_arr = image_gradients_arr[np.any(image_gradients_arr != 0., axis=-1), ...]
        print('[ ] Average non-zero input gradient: {}'.format(
            np.mean(image_gradients_arr, axis=0)))


    print('[ ] Rendering with radiative backprop on the GPU...')
    renderer.delta_render(12345, image_gradients)
    print('[+] Done!')
    # Extract gradients
    gradients = { k: ek.gradient(v) for k, v in params.items() }

    # Compare against autodiff-based differentiable rendering
    scene = load_scene(config['scene'])
    params = get_differentiable_parameters(scene)
    params.keep(config['params'])
    opt = Adam(params, lr=1e-2)

    print('[ ] Rendering with autodiff on the GPU...')
    _, ref_gradients = reference_differentiable_render(scene, reference_b, opt, loss_function, spp=4)
    print('[+] Done!')
    assert set(ref_gradients.keys()) == set(gradients.keys())

    for k, v in gradients.items():
        print('- {}:\n   {}\n   vs\n   {}\n'.format(k, v, ref_gradients[k]))


@pytest.mark.skip('Bindings not available for Enoki PTX extraction')
def test03_enoki_cuda_get_ptx():
    ek.cuda_eval()
    a = ek.FloatC([1, 2.3, 3, 4., 5.])
    b = ek.sqr(a)
    c = -b
    ptx = ek.cuda_get_ptx("custom_ptx")
    print(ptx)
    assert ".visible .func custom_ptx" in ptx


@pytest.mark.parametrize('use_interleaving', [False, True])
@pytest.mark.parametrize('config_name',
                         ('cbox-rgb', 'cbox-rgb-approx', 'cbox-textured',
                          'rb-unit-scattering-homogeneous',
                          'rb-unit-scattering-heterogeneous',
                          'rb-homogeneous-noise-opt', 'rb-homogeneous-noise-opt-hidden-emitters',
                          'rb-heterogeneous-noise-opt',
                          'smoke-plume', 'smoke-plume-approx',
                          'living-room-01',
                          ))
def test04_radiative_backprop_optimization(config_name, use_interleaving):
    """Optimization of RGB colors or textures in the cbox scene."""
    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)
    if use_interleaving:
        output_dir += '_interleaved'
    os.makedirs(output_dir, exist_ok=True)

    # Render reference image, if missing
    ref_fname = config['ref']
    if not os.path.isfile(ref_fname):
        print('[ ] Rendering reference image with OptiX on the GPU...')
        ref_scene = config.get('ref_scene', config['scene'])
        scene = load_scene(ref_scene, parameters=config['ref_scene_params'])
        renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
        # ref_b = renderer.render(seed=1234, spp=1024)
        ref_b = renderer.render(seed=1234, spp=8192)
        ref_b.write(ref_fname)
        print('[+] Saved reference image: {}'.format(ref_fname))
    reference_b = Bitmap(ref_fname)
    reference_b.write(join(output_dir, 'ref.exr'))

    scene_params = config.get('scene_params', [])
    scene = load_scene(config['scene'], parameters=scene_params)
    params = get_differentiable_parameters(scene)
    params.keep(config['params'])

    loss_function = l2_loss
    opt = Adam(params, lr=config.get('learning_rate', 2e-2))
    forward_spp = config.get('forward_spp', 32)
    backward_spp = config.get('backward_spp', forward_spp)
    max_iterations = config.get('max_iterations', 150)

    integrator_type = integrator_type_for_scene(scene, with_gradients=True, config=config)

    renderer = OptiXRenderer(scene, integrator_type, params=params)
    if config.get('use_approx_li_1', False):
        print('[i] Using approximate Li = 1')
        renderer.set_use_approx_li_1(True)

    metadata = { k: [] for k in ['timings', 'timings_forward', 'timings_backward', 'loss_values'] }
    cache = None

    if use_interleaving:
        # Render current state for initial image gradients
        current_b = renderer.render(0, spp=forward_spp)
        gpu_image_fname = join(output_dir, 'initial.exr')
        current_b.write(gpu_image_fname)

    for it_i in range(max_iterations):
        t0 = time.time()

        # Render current state (forward pass)
        seed_forward = it_i
        if not use_interleaving:
            current_b = renderer.render(seed_forward, spp=forward_spp)
            gpu_image_fname = join(output_dir, 'out_{:04d}.exr'.format(it_i))
            current_b.write(gpu_image_fname)
        t_end_forward = time.time()

        # Compute image gradients
        image_gradients, loss, cache = compute_image_gradients(
            reference_b, current_b, loss_function, cache=cache)

        seed_backward = max_iterations + seed_forward + 1
        img_or_none = renderer.delta_render(seed_backward, image_gradients,
                                            spp=backward_spp, return_image=use_interleaving)

        if use_interleaving:
            # To save time, will use image rendered with current parameter to
            # compute loss at the next step, even though it means we won't
            # have the update values in there.
            current_b = img_or_none
            gpu_image_fname = join(output_dir, 'out_{:04d}.exr'.format(it_i))
            current_b.write(gpu_image_fname)

        # for k, v in params.items():
        #     print('{}: value abs sum = {}, gradient abs sum = {}'.format(
        #         k, ek.hsum(ek.abs(v)), ek.hsum(ek.abs(ek.gradient(v)))))

        # Apply gradient step
        opt.step()

        # Ensure valid gradients
        ensure_valid_values(params)

        # Re-upload globals to the correct OptiX constants
        # TODO: this shouldn't be needed
        renderer.upload_bsdf_globals()

        t_end_backward = time.time()
        metadata['timings'].append(t_end_backward - t0)
        metadata['timings_forward'].append(t_end_forward - t0)
        metadata['timings_backward'].append(t_end_backward - t_end_forward)
        metadata['loss_values'].append(loss.numpy()[0])

        print("[{:03d}]: loss = {:.6f}, took {:.2f}s (forward: {:.4f}s, backward: {:.4f}s)".format(
            it_i, metadata['loss_values'][-1], metadata['timings'][-1],
            metadata['timings_forward'][-1], metadata['timings_backward'][-1]), end='\n')

    # Save metadata about the run
    fname = join(output_dir, 'metadata.csv')
    metadata['iteration'] = list(range(max_iterations))
    save_info_csv(fname, **metadata)
    print('[+] Saved run metadata to: {}'.format(fname))



@pytest.mark.parametrize('config_name',
                         ('cbox-rgb', 'rb-unit-scattering-homogeneous'))
def test05_return_image(config_name):
    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)
    os.makedirs(output_dir, exist_ok=True)
    nnz_expected = 6 if 'cbox' in 'config_name' else 1
    # Volumes are pretty noisy, it's recommended to check visually as well
    spp = 32 if 'cbox' in 'config_name' else 512
    tolerance = 1e-4 if 'cbox' in 'config_name' else 2e-1

    def make_renderer(with_gradients):
        scene = load_scene(config['scene'])
        int_tp = integrator_type_for_scene(scene, with_gradients=with_gradients)
        if with_gradients:
            params = get_differentiable_parameters(scene)
            params.keep(config['params'])
            return OptiXRenderer(scene, int_tp, params), params
        else:
            return OptiXRenderer(scene, int_tp)


    print('[ ] Rendering with OptiX path tracer for reference...')
    renderer = make_renderer(False)
    normal_b = renderer.render(seed=1234, spp=spp)
    fname = join(output_dir, 'rb_path.exr')
    normal_b.write(fname)
    print('[+] Saved OptiX path tracer render: {}'.format(fname))

    print('[ ] Rendering with RB integrator, forward mode...')
    renderer, params = make_renderer(True)
    forward_b = renderer.render(seed=1234, spp=spp)
    f_arr = np.array(forward_b, copy=False)
    fname = join(output_dir, 'rb_forward.exr')
    forward_b.write(fname)
    print('[+] Saved forward mode render: {}'.format(fname))

    # Fake gradients image
    image_gradients = Bitmap(np.ones(f_arr.shape, dtype=np.float32))

    print('[ ] Rendering with RB integrator, backward mode without image...')
    probably_none = renderer.delta_render(1234, image_gradients, spp=spp)
    assert probably_none is None
    gradients_without_image = {}
    nnz = 0
    for k, v in params.items():
        gradients_without_image[k] = ek.gradient(v).numpy()
        if np.any(np.abs(gradients_without_image[k]) > 0.):
            nnz += 1
        ek.set_gradient(params[k], type(ek.detach(v)).zero(ek.slices(v)))
    assert nnz >= nnz_expected
    print('[+] Done, gradients were produced but no image')


    print('[ ] Rendering with RB integrator, backward mode with image...')
    renderer, params = make_renderer(True)
    backward_b = renderer.delta_render(1234, image_gradients, spp=spp, return_image=True)
    b_arr = np.array(backward_b, copy=False)
    assert isinstance(backward_b, Bitmap)
    fname = join(output_dir, 'rb_backward.exr')
    backward_b.write(fname)
    print('[+] Saved forward mode render: {}'.format(fname))
    # Should match up to noise, both forward and the gradients
    assert np.allclose(f_arr, b_arr, atol=tolerance)
    nnz = 0
    for k, v in params.items():
        grad = ek.gradient(v).numpy()
        if np.any(np.abs(grad) > 0.):
            nnz += 1
        assert np.allclose(gradients_without_image[k], grad, rtol=1e-4)
    assert nnz >= nnz_expected
    print('[+] Done, gradients *and* image were produced')



@pytest.mark.parametrize('config_name',
                         ('cbox-rgb', 'rb-unit-scattering-homogeneous'))
def test06_rb_approx_learning_rate(config_name):
    """Try and determine a scaling factor from the normal RB gradients and RB gradients
    with approximate Li = 1 term."""
    import pandas as pd
    from test_rb_gradients import gradients_to_df

    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, 'rb_gradients_scaling_' + config_name)
    os.makedirs(output_dir, exist_ok=True)

    # Reference image and scene loading
    reference_b = Bitmap(config['ref'])
    reference_b.write(join(output_dir, 'ref.exr'))

    forward_spp = config.get('forward_spp', 32)
    backward_spp = config.get('backward_spp', forward_spp)
    max_iterations = 20
    loss_function = l2_loss
    def prepare_integrator(allow_approx=True):
        scene_params = config.get('scene_params', [])
        scene = load_scene(config['scene'], parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])

        # Optimization setup
        opt = Adam(params, lr=config.get('learning_rate', 2e-2))
        integrator_type = integrator_type_for_scene(scene, with_gradients=True, config=config)
        renderer = OptiXRenderer(scene, integrator_type, params=params)
        if use_approx and allow_approx:
            print('[i] Using approximate Li = 1')
            renderer.set_use_approx_li_1(True)
        return renderer, params, opt


    df = pd.DataFrame(columns=['method', 'use_approx', 'iteration', 'loss',
                               'param_name', 'param_gradient'])
    series = pd.Series(index=df.columns)

    for use_approx in [False, True]:
        series['method'] = 'rb_approx' if use_approx else 'rb'
        series['use_approx'] = use_approx
        series['config_name'] = config_name
        renderer, params, opt = prepare_integrator()

        cache = None
        for it_i in range(max_iterations):
            series['iteration'] = it_i

            # Render current state (forward pass)
            seed_forward = it_i
            current_b = renderer.render(seed_forward, spp=forward_spp)
            gpu_image_fname = join(output_dir, 'out_{:04d}.exr'.format(it_i))
            current_b.write(gpu_image_fname)

            # Compute image gradients
            image_gradients, loss, cache = compute_image_gradients(
                reference_b, current_b, loss_function, cache=cache)
            series['loss'] = loss.numpy().squeeze()

            seed_backward = max_iterations + seed_forward + 1
            renderer.delta_render(seed_backward, image_gradients, spp=backward_spp)

            gradients = {k: ek.gradient(v) for k, v in params.items()}
            grad_df = gradients_to_df({series['method']: gradients})
            for i, row in grad_df.iterrows():
                series['param_name'] = row['Param']
                series['param_gradient'] = row['Value']
                df = df.append(series, ignore_index=True)

            # Apply gradient step
            opt.step()
            ensure_valid_values(params)
            renderer.upload_bsdf_globals()
            print('[{:04d}] Collected {} gradients'.format(it_i, len(grad_df)), end='\r')

    # Save metadata about the run
    fname = join(output_dir, 'gradients.csv')
    with open(fname, 'w') as f:
        f.write(df.to_csv())
    print('[+] Saved gradients data to: {}'.format(fname))



@pytest.mark.parametrize('config_name',
                         ('cbox-rgb', 'cbox-microfacet', 'rb-unit-scattering-homogeneous'))
def test07_eastimate_learning_rate_scaling(config_name):
    from approx_utils import estimate_learning_rate_scaling

    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, 'rb_gradients_scaling_' + config_name)
    os.makedirs(output_dir, exist_ok=True)

    # Reference image and scene loading
    reference_b = Bitmap(config['ref'])
    reference_b.write(join(output_dir, 'ref.exr'))

    loss_function = l2_loss
    def prepare_integrator(allow_approx=True):
        scene_params = config.get('scene_params', [])
        scene = load_scene(config['scene'], parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])

        # Optimization setup
        opt = Adam(params, lr=config.get('learning_rate', 2e-2))
        integrator_type = integrator_type_for_scene(scene, with_gradients=True, config=config)
        renderer = OptiXRenderer(scene, integrator_type, params=params)
        if allow_approx:
            print('[i] Using approximate Li = 1')
            renderer.set_use_approx_li_1(True)
        return renderer, params, opt

    def compute_loss(current_b, cache=None):
        return compute_image_gradients(reference_b, current_b, loss_function, cache=cache)

    ratio, _, _ = estimate_learning_rate_scaling(config, output_dir, prepare_integrator,
                                                 compute_loss, spp_large=512)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(prog='test_radiative_backprop.py')
    parser.add_argument('config_name', type=str, help='Configuration (scene) to optimize')
    args = parser.parse_args()

    test04_radiative_backprop_optimization(args.config_name, use_interleaving=False)
