import os
from os.path import basename, join, realpath

import mitsuba
from mitsuba.optix import OptiXRenderer
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
from mitsuba.render.autodiff import get_differentiable_parameters
import numpy as np
import pytest

from render_helpers import load_scene, integrator_type_for_scene
from config import CONFIGS, OUTPUT_DIR, DIFF_SCENE_DIR

VOLUMES_DEPTH_TEST_SCENE = realpath(join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-depth.xml'))

MAX_DEPTH_VALUES = [1, 2, 3, 4, 6, 8]
HIDE_EMITTERS_VALUES = [False, True]
COMMON_SCENE_PARAMETERS = [
    ('spp', str(1024))
]

def test01_render_mitsuba_cpu():
    output_dir = join(OUTPUT_DIR, 'rb-depth-test')
    os.makedirs(output_dir, exist_ok=True)

    counter = 0
    for hide_emitters in HIDE_EMITTERS_VALUES:
        for max_depth in MAX_DEPTH_VALUES:
            parameters = COMMON_SCENE_PARAMETERS + [
                ('max_depth', str(max_depth)),
                ('hide_emitters', str(hide_emitters)),
            ]

            scene = load_scene(VOLUMES_DEPTH_TEST_SCENE, parameters=parameters)
            integrator = scene.integrator()
            assert integrator.max_depth() == max_depth
            assert integrator.hide_emitters() == hide_emitters
            assert scene.color_mode() == EColorMode.ERGBColor

            integrator.render(scene, vectorize=False)
            b = scene.film().bitmap()
            fname = join(output_dir, '{:02d}-depth-test-mitsuba-md-{}-hide-emitters-{}.exr'.format(
                counter, max_depth, str(hide_emitters).lower()))
            b.convert(Bitmap.ERGB, Struct.EFloat32, False).write(fname)
            print('[+] Rendered: {}'.format(fname))

            counter += 1


@pytest.mark.parametrize('with_gradients', (False, True))
def test02_render_optix_pt_or_rb(with_gradients):
    output_dir = join(OUTPUT_DIR, 'rb-depth-test')
    os.makedirs(output_dir, exist_ok=True)

    counter = 0
    for hide_emitters in HIDE_EMITTERS_VALUES:
        for max_depth in MAX_DEPTH_VALUES:
            parameters = COMMON_SCENE_PARAMETERS + [
                ('max_depth', str(max_depth)),
                ('hide_emitters', str(hide_emitters)),
            ]

            scene = load_scene(VOLUMES_DEPTH_TEST_SCENE, parameters=parameters)
            assert scene.integrator().max_depth() == max_depth
            assert scene.integrator().hide_emitters() == hide_emitters
            assert scene.color_mode() == EColorMode.ERGBColor

            int_type = integrator_type_for_scene(scene, with_gradients=with_gradients)
            int_name = 'rb' if with_gradients else 'pt'
            # int_name = 'rb_rec' if with_gradients else 'pt'  # Enable this for recursion mode

            if with_gradients:
                params = get_differentiable_parameters(scene)
                params.keep([])
                renderer = OptiXRenderer(scene, int_type, params=params)
            else:
                renderer = OptiXRenderer(scene, int_type)


            b = renderer.render(seed=1234)

            fname = join(output_dir, '{:02d}-depth-test-{}-md-{}-hide-emitters-{}.exr'.format(
                counter, int_name, max_depth, str(hide_emitters).lower()))
            b.convert(Bitmap.ERGB, Struct.EFloat32, False).write(fname)
            print('[+] Rendered: {}'.format(fname))

            counter += 1
