import os
from os.path import join, dirname, realpath
import pytest
import re
import time

import enoki as ek
import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode, Log, EInfo
from mitsuba.optix import OptiXRenderer
from mitsuba.render.autodiff import get_differentiable_parameters, Adam
import numpy as np

from render_helpers import load_scene, l2_loss, reference_differentiable_render, compute_image_gradients, finite_differences_render, integrator_type_for_scene
from config import CONFIGS, OUTPUT_DIR

INTERESTING_CONFIGS = [
    'cbox-rgb',
    'cbox-rgb-approx',
    'cbox-textured',
    'simpler',
    'cbox-microfacet',
    'rb-unit-absorptive',
    'rb-unit-scattering-homogeneous',
    'rb-unit-albedo',
    'rb-homogeneous-noise-opt',
    'rb-heterogeneous-noise-opt',
]
for k in INTERESTING_CONFIGS:
    assert k in CONFIGS

ID_RE = r'(?:BSDF|Medium)\[id=\"([a-zA-Z0-9_]+)\"'


def enable_gradients(params):
    for k, _ in params.items():
        ek.set_requires_gradient(params[k])


def print_gradients(gradients):
    for k, v in gradients.items():
        Log(EInfo, '- {:<50s}: {}'.format(k, v))


def plots_gradients(df):
    import matplotlib.pyplot as plt
    import seaborn as sns

    fig, ax = plt.subplots(2, 1, figsize=(12, 2 * 6))
    sns.barplot(x='Param', y='Value', hue='Method', data=df, ax=ax[0])
    ax[0].set_xticklabels(ax[0].get_xticklabels(), rotation=45,
                          horizontalalignment='right')

    methods = df['Method'].unique()
    if len(methods) > 1:
        ref_method = methods[0]

        df2 = df.copy()
        df2['Ratio'] = df2['Value'].copy()
        df2_idx = df.set_index(['Method', 'Param'])
        for i, row in df2.iterrows():
            if row['Method'] == ref_method:
                continue

            denom = df2_idx.loc[ref_method, row['Param']]['Value'].squeeze()
            if row['Value'] == 0 and denom == 0:
                df2.loc[i, 'Ratio'] = 1.
            else:
                df2.loc[i, 'Ratio'] = row['Value'] / denom
            df2.loc[i, 'Method'] = '{} / {}'.format(row['Method'], ref_method)
            # Log(EInfo, '{} = {} / {}  for param = {}'.format(
            #     row['Ratio'], row['Value'], denom, row['Param']
            #     ))
        df2 = df2[df2['Method'] != ref_method]

        sns.barplot(x='Param', y='Ratio', hue='Method', data=df2, ax=ax[1])
        ax[1].set_xticklabels(ax[1].get_xticklabels(), rotation=45,
                              horizontalalignment='right')
    return fig, ax


def plots_texture_gradients(gradients_per_method):
    import matplotlib.pyplot as plt
    from render_helpers import add_colorbar

    n_params = None
    n_methods = 0

    images = {}
    to_plot = {}
    bounds = {}

    for method, gradients in gradients_per_method.items():
        if gradients is None:
            continue
        if n_params is None:
            n_params = len(gradients)
        assert n_params == len(gradients)
        n_methods += 1
        to_plot[method] = {}
        images[method] = {}


        for k, values_c in gradients.items():
            slices = ek.slices(values_c)
            # Assume square RGB texture
            side = int(np.sqrt(slices // 3))
            assert side * side * 3 == slices

            values = values_c.numpy().reshape((side, side, 3))

            matches = re.search(ID_RE, k)
            assert matches, k
            obj_id = matches.groups(1)[0]
            images[method][obj_id] = Bitmap(values.copy(), pixel_format=Bitmap.ERGB)

            # TODO: handle R, G, B gradients
            values = values[:, :, 1]
            to_plot[method][obj_id] = values

            vmin, vmax = bounds.get(obj_id, (np.inf, -np.inf))
            # Ignore outlier gradient values
            clip = np.quantile(np.abs(values), 0.95)
            values_clipped = np.clip(values, -clip, clip)

            vmin = min(vmin, np.min(values_clipped))
            vmax = max(vmax, np.max(values_clipped))
            bounds[obj_id] = (vmin, vmax)

    # Symmetrize bounds
    for k, (vmin, vmax) in bounds.items():
        abs_v = np.max(np.abs([vmin, vmax]))
        bounds[k] = (-abs_v, abs_v)

    fig, ax = plt.subplots(n_params, n_methods, figsize=(n_methods * 6, n_params * 6))
    if n_methods == 1 or n_params == 1:
        ax = np.atleast_2d(ax)

    for j, (method, gradients) in enumerate(to_plot.items()):
        for i, (param_name, gradient) in enumerate(gradients.items()):
            vmin, vmax = bounds[param_name]
            widget = ax[i, j].imshow(gradient, cmap='coolwarm', vmin=vmin, vmax=vmax)

            if i == 0:
                ax[i, j].set_title(method)
            if j == 0:
                ax[i, j].set_ylabel(param_name)
            elif j == n_methods - 1:
                add_colorbar(fig, ax[i, j], widget)

    return fig, ax, images


def gradients_to_df(gradients_per_method):
    import pandas as pd
    entry_names = ['x', 'y', 'z', 'w']

    df = pd.DataFrame(columns=['Method', 'Param', 'Value'])

    for method, gradients_list in gradients_per_method.items():
        if gradients_list is None:
            continue

        series = pd.Series(index=df.columns)
        series['Method'] = method
        gradients_list = (gradients_list
                          if isinstance(gradients_list, (list, tuple))
                          else [gradients_list])

        for gradients in gradients_list:
            for k, grad in gradients.items():
                matches = re.search(ID_RE, k)
                assert matches, k
                obj_id = matches.groups(1)[0]

                if isinstance(grad, ek.FloatC):
                    if len(grad) == 1:
                        series['Param'] = obj_id
                        series['Value'] = grad.numpy().squeeze()
                        df = df.append(series, ignore_index=True)
                    else:
                        assert len(grad) <= 50, "Too many entries in parameter {}gradient for plotting: {} entries".format(
                            k, len(grad))
                        # Split entries into individual parameters (works for low-res data)
                        grad_values = grad.numpy().squeeze()
                        for i in range(len(grad)):
                            series['Param'] = '{}_{:02d}'.format(obj_id, i)
                            series['Value'] = grad_values[i]
                            df = df.append(series, ignore_index=True)

                elif isinstance(grad, ek.Vector3fC):
                    for i in range(len(grad)):
                        series['Param'] = obj_id + '.' + entry_names[i]
                        series['Value'] = grad[i].numpy().squeeze()
                        df = df.append(series, ignore_index=True)
                else:
                    raise ValueError(
                        'Unsupported gradient type {} for param {} '
                        '(method {}): {}'.format(type(grad), k, method, grad))

    return df


def accumulate_gradient_observations(method, base_seed, total_spp, pass_count=None):
    """Calls `method` multiple times and accumulates the resulting gradients."""
    pass_count = pass_count or total_spp
    spp_per_pass = total_spp // pass_count
    assert spp_per_pass * pass_count == total_spp

    accumulated = {}
    for i in range(pass_count):
        # TODO: make sure the passes are not correlated!
        seed = base_seed + i * pass_count * total_spp
        gradients = method(spp=spp_per_pass, seed=seed)
        for k, v in gradients.items():
            if k not in accumulated:
                accumulated[k] = []
            accumulated[k].append(v)

    for k, l in accumulated.items():
        assert len(l) == pass_count
    return accumulated


@pytest.mark.parametrize('current_config', INTERESTING_CONFIGS)
def test01_gradient_values(current_config, enable_fd=True, enable_rb_fd=True,
                           enable_ad=True, force_spp=None, figure_fname=None):
    if isinstance(current_config, str):
        config = CONFIGS[current_config]
    else:
        config = current_config
        current_config = config['name']

    has_volumes = config.get('has_volumes', False)
    is_heterogeneous = config.get('is_heterogeneous', False)
    scene_params = config.get('scene_params', [])
    base_seed = 1234
    fd_epsilon = config.get('fd_epsilon', 1e-2)

    scene_fname = config['scene']
    reference_b = Bitmap(config['ref'])
    total_spp = force_spp or config.get('backward_spp', 64)
    loss_function = l2_loss
    def get_integrator_type(scene):
        return integrator_type_for_scene(scene, with_gradients=True, config=config)

    output_dir = join(OUTPUT_DIR, 'test-gradients-' + current_config)
    os.makedirs(output_dir, exist_ok=True)

    def compute_rb_gradients(spp, seed):
        # Load scene
        scene = load_scene(scene_fname, parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])
        renderer = OptiXRenderer(scene, get_integrator_type(scene),
                                 params=params)
        if config.get('use_approx_li_1', False):
            Log(EInfo, '[i] Using approximate Li = 1')
            renderer.set_use_approx_li_1(True)

        # Render current state (forward pass)
        current_b = renderer.render(seed, spp=total_spp)
        # Save forward pass
        fwd_fname = join(output_dir, 'rb_forward.exr')
        current_b.convert(Bitmap.ERGB, Struct.EFloat32, False).write(fwd_fname)

        # Compute image gradients
        image_gradients, _, _ = compute_image_gradients(
            reference_b, current_b, loss_function)

        if len(image_gradients) <= 16:
            Log(EInfo, 'Image gradients (flattened): {}'.format(image_gradients))

        renderer.upload_bsdf_globals()

        Log(EInfo, '[ ] Running radiative backropagation...')
        enable_gradients(params)
        renderer.delta_render(seed + 1, image_gradients, spp=spp)
        Log(EInfo, '[+] Done.')

        return {k: ek.gradient(v) for k, v in params.items()}

    def compute_autodiff_gradients(spp, seed):
        # Load scene
        scene = load_scene(scene_fname, parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])
        opt = Adam(params, lr=np.nan)

        if has_volumes:
            if is_heterogeneous:
                spp_per_pass = min(32, spp)
            else:
                # spp_per_pass = min(48, spp)
                spp_per_pass = min(64, spp)
        elif scene.integrator().max_depth() > 2:
            spp_per_pass = min(8, spp)
            # spp_per_pass = min(1, spp)
        else:
            spp_per_pass = min(8, spp)
        pass_count = int(np.ceil(spp / spp_per_pass))
        Log(EInfo, '[ ] Running autodiff-based differentiable rendering ({} x {}spp)...'.format(
            pass_count, spp_per_pass))
        _, gradients = reference_differentiable_render(scene, reference_b, opt,
                                                       loss_function=loss_function,
                                                       pass_count=pass_count, spp=spp_per_pass,
                                                       gradient_spp=spp_per_pass,
                                                       base_seed=seed,
                                                       output_dir=output_dir)
        Log(EInfo, '[+] Done.')
        return gradients

    def compute_finite_differences_gradients(spp, seed):
        # Load scene
        scene = load_scene(scene_fname, parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])
        enable_gradients(params)
        opt = Adam(params, lr=np.nan)

        if has_volumes:
            spp_per_pass = min(64, spp)
        else:
            spp_per_pass = min(8, spp)
        pass_count = int(np.ceil(spp / spp_per_pass))
        Log(EInfo, '[ ] Running finite differences-based rendering '
              '({} x {}spp x {} params, may be slow)...'.format(pass_count, spp_per_pass, len(config['params'])))
        gradients = finite_differences_render(
            scene, reference_b, opt, loss_function=loss_function,
            epsilon=fd_epsilon,
            pass_count=pass_count, spp=spp_per_pass, base_seed=seed,
            output_dir=output_dir)
        Log(EInfo, '[+] Done.')
        return gradients

    def compute_finite_differences_rb_gradients(spp, seed):
        # Load scene
        scene = load_scene(scene_fname, parameters=scene_params)
        params = get_differentiable_parameters(scene)
        params.keep(config['params'])
        enable_gradients(params)
        opt = Adam(params, lr=np.nan)

        empty_params = mitsuba.render.autodiff.DifferentiableParameters()
        renderer = OptiXRenderer(scene, get_integrator_type(scene),
                                 params=empty_params)

        def rb_render_function(this_scene, pixel_format, spp):
            nonlocal scene
            assert this_scene == scene
            assert pixel_format == Bitmap.ERGB
            renderer.upload_bsdf_globals()
            return renderer.render(seed + 1, spp=spp)

        Log(EInfo, '[ ] Running finite differences-based rendering with RB '
              '(1 x {}spp x {} params, may be slow)...'.format(spp, len(config['params'])))
        gradients = finite_differences_render(
            scene, reference_b, opt, loss_function=loss_function,
            epsilon=fd_epsilon, pass_count=1, spp=spp, base_seed=seed,
            output_dir=output_dir,
            render_function=rb_render_function, method_name='fd_rb')
        Log(EInfo, '[+] Done.')
        return gradients

    # ----- Radiative Backprop method
    # rb_gradients = None
    rb_gradients = compute_rb_gradients(spp=total_spp, seed=base_seed)
    if rb_gradients is not None:
        Log(EInfo, '\nRadiative backropagation:')
        print_gradients(rb_gradients)
        Log(EInfo, '\n')

    # ----- Autodiff method
    ad_gradients = None
    if enable_ad:
        ad_gradients = compute_autodiff_gradients(spp=total_spp, seed=base_seed)
        if ad_gradients is not None:
            Log(EInfo, '\nAutodiff-based:')
            print_gradients(ad_gradients)
            Log(EInfo, '\n')

    # ----- Finite differences method
    fd_gradients = None
    if enable_fd:
        fd_gradients = compute_finite_differences_gradients(spp=total_spp, seed=base_seed)
        if fd_gradients is not None:
            Log(EInfo, '\nFinite differences-based:')
            print_gradients(fd_gradients)
        Log(EInfo, '\n')

    # ----- Finite differences of RB method
    fd_rb_gradients = None
    if enable_rb_fd:
        fd_rb_gradients = compute_finite_differences_rb_gradients(spp=total_spp, seed=base_seed)
        if fd_rb_gradients is not None:
            Log(EInfo, '\nFinite differences-based with RB:')
            print_gradients(fd_rb_gradients)
            Log(EInfo, '\n')

    # ----- Summary
    gradients_per_method = {
        'rb': rb_gradients,
        'ad': ad_gradients,
        'fd': fd_gradients,
        'fd_rb': fd_rb_gradients,
    }
    is_texture = 'bitmap' in config['params'][0]  # TODO: better support
    figure_fname = figure_fname or join(output_dir, 'gradients-comparison-{}.pdf'.format(current_config))
    if is_texture:
        fig, _, images = plots_texture_gradients(gradients_per_method)
        for method, ims in images.items():
            for obj_id, image in ims.items():
                im_fname = '{}_{}_{}.exr'.format(
                    os.path.splitext(figure_fname)[0], method, obj_id)
                image.write(im_fname)

    else:
        df = gradients_to_df(gradients_per_method)
        fig, _ = plots_gradients(df)
        df_fname = os.path.splitext(figure_fname)[0] + '.csv'
        with open(df_fname, 'w') as f:
            f.write(df.to_csv(index=False))

    fig.savefig(figure_fname, bbox_inches='tight', dpi=200)
    # plt.show()
    Log(EInfo, '[+] Gradients comparison figure saved to: {}'.format(figure_fname))



if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(prog='test_rb_gradients.py')
    parser.add_argument('config_name', type=str, help='Configuration (scene) to test on')
    parser.add_argument('--no-fd', action='store_true', help='Disable FD-based validation')
    parser.add_argument('--no-fd-rb', action='store_true', help='Disable RB-FD-based validation')
    parser.add_argument('--spp', type=int, default=None, help='Number of spp to use')
    args = parser.parse_args()

    test01_gradient_values(args.config_name, enable_fd=not args.no_fd,
                           enable_rb_fd=not args.no_fd_rb, force_spp=args.spp)
