from copy import deepcopy
import os
from os.path import join, dirname, realpath, splitext
import pytest

import enoki as ek
import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
from mitsuba.render.autodiff import DifferentiableParameters
import numpy as np

from render_helpers import load_scene, l2_loss, reference_differentiable_render, compute_image_gradients, finite_differences_render
from config import CONFIGS, OUTPUT_DIR, DIFF_SCENE_DIR, RES_SCENE_DIR

from test_rb_gradients import test01_gradient_values as compute_gradients_plot


parametrization = pytest.mark.parametrize(
    'hide_emitters, emitter_type',
    [(False, 'envmap'), (True, 'envmap'), (False, 'constant'), (True, 'constant'), ],
)
MAX_DEPTH_VALUES = [1, 2, 3, 4, 6, 8]
DENSITY_VALUES = [0.5, 1.0, 5,
                  0.01, 0.1, 2, 3, 10, 50]
ENABLE_AD = True
# ENABLE_AD = False
ENABLE_FD = True
# ENABLE_FD = False
ENABLE_RB_FD = True

FIGURE_FNAME_FMT = realpath(join(OUTPUT_DIR,
    'rb-gradients/{check_type}/d-{density}-md-{max_depth}-emitters-{hide_emitters}-{emitter_type}{extras}.pdf'))
FIGURE_FNAME_SURFACE_FMT = realpath(join(OUTPUT_DIR,
    'rb-gradients/{check_type}/md-{max_depth}{extras}.pdf'))
SKIP_EXISTING = True


def compute_ref_if_needed(config):
    from mitsuba.optix import OptiXRenderer
    from render_helpers import integrator_type_for_scene

    ref_fname = config['ref']
    if not os.path.isfile(ref_fname):
        print('[ ] Rendering reference image with OptiX...')
        scene = load_scene(config['scene'], parameters=config['ref_scene_params'])
        renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
        current_b = renderer.render(seed=1234, spp=config.get('ref_spp', 1024))
        # current_b = renderer.render(seed=1234, spp=8192)
        current_b.write(ref_fname)
        print('[+] Saved reference image: {}'.format(ref_fname))

def render_forward(config, filename, seed=1234):
    from mitsuba.optix import OptiXRenderer
    from render_helpers import integrator_type_for_scene
    scene = load_scene(config['scene'], parameters=config['scene_params'])

    params = DifferentiableParameters()
    int_type = integrator_type_for_scene(scene, with_gradients=False, config=config)
    renderer = OptiXRenderer(scene, int_type, params=params)

    current_b = renderer.render(seed, spp=config['forward_spp'])
    current_b.convert(Bitmap.ERGB, Struct.EFloat32, False).write(filename)


@parametrization
def test01_purely_absorptive(hide_emitters, emitter_type):
    # if hide_emitters:
    #     pytest.skip('Absorptive emitters are not visible on dark background')

    # logger = mitsuba.core.Thread.thread().logger()
    # logger.set_log_level(mitsuba.core.EInfo)
    spp = 128
    scene_fname = 'medium-opt/rb-unit-absorptive-homogeneous.xml'
    config = {
        'name': 'rb-unit-absorptive-homogeneous',
        'scene': join(DIFF_SCENE_DIR, scene_fname),
        'ref': join(DIFF_SCENE_DIR, splitext(scene_fname)[0] + '_ref.exr'),
        'params': ['/Scene/Mesh[id="volume_bbox"]/Medium[id="medium1"]/density'],
        'has_volumes': True,
        'is_heterogeneous': False,
        'ref_scene_params': [('density', str(2))],
        'forward_spp': spp,
        'backward_spp': spp,
    }
    compute_ref_if_needed(config)

    for density in DENSITY_VALUES:
        for max_depth in MAX_DEPTH_VALUES[1:]:
            config['scene_params'] = [
                ('hide_emitters', str(hide_emitters).lower()),
                ('emitter_type', str(emitter_type)),
                ('density', str(density)),
                ('max_depth', str(max_depth)),
            ]
            figure_fname = FIGURE_FNAME_FMT.format(
                check_type='absorptive-homogeneous', density=density, max_depth=max_depth,
                hide_emitters=hide_emitters, emitter_type=emitter_type,
                extras=''
            )
            os.makedirs(dirname(figure_fname), exist_ok=True)

            if SKIP_EXISTING and os.path.isfile(figure_fname):
                continue

            print('[ ] Starting: {}, d = {}, md = {}, emitters = {}, type = {}'.format(
                config['name'], density, max_depth, hide_emitters, emitter_type))
            # Render an illustrative image for this combination
            render_forward(config, splitext(figure_fname)[0] + '.exr')

            # Compute gradients
            compute_gradients_plot(config, enable_ad=ENABLE_AD, enable_fd=ENABLE_FD,
                                   enable_rb_fd=ENABLE_RB_FD, figure_fname=figure_fname)
            print('[+] Saved: {}'.format(figure_fname))

@parametrization
def test02_purely_scattering(hide_emitters, emitter_type):
    pass


@parametrization
def test03_heterogeneous_medium_density_scattering(hide_emitters, emitter_type):
    if hide_emitters:
        pytest.skip('No time to run this yet')
    if emitter_type != 'envmap':
        pytest.skip('No time to run this yet')

    spp = 128
    density_value_initial = 0.1
    config = {
        'name': 'rb-unit-scattering-heterogeneous',
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-heterogeneous.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-heterogeneous_ref.exr'),
        'params': [
            '/Scene/Mesh[id="volume_bbox"]/Medium[id="medium1"]/Texture3D[id="density"]/data',
        ],
        'has_volumes': True,
        'is_heterogeneous': True,
        'scene_params': [
            ('emitter_type', 'envmap'),
            ('max_depth', str(8)),
            ('volume_filename', 'textures/constant-side-3-value-{}.vol'.format(density_value_initial)),
        ],
        'ref_scene_params': [
            ('emitter_type', 'envmap'),
            ('max_depth', str(8)),
            ('volume_filename', 'textures/noise-side-3.vol'),
        ],
        'max_iterations': 10,
        'forward_spp': spp,
        'backward_spp': spp,
    }
    compute_ref_if_needed(config)

    for max_depth in MAX_DEPTH_VALUES[1:]:
        config['scene_params'] = [
            ('hide_emitters', str(hide_emitters).lower()),
            ('emitter_type', str(emitter_type)),
            ('max_depth', str(max_depth)),
        ]
        figure_fname = FIGURE_FNAME_FMT.format(
            check_type=config['name'], density=density_value_initial, max_depth=max_depth,
            hide_emitters=hide_emitters, emitter_type=emitter_type,
            extras=''
        )
        os.makedirs(dirname(figure_fname), exist_ok=True)

        if SKIP_EXISTING and os.path.isfile(figure_fname):
            continue

        print('[ ] Starting: {}, d = {}, md = {}, emitters = {}, type = {}'.format(
            config['name'], density_value_initial, max_depth, hide_emitters, emitter_type))
        # Render an illustrative image for this combination
        render_forward(config, splitext(figure_fname)[0] + '.exr')

        # Compute gradients
        compute_gradients_plot(config, enable_ad=ENABLE_AD, enable_fd=ENABLE_FD,
                               enable_rb_fd=ENABLE_RB_FD, figure_fname=figure_fname)
        print('[+] Saved: {}'.format(figure_fname))



@pytest.mark.parametrize('config_name', ('cbox-rgb', 'cbox-envmap', 'cbox-envmap-only'))
def test06_diffuse_rgb(config_name):
    from config import CBOX_PARAM_NAMES

    # logger = mitsuba.core.Thread.thread().logger()
    # logger.set_log_level(mitsuba.core.EWarn)
    spp = 128
    scene_fname = CONFIGS[config_name]['scene']
    config = {
        'name': config_name,
        'scene': scene_fname,
        'ref_scene': CONFIGS[config_name]['ref_scene'],
        'ref': splitext(scene_fname)[0] + '_ref_256.exr',
        'params': CBOX_PARAM_NAMES,
        'ref_scene_params': [('res', str(256)), ('max_depth', str(8))],
        'forward_spp': spp,
        'backward_spp': spp,
    }
    compute_ref_if_needed(config)

    for max_depth in MAX_DEPTH_VALUES[1:]:
        config['scene_params'] = [
            ('max_depth', str(max_depth)),
            ('res', str(256)),
        ]
        figure_fname = FIGURE_FNAME_SURFACE_FMT.format(
            check_type=config_name, max_depth=max_depth, extras='')
        os.makedirs(dirname(figure_fname), exist_ok=True)

        if SKIP_EXISTING and os.path.isfile(figure_fname):
            continue

        print('[ ] Starting: {}, md = {}'.format(
            config['name'], max_depth))
        # Render an illustrative image for this combination
        render_forward(config, splitext(figure_fname)[0] + '.exr')
        # Compute gradients
        compute_gradients_plot(config, enable_ad=ENABLE_AD, enable_fd=ENABLE_FD,
                                enable_rb_fd=ENABLE_RB_FD, figure_fname=figure_fname)
        print('[+] Saved: {}'.format(figure_fname))


@pytest.mark.parametrize('config_name', ('object-envmap-diffuse',))
def test07_envmap_only(config_name):
    logger = mitsuba.core.Thread.thread().logger()
    logger.set_log_level(mitsuba.core.EWarn)
    spp = 128
    config = deepcopy(CONFIGS[config_name])
    compute_ref_if_needed(config)
    original_params = deepcopy(config['scene_params'])

    for max_depth in MAX_DEPTH_VALUES:
        config['scene_params'] = original_params + [
            ('max_depth', str(max_depth)),
            ('res', str(256)),
        ]
        config['name'] = config_name
        figure_fname = FIGURE_FNAME_SURFACE_FMT.format(
            check_type=config_name, max_depth=max_depth, extras='')
        os.makedirs(dirname(figure_fname), exist_ok=True)

        if SKIP_EXISTING and os.path.isfile(figure_fname):
            continue

        print('[ ] Starting: {}, md = {}'.format(config_name, max_depth))
        # Render an illustrative image for this combination
        render_forward(config, splitext(figure_fname)[0] + '.exr')
        # Compute gradients
        compute_gradients_plot(config, enable_ad=ENABLE_AD, enable_fd=ENABLE_FD,
                                enable_rb_fd=ENABLE_RB_FD, figure_fname=figure_fname)
        print('[+] Saved: {}'.format(figure_fname))


@pytest.mark.skip(reason='Not implemented')
def test08_diffuse_texture():
    pass

@pytest.mark.skip(reason='Not implemented')
def test09_microfacet_alpha():
    pass
