#include <iostream>

#include <enoki/autodiff.h>
#include <enoki/cuda.h>
#include <mitsuba/core/argparser.h>
#include <mitsuba/core/bitmap.h>
#include <mitsuba/core/filesystem.h>
#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/fstream.h>
#include <mitsuba/core/jit.h>
#include <mitsuba/core/logger.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/core/profiler.h>
#include <mitsuba/core/thread.h>
#include <mitsuba/core/util.h>
#include <mitsuba/core/vector.h>
#include <mitsuba/core/xml.h>
#include <mitsuba/mitsuba.h>
#include <mitsuba/render/common.h>
#include <mitsuba/render/integrator.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/texture3d.h>

using namespace mitsuba;
using namespace enoki;
using FloatC = CUDAArray<float>;
using FloatD = DiffArray<FloatC>;
using UInt32D = replace_scalar_t<FloatD, uint32_t>;
using UInt64C = replace_scalar_t<FloatC, uint64_t>;
using Point3fC = mitsuba::Point<FloatC, 3>;


#define EXAMPLE_VOLUME_FILE "/home/merlin/Code/differentiable-rendering/scenes/medium-opt/textures/smoke.vol"


std::string extract_grid3d_eval() {
    using Interaction3fC = Interaction<Point3fC>;

    ref<Object> obj = xml::load_string(R"(
        <texture3d type="grid3d" version="2.0.0">
            <string name="filename" value=")" EXAMPLE_VOLUME_FILE R"("/>
            <float name="max_value" value="1.0"/>
        </texture3d>
    )");
    ref<Texture3D> texture = dynamic_cast<Texture3D *>(obj.get());

    Interaction3fC it = zero<Interaction3fC>(1);
    set_label(it, "it");
    texture->set_labels("grid");
    cuda_eval();

    cuda_create_ptx_module_context();

    /* grid3d_eval */ {
        cuda_start_recording_ptx_function("grid3d_eval");

        auto result = texture->eval(it);
        set_label(result, "result");

        cuda_set_inputs(it);
        cuda_set_outputs(result);
        cuda_stop_recording_ptx_function();
    }

    char* ptx_c = cuda_get_ptx_module();
    if (ptx_c == nullptr)
        throw std::runtime_error("Could not extract function");
    std::string ptx = ptx_c;

    free(ptx_c);
    cuda_destroy_ptx_module_context();
    return ptx;
}

std::string extract_homogeneous_medium_eval() {
    using Ray3fC = mitsuba::Ray<Point3fC>;

    ref<Object> obj = xml::load_string(R"(
        <medium type="homogeneous" version="2.0.0">
            <float name="density" value="0.5"/>
            <rgb name="albedo" value="0.5, 0.8, 0.95"/>
        </medium>
    )");
    ref<Medium> medium = dynamic_cast<Medium *>(obj.get());

    Ray3fC ray = zero<Ray3fC>(1);
    UInt32D index = zero<UInt32D>(1);
    set_label(ray, "ray");
    set_label(index, "index");
    medium->set_labels("medium");
    cuda_eval();

    cuda_create_ptx_module_context();

    /* homogeneous_eval_transmittance */ {
        cuda_start_recording_ptx_function("homogeneous_eval_transmittance");

        auto result = medium->eval_transmittance(ray, nullptr, index);
        set_label(result, "result");

        cuda_set_inputs(ray, index);
        cuda_set_outputs(result);
        cuda_stop_recording_ptx_function();
    }

    char* ptx_c = cuda_get_ptx_module();
    if (ptx_c == nullptr)
        throw std::runtime_error("Could not extract function");
    std::string ptx = ptx_c;

    free(ptx_c);
    cuda_destroy_ptx_module_context();
    return ptx;
}

std::string extract_heterogeneous_medium_eval() {
    using Ray3fC = mitsuba::Ray<Point3fC>;

    ref<Object> obj = xml::load_string(R"(
        <medium type="heterogeneous" version="2.0.0">
            <float name="density_scale" value="5"/>

            <spectrum name="sigma_t" type="uniform">
                <float name="value" value="1.0"/>
            </spectrum>

            <phase type="isotropic" />
            <float name="raymarching_step_scaling" value="0.5"/>

            <texture3d type="constant3d" name="albedo">
                <rgb name="color" value="0.2, 0.5, 0.8"/>
            </texture3d>

            <texture3d type="grid3d" name="density">
                <string name="filename" value=")" EXAMPLE_VOLUME_FILE R"("/>
                <float name="max_value" value="1.0"/>
            </texture3d>

            <boolean name="sample_emitters" value="true"/>
            <boolean name="use_ratio_tracking" value="false"/>
            <boolean name="use_raymarching" value="true"/>
        </medium>
    )");
    ref<Medium> medium = dynamic_cast<Medium *>(obj.get());

    Ray3fC ray = zero<Ray3fC>(1);
    UInt32D index = zero<UInt32D>(1);
    set_label(ray, "ray");
    set_label(index, "index");
    medium->set_labels("medium");
    cuda_eval();

    cuda_create_ptx_module_context();

    /* heterogeneous_eval_transmittance */ {
        cuda_start_recording_ptx_function("heterogeneous_eval_transmittance");

        auto result = medium->eval_transmittance(ray, nullptr, index);
        set_label(result, "result");

        cuda_set_inputs(ray, index);
        cuda_set_outputs(result);
        cuda_stop_recording_ptx_function();
    }

    char* ptx_c = cuda_get_ptx_module();
    if (ptx_c == nullptr)
        throw std::runtime_error("Could not extract function");
    std::string ptx = ptx_c;

    free(ptx_c);
    cuda_destroy_ptx_module_context();
    return ptx;
}

void print_ptx(const std::string &ptx) {
    std::cout << "////////////////////////////*/"    << std::endl;
    std::cout << "///     ____ _______  __    */"    << std::endl;
    std::cout << "///    |  _ \\_   _\\ \\/ /    */" << std::endl;
    std::cout << "///    | |_) || |  \\  /     */"   << std::endl;
    std::cout << "///    |  __/ | |  /  \\     */"   << std::endl;
    std::cout << "///    |_|    |_| /_/\\_\\    */"  << std::endl;
    std::cout << "///                         */"    << std::endl;
    std::cout << "////////////////////////////*/"    << std::endl << std::endl;
    std::cout << ptx << std::endl;
}

int main() {
    Jit::static_initialization();
    Class::static_initialization();
    Thread::static_initialization();
    Logger::static_initialization();
    Bitmap::static_initialization();
    Profiler::static_initialization();
    librender_nop();

    print_ptx(extract_grid3d_eval());
    print_ptx(extract_homogeneous_medium_eval());
    print_ptx(extract_heterogeneous_medium_eval());

    std::cout << "Test complete." << std::endl;

    Profiler::static_shutdown();
    Bitmap::static_shutdown();
    Logger::static_shutdown();
    Thread::static_shutdown();
    Class::static_shutdown();
    Jit::static_shutdown();
}
