#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/phase.h>
#include <mitsuba/render/spectrum.h>

NAMESPACE_BEGIN(mitsuba)

class HGPhaseFunction final : public PhaseFunction {
public:
    HGPhaseFunction(const Properties &props) {
        m_g = props.float_("g", 0.8f);
        if (m_g >= 1 || m_g <= -1)
            Log(EError, "The asymmetry parameter must lie in the interval (-1, 1)!");
    }

    template <typename MediumInteraction, typename Point2,
              typename Value   = typename Point2::Value,
              typename Vector3 = typename MediumInteraction::Vector3>
    MTS_INLINE Vector3 sample_impl(const MediumInteraction & /* mi */,
                                   const Point2 &sample,
                                   mask_t<Value> /*active*/) const {
        Value cos_theta;
        if (std::abs(m_g) < math::Epsilon) {
            cos_theta = 1 - 2 * sample.x();
        } else {
            Value sqr_term = (1 - m_g * m_g) / (1 - m_g + 2 * m_g * sample.x());
            cos_theta = (1 + m_g * m_g - sqr_term * sqr_term) / (2 * m_g);
        }

        Value sin_theta = enoki::safe_sqrt(1.0f - cos_theta * cos_theta);
        auto [sin_phi, cos_phi] = enoki::sincos(2 * math::Pi * sample.y());
        return Vector3(sin_theta * cos_phi, sin_theta * sin_phi, cos_theta);
    }

    template <typename Vector3, typename Value = value_t<Vector3>>
    MTS_INLINE Value eval_impl(const Vector3 &wo,
                               mask_t<Value> /*active*/) const {
        Value temp = 1.0f + m_g * m_g - 2.0f * m_g * wo.z();
        return math::InvFourPi * (1 - m_g * m_g) / (temp * enoki::sqrt(temp));
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "HGPhaseFunction[" << std::endl
            << "  g = " << string::indent(m_g) << std::endl
            << "]";
        return oss.str();
    }

    MTS_IMPLEMENT_PHASE_FUNCTION_ALL()
    MTS_DECLARE_CLASS()
private:
    Float m_g;
};

MTS_IMPLEMENT_CLASS(HGPhaseFunction, PhaseFunction)
MTS_EXPORT_PLUGIN(HGPhaseFunction, "Henyey-Greenstein phase function")

NAMESPACE_END(mitsuba)
