"""
Self-contained optimization script.
"""

#!/usr/bin/python3

import os
from os.path import realpath, dirname, join
import time

import numpy as np
import enoki as ek
import mitsuba
from mitsuba.core import Bitmap, Struct
from mitsuba.core.xml import load_file
from mitsuba.render import ImageBlock
from mitsuba.render.autodiff import (get_differentiable_parameters, render, Adam)
from mitsuba.test.util import fresolver_append_path


@fresolver_append_path
def load_scene(filename, *args, **kwargs):
    """Prepares the file resolver and loads a Mitsuba scene from the given path."""
    scene = load_file(filename, *args, **kwargs)
    assert scene is not None
    return scene


def render_reference(scene_path, output, spp=1024):
    scene = load_scene(scene_path, parameters=[('spp', str(spp))])
    scene.integrator().render(scene, vectorize=False)
    film = scene.film()
    film.set_destination_file(output, 32)
    film.develop()
    print('[+] Rendered reference image at {} spp: {}'.format(spp, output))


def load_reference(filename):
    ref = Bitmap(filename).convert(Bitmap.ERGB, Struct.EFloat32, False)
    return linearize_image(ref), ref


def linearize_image(image):
    if isinstance(image, ImageBlock):
        X, Y, Z, _, W = image.bitmap_d()
    else:
        assert isinstance(image, Bitmap)
        arr = np.array(image, copy=False)
        channels = [ek.FloatD(arr[:, :, k].flatten()) for k in range(image.channel_count())]

        if image.pixel_format() == Bitmap.ERGB:
            # No conversion to do
            return ek.Vector3fD(*channels)

        assert len(channels) == 5
        X, Y, Z, _, W = channels

    # Apply normalization weight
    W = ek.select(ek.eq(W, 0), 0, ek.rcp(W))
    X *= W
    Y *= W
    Z *= W
    # XYZ to RGB conversion
    return ek.Vector3fD(
        3.240479 * X + -1.537150 * Y + -0.498535 * Z,
        -0.969256 * X + 1.875991 * Y + 0.041556 * Z,
        0.055648 * X + -0.204043 * Y + 1.057311 * Z
    )

def refresh_requires_gradient(params):
    """Note: this is a temporary workaround that will not be needed in the future."""
    print(params)
    for k, _ in params.items():
        ek.set_requires_gradient(params[k], True)
        parent = params.parent(k)
        if parent is not None:
            parent.parameters_changed()


def optimize_scene(scene, parameters_to_optimize, reference_image,
                   spp=16, learning_rate=4e-2,
                   output_dir=None, max_iterations=100):
    output_dir = output_dir or realpath(join(dirname(scene_path), 'output'))
    print('[ ] Output directory: {}'.format(output_dir))
    os.makedirs(output_dir, exist_ok=True)

    # Load reference image
    ref_flat, ref_bitmap = load_reference(reference_image)
    ref_bitmap.write(join(output_dir, 'ref.exr'))
    # Get differentiable parameters and retain only the selected ones
    params = get_differentiable_parameters(scene)
    params.keep(parameters_to_optimize)

    # Reset optimized parameters to an initial guess
    for k, v in params.items():
        params[k] = type(v).full(0.5, ek.slices(v))
    refresh_requires_gradient(params)

    # Create optimizer (e.g. SGD or Adam)
    opt = Adam(params, lr=learning_rate)

    # Optimization loop
    for i in range(max_iterations):
        t0 = time.time()
        rendered = render(scene, pixel_format=Bitmap.ERGB, spp=spp)
        # Linearize rendered image before computing the loss
        rendered_flat = linearize_image(rendered)
        # L2 loss
        loss = ek.hsum(ek.hsum(ek.sqr(ref_flat - rendered_flat))) / ek.slices(ref_flat)

        # Backpropagate gradients of the loss w.r.t. inputs and take a descent step
        ek.backward(loss)
        opt.step()
        refresh_requires_gradient(params)

        # Save current image & print current loss
        out_bitmap = rendered.bitmap().convert(Bitmap.ERGB, Struct.EFloat32, False)
        out_bitmap.write(join(output_dir, 'out_{:04d}.exr').format(i))
        loss_v = loss.numpy()[0]
        print('[{:03d}] Loss: {:.6f}, took {:.2f}s'.format(i, loss_v, time.time() - t0), end='\r')

    # Final values
    print('\nFinal loss: {:.6f}'.format(loss_v))
    print('Final parameter values:')
    for k, v in params.items():
        print('- {}: {}'.format(k, v))


if __name__ == '__main__':
    scene_dir = realpath(join(dirname(__file__), '../../../resources/data/scenes/cbox'))
    scene_path = join(scene_dir, 'cbox-rgb.xml')
    reference_image = join(scene_dir, 'cbox-rgb-ref.exr')

    # Render reference image first if not present
    if not os.path.isfile(reference_image):
        print('[ ] Reference image file missing, rendering...')
        render_reference(scene_path, reference_image, spp=4096)

    # Load scene and optimize for the desired parameters
    scene = load_scene(scene_path)
    optimize_scene(
        scene,
        parameters_to_optimize=[
            '/Scene/BSDF[id="box"]/ContinuousSpectrum/coeff',
            '/Scene/BSDF[id="green"]/ContinuousSpectrum/coeff',
            '/Scene/BSDF[id="light"]/ContinuousSpectrum/coeff',
            '/Scene/BSDF[id="red"]/ContinuousSpectrum/coeff',
            '/Scene/BSDF[id="white"]/ContinuousSpectrum/coeff',
        ],
        reference_image=reference_image,
        # Reduce the number of samples per pixels if encountering out-of-memory errors
        spp=16,
    )
