#include <mitsuba/core/properties.h>
#include <mitsuba/core/transform.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/autodiff.h>
#include <mitsuba/render/srgb.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <mitsuba/render/scene.h>
#include <optix.h>
#include <optix_cuda_interop.h>
#endif

NAMESPACE_BEGIN(mitsuba)

class SRGBSpectrum final : public ContinuousSpectrum {
public:
    using Color3fC = Color<FloatC, 3>;

    SRGBSpectrum(const Properties &props)
        : ContinuousSpectrum(props) {
        Color3f color = props.color("color");
        m_rgb_mode = ((EColorMode) props.int_("color_mode")) == ERGBColor;
        if (any(color < 0 || color > 1) && !m_rgb_mode)
            Throw("Invalid RGB reflectance value %s, must be in the range [0, 1]!", color);

        m_coeff = m_rgb_mode ? Vector3f(color) : srgb_model_fetch(color);

#if defined(MTS_ENABLE_AUTODIFF)
        m_color_rgb_d = color;
        m_coeff_d     = m_rgb_mode ? Vector3fD(m_color_rgb_d) : srgb_model_fetch_d(m_color_rgb_d);
#endif
    }

    template <typename Value, typename Mask = mask_t<Value>>
    MTS_INLINE Value eval_impl(Value wavelengths, Mask /*active*/) const {
        if constexpr (is_cuda_array_v<Value> && !is_diff_array_v<Value>) {
            if (!m_rgb_mode)
                NotImplementedError("OptiX-style rendering in non-RGB mode");
        }

        if (m_rgb_mode) {
            // TODO: avoid this ugly duplication
            if constexpr (is_cuda_array_v<Value> && !is_diff_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
                auto v = detach(m_color_rgb_d);
                return Value(v.x(), v.y(), v.z(), 0.f);
#endif
            } else if constexpr (is_diff_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
                return Value(m_color_rgb_d.x(), m_color_rgb_d.y(), m_color_rgb_d.z(),
                            // Workaround needed for `set_gradient()` on that value
                            0.f * m_color_rgb_d.z());
#endif
            } else {
                return Value(m_coeff.x(), m_coeff.y(), m_coeff.z(), 0.f);
            }
        }

        return srgb_model_eval(coeff<Value>(), wavelengths);
    }

    Float mean() const override {
        if (m_rgb_mode)
            return enoki::mean(m_coeff);

        return srgb_model_mean(m_coeff);
    }

    std::string to_string() const override {
        return tfm::format("SRGBSpectrum[coeff = %f, %f, %f]", m_coeff);
    }


#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override {
        dp.put(this, "coeff", m_coeff_d);
        dp.put(this, "rgb", m_color_rgb_d);
    }

    void parameters_changed() override {
        // m_color_rgb_d = enoki::clamp(m_color_rgb_d, 0.f, 1.f);
        m_coeff_d = m_rgb_mode ? Vector3fD(m_color_rgb_d) : srgb_model_fetch_d(m_color_rgb_d);
        m_coeff = slice(m_coeff_d, 0);
        m_needs_reupload = true;
    }

    void set_labels(const std::string &prefix) const override {
        set_label(m_color_rgb_d, (prefix + ".m_color_rgb").c_str());
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        if (!m_needs_reupload)
            return;

        cuda_eval();
        // TODO: it would be much nicer to avoid this round-trip
        Color3f rgb;
        for (size_t k = 0; k < Color3f::Size; ++k)
            cuda_memcpy_from_device(&rgb[k], m_color_rgb_d[k].data(), 1 * sizeof(Float));

        RTvariable var;
        for (size_t k = 0; k < Color3f::Size; ++k) {
            auto name = tfm::format("%s_m_color_rgb_%d", prefix, k);
            // TODO: add rt_check
            var = get_or_create_optix_var(name, context, material);
            rtVariableSet1f(var, rgb[k]);
        }

        m_needs_reupload = false;
    }
#endif

    MTS_IMPLEMENT_SPECTRUM_EVAL_ALL()
    MTS_AUTODIFF_GETTER(coeff, m_coeff, m_coeff_d)
    MTS_DECLARE_CLASS()

private:
    Vector3f m_coeff;
    bool m_rgb_mode = false;

#if defined(MTS_ENABLE_AUTODIFF)
    Color3fD m_color_rgb_d;
    Vector3fD m_coeff_d;
    mutable bool m_needs_reupload = true;
#endif
};

MTS_IMPLEMENT_CLASS(SRGBSpectrum, ContinuousSpectrum)
MTS_EXPORT_PLUGIN(SRGBSpectrum, "sRGB spectrum")

NAMESPACE_END(mitsuba)
