#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/autodiff.h>
#include <mitsuba/core/properties.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#include <optix_cuda_interop.h>
#include <mitsuba/render/scene.h>
#endif

NAMESPACE_BEGIN(mitsuba)
/**
 * \brief Spectrum that takes on a constant value between
 * \c MTS_WAVELENGTH_MIN * and \c MTS_WAVELENGTH_MAX
 */
class UniformSpectrum final : public ContinuousSpectrum {
public:
    UniformSpectrum(const Properties &props)
        : ContinuousSpectrum(props) {
        m_value = props.float_("value");
        m_rgb_mode = ((EColorMode) props.int_("color_mode")) == ERGBColor;

#if defined(MTS_ENABLE_AUTODIFF)
        m_value_d = m_value;
#endif
    }

    template <typename Value>
    MTS_INLINE Value eval_impl(Value lambda, mask_t<Value> active) const {
        if (!m_rgb_mode)
            active &= (lambda >= MTS_WAVELENGTH_MIN) && (lambda <= MTS_WAVELENGTH_MAX);
        return select(active, Value(value<Value>()), Value(0.f));
    }

    template <typename Value>
    MTS_INLINE Value pdf_impl(Value lambda, mask_t<Value> active) const {
        Value pdf = 1.f;
        if (!m_rgb_mode) {
            active &= (lambda >= MTS_WAVELENGTH_MIN) && (lambda <= MTS_WAVELENGTH_MAX);
            pdf = 1.f / (MTS_WAVELENGTH_MAX - MTS_WAVELENGTH_MIN);
        }
        return select(active, pdf, Value(0.f));
    }

    template <typename Value>
    MTS_INLINE std::pair<Value, Value> sample_impl(Value sample, mask_t<Value> /*unused*/) const {
        if (m_rgb_mode) {
            // Wavelengths corresponding to the peak of the CIE RGB color matching curves.
            // Note: they should not be used anyway when rendering in RGB mode.
            return {
                Value(602.785f, 539.285f, 445.772f, 0.f),
                Value(value<Value>() * 1.f)
            };
        } else {
            return {
                MTS_WAVELENGTH_MIN + (MTS_WAVELENGTH_MAX - MTS_WAVELENGTH_MIN) * sample,
                Value(value<Value>() * (MTS_WAVELENGTH_MAX - MTS_WAVELENGTH_MIN))
            };
        }
    }

    Float mean() const override { return m_value; }

    Float max() const override { return m_value; }

#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override {
        dp.put(this, "value", m_value_d);
    }

    void parameters_changed() override {
        cuda_eval();
        // Copy back value to the CPU
        cuda_memcpy_from_device(&m_value, m_value_d.data(), 1 * sizeof(Float));
        cuda_eval();
    }

    void set_labels(const std::string &prefix) const override {
        set_label(m_value_d, (prefix + ".m_value").c_str());
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        auto name = prefix + "_m_value";
        RTvariable var = get_or_create_optix_var(name, context, material);
        rtVariableSet1f(var, m_value);
    }
#endif

    std::string to_string() const override {
        if (m_id.empty())
            return tfm::format("UniformSpectrum[value=%f]", m_value);
        else
            return tfm::format("UniformSpectrum[id=%d, value=%f]", m_id, m_value);
    }

    MTS_IMPLEMENT_SPECTRUM_ALL()
    MTS_AUTODIFF_GETTER(value, m_value, m_value_d)
    MTS_DECLARE_CLASS()

private:
    Float m_value;
    bool m_rgb_mode = false;

#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_value_d;
#endif
};

MTS_IMPLEMENT_CLASS(UniformSpectrum, ContinuousSpectrum)
MTS_EXPORT_PLUGIN(UniformSpectrum, "Uniform spectrum")

NAMESPACE_END(mitsuba)
