#include <mitsuba/core/bitmap.h>
#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/spectrum.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/srgb.h>

#if defined(MTS_ENABLE_AUTODIFF)
#include <optix.h>
#include <optix_cuda_interop.h>
#include <mitsuba/render/scene.h>
#endif

NAMESPACE_BEGIN(mitsuba)

class BitmapTexture final : public ContinuousSpectrum {
public:
    BitmapTexture(const Properties &props)
        : ContinuousSpectrum(props) {
        m_transform = props.transform("to_uv", Transform4f()).extract();
        m_color_mode = (EColorMode) props.int_("color_mode");

        auto fs = Thread::thread()->file_resolver();
        fs::path file_path = fs->resolve(props.string("filename"));
        m_name = file_path.filename().string();
        Log(EDebug, "Loading bitmap texture from \"%s\" ..", m_name);

        m_bitmap_rgb = new Bitmap(file_path);

        /* Convert to linear RGBA float bitmap, will be converted
           into spectral profile coefficients below */
        m_bitmap_rgb = m_bitmap_rgb->convert(Bitmap::ERGB, Bitmap::EFloat, false);

        Float upscale = props.float_("upscale", 1.f);
        if (upscale != 1.f) {
            Vector2s old_size = m_bitmap_rgb->size(),
                     new_size = upscale * old_size;

            ref<ReconstructionFilter> rfilter =
                PluginManager::instance()->create_object<ReconstructionFilter>(
                    Properties("tent"));

            Log(EInfo, "Upsampling bitmap texture \"%s\" to %ix%i ..", m_name,
                       new_size.x(), new_size.y());
            m_bitmap_rgb = m_bitmap_rgb->resample(new_size, rfilter);
        }

        initialize_from_rgb(*m_bitmap_rgb);
    }

    template <typename SurfaceInteraction, typename Mask,
              typename Value    = typename SurfaceInteraction::Value,
              typename Spectrum = mitsuba::Spectrum<Value>>
    MTS_INLINE Spectrum eval_impl(const SurfaceInteraction &it,
                                  Mask active) const {
        using UInt32   = uint32_array_t<Value>;
        using Point2u  = Point<UInt32, 2>;
        using Point2f  = Point<Value, 2>;
        using Vector3f = Vector<Value, 3>;

        if constexpr (is_cuda_array_v<Value> && !is_diff_array_v<Value>) {
            if (m_color_mode != ERGBColor)
                NotImplementedError("OptiX-style rendering in non-RGB mode");
        }

        auto uv = m_transform.transform_affine(it.uv);
        uv -= floor(uv);
        uv *= Vector2f(m_bitmap->size() - 1u);

        Point2u pos = min(Point2u(uv), m_bitmap->size() - 2u);

        Point2f w1 = uv - Point2f(pos),
                w0 = 1.f - w1;

        UInt32 index = pos.x() + pos.y() * (uint32_t) m_bitmap->size().x();

        Vector3f v00, v10, v01, v11;
        if constexpr (is_cuda_array_v<Value> && !is_diff_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
            // TODO: avoid this duplication
            FloatC bitmap_c = detach(m_bitmap_d);
            uint32_t width = (uint32_t) m_bitmap->size().x();
            v00 = gather<Vector3f>(bitmap_c, index, active);
            v10 = gather<Vector3f>(bitmap_c, index + 1u, active);
            v01 = gather<Vector3f>(bitmap_c, index + width, active);
            v11 = gather<Vector3f>(bitmap_c, index + width + 1u, active);
#endif
        } else if constexpr (is_diff_array_v<Value>) {
#if defined(MTS_ENABLE_AUTODIFF)
            uint32_t width = (uint32_t) m_bitmap->size().x();
            v00 = gather<Vector3f>(m_bitmap_d, index, active);
            v10 = gather<Vector3f>(m_bitmap_d, index + 1u, active);
            v01 = gather<Vector3f>(m_bitmap_d, index + width, active);
            v11 = gather<Vector3f>(m_bitmap_d, index + width + 1u, active);
#endif
        } else {
            uint32_t width = (uint32_t) m_bitmap->size().x() * 3;
            const Float *ptr = (const Float *) m_bitmap->data();

            v00 = gather<Vector3f, 0, true>(ptr, index, active);
            v10 = gather<Vector3f, 0, true>(ptr + 3, index, active);
            v01 = gather<Vector3f, 0, true>(ptr + width, index, active);
            v11 = gather<Vector3f, 0, true>(ptr + width + 3, index, active);
        }

        if (m_color_mode == ERGBColor) {
            Vector3f v0 = fmadd(w0.x(), v00, w1.x() * v10),
                     v1 = fmadd(w0.x(), v01, w1.x() * v11),
                     v  = fmadd(w0.y(), v0, w1.y() * v1);
            return Spectrum(v.x(), v.y(), v.z(),
                            // Workaround needed for `set_gradient()` on that value
                            0.f * v.z());
        }

        Spectrum s00 = srgb_model_eval(v00, it.wavelengths),
                 s10 = srgb_model_eval(v10, it.wavelengths),
                 s01 = srgb_model_eval(v01, it.wavelengths),
                 s11 = srgb_model_eval(v11, it.wavelengths);
        Spectrum s0 = fmadd(w0.x(), s00, w1.x() * s10),
                 s1 = fmadd(w0.x(), s01, w1.x() * s11);
        return fmadd(w0.y(), s0, w1.y() * s1);
    }

    Float mean() const override { return m_mean; }

#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override {
        dp.put(this, "bitmap", m_bitmap_d);
    }

    void set_labels(const std::string &prefix) const override {
        auto var_id = cuda_var_register_ptr(m_bitmap_d.data());
        cuda_var_set_label(var_id, (prefix + ".m_bitmap_d_data").c_str());
    }

    void parameters_changed() override {
        m_needs_reupload = true;
    }

    void set_globals(const std::string &prefix, const RTcontext & context,
                     const RTmaterial &material) const override {
        if (!m_needs_reupload)
            return;

        cuda_eval();
        auto name = tfm::format("%s_m_bitmap_d_data", prefix);
        RTvariable var = get_or_create_optix_var(name, context, material);
        const auto *ptr_value = (float *) m_bitmap_d.data();
        rtVariableSetUserData(var, sizeof(ptr_value), &ptr_value);
        m_needs_reupload = false;
    }
#endif

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "Bitmap[" << std::endl
            << "  name = \"" << m_name << "\"," << std::endl
            << "  bitmap = " << string::indent(m_bitmap) << std::endl
            << "]";
        return oss.str();
    }

    MTS_IMPLEMENT_TEXTURE_ALL()
    MTS_DECLARE_CLASS()

protected:
    void initialize_from_rgb(const Bitmap &bitmap_rgb) {
        m_bitmap = new Bitmap(bitmap_rgb);

        Float *ptr  = (Float *) m_bitmap->data();
        double mean = 0.0;
        for (size_t y = 0; y < m_bitmap->size().y(); ++y) {
            for (size_t x = 0; x < m_bitmap->size().x(); ++x) {
                Color3f rgb = load_unaligned<Vector3f>(ptr);
                if (m_color_mode == EMonochrome)
                    rgb = luminance(rgb);

                // Fetch spectral fit for given sRGB color value (3 coefficients)
                Vector3f coeff;
                if (m_color_mode == ERGBColor) {
                    coeff = rgb;
                    mean += (double) enoki::mean(coeff);
                } else {
                    coeff = srgb_model_fetch(rgb);
                    mean += (double) srgb_model_mean(coeff);
                }

                // Overwrite the pixel value with the coefficients
                store_unaligned(ptr, coeff);
                ptr += 3;
            }
        }
        m_mean = Float(mean / hprod(m_bitmap->size()));

#if defined(MTS_ENABLE_AUTODIFF)
        auto *data = (Float *) m_bitmap->data();
        // Spectral upsampling was already carried out if needed, just copy to the GPU
        m_bitmap_d = FloatC::copy(data, m_bitmap->channel_count() * m_bitmap->pixel_count());
#endif
    }

protected:
    ref<Bitmap> m_bitmap;
    ref<Bitmap> m_bitmap_rgb;

#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_bitmap_d;
    mutable bool m_needs_reupload = true;
#endif

    std::string m_name;
    Transform3f m_transform;
    Float m_mean;
    EColorMode m_color_mode = ESpectral;
};

MTS_IMPLEMENT_CLASS(BitmapTexture, ContinuousSpectrum)
MTS_EXPORT_PLUGIN(BitmapTexture, "Bitmap texture")

NAMESPACE_END(mitsuba)
