#include <fstream>
#include <sstream>

#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/string.h>
#include <mitsuba/core/thread.h>
#include <mitsuba/core/transform.h>
#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/texture3d.h>
#include "volume_data.h"

#if defined(MTS_ENABLE_AUTODIFF)
#include <mitsuba/render/scene.h>
#include <optix.h>
#endif

NAMESPACE_BEGIN(mitsuba)

/**
 * Interpolated 3D grid texture of floating point values.
 * Values can be read from the Mitsuba 1 binary volume format, or a simple
 * comma-separated string (for testing).
 *
 * The data must be ordered so that the following C-style (row-major) indexing
 * operation makes sense after the file has been mapped into memory:
 *     data[((zpos*yres + ypos)*xres + xpos)*channels + chan]}
 *     where (xpos, ypos, zpos, chan) denotes the lookup location.
 */
class Grid3D final : public Grid3DBase {
public:
    MTS_AUTODIFF_GETTER(data, m_data, m_data_d)

public:
    explicit Grid3D(const Properties &props) : Grid3DBase(props) {
        VolumeMetadata meta;
        if (props.has_property("filename")) {
            if (props.has_property("side") || props.has_property("nx") ||
                props.has_property("ny") || props.has_property("nz"))
                Throw("Grid dimensions are already specified in the binary volume "
                      "file, cannot override them with properties side, nx, ny or "
                      "nz.");

            meta = read_binary_volume_data<1>(props.string("filename"), &m_data);
        } else {
            meta = parse_string_grid(props, &m_data);
        }

        set_metadata(meta, props.bool_("use_grid_bbox", false));
        if (props.has_property("max_value")) {
            m_fixed_max    = true;
            m_metadata.max = props.float_("max_value");
        }

#if defined(MTS_ENABLE_AUTODIFF)
        // Copy parsed data over to the GPU
        m_data_d = CUDAArray<Float>::copy(m_data.data(), m_size);
#endif
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void put_parameters(DifferentiableParameters &dp) override { dp.put(this, "data", m_data_d); }

    /// Note: this assumes that the size has not changed.
    void parameters_changed() override {
        Grid3DBase::parameters_changed();

        m_metadata.mean = (double) hsum(detach(m_data_d))[0] / (double) m_size;
        if (!m_fixed_max)
            m_metadata.max = hmax(detach(m_data_d))[0];

        m_needs_reupload = true;
    }

    size_t data_size() const override { return m_data_d.size(); }
#endif

    /**
     * Taking a 3D point in [0, 1)^3, estimates the grid's value at that
     * point using trilinear interpolation.
     *
     * The passed `active` mask must disable lanes that are not within the
     * domain.
     */
    template <bool with_gradient, typename Point3, typename Value = value_t<Point3>>
    MTS_INLINE auto trilinear_interpolation(Point3 p, mask_t<Value> active) const {
        using Index    = uint32_array_t<Value>;
        using Index3   = uint32_array_t<Point3>;
        using Spectrum = mitsuba::Spectrum<Value>;
        using Vector3  = Vector<Value, 3>;

        auto nx       = Grid3D::nx<Value>();
        auto ny       = Grid3D::ny<Value>();
        auto nz       = Grid3D::nz<Value>();
        auto z_offset = Grid3D::z_offset<Value>();

        auto max_coordinates = Point3(nx - 1.f, ny - 1.f, nz - 1.f);
        p *= max_coordinates;

        // Integer part (clamped to include the upper bound)
        Index3 pi = enoki::floor2int<Index3>(p);
        pi        = clamp(pi, 0, max_coordinates - 1);
        // pi[active] = clamp(pi, 0, max_coordinates - 1);

        // Fractional part
        Point3 f  = p - Point3(pi),
               rf = 1.f - f;
        // TODO: this check is redundant, right?
        // active &= all(pi >= 0 && (pi + 1) < Index3(nx, ny, nz));

        auto wgather = [&](const Index &index) {
            if constexpr (!is_cuda_array_v<Index>) {
                return gather<Value>(m_data.data(), index, active);
            }
#if defined(MTS_ENABLE_AUTODIFF)
            else if constexpr (!is_diff_array_v<Index>) {
                return gather<Value>(detach(m_data_d), index, active);
            }
            else {
                return gather<Value>(m_data_d, index, active);
            }
#endif
        };

        // (z * ny + y) * nx + x
        Index index = fmadd(fmadd(pi.z(), ny, pi.y()), nx, pi.x());

        auto index_z    = index + z_offset;
        auto index1     = index + 1;
        auto index_z_nx = index_z + nx;
        Value d000 = wgather(index),
              d001 = wgather(index1),
              d010 = wgather(index + nx),
              d011 = wgather(index1 + nx),

              d100 = wgather(index_z),
              d101 = wgather(index_z + 1),
              d110 = wgather(index_z_nx),
              d111 = wgather(index_z_nx + 1);

        Value d00 = fmadd(d000, rf.x(), d001 * f.x()),
              d01 = fmadd(d010, rf.x(), d011 * f.x()),
              d10 = fmadd(d100, rf.x(), d101 * f.x()),
              d11 = fmadd(d110, rf.x(), d111 * f.x());

        Value d0 = fmadd(d00, rf.y(), d01 * f.y()),
              d1 = fmadd(d10, rf.y(), d11 * f.y());

        Value d = fmadd(d0, rf.z(), d1 * f.z());

        if constexpr (!with_gradient) {
            return Spectrum(d);
        } else {
            Value gx0 = fmadd(d001 - d000, rf.y(), (d011 - d010) * f.y()),
                  gx1 = fmadd(d101 - d100, rf.y(), (d111 - d110) * f.y()),
                  gy0 = fmadd(d010 - d000, rf.x(), (d011 - d001) * f.x()),
                  gy1 = fmadd(d110 - d100, rf.x(), (d111 - d101) * f.x()),
                  gz0 = fmadd(d100 - d000, rf.x(), (d101 - d001) * f.x()),
                  gz1 = fmadd(d110 - d010, rf.x(), (d111 - d011) * f.x());

            // Smaller grid cells means variation is faster (-> larger gradient)
            Vector3 gradient(fmadd(gx0, rf.z(), gx1 * f.z()) * (nx - 1),
                             fmadd(gy0, rf.z(), gy1 * f.z()) * (ny - 1),
                             fmadd(gz0, rf.y(), gz1 * f.y()) * (nz - 1));
            return std::pair(Spectrum(d), gradient);
        }
    }

    template <bool with_gradient, typename Interaction,
              typename Value = typename Interaction::Value>
    MTS_INLINE auto eval_impl(const Interaction &it, mask_t<Value> active) const {
        using Spectrum = mitsuba::Spectrum<Value>;
        using Vector3  = Vector<Value, 3>;

        auto p = m_world_to_local * it.p;
        // TODO: is this really not needed?
        // active &= all((p >= 0) && (p <= 1));

        if constexpr (with_gradient) {
            if (none_or<false>(active))
                return std::make_pair(zero<Spectrum>(), zero<Vector3>());
            auto [result, gradient] = trilinear_interpolation<true>(p, active);
            return std::make_pair(select(active, result, zero<Spectrum>()),
                                  select(active, gradient, zero<Vector3>()));
        } else {
            if (none_or<false>(active))
                return zero<Spectrum>();
            return trilinear_interpolation<false>(p, active);
        }
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        Grid3DBase::set_labels(prefix);
        set_label(m_data_d, (prefix + ".m_data_d").c_str());
        auto var_id = cuda_var_register_ptr(m_data_d.data());
        cuda_var_set_label(var_id, (prefix + ".m_data_d_data").c_str());
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        if (!m_needs_reupload)
            return;

        Grid3DBase::set_globals(prefix, context, material);
        cuda_eval();

        RTvariable var;
        // --- Data pointer
        std::string name = prefix + "_m_data_d_data";
        var = get_or_create_optix_var(name, context, material);
        const auto *ptr_value = (float *) m_data_d.data();
        rtVariableSetUserData(var, sizeof(ptr_value), &ptr_value);

        m_needs_reupload = false;
    }
#endif

    MTS_IMPLEMENT_TEXTURE_3D()
    MTS_DECLARE_CLASS()

protected:
    FloatX m_data;
    bool m_fixed_max = false;

#if defined(MTS_ENABLE_AUTODIFF)
    FloatD m_data_d;
    mutable bool m_needs_reupload = true;
#endif
};

MTS_IMPLEMENT_CLASS(Grid3DBase, Texture3D)
MTS_IMPLEMENT_CLASS(Grid3D, Grid3DBase)
MTS_EXPORT_PLUGIN(Grid3D, "Grid 3D texture with interpolation")

NAMESPACE_END(mitsuba)
